from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_audit.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
import datetime
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_reqinfo, trackme_idx_for_tenant

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerAudit_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerAudit_v2, self).__init__(command_line, command_arg, logger)

    def get_resource_group_desc_audit(self, request_info, **kwargs):
        response = {
            "resource_group_name": "audit",
            "resource_group_desc": "These endpoints provide endpoints to generate audit events in the TrackMe tenants, these are used internally and can as well be actioned to generate additional audit events in TrackMe sub-systems",
        }

        return {"payload": response, "status": 200}

    # Register an audit event
    def post_audit_event(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                action = resp_dict["action"]
                user = resp_dict["user"]
                change_type = resp_dict["change_type"]
                object_name = resp_dict["object"]
                object_category = resp_dict["object_category"]
                object_attrs = resp_dict["object_attrs"]
                result = resp_dict["result"]
                comment = resp_dict["comment"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint allows registering a TrackMe audit event:",
                "resource_desc": "Audit a new event",
                "resource_spl_example": '| trackme url="/services/trackme/v2/audit/audit_event" mode="post", body="{\'tenant_id\': '
                + "'mytenant', 'action': 'success', 'user': 'joe', 'change_type': 'disable entity', 'object': 'network:pan:traffic', 'object_category': 'splk-dsm', "
                + "'object_attrs': '<object attributes>', 'result': '<operation results>'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "action": "The value for action, expected values: success | failure",
                        "user": "The username associated with this audit event",
                        "change_type": "The type of change",
                        "object": "The object name related to the audit event",
                        "object_category": "The category related to the object",
                        "object_attrs": "The object attributes, such as the full record as registered when the change is performed",
                        "result": "The result of the change operation, either a human readable message or the server technical answer",
                        "comment": "A comment provided by the user or the automation system for that change",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # TrackMe reqinfo
        reqinfo_trackme = trackme_reqinfo(
            request_info.system_authtoken, request_info.server_rest_uri
        )
        trackmeconf = reqinfo_trackme["trackme_conf"]["index_settings"]

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # Register the new audit event using libs
        """
        def trackme_audit_event(session_key, tenant_id, user, action, change_type, object_name, object_category, object_attrs, result, comment):
        """
        try:
            # Define Meta
            splunk_sourcetype = "trackme:audit"
            splunk_source = "trackme:audit"
            splunk_host = request_info.server_servername

            audit_record = json.dumps(
                {
                    "_time": str(time.time()),
                    "timeStr": str(datetime.datetime.now()),
                    "tenant_id": tenant_id,
                    "user": user,
                    "action": action,
                    "change_type": change_type,
                    "object": object_name,
                    "object_category": object_category,
                    "object_attrs": object_attrs,
                    "result": result,
                    "comment": comment,
                }
            )

            if tenant_id != "all":
                # get the target index
                tenant_indexes = trackme_idx_for_tenant(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                )

                # index the audit record in the tenant audit index
                target = service.indexes[tenant_indexes["trackme_audit_idx"]]

            else:
                # index the audit record in the global audit index
                target = service.indexes[trackmeconf["trackme_audit_idx"]]

            # submit the audit record
            target.submit(
                event=str(audit_record),
                source=str(splunk_source),
                sourcetype=str(splunk_sourcetype),
                host=str(splunk_host),
            )

            return {
                "payload": {
                    "action": "success",
                    "result": "The audit record was registered successfully",
                },
                "status": 200,
            }

        except Exception as e:
            response = {
                "action": "failure",
                "response": "audit event has failed",
                "exception": str(e),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Register a flip event
    def post_flip_event(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_name = resp_dict["object"]
                object_category = resp_dict["object_category"]
                object_state = resp_dict["object_state"]
                object_previous_state = resp_dict["object_previous_state"]
                latest_flip_time = resp_dict["latest_flip_time"]
                latest_flip_state = resp_dict["latest_flip_state"]
                result = resp_dict["result"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint allows registering a TrackMe flipping event:",
                "resource_desc": "Register a flipping event",
                "resource_spl_example": '| trackme url="/services/trackme/v2/vtenants/flip_event" mode="post", body="{\'tenant_id\': '
                + "'mytenant', 'object': 'network:pan:traffic', 'object_category': 'splk-dsm', "
                + "'object_state': 'red', 'object_previous_state': 'green', 'latest_flip_time': '1689613080', 'latest_flip_state': 'red', 'result': 'my message'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "object": "The object name related to the audit event",
                        "object_category": "The category related to the object",
                        "object_state": "The object state",
                        "object_previous_state": "The object previous state",
                        "latest_flip_time": "The latest flip time",
                        "latest_flip_state": "The latest flip state",
                        "result": "The result of the change operation, either a human readable message or the server technical answer",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # TrackMe reqinfo
        reqinfo_trackme = trackme_reqinfo(
            request_info.system_authtoken, request_info.server_rest_uri
        )
        trackmeconf = reqinfo_trackme["trackme_conf"]["index_settings"]

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # Register the new audit event using libs
        try:
            # Define Meta
            splunk_sourcetype = "trackme:flip"
            splunk_source = "flip_state_change_tracking"
            splunk_host = request_info.server_servername

            audit_record = json.dumps(
                {
                    "_time": str(time.time()),
                    "timeStr": str(datetime.datetime.now()),
                    "tenant_id": tenant_id,
                    "object": object_name,
                    "object_category": object_category,
                    "object_state": object_state,
                    "object_previous_state": object_previous_state,
                    "latest_flip_time": latest_flip_time,
                    "latest_flip_state": latest_flip_state,
                    "result": result,
                }
            )

            # get the target index
            tenant_indexes = trackme_idx_for_tenant(
                request_info.system_authtoken, request_info.server_rest_uri, tenant_id
            )

            # index the audit record
            target = service.indexes[tenant_indexes["trackme_summary_idx"]]
            target.submit(
                event=str(audit_record),
                source=str(splunk_source),
                sourcetype=str(splunk_sourcetype),
                host=str(splunk_host),
            )

            return {
                "payload": {
                    "action": "success",
                    "result": "The flipping record was registered successfully",
                },
                "status": 200,
            }

        except Exception as e:
            response = {
                "action": "failure",
                "response": "flipping event has failed",
                "exception": str(e),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Register a state event
    def post_state_event(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict.get("tenant_id", None)
                if not tenant_id:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "tenant_id is required",
                        },
                        "status": 500,
                    }
                index = resp_dict.get("index", "trackme_summary")
                sourcetype = resp_dict.get("sourcetype", "trackme:state")
                source = resp_dict.get("source", "trackme:api")
                record = resp_dict.get("record", None)
                if not record:
                    return {
                        "payload": {
                            "action": "failure",
                            "response": "record is required",
                        },
                        "status": 500,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint allows registering a TrackMe state event:",
                "resource_desc": "Register a state event",
                "resource_spl_example": '| trackme url="/services/trackme/v2/vtenants/state_event" mode="post", body="{\'tenant_id\': '
                + "'mytenant', 'source': 'mysource', 'record': 'myrecord'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "index": "Recommended, the index target, if not specified, defaults to trackme_summary",
                        "sourcetype": "Optional, The value for the sourcetype Splunk Metadata, defaults to trackme:state if not specified",
                        "source": "Optional, The value for the source Splunk Metadata, defaults to trackme:api if not specified",
                        "record": "The record to be indexed, can be submitted as a string or a JSON object",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # if the record is a JSON object, pass it through json.dumps, otherwise just use the string
        if isinstance(record, dict):
            record = json.dumps(record)

        # index the audit record
        target = service.indexes[index]
        try:
            target.submit(
                event=record,
                source=source,
                sourcetype=sourcetype,
                host=request_info.server_servername,
            )

            return {
                "payload": {
                    "action": "success",
                    "result": "The state record was registered successfully",
                },
                "status": 200,
            }
        except Exception as e:
            response = {
                "action": "failure",
                "response": "state event has failed",
                "exception": str(e),
            }
            logging.error(json.dumps(response, indent=2))
            return {
                "payload": response,
                "status": 500,
            }
