from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_cim_tracking.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from collections import OrderedDict
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append libs
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel

# import Splunk lbs
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerSplkCimTracking_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkCimTracking_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_cim(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_cim",
            "resource_group_desc": "Endpoints specific to the splk-cim TrackMe component (Splunk Common Information Model compliance monitoring, read only operations)",
        }

        return {"payload": response, "status": 200}

    # get the component table
    def post_get_cim_table(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        key_id = None
        object = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                key_id = resp_dict["key_id"]
                object = resp_dict["object"]
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint retrievs the component table, it requires a POST call with the following options:",
                "resource_desc": "Get the entity table",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_cim/get_cim_table\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'key_id': '*'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "key_id": "key id, use a * to match all entities",
                        "object": "cim entity name, use a * to match all entities",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = (
            '| `get_splk_cim_table("'
            + str(tenant_id)
            + '", "'
            + str(key_id)
            + '*", "'
            + str(object)
            + '")` | `trackme_cim_emoji`'
        )

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.export(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            return {"payload": query_results, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # get all records
    def post_cim_tracker_show(self, request_info, **kwargs):
        """
        | trackme mode=post url="/services/trackme/v2/splk_cim/cim_tracker_show" body="{'tenant_id': 'mytenant'}"
        """

        # Declare
        tenant_id = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint retrieves all records for the CIM tracker collection, it requires a POST call with the following information:",
                "resource_desc": "Get the cim entities KVstore collection records",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_cim/cim_tracker_show\" body=\"{'tenant_id': 'mytenant'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = "kv_trackme_cim_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            return {"payload": collection.data.query(), "status": 200}

        except Exception as e:
            logging.error('Warn: exception encountered="{}"'.format(str(e)))
            return {"payload": 'Warn: exception encountered="{}"'.format(str(e))}

    # Returns the CIM entity rules
    def post_cim_tracker_rules(self, request_info, **kwargs):
        # By key
        describe = False
        tenant_id = None
        object = None
        context = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object = resp_dict["object"]
                # replace chars as would a creation of a new entity do
                object = object.lower().replace(" ", "_").replace("-", "_")

                # context: the context is used for simulation purposes versus live context
                # This is an optional argument, and defaults to live
                # swiching to simulation creates the records in a secondary KVstore that is not used for true alerting purpose
                # As well, the scheduled report will not be created in simulation context
                try:
                    context = resp_dict["context"]
                except Exception as e:
                    context = "live"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint retrieves tracking rules for a given object:",
                "resource_desc": "Get cim tracking rules for a given entity",
                "resource_spl_example": "| trackme mode=get url=\"/services/trackme/v2/splk_cim/cim_tracker_rules\" body=\"{'tenant_id': 'mytenant', 'object': 'auth001', 'context': 'live'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object": "name of the cim entity, the name of the tracker will be derived from it",
                        "context": "OPTIONAL: context is used for simulation purposes, defaults to live (valid option: live | simulation)",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        if context == "live":
            collection_name = "kv_trackme_cim_tenant_" + str(tenant_id)
        elif context == "simulation":
            collection_name = "kv_trackme_cim_simulation_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # Define the KV query
        query_string = {
            "object": object,
        }

        # Get the record
        try:
            kvrecord = collection.data.query(query=json.dumps(query_string))[0]
        except Exception as e:
            kvrecord = None

        # Render result
        if kvrecord:
            logging.debug(
                'tenant_id="{}", object="{}", success for record="{}"'.format(
                    tenant_id, object, kvrecord
                )
            )
            return {"payload": kvrecord.get("cim_tracking_rules"), "status": 200}

        # Render empty result
        else:
            logging.info(
                'tenant_id="{}", resource does not exist, object="{}"'.format(
                    tenant_id, object
                )
            )
            return {"payload": {}, "status": 200}
