from __future__ import absolute_import, division, print_function, unicode_literals
from concurrent.futures import process

__name__ = "trackme_rest_handler_allowlist.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
import re
import time

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event, dataset_update_cache
import splunklib.client as client


class TrackMeHandlerSplkDataSet_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkDataSet_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_dataset(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_dataset/write",
            "resource_group_desc": "These endpoints provide capabilities to manage the data set scope of tenants running the splk-dsm/dhm/mhm, to restrict and configure the scope of data available to the tenants (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add new allowlist index if does not exist yet
    def post_dataset_add(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_dataset/write/dataset_add\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'object_category': 'index', 'object': 'main'}\"
        """

        # Declare
        tenant_id = None
        component = None
        object_category = None
        object_value = None
        # note: action is always block
        action = "block"
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                component = resp_dict["component"]
                if component not in ("dsm", "dhm", "mhm"):
                    return {
                        "payload": 'Invalid component="{}"'.format(component),
                        "status": 500,
                    }
                object_category = resp_dict["object_category"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint adds a new allow list record for data sources, it requires a POST call with the following information:",
                "resource_desc": "Add a new object to the dataset for that component",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_dataset/write/dataset_add\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'object_category': 'index', 'object': 'main'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "component": "The component, valid options are: dsm | dhm | mhm",
                        "object_category": "The object category, valid options are: index | source | sourcetype | object",
                        "object": "The object name, example the name of the index or an index pattern with wildcards",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "$and": [
                {
                    "object_category": object_category,
                    "object": object_value,
                    "action": action,
                }
            ]
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = (
                "kv_trackme_" + str(component) + "_allowlist_tenant_" + str(tenant_id)
            )
            collection = service.kvstore[collection_name]

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                kvrecord = collection.data.query(query=json.dumps(query_string))[0]
                key = kvrecord.get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already
                logging.error(
                    'conflict, this object exists already, record="{}"'.format(
                        json.dumps(kvrecord, indent=0)
                    )
                )

                return {
                    "payload": {
                        "action": "failure",
                        "response": "conflict, this object exists already",
                        "record": kvrecord,
                    },
                    "status": 500,
                }

            else:
                # This record does not exist yet
                try:
                    # Define if the pattern is a regex and store this information in the collection
                    r = re.match("[\\\\|\\?|\\$|\\^|\\[|\\]|\\{|\\}|\\+]", object_value)
                    r2 = re.findall("\\.[\\*|\\+]", object_value)

                    if r or r2:
                        is_rex = "true"
                    else:
                        is_rex = "false"

                    # Insert the record
                    collection.data.insert(
                        json.dumps(
                            {
                                "object_category": object_category,
                                "object": object_value,
                                "action": action,
                                "is_rex": is_rex,
                            }
                        )
                    )

                    # Get record
                    kvrecord = collection.data.query(query=json.dumps(query_string))[0]

                    # Audit
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "add allowlist index",
                        str(object_value),
                        str(component),
                        json.dumps(kvrecord, indent=0),
                        "The object was successfulled added to allow list",
                        str(update_comment),
                    )

                    # run update cache
                    try:
                        cache_update = dataset_update_cache(
                            request_info.session_key,
                            request_info.server_rest_uri,
                            tenant_id,
                            component,
                        )
                        logging.debug("update cache successful")
                    except Exception as e:
                        logging.error(
                            'failed to run the update datset cache with exception="{}"'.format(
                                str(e)
                            )
                        )

                    # Log and return response
                    logging.info(
                        'action="success", record="{}"'.format(
                            json.dumps(kvrecord, indent=0)
                        )
                    )
                    return {
                        "payload": {
                            "action": "success",
                            "response": "the new record was successfully added",
                            "record": kvrecord,
                        },
                        "status": 200,
                    }

                except Exception as e:
                    response = {
                        "action": "failure",
                        "response": 'an exception was encountered, exception="{}"'.format(
                            str(e)
                        ),
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Delete records from the collection
    def post_dataset_del(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_dataset/write/dataset_del\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'keys_list': '<redacted_list_of_csv_keyids>'}\"
        """

        # Declare
        tenant_id = None
        component = None
        keys_list = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                component = resp_dict["component"]
                if component not in ("dsm", "dhm", "mhm"):
                    return {
                        "payload": 'Invalid component="{}"'.format(component),
                        "status": 500,
                    }
                keys_list = resp_dict["keys_list"]
                # Handle as a CSV list of keys
                keys_list = keys_list.split(",")

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint deletes an allow list record for data sources, it requires a POST call with the following information:",
                "resource_desc": "Add a new object to the dataset for that component",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_dataset/write/dataset_del\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'keys_list': '<redacted_list_of_csv_keyids>'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "component": "The component, valid options are: dsm | dhm | mhm",
                        "keys_list": "List of record keys seperated by a comma of the records to be deleted from the collection",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = (
            "kv_trackme_" + str(component) + "_allowlist_tenant_" + str(tenant_id)
        )
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # loop
        for item in keys_list:
            # Define the KV query
            query_string = {
                "_key": item,
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                kvrecord = collection.data.query(query=json.dumps(query_string))[0]
                key = kvrecord.get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # Remove and audit
                try:
                    # Remove the record
                    collection.data.delete(json.dumps({"_key": key}))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "delete allowlist index",
                        str(item),
                        str(component),
                        str(kvrecord),
                        "The index was deleted from allow list successfully",
                        str(update_comment),
                    )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": kvrecord,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete allowlist index",
                        str(item),
                        str(component),
                        str(kvrecord),
                        str(e),
                        str(update_comment),
                    )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": kvrecord,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "failure",
                    "delete allowlist index",
                    str(item),
                    str(component),
                    str(kvrecord),
                    "HTTP 404 NOT FOUND",
                    str(update_comment),
                )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # run update cache
        try:
            cache_update = dataset_update_cache(
                request_info.session_key,
                request_info.server_rest_uri,
                tenant_id,
                component,
            )
            logging.debug("update cache successful")
        except Exception as e:
            logging.error(
                'failued to run the update datset cache with exception="{}"'.format(
                    str(e)
                )
            )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update records
    def post_dataset_update(self, request_info, **kwargs):
        """
        | trackme mode=post url=/services/trackme/v2/splk_dataset/write/dataset_update body="
        {'tenant_id': 'mytenant', 'records_list': '[{\'object_category\':\'index\',\'object\':\'test2\',\'action\':\'block\',\'_key\':\'61fe64224aa485576f72b0a0\'},
        {\'object_category\':\'index\',\'object\':\'test3\',\'action\':\'block\',\'_key\':\'61fe64224aa485576f72b0a0\'}]'}"
        """

        # Declare
        tenant_id = None
        component = None
        records_list = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                component = resp_dict["component"]
                if component not in ("dsm", "dhm", "mhm"):
                    return {
                        "payload": 'Invalid component="{}"'.format(component),
                        "status": 500,
                    }
                records_list = resp_dict["records_list"]
                # Handle as a CSV list of keys
                records_list = json.loads(records_list)

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint updates records, it requires a POST call with the following information:",
                "resource_desc": "Update a list of objects for that component",
                "resource_spl_example": '| trackme mode=post url=/services/trackme/v2/splk_dataset/write/dataset_update body="'
                + "{'tenant_id': 'mytenant', 'records_list': '[{'object_category':'index','object':'test2','action':'block','_key':'61fe64224aa485576f72b0a0'},"
                + "{'object_category':'index','object':'test3','action':'block','_key':'61fe64224aa485576f72b0a0'}]'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "records_list": "JSON records to be updated",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = (
            "kv_trackme_" + str(component) + "_allowlist_tenant_" + str(tenant_id)
        )
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # debug
        logging.info('records_list="{}"'.format(json.dumps(records_list, indent=0)))

        # loop
        for item in records_list:
            # debug
            logging.info('item="{}"'.format(item))

            # Define the KV query
            query_string = {
                "_key": item.get("_key"),
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                kvrecord = collection.data.query(query=json.dumps(query_string))[0]
                key = kvrecord.get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # Update and audit
                try:
                    # Define if the pattern is a regex and store this information in the collection
                    r = re.match(
                        "[\\\\|\\?|\\$|\\^|\\[|\\]|\\{|\\}|\\+]", item.get("object")
                    )
                    r2 = re.findall("\\.[\\*|\\+]", item.get("object"))

                    if r or r2:
                        is_rex = "true"
                    else:
                        is_rex = "false"

                    # Update the record
                    collection.data.update(
                        str(key),
                        json.dumps(
                            {
                                "object_category": item.get("object_category"),
                                "object": item.get("object"),
                                "action": item.get("action"),
                                "is_rex": is_rex,
                            }
                        ),
                    )

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "update allowlist index",
                        str(item),
                        str(component),
                        json.dumps(item, indent=1),
                        "The index was deleted from allow list successfully",
                        str(update_comment),
                    )

                    result = {
                        "action": "update",
                        "result": "success",
                        "record": kvrecord,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete allowlist index",
                        str(item),
                        str(component),
                        str(kvrecord),
                        str(e),
                        str(update_comment),
                    )

                    result = {
                        "action": "update",
                        "result": "failure",
                        "record": kvrecord,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "failure",
                    "delete allowlist index",
                    str(item),
                    str(component),
                    str(kvrecord),
                    "HTTP 404 NOT FOUND",
                    str(update_comment),
                )

                result = {
                    "action": "update",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # run update cache
        try:
            cache_update = dataset_update_cache(
                request_info.session_key,
                request_info.server_rest_uri,
                tenant_id,
                component,
            )
            logging.debug("update cache successful")
        except Exception as e:
            logging.error(
                'failued to run the update datset cache with exception="{}"'.format(
                    str(e)
                )
            )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}
