from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_splk_dhm.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
import time

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel

# TrackMe splk-feeds libs
from trackme_libs_splk_feeds import (
    splk_dhm_return_entity_info,
    splk_dhm_return_searches,
)

# Splunk libs
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerSplkDhm_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkDhm_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_dhm(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_dhm",
            "resource_group_desc": "Endpoints specific to the splk-dhm TrackMe component (Splunk Data Hosts monitoring, read only operations)",
        }

        return {"payload": response, "status": 200}

    # Get entity info
    def post_dh_entity_info(self, request_info, **kwargs):
        """
        | trackme url="/services/trackme/v2/splk_dhm/dh_entity_info" mode="post" body="{'tenant_id': 'mytenant', 'object': 'remote|account:lab|firewall.pan.amer.design.node1'}"
        """

        # init
        object_value = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint returns various information related to the entity, it requires a POST call with the following information:",
                "resource_desc": "Return the entity main information, these normalized information are used to build dynamically the various queries through the user interfaces",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dhm/dh_entity_info\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'object': 'remote|account:lab|firewall.pan.amer.design.node1'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object": "The entity identifier",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dhm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # Define the KV query
        query_string = {"object": object_value}

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            kvrecord = collection.data.query(query=json.dumps(query_string))[0]
            key = kvrecord.get("_key")

        except Exception as e:
            key = None

            return {
                "payload": {
                    "action": "failure",
                    "response": 'the entity with identifier="{}" was not found'.format(
                        object_value
                    ),
                },
                "status": 404,
            }

        # proceed
        if key:
            # init
            response = {}

            try:
                # Get entity info
                entity_info = splk_dhm_return_entity_info(kvrecord)

                # Add
                response["entity_info"] = entity_info

                # log debug
                logging.debug(
                    'function splk_dhm_return_entity_info, entity_info="{}"'.format(
                        json.dumps(entity_info, indent=2)
                    )
                )

                # Add to dict
                entity_info["index"] = kvrecord.get("data_index")
                entity_info["sourcetype"] = kvrecord.get("data_sourcetype")
                # set the search constraint
                if entity_info["breakby_key"] != "none":
                    entity_info["search_constraint"] = (
                        "index IN ("
                        + kvrecord.get("data_index")
                        + ") sourcetype IN ("
                        + kvrecord.get("data_sourcetype")
                        + ") "
                        + entity_info["breakby_key"]
                        + "="
                        + '"'
                        + entity_info["breakby_value"]
                        + '"'
                    )
                else:
                    entity_info["search_constraint"] = (
                        "index IN ("
                        + kvrecord.get("data_index")
                        + ") sourcetype IN ("
                        + kvrecord.get("data_sourcetype")
                        + ") "
                        + 'host="'
                        + str(object_value)
                        + '"'
                    )

                # get entity searches
                entity_searches = splk_dhm_return_searches(
                    tenant_id, object_value, entity_info
                )

                # log debug
                logging.debug(
                    'function splk_dhm_return_searches, entity_searches="{}"'.format(
                        json.dumps(entity_searches, indent=2)
                    )
                )

                # add
                entity_info["splk_dhm_raw_search"] = entity_searches.get(
                    "splk_dhm_raw_search"
                )

                # single stats: manage both splunk query and trackme metric based query
                entity_info[
                    "splk_dhm_overview_splunk_single_stats"
                ] = entity_searches.get("splk_dhm_overview_root_search")
                entity_info["splk_dhm_overview_trackme_single_stats"] = (
                    "| mstats avg(trackme.splk.feeds.avg_latency_5m) as avg_latency, "
                    + "avg(trackme.splk.feeds.perc95_latency_5m) as perc95_latency "
                    + "where `trackme_metrics_idx("
                    + str(tenant_id)
                    + ')` tenant_id="'
                    + str(tenant_id)
                    + '" object_category="splk-dhm" object="'
                    + str(object_value)
                    + '"'
                    + "\n| appendcols [ | inputlookup trackme_dhm_tenant_"
                    + str(tenant_id)
                    + ' where object="'
                    + str(object_value)
                    + '" | eval event_delay=now()-data_last_time_seen | table event_delay ]'
                )

                # add both options for trackme and splunk
                entity_info[
                    "splk_dhm_overview_splunk_pie_root_search"
                ] = entity_searches.get("splk_dhm_overview_pie_root_search")
                entity_info[
                    "splk_dhm_overview_trackme_pie_root_search"
                ] = f'| mstats latest(trackme.splk_dhm.last_eventcount) as count where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="splk-dhm" object="{object_value}" by object, idx, st | rename idx as index, st as sourcetype'

                # timechart view
                entity_info["splk_dhm_overview_splunk_timechart"] = entity_searches.get(
                    "splk_dhm_overview_timechart"
                )
                entity_info["splk_dhm_overview_trackme_timechart"] = (
                    "| mstats avg(trackme.splk.feeds.lag_event_sec) as lag_event_sec, avg(trackme.splk.feeds.latest_eventcount_5m) as latest_eventcount_5m, "
                    + "avg(trackme.splk.feeds.avg_latency_5m) as avg_latency_5m where `trackme_metrics_idx("
                    + str(tenant_id)
                    + ")` "
                    + 'tenant_id="'
                    + str(tenant_id)
                    + '" object_category="splk-dhm" object="'
                    + str(object_value)
                    + '" by object span=5m'
                    + "| timechart span=5m avg(avg_latency_5m) as avg_latency_5m, avg(latest_eventcount_5m) as latest_eventcount_5m, avg(lag_event_sec) as lag_event_sec"
                )

                # add object and key
                entity_info["object"] = object_value
                entity_info["key"] = key

                # render response
                return {"payload": entity_info, "status": 200}

            except Exception as e:
                # render response
                msg = 'An exception was encountered while processing entity get info, exception="{}"'.format(
                    str(e)
                )
                logging.error(msg)
                return {
                    "payload": {
                        "action": "failure",
                        "response": msg,
                    },
                    "status": 500,
                }

    # get the component table
    def post_dh_get_table(self, request_info, **kwargs):
        """
        | trackme url="/services/trackme/v2/splk_dhm/dh_get_table" mode="post" body="{'tenant_id': 'mytenant', 'mode': 'compact', 'key_id': '*'}"
        """

        describe = False
        tenant_id = None
        mode = None
        key_id = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                mode = resp_dict["mode"]
                key_id = resp_dict["key_id"]
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint retrievs the component table, it requires a POST call with the following options:",
                "resource_desc": "Get the entity table",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dhm/dh_get_table\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'mode': 'minimal', 'key_id': '*'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "mode": "the summary mode, valid options are: minimal (default) / compact / full",
                        "key_id": "key id, use a * to match all entities",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = (
            "| `get_splk_dhm_table("
            + str(tenant_id)
            + ","
            + str(mode)
            + ","
            + str(key_id)
            + ', "*")`'
        )

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.export(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            return {"payload": query_results, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}
