from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_splk_dsm.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
import time

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import (
    trackme_getloglevel,
    trackme_audit_event,
    trackme_register_tenant_component_summary,
)

# import Splunk libs
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerSplkDsm_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkDsm_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_dsm(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_dsm/write",
            "resource_group_desc": "Endpoints specific to the splk-dsm TrackMe component (Splunk Data Sources monitoring, power operations)",
        }

        return {"payload": response, "status": 200}

    # Bulk edit (to be used from the inline Tabulator)
    def post_ds_bulk_edit(self, request_info, **kwargs):
        """
        This function performs a bulk edit on given json data.
        :param request_info: Contains request related information
        :param kwargs: Other keyword arguments
        :return: Status and payload of the bulk edit operation
        """

        # perf counter
        start_time = time.time()

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                json_data = resp_dict["json_data"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint performs a bulk edit, it requires a POST call with the following information:",
                "resource_desc": "Perform a bulk edit to one or more entities",
                "resource_spl_example": '| trackme url="/services/trackme/v2/splk_dsm/write/ds_bulk_edit" mode="post" body="{\'tenant_id\':\'mytenant\', '
                + "'json_data':[{'keyid':'b55658d1fc032ea3e1ecfc9eb60ad070','object':'netscreen:netscreen:firewall',"
                + "'alias':'netscreen:netscreen:firewall','priority':'high','monitored_state':'enabled',"
                + "'data_override_lagging_class':'false','data_lag_alert_kpis':'all_kpis','data_max_lag_allowed':'3600',"
                + "'data_max_delay_allowed':'3600'}]}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "json_data": "The JSON array object",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        failures_count = 0

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # get all records
        get_collection_start = time.time()
        collection_records = []
        collection_records_keys = set()

        end = False
        skip_tracker = 0
        while end == False:
            process_collection_records = collection.data.query(skip=skip_tracker)
            if len(process_collection_records) != 0:
                for item in process_collection_records:
                    if item.get("_key") not in collection_records_keys:
                        collection_records.append(item)
                        collection_records_keys.add(item.get("_key"))
                skip_tracker += 5000
            else:
                end = True

        # turn it into a dict for fast operations
        collection_dict = {record["object"]: record for record in collection_records}

        # len collection_records
        collection_records_len = len(collection_records)

        logging.info(
            f'context="perf", get collection records, no_records="{collection_records_len}", run_time="{round((time.time() - get_collection_start), 3)}", collection="{collection_name}"'
        )

        # final records
        entities_list = []
        final_records = []

        # error counters and exceptions
        failures_count = 0
        exceptions_list = []

        # loop and proceed
        for json_record in json_data:
            keys_to_check = [
                "data_max_lag_allowed",
                "data_max_delay_allowed",
                "data_lag_alert_kpis",
                "monitored_state",
                "alias",
                "data_override_lagging_class",
                "priority",
            ]

            object = json_record["object"]

            try:
                if object in collection_dict:
                    current_record = collection_dict[object]

                    is_different = False
                    for key in keys_to_check:
                        new_value = json_record.get(key)
                        if current_record.get(key) != new_value:
                            is_different = True
                            current_record[key] = new_value

                    if is_different:
                        current_record["mtime"] = time.time()

                        # Add for batch update
                        final_records.append(current_record)

                        # Add for reporting
                        entities_list.append(current_record.get("object"))
                else:
                    raise KeyError("Resource not found for object {}".format(object))

            except Exception as e:
                # increment counter
                failures_count += 1
                exceptions_list.append(
                    'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    )
                )

        # batch update/insert
        batch_update_collection_start = time.time()

        # process by chunk
        chunks = [final_records[i : i + 500] for i in range(0, len(final_records), 500)]
        for chunk in chunks:
            try:
                collection.data.batch_save(*chunk)
            except Exception as e:
                logging.error(f'KVstore batch failed with exception="{str(e)}"')
                failures_count += 1
                exceptions_list.append(str(e))

        # calculate len(final_records) once
        final_records_len = len(final_records)

        # perf counter for the batch operation
        logging.info(
            f'context="perf", batch KVstore update terminated, no_records="{final_records_len}", run_time="{round((time.time() - batch_update_collection_start), 3)}"'
        )

        # Record an audit change
        for record in final_records:
            status = "success" if failures_count == 0 else "failure"
            message = (
                "Entity was updated successfully"
                if failures_count == 0
                else "Entity bulk update has failed"
            )
            trackme_audit_event(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                request_info.user,
                status,
                "inline bulk edit",
                record.get("object"),
                "splk-dsm",
                json.dumps(record, indent=1)
                if failures_count == 0
                else exceptions_list,
                message,
                str(update_comment),
            )

        if failures_count == 0:
            # Call the trackme_register_tenant_component_summary
            trackme_register_tenant_component_summary(
                request_info.session_key,
                request_info.server_rest_port,
                tenant_id,
                "dsm",
            )

            req_summary = {
                "process_count": final_records_len,
                "failures_count": failures_count,
                "entities_list": entities_list,
            }
            logging.info(
                f'entity bulk edit was successfull, no_modified_records="{final_records_len}", no_records="{collection_records_len}", run_time="{round((time.time() - start_time), 3)}", collection="{collection_name}", results="{json.dumps(req_summary, indent=1)}"'
            )
            return {"payload": req_summary, "status": 200}

        else:
            req_summary = {
                "process_count": final_records_len,
                "failures_count": failures_count,
                "entities_list": entities_list,
                "exceptions": exceptions_list,
            }
            logging.error(
                f'entity bulk edit has failed, no_modified_records="{final_records_len}", no_records="{collection_records_len}", run_time="{round((time.time() - start_time), 3)}", collection="{collection_name}", results="{json.dumps(req_summary, indent=1)}"'
            )
            return {"payload": req_summary, "status": 500}

    # Enable/Disable monitoring by object name
    def post_ds_monitoring(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                action = resp_dict["action"]
                if not action in ("enable", "disable"):
                    return {
                        "payload": "Invalid option for action, valid options are: enable | disable",
                        "status": 500,
                    }
                else:
                    if action == "enable":
                        action_value = "enabled"
                    elif action == "disable":
                        action_value = "disabled"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint enables data monitoring for an existing data source by the data source name (object), it requires a POST call with the following information:",
                "resource_desc": "Enable/Disable monitoring for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_monitoring\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'action': 'disable', 'object_list': 'netscreen:netscreen:firewall,wineventlog:WinEventLog'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "action": "The action to be performed, valid options are: enable | disable",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]
                object = kvrecord.get("object")
                kvrecord["monitored_state"] = action_value
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "manage monitoring enablement",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "Monitoring status was updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # Call the trackme_register_tenant_component_summary
        trackme_register_tenant_component_summary(
            request_info.session_key, request_info.server_rest_port, tenant_id, "dsm"
        )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update priority by object name
    def post_ds_update_priority(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
                else:
                    describe = False
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                priority = resp_dict["priority"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint defines the priority for an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Update priority for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_priority\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'priority': 'high', 'object_list': 'netscreen:netscreen:firewall,wineventlog:WinEventLog'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "priority": "the value for priority, valid options are low / medium / high",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                object = kvrecord.get("object")
                kvrecord["priority"] = priority
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "modify priority",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "Priority was updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # Call the trackme_register_tenant_component_summary
        trackme_register_tenant_component_summary(
            request_info.session_key, request_info.server_rest_port, tenant_id, "dsm"
        )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update lagging policy by object name
    def post_ds_update_lag_policy(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                try:
                    data_lag_alert_kpis = resp_dict[
                        "data_lag_alert_kpis"
                    ]  # all_kpis / lag_ingestion_kpi / lag_event_kpi
                    if not data_lag_alert_kpis in (
                        "all_kpis",
                        "lag_ingestion_kpi",
                        "lag_event_kpi",
                    ):
                        data_lag_alert_kpis = None
                except Exception as e:
                    data_lag_alert_kpis = None

                try:
                    data_max_lag_allowed = int(
                        resp_dict["data_max_lag_allowed"]
                    )  # seconds
                except Exception as e:
                    data_max_lag_allowed = None

                try:
                    data_max_delay_allowed = int(
                        resp_dict["data_max_delay_allowed"]
                    )  # seconds
                except Exception as e:
                    data_max_delay_allowed = None

                try:
                    data_override_lagging_class = resp_dict[
                        "data_override_lagging_class"
                    ]  # true / false
                    if not data_override_lagging_class in ("true", "false"):
                        data_override_lagging_class = None
                except Exception as e:
                    data_override_lagging_class = None

                try:
                    allow_adaptive_delay = resp_dict[
                        "allow_adaptive_delay"
                    ]  # true / false
                    if not allow_adaptive_delay in ("true", "false"):
                        allow_adaptive_delay = None
                except Exception as e:
                    allow_adaptive_delay = None

                try:
                    future_tolerance = resp_dict["future_tolerance"]
                except Exception as e:
                    future_tolerance = "system"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint configures the lagging policy for an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Update lag policies for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_lag_policy\" mode=\"post\" body=\"{'tenant_id':'mytenant','object_list':'netscreen:netscreen:firewall','data_max_lag_allowed':'900','data_max_delay_allowed':'1600','data_lag_alert_kpis':'all_kpis','data_override_lagging_class':'false'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "data_lag_alert_kpis": "OPTIONAL, KPIs policy to be applied, valid options are all_kpis / lag_ingestion_kpi / lag_event_kpi",
                        "data_max_lag_allowed": "OPTIONAL, maximal accepted lagging value in seconds, must be an integer",
                        "data_max_delay_allowed": "OPTIONAL, maximal accepted delay value in seconds, must be an integer",
                        "data_override_lagging_class": "OPTIONAL, overrides lagging classes, valid options are true / false",
                        "future_tolerance": "OPTIONAL, the negative value for future tolerance, specify system to rely on the system level setting, disabled to allow data in the future up to 7 days without affecting the status of the entity",
                        "allow_adaptive_delay": "OPTIONAL, allow adaptive delay, valid options are true / false",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]
            object = kvrecord.get("object")
            if data_max_lag_allowed:
                kvrecord["data_max_lag_allowed"] = data_max_lag_allowed
            if data_max_delay_allowed:
                kvrecord["data_max_delay_allowed"] = data_max_delay_allowed
            if data_lag_alert_kpis:
                kvrecord["data_lag_alert_kpis"] = data_lag_alert_kpis
            if data_override_lagging_class:
                kvrecord["data_override_lagging_class"] = data_override_lagging_class
            if allow_adaptive_delay:
                kvrecord["allow_adaptive_delay"] = allow_adaptive_delay
            kvrecord["future_tolerance"] = future_tolerance
            kvrecord["mtime"] = time.time()

            try:
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "modify monitoring lag policy",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "Lag policy was updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                logging.error(str(e))

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update min dcount host by object name
    def post_ds_update_min_dcount_host(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                min_dcount_host = resp_dict["min_dcount_host"]
                # We need to accept the string any (to disable the fearure) or an integer
                if str(min_dcount_host) not in ("any"):
                    # anything else than "any" or an integer will fail here
                    min_dcount_host = int(min_dcount_host)

                # min_dcount_field is optional, if not provided we will use latest_dcount_host_5m
                try:
                    min_dcount_field = resp_dict["min_dcount_field"]
                    if not min_dcount_field in (
                        "avg_dcount_host_5m",
                        "latest_dcount_host_5m",
                        "perc95_dcount_host_5m",
                        "stdev_dcount_host_5m",
                    ):
                        min_dcount_field = "latest_dcount_host_5m"
                except Exception as e:
                    min_dcount_field = "latest_dcount_host_5m"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint configures the minimal number of distinct hosts count for an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Update lag policies for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_min_dcount_host\" mode=\"post\" body=\"{'tenant_id': 'mytenant', 'min_dcount_host': '0', 'min_dcount_field': 'stdev_dcount_host_5m', 'object_list': 'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "min_dcount_host": "minimal accepted number of distinct count hosts, must be an integer or the string any",
                        "min_dcount_field": "The dictinct count metric to be used for this entity, valid options are: avg_dcount_host_5m, latest_dcount_host_5m, perc95_dcount_host_5m, stdev_dcount_host_5m",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        # Loop
        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                object = kvrecord.get("object")
                kvrecord["min_dcount_host"] = min_dcount_host
                kvrecord["min_dcount_field"] = min_dcount_field
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "update min disctinct host count",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "The min disctinct count number of hosts was updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                logging.error(str(e))

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update monitoring week days by object name
    def post_ds_update_wdays(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                # Week days monitoring can be:
                # manual:all_days / manual:monday-to-friday / manual:monday-to-saturday / [ 0, 1, 2, 3, 4, 5, 6 ] where Sunday is 0
                data_monitoring_wdays = resp_dict["data_monitoring_wdays"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint configures the week days monitoring rule for an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Update week days monitoring for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_wdays\" mode=\"post\" body=\"{'tenant_id':'mytenant','object_list':'netscreen:netscreen:firewall','data_monitoring_wdays':'manual:1,2,3,4,5'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "data_monitoring_wdays": "the week days rule, valid options are manual:all_days / manual:monday-to-friday / manual:monday-to-saturday / [ 0, 1, 2, 3, 4, 5, 6 ] where Sunday is 0",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                object = kvrecord.get("object")
                kvrecord["data_monitoring_wdays"] = data_monitoring_wdays
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "modify week days monitoring",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "Week days monitoring was updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                logging.error(str(e))

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update monitoring hours ranges by object name
    def post_ds_update_hours_ranges(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                data_monitoring_hours_ranges = resp_dict["data_monitoring_hours_ranges"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint configures the week days monitoring rule for an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Update hours of monitoring for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_hours_ranges\" mode=\"post\" body=\"{'tenant_id':'mytenant', 'object_list':'netscreen:netscreen:firewall', 'data_monitoring_hours_ranges':'manual:8,9,10,11,12,13,14,15,16,17'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "data_monitoring_hours_ranges": "the hours ranges rule, valid options are manual:all_ranges / manual:08h-to-20h / [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 ] where 00h00 to 01h59 is 0",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                object = kvrecord.get("object")
                kvrecord["data_monitoring_hours_ranges"] = data_monitoring_hours_ranges
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "modify hours ranges monitoring",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(collection.data.query_by_id(key), indent=1)),
                    "Monitoring hours ranges were updated successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                logging.error(str(e))

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Remove object entities
    def post_ds_delete(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                deletion_type = resp_dict["deletion_type"]
                if deletion_type not in ("temporary", "permanent"):
                    return {
                        "payload": {
                            "error": "Invalid option for deletion_type, valid options are: tempoary | permanent"
                        },
                        "status": 500,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint performs a temporary deletion of an existing data source, it requires a POST call with the following information:",
                "resource_desc": "Delete one or more entities, either temporarily or permanently",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_delete\" mode=\"post\" body=\"{'tenant_id':'mytenant', 'deletion_type': 'temporary', 'object_list':'netscreen:netscreen:firewall,wineventlog:WinEventLog'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "deletion_type": "The type of deletion, valid options are: temporary | permanent",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # Permanently deleted objects
        collection_perm_deleted_name = (
            "kv_trackme_common_permanently_deleted_objects_tenant_" + str(tenant_id)
        )
        collection_perm_deleted = service.kvstore[collection_perm_deleted_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        for key in keys_list:
            try:
                # Remove the record
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]
                object = kvrecord.get("object")
                collection.data.delete(json.dumps({"_key": key}))

                if deletion_type == "permanent":
                    # Register a new permanently deleted object
                    collection_perm_deleted.data.insert(
                        json.dumps(
                            {
                                "ctime": str(time.time()),
                                "object": str(object),
                                "object_category": "splk-dsm",
                            }
                        )
                    )

                # Record an audit change
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "delete " + str(deletion_type),
                    str(object),
                    "splk-dsm",
                    str(json.dumps(kvrecord, indent=1)),
                    "Entity was temporay deleted successfully",
                    str(update_comment),
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "delete",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully deleted'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "delete",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to remove the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # Call the trackme_register_tenant_component_summary
        trackme_register_tenant_component_summary(
            request_info.session_key, request_info.server_rest_port, tenant_id, "dsm"
        )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Enable/Disable Data Sampling by object name
    def post_ds_manage_data_sampling(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_list = resp_dict["object_list"]
                # Handle as a CSV list of keys
                if not isinstance(object_list, list):
                    object_list = object_list.split(",")
                action = resp_dict["action"]
                if action not in ("enable", "disable", "reset", "run"):
                    return {
                        "payload": 'invalid value for action="{}", valid options are: enable | disable | reset| run',
                        "status": 200,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint allows management of the data sampling feature for an existing data source by the data source name (object), it requires a POST call with the following information:",
                "resource_desc": "Enable/Disable/Reset/Run Data Sampling for one or more entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_manage_data_sampling\" mode=\"post\" body=\"{'tenant_id':'mytenant', 'action': 'reset', 'object_list': 'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "Comma separated list of entities",
                        "action": "the action to be performed, valid options are: enable | disable | reset | run",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Set the value for data_sample_feature
        if action == "enable":
            data_sample_feature = "enabled"
        elif action == "disable":
            data_sample_feature = "disabled"
        elif action == "reset":
            data_sample_feature = "enabled"
        elif action == "run":
            data_sample_feature = "enabled"

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collections

        # component
        collection_name = "kv_trackme_dsm_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # sampling
        collection_sampling_name = "kv_trackme_dsm_data_sampling_tenant_" + str(
            tenant_id
        )
        collection_sampling = service.kvstore[collection_sampling_name]

        for object in object_list:
            try:
                # check if we have a KVrecord already for this object
                query_string = {
                    "$and": [
                        {
                            "object": object,
                        }
                    ]
                }

                # record from the component
                try:
                    # try get to get the key
                    kvrecord = collection.data.query(query=(json.dumps(query_string)))[
                        0
                    ]
                    key = kvrecord.get("_key")
                except Exception as e:
                    key = None

                # record from sampling
                try:
                    # try get to get the key
                    kvrecord_sampling = collection_sampling.data.query(
                        query=(json.dumps(query_string))
                    )[0]
                    key_sampling = kvrecord_sampling.get("_key")
                except Exception as e:
                    key_sampling = None

                # Render result
                if key:
                    # if action is enable / disable
                    if action in ("enable", "disable"):
                        try:
                            # Update the record
                            kvrecord_sampling[
                                "data_sample_feature"
                            ] = data_sample_feature
                            collection_sampling.data.update(
                                str(key), json.dumps(kvrecord_sampling)
                            )

                            # Record an audit change
                            record = {
                                "object": str(object),
                                "data_sample_feature": str(data_sample_feature),
                            }
                            trackme_audit_event(
                                request_info.system_authtoken,
                                request_info.server_rest_uri,
                                tenant_id,
                                request_info.user,
                                "success",
                                f"{action} data sampling",
                                str(object),
                                "splk-dsm",
                                str(json.dumps(record, indent=1)),
                                "Data sampling was managed successfully",
                                str(update_comment),
                            )

                            logging.info(
                                'tenant_id="{}", The object was successfully updated'.format(
                                    tenant_id, object
                                )
                            )

                            # increment counter
                            processed_count += 1
                            succcess_count += 1
                            failures_count += 0

                            # append for summary
                            result = {
                                "object": object,
                                "action": action,
                                "result": "success",
                                "message": 'tenant_id="{}", The object was successfully updated'.format(
                                    tenant_id, object
                                ),
                            }
                            records.append(result)

                        except Exception as e:
                            logging.error(
                                'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                                    tenant_id, object, str(e)
                                )
                            )

                            # increment counter
                            processed_count += 1
                            succcess_count += 0
                            failures_count += 1

                            result = {
                                "object": object,
                                "action": action,
                                "result": "failure",
                                "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                                    tenant_id, object, str(e)
                                ),
                            }
                            records.append(result)

                    # if action is reset
                    elif action in ("reset", "run"):
                        try:
                            # Only if reset
                            if action == "reset":
                                # don't fail if reset fails
                                try:
                                    # reset the sampling record
                                    current_data_sampling_nr = kvrecord.get("kvrecord")
                                    kvrecord_sampling = {
                                        "anomaly_reason": "pending",
                                        "current_detected_format": ["pending"],
                                        "data_sample_anomaly_detected": 0,
                                        "data_sample_feature": "pending",
                                        "data_sample_mtime": "pending",
                                        "data_sample_status_colour": "yellow",
                                        "data_sample_status_message": json.dumps(
                                            {
                                                "state": "pending",
                                                "desc": "Data Sampling is pending and was reset for this entity",
                                            }
                                        ),
                                        "data_sampling_nr": current_data_sampling_nr,
                                        "direction": "none",
                                        "data_sample_feature": "enabled",
                                        "mtime": "pending",
                                        "object": object,
                                        "previous_detected_format": ["pending"],
                                    }
                                    collection_sampling.data.update(
                                        str(key_sampling), json.dumps(kvrecord_sampling)
                                    )
                                    logging.info(
                                        'tenant_id="{}", object="{}", action="{}", KVstore sampling record was successfully reset'.format(
                                            tenant_id, object, action
                                        )
                                    )

                                    # audit
                                    trackme_audit_event(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        request_info.user,
                                        "success",
                                        "reset data sampling",
                                        str(object),
                                        "splk-dsm",
                                        str(json.dumps(kvrecord_sampling, indent=1)),
                                        "Data sampling was reset successfully",
                                        str(update_comment),
                                    )

                                except Exception as e:
                                    logging.error(
                                        'tenant_id="{}", object="{}", failed to update the KVstore sampling record with exception="{}"'.format(
                                            tenant_id, object, str(e)
                                        )
                                    )

                                try:
                                    # reset the record
                                    kvrecord["isAnomaly"] = 0
                                    kvrecord["data_sample_lastrun"] = 0

                                    collection.data.update(
                                        str(key), json.dumps(kvrecord)
                                    )
                                    logging.info(
                                        'tenant_id="{}", object="{}", action="{}", KVstore entity record was successfully reset'.format(
                                            tenant_id, object, action
                                        )
                                    )

                                except Exception as e:
                                    logging.error(
                                        'tenant_id="{}", object="{}", failed to update the entity KVstore record with exception="{}"'.format(
                                            tenant_id, object, str(e)
                                        )
                                    )

                            # Define the SPL query
                            kwargs_search = {
                                "app": "trackme",
                                "earliest_time": "-5m",
                                "latest_time": "now",
                                "output_mode": "json",
                                "count": 0,
                            }
                            searchquery = (
                                "| trackmesamplingexecutor tenant_id="
                                + str(tenant_id)
                                + ' object="'
                                + object
                                + '"'
                                + "\n| `trackme_data_sampling_engine("
                                + str(tenant_id)
                                + ")`"
                                + "\n| `trackme_data_sampling_set_obfuscation_mode("
                                + str(tenant_id)
                                + ")`"
                                + "\n| outputlookup append=t key_field=key trackme_dsm_data_sampling_tenant_"
                                + str(tenant_id)
                                + "\n| fields - raw_sample"
                            )

                            # log debug
                            logging.debug(
                                'tenant_id="{}", object="{}", searchquery="{}"'.format(
                                    tenant_id, object, searchquery
                                )
                            )

                            query_results = []
                            # spawn the search and get the results
                            searchresults = service.jobs.oneshot(
                                searchquery, **kwargs_search
                            )
                            reader = results.JSONResultsReader(searchresults)
                            for item in reader:
                                if isinstance(item, dict):
                                    query_results.append(item)

                            # increment counter
                            processed_count += 1
                            succcess_count += 1
                            failures_count += 0

                            # append for summary
                            result = {
                                "object": object,
                                "action": action,
                                "result": "success",
                                "message": 'tenant_id="{}", The action="{}" was successfully performed'.format(
                                    tenant_id, action, object
                                ),
                                "searchquery": searchquery,
                                "results": query_results,
                            }
                            records.append(result)

                        except Exception as e:
                            logging.error(
                                'tenant_id="{}", failed to perform the action="{}" the entity, object="{}", exception="{}"'.format(
                                    tenant_id, action, object, str(e)
                                )
                            )

                            # increment counter
                            processed_count += 1
                            succcess_count += 0
                            failures_count += 1

                            result = {
                                "object": object,
                                "action": action,
                                "result": "failure",
                                "exception": 'tenant_id="{}", failed to perform the action="{}" for the entity, object="{}", exception="{}"'.format(
                                    tenant_id, action, object, str(e)
                                ),
                            }
                            records.append(result)

                # requested entity could not be found
                else:
                    logging.error(
                        'tenant_id="{}", entity not found, object="{}"'.format(
                            tenant_id, object
                        )
                    )

                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    result = {
                        "object": object,
                        "action": action,
                        "result": "failure",
                        "exception": 'tenant_id="{}", entity not found, object="{}"'.format(
                            tenant_id, object
                        ),
                    }
                    records.append(result)

            except Exception as e:
                logging.error(
                    'tenant_id="{}", general exception, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    )
                )

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": action,
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", general exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update number of records per sample by object name
    def post_ds_update_data_sampling_records_nr(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]

                # handle object_list / keys_list
                object_list = resp_dict.get("object_list", None)
                if object_list:
                    if not isinstance(object_list, list):
                        object_list = object_list.split(",")

                keys_list = resp_dict.get("keys_list", None)
                if keys_list:
                    if not isinstance(keys_list, list):
                        keys_list = keys_list.split(",")

                if not object_list and not keys_list:
                    return {
                        "payload": {
                            "error": "either object_list or keys_list must be provided"
                        },
                        "status": 500,
                    }

                data_sampling_nr = resp_dict["data_sampling_nr"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint enables the data sampling feature for an existing data source by the data source name (object), it requires a POST call with the following information:",
                "resource_desc": "Update the number of Data sampling records for a comma separated list of entities",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/splk_dsm/write/ds_update_data_sampling_records_nr\" mode=\"post\" body=\"{'tenant_id':'mytenant', 'data_sampling_nr': '100', 'object_list': 'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "object_list": "List of object entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "keys_list": "List of key entities, provided as a comma separated list of fields, you can provid object_list or keys_list",
                        "data_sampling_nr": "number of records to be sampled per data source and data sampling execution (defaults to 100 at first sampling, then 50)",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        collection_name = "kv_trackme_dsm_data_sampling_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # loop and proceed
        if object_list:
            keys_list = []
            for object in object_list:
                try:
                    kvrecord = collection.data.query(
                        query=json.dumps({"object": object})
                    )[0]
                    key = kvrecord.get("_key")
                    keys_list.append(key)
                except Exception as e:
                    key = None

        # Loop
        for key in keys_list:
            try:
                kvrecord = collection.data.query(query=json.dumps({"_key": key}))[0]

                # Update the record
                object = kvrecord.get("object")
                kvrecord["data_sampling_nr"] = data_sampling_nr
                kvrecord["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(kvrecord))

                # Record an audit change
                record = {
                    "object": str(object),
                    "data_sampling_nr": str(data_sampling_nr),
                }
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "update data sampling number of records",
                    str(object),
                    "splk-dsm",
                    str(json.dumps(record, indent=1)),
                    "Data sampling number of records was updated successfully",
                    str(update_comment),
                )

                logging.info(
                    "success for record="
                    + json.dumps(collection.data.query_by_id(key), indent=1)
                )

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "object": object,
                    "action": "update",
                    "result": "success",
                    "message": 'tenant_id="{}", The object was successfully updated'.format(
                        tenant_id, object
                    ),
                }
                records.append(result)

            except Exception as e:
                logging.error(str(e))

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "object": object,
                    "action": "update",
                    "result": "failure",
                    "exception": 'tenant_id="{}", failed to update the entity, object="{}", exception="{}"'.format(
                        tenant_id, object, str(e)
                    ),
                }
                records.append(result)

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}
