from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_lagging_classes.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkLaggingClasses_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkLaggingClasses_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_lagging_classes(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_lagging_classes",
            "resource_group_desc": "Endpoints related to the management of lagging classes for splk-feeds components (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add new policy
    def post_lagging_classes_add(self, request_info, **kwargs):
        """
        | trackme mode=post url="/services/trackme/v2/splk_lagging_classes/write/lagging_classes_add" body="{'tenant_id':'mytenant','lagging_class_type':'events','name':'linux_secure','level':'sourcetype','object':'all','value_lag':'900','value_delay':'3600'}"
        """

        # Declare
        tenant_id = None
        name = None
        level = None
        object_value = None
        value_lag = None
        value_delay = None

        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                name = resp_dict["name"]
                level = resp_dict["level"]
                object_value = resp_dict["object"]
                value_lag = int(resp_dict["value_lag"])
                value_delay = int(resp_dict["value_delay"])

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new lagging class policy, it requires a POST call with the following data:",
                "resource_desc": "Create a new lagging class policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_add\" body=\"{'tenant_id':'mytenant','lagging_class_type':'events','name':'linux_secure','level':'sourcetype','object':'all','value_lag':'900','value_delay':'3600'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant idenfitier",
                        "name": "name of the lagging class",
                        "level": "which level the lagging class is based on, valid otions are: sourcetype / index / priority",
                        "object": "which type of objects the lagging class is applied to, valid options are: splk-dsm / splk-dhm / all",
                        "value_lag": "the lagging value in seconds, an integer is expected",
                        "value_delay": "the delay value in seconds, an integer is expected",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "$and": [
                {
                    "name": name,
                    "object": object_value,
                    "level": level,
                }
            ]
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = "kv_trackme_common_lagging_classes_tenant_" + str(
                tenant_id
            )
            collection = service.kvstore[collection_name]

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if (
                key
                and level in ("sourcetype", "index", "priority")
                and object_value in ("splk-dsm", "splk-dhm", "all")
            ):
                # conflict, the same entity exists already
                logging.error(
                    'tenant_id="{}", conflict the same object exists already, name="{}"'.format(
                        tenant_id, name
                    )
                )
                return {
                    "payload": 'tenant_id="{}", conflict the same object exists already, name="{}"'.format(
                        tenant_id, name
                    ),
                    "status": 500,
                }

            elif level in ("sourcetype", "index", "priority") and object_value in (
                "splk-dsm",
                "splk-dhm",
                "all",
            ):
                # This record does not exist yet

                # Perform and audit
                try:
                    # Insert the record
                    collection.data.insert(
                        json.dumps(
                            {
                                "name": name,
                                "level": level,
                                "object": object_value,
                                "value_lag": str(value_lag),
                                "value_delay": str(value_delay),
                            }
                        )
                    )

                    # Get record
                    record = json.dumps(
                        collection.data.query(query=json.dumps(query_string)), indent=1
                    )

                    # Audit
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "create lagging class",
                            str(name),
                            "lagging_class",
                            str(record),
                            "The lagging class was created successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                except Exception as e:
                    response = {
                        "action": "failure",
                        "response": 'an exception was encountered, exception="{}"'.format(
                            str(e)
                        ),
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

                logging.info("success for record=" + str(record))
                return {"payload": str(record), "status": 200}

            else:
                logging.error(
                    "Error: data submitted is incorrect, valid options for level are sourcetype / index / priority, valid options for object are splk-dsm / splk-dhm / all"
                )
                return {
                    "payload": "Error: data submitted is incorrect, valid options for level are sourcetype / index / priority, valid options for object are splk-dsm / splk-dhm / all",
                    "status": 400,
                }

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Add new policy
    def post_lagging_classes_metrics_add(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_metrics_add\" body=\"{'tenant_id':'mytenant','metric_category':'spl','metric_max_lag_allowed':'900'}\"
        """

        # Declare
        tenant_id = None
        metric_category = None
        metric_max_lag_allowed = None
        query_string = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                metric_category = resp_dict["metric_category"]
                metric_max_lag_allowed = resp_dict["metric_max_lag_allowed"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new tag policy, it requires a POST call with the following data:",
                "resource_desc": "Create a new lagging class policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_metrics_add\" body=\"{'tenant_id':'mytenant','metric_category':'spl','metric_max_lag_allowed':'900'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "metric_category": "name of the metric category",
                        "metric_max_lag_allowed": "the lagging value in seconds, an integer is expected",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Data collection
            collection_name = "kv_trackme_mhm_lagging_classes_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            # Define the KV query
            query_string = {
                "metric_category": metric_category,
            }

            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # conflict, the same entity exists already
                logging.error(
                    'tenant_id="{}", conflict the same object exists already, metric_category="{}"'.format(
                        tenant_id, metric_category
                    )
                )
                return {
                    "payload": 'tenant_id="{}", conflict the same object exists already, metric_category="{}"'.format(
                        tenant_id, metric_category
                    ),
                    "status": 500,
                }

            else:
                # This record does not exist yet

                new_kvrecord = {
                    "metric_category": metric_category,
                    "metric_max_lag_allowed": metric_max_lag_allowed,
                }

                # Perform and audit
                try:
                    # Insert the record
                    collection.data.insert(json.dumps(new_kvrecord))

                    # Audit
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "create metric sla policy",
                            str(metric_category),
                            "metric_sla_policy",
                            new_kvrecord,
                            "The metric SLA policy was created successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                except Exception as e:
                    response = {
                        "action": "failure",
                        "response": 'an exception was encountered, exception="{}"'.format(
                            str(e)
                        ),
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

                response = {
                    "action": "success",
                    "response": "lagging class added successfully",
                    "record": new_kvrecord,
                }
                return {"payload": response, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Delete records from the collection
    def post_lagging_classes_del(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_del\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"
        """

        # Declare
        tenant_id = None
        lagging_class_type = None
        keys_list = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                lagging_class_type = resp_dict["lagging_class_type"]
                if lagging_class_type not in ("events, metrics"):
                    logging.error(
                        'tenant_id="{}", invalid option lagging_class_type="{}"'.format(
                            tenant_id, lagging_class_type
                        )
                    )
                    return {
                        "payload": 'tenant_id="{}", invalid option lagging_class_type="{}"'.format(
                            tenant_id, lagging_class_type
                        ),
                        "status": 500,
                    }
                keys_list = resp_dict["keys_list"]
                # Handle as a CSV list of keys
                keys_list = keys_list.split(",")

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint deletes lagging classes, it requires a POST call with the following information:",
                "resource_desc": "Delete one or more lagging classes",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_del\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "lagging_class_type": "The type of lagging classes, valid options are: events | metrics",
                        "keys_list": "List of record keys seperated by a comma of the records to be deleted from the collection",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        if lagging_class_type == "events":
            collection_name = "kv_trackme_common_lagging_classes_tenant_" + str(
                tenant_id
            )
        elif lagging_class_type == "metrics":
            collection_name = "kv_trackme_mhm_lagging_classes_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # loop
        for item in keys_list:
            # Define the KV query
            query_string = {
                "_key": item,
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already

                # Store the record for audit purposes
                record = str(json.dumps(collection.data.query_by_id(key), indent=1))

                # Remove and audit
                try:
                    # Remove the record
                    collection.data.delete(json.dumps({"_key": key}))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "delete lagging class",
                            str(item),
                            "all",
                            str(record),
                            "The lagging class was deleted successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "delete lagging class",
                            str(item),
                            "all",
                            str(record),
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete lagging class",
                        str(item),
                        "all",
                        str(record),
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update records
    def post_lagging_classes_update(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/lagging_classes_del\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"
        """

        # Declare
        tenant_id = None
        lagging_class_type = None
        records_list = None
        query_string = None

        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                lagging_class_type = resp_dict["lagging_class_type"]
                if lagging_class_type not in ("events, metrics"):
                    logging.error(
                        'tenant_id="{}", invalid option lagging_class_type="{}"'.format(
                            tenant_id, lagging_class_type
                        )
                    )
                    return {
                        "payload": 'tenant_id="{}", invalid option lagging_class_type="{}"'.format(
                            tenant_id, lagging_class_type
                        ),
                        "status": 500,
                    }
                component = resp_dict["component"]
                if component not in ("dsm", "dhm", "mhm"):
                    return {
                        "payload": 'Invalid component="{}"'.format(component),
                        "status": 500,
                    }
                records_list = resp_dict["records_list"]
                records_list = json.loads(records_list)

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint updates records, it requires a POST call with the following information:",
                "resource_desc": "Update one or more lagging classes",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_lagging_classes/write/post_lagging_classes_update\" body=\"{'tenant_id':'mytenant','lagging_class_type':'metrics','component':'splk-mhm','keys_list':'63716c098336d473a8152f30'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "lagging_class_type": "The type of lagging classes, valid options are: events | metrics",
                        "records_list": "JSON records to be updated",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        if lagging_class_type == "events":
            collection_name = "kv_trackme_common_lagging_classes_tenant_" + str(
                tenant_id
            )
        elif lagging_class_type == "metrics":
            collection_name = "kv_trackme_mhm_lagging_classes_tenant_" + str(tenant_id)
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # debug
        logging.info('records_list="{}"'.format(json.dumps(records_list, indent=0)))

        # loop
        for item in records_list:
            # debug
            logging.info('item="{}"'.format(item))

            # Define the KV query
            query_string = {
                "_key": item.get("_key"),
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already

                # Store the record for audit purposes
                record = str(json.dumps(collection.data.query_by_id(key), indent=1))

                # Update and audit
                try:
                    # Update the record

                    if lagging_class_type == "events":
                        collection.data.update(
                            str(key),
                            json.dumps(
                                {
                                    "name": item.get("name"),
                                    "level": item.get("level"),
                                    "object": item.get("object"),
                                    "value_lag": item.get("value_lag"),
                                    "value_delay": item.get("value_delay"),
                                }
                            ),
                        )
                    elif lagging_class_type == "metrics":
                        collection.data.update(
                            str(key),
                            json.dumps(
                                {
                                    "metric_category": item.get("metric_category"),
                                    "metric_max_lag_allowed": item.get(
                                        "metric_max_lag_allowed"
                                    ),
                                }
                            ),
                        )

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "update lagging classes",
                            str(item),
                            str(component),
                            json.dumps(item, indent=1),
                            "The lagging classs was updated successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "update lagging classes",
                            str(item),
                            str(component),
                            str(record),
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "update lagging classes",
                        str(item),
                        str(component),
                        str(record),
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}
