from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_logical_groups.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkLogicalGroups_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkLogicalGroups_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_logical_groups(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_logical_groups/write",
            "resource_group_desc": "Endpoints related to the management of logical groups (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add a new group
    def post_logical_groups_add_grp(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_add_grp\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'grp-lb001', 'object_group_members': 'lb-001,lb-002', 'object_group_min_green_percent': '50'}\"
        """

        # define
        tenant_id = None
        object_group_name = None
        object_group_members = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_group_name = resp_dict["object_group_name"]

                try:
                    object_group_members = resp_dict["object_group_members"]
                    # object_group_members is expected as a comma separted list of values
                    # We accept comma with or without a space after the seperator, let's remove any space after the separator
                    object_group_members = object_group_members.replace(", ", ",")
                    # Split by the separator
                    object_group_members = object_group_members.split(",")
                except Exception as e:
                    object_group_members = []

                # group min percentage is optional and set to 50% if not provided
                try:
                    object_group_min_green_percent = resp_dict[
                        "object_group_min_green_percent"
                    ]
                except Exception as e:
                    object_group_min_green_percent = "50"

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new logical group, if the group exists already, members mentioned in the query (if any) will be added to the group, it requires a POST call with the following data required:",
                "resource_desc": "Create a new logical group, update if exists already",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_add_grp\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'grp-lb001', 'object_group_members': 'lb-001,lb-002', 'object_group_min_green_percent': '50'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "object_group_name": "name of the logical group to be created",
                        "object_group_members": "OPTIONAL: comma separated list of the group members, if not provided and the group exists already the current value will be preserved, will be set to empty group if the the group does not exist yet",
                        "object_group_min_green_percent": "OPTIONAL: minimal percentage of hosts that need to be green for the logical group to be green, if unset defaults to 50. Recommended options for this value: 12.5 / 33.33 / 50",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Retrieve from data
        resp_dict = json.loads(str(request_info.raw_args["payload"]))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "object_group_name": object_group_name,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            collection_name = "kv_trackme_common_logical_group_tenant_" + str(tenant_id)
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )
            collection = service.kvstore[collection_name]

            # update time for the object
            object_group_mtime = time.time()

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None
                record = json.dumps(
                    {
                        "object_group_name": object_group_name,
                        "object_group_members": object_group_members,
                        "object_group_min_green_percent": str(
                            object_group_min_green_percent
                        ),
                        "object_group_mtime": str(object_group_mtime),
                    }
                )

            # if the group exists already, retrieve its existing members and add our new members
            if key:
                # set an action_desc
                action_desc = "update"

                # create a new list
                object_group_members_new = []

                # add existing members
                for object_group_existing_member in record[0].get(
                    "object_group_members"
                ):
                    object_group_members_new.append(object_group_existing_member)

                # add our new members, if not in there already
                for object_group_new_member in object_group_members:
                    if object_group_new_member not in object_group_members_new:
                        object_group_members_new.append(object_group_new_member)

            else:
                # set an action_desc
                action_desc = "create"

            # proceed
            try:
                if key:
                    # Update the record
                    collection.data.update(
                        str(key),
                        json.dumps(
                            {
                                "object_group_name": object_group_name,
                                "object_group_members": object_group_members_new,
                                "object_group_min_green_percent": str(
                                    object_group_min_green_percent
                                ),
                                "object_group_mtime": str(object_group_mtime),
                            }
                        ),
                    )

                else:
                    # Insert the record
                    collection.data.insert(
                        json.dumps(
                            {
                                "object_group_name": object_group_name,
                                "object_group_members": object_group_members,
                                "object_group_min_green_percent": str(
                                    object_group_min_green_percent
                                ),
                                "object_group_mtime": str(object_group_mtime),
                            }
                        )
                    )

                # Record an audit change
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "Logical group add",
                        str(object_group_name),
                        "logical_group",
                        collection.data.query(query=json.dumps(query_string)),
                        "The logical group was added successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                # render response
                response = {
                    "action": "success",
                    "action_desc": action_desc,
                    "record": collection.data.query(query=json.dumps(query_string)),
                }

                logging.info(json.dumps(response, indent=2))
                return {"payload": response, "status": 200}

            except Exception as e:
                # Record an audit change
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "Logical group add",
                        str(object_group_name),
                        "logical_group",
                        collection.data.query(query=json.dumps(query_string)),
                        "The logical group failed to be created",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                # render response
                response = {
                    "action": "failure",
                    "action_desc": action_desc,
                    "response": 'failed to {} the KVstore record, exception="{}"'.format(
                        action_desc, str(e)
                    ),
                }
                logging.error(json.dumps(response, indent=2))
                return {"payload": response, "status": 500}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Delete group
    def post_logical_groups_del_grp(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_del_grp\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'grp-lb001'}\"
        """

        # define
        tenant_id = None
        object_group_name = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_group_name = resp_dict["object_group_name"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint deletes a logical group, it requires a POST call with the following data required:",
                "resource_desc": "Delete a logical group and all membership related to it",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_del_grp\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'grp-lb001'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "object_group_name": "name of the logical group to be removed",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "object_group_name": object_group_name,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            collection_name = "kv_trackme_common_logical_group_tenant_" + str(tenant_id)
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )
            collection = service.kvstore[collection_name]

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if not key:
                response = {
                    "action": "failure",
                    "response": "an exception was encountered, this group does not exist",
                }
                logging.error(json.dumps(response))
                return {"payload": response, "status": 500}

            else:
                # Proceed
                try:
                    # Remove the record
                    collection.data.delete(json.dumps({"_key": key}))

                    # Record an audit change
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "Logical group delete",
                            str(object_group_name),
                            "logical_group",
                            collection.data.query(query=json.dumps(query_string)),
                            "The logical group was deleted successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    # render response
                    response = {
                        "action": "success",
                        "action_desc": "delete",
                        "response": 'the record for object_group_name="{}" was deleted successfully'.format(
                            object_group_name
                        ),
                        "record": record,
                    }

                    logging.info(json.dumps(response, indent=2))
                    return {"payload": response, "status": 200}

                except Exception as e:
                    # render response
                    response = {
                        "action": "failure",
                        "action_desc": "delete",
                        "response": 'failed to delete the KVstore record with exception="{}"'.format(
                            str(e)
                        ),
                    }

                    logging.info(json.dumps(response, indent=2))
                    return {"payload": response, "status": 500}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Associate a logical group with a member
    def post_logical_groups_associate_group(self, request_info, **kwargs):
        """
        | trackme mode=\"post\" url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_associate_group\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'object_group_name': 'grp-lb001', 'action': 'unassociate', 'object_list': 'lb-002,lb-003'}\"
        """

        # define
        tenant_id = None
        object_list = None
        object_group_name = None
        object_group_min_green_percent = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                action = resp_dict["action"]
                object_list = resp_dict["object_list"]
                # turns into a list
                object_list = object_list.split(",")
                object_group_name = resp_dict["object_group_name"]
                # optional
                try:
                    object_group_min_green_percent = resp_dict[
                        "object_group_min_green_percent"
                    ]
                except Exception as e:
                    object_group_min_green_percent = 50

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint associates/unassociates an object with a logical group, "
                + "if action is associate and the logical group does not exist, it will be created. "
                "it requires a POST call with the following data required:",
                "resource_desc": "Manage association on a logical group",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_logical_groups/write/logical_groups_associate_group\" body=\"{'tenant_id': 'mytenant', 'object_group_name': 'object_group_name': 'grp-lb001', 'action': 'unassociate', 'object_list': 'lb-002,lb-003'}\"",
                "options": [
                    {
                        "tenant_id": "The tenant identifier",
                        "object_group_name": "The logical group name",
                        "action": "The action to be performed, valid options are: associate | unassociate",
                        "object_list": "comma seprated list of entities to be associated with or unassociated from this logical group",
                        "object_group_min_green_percent": "OPTIONAL, if the group does not exist and is created, optionally specify the min green percentage (defaults to 50)",
                        "update_comment": "OPTIONAL: a comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Retrieve from data
        resp_dict = json.loads(str(request_info.raw_args["payload"]))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        try:
            # Define the KV query
            query_string = {
                "object_group_name": object_group_name,
            }

            # collection
            collection_name = "kv_trackme_common_logical_group_tenant_" + str(tenant_id)
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )
            collection = service.kvstore[collection_name]

            # Get the record
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result

            # if unassociate but there is no record, then there is nothing we can do
            if not key and action == "unassociate":
                response = {
                    "action": "failure",
                    "response": 'object_group_name="{}" not found'.format(
                        object_group_name
                    ),
                }
                logging.error(json.dumps(response))
                return {"payload": response, "status": 404}

            # else if there is no record but action is associate
            elif not key and action == "associate":
                # set new record
                record = {
                    "object_group_name": object_group_name,
                    "object_group_min_green_percent": object_group_min_green_percent,
                    "object_group_members": object_list,
                    "object_group_mtime": time.time(),
                }

                # Update the record
                try:
                    collection.data.insert(json.dumps(record))

                    # register an audit event
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "create and associate logical group",
                            str(object_group_name),
                            "logical_group",
                            collection.data.query(query=json.dumps(query_string)),
                            "The logical group was created successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    # render
                    response = {
                        "action": "success",
                        "response": 'A new logical group object_group_name="{}" was successfully created and associated with object_list="{}"'.format(
                            object_group_name, object_list
                        ),
                        "record": collection.data.query(query=json.dumps(query_string)),
                    }
                    logging.info(json.dumps(response))
                    return {"payload": response, "status": 200}

                except Exception as e:
                    response = {
                        "action": "failure",
                        "response": 'an exception was encountered, exception="{}"'.format(
                            str(e)
                        ),
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

            else:
                # get current list of associated objects with that logical group, if any
                object_new_list = []
                try:
                    object_current_list = record[0].get("object_group_members")
                except Exception as e:
                    object_current_list = None

                # Loop if we have a value
                if object_current_list:
                    if isinstance(object_current_list, list):
                        for object_value in object_current_list:
                            object_new_list.append(object_value)
                    else:
                        object_new_list.append(record[0].get("object_group_members"))

                # Loop through the submitted list of object, and act depending on the requested action
                for object_value in object_list:
                    if action == "associate":
                        if object_value not in object_new_list:
                            object_new_list.append(object_value)
                    elif action == "unassociate":
                        if object_value in object_new_list:
                            object_new_list.remove(object_value)

                # set new record
                record = {
                    "object_group_name": object_group_name,
                    "object_group_min_green_percent": record[0].get(
                        "object_group_min_green_percent"
                    ),
                    "object_group_members": object_new_list,
                    "object_group_mtime": time.time(),
                }

                # Update the record
                try:
                    collection.data.update(str(key), record)

                    # Record an audit change
                    for object_value in object_list:
                        try:
                            trackme_audit_event(
                                request_info.system_authtoken,
                                request_info.server_rest_uri,
                                tenant_id,
                                request_info.user,
                                "success",
                                "associate logical group",
                                str(object_value),
                                "logical_group",
                                collection.data.query_by_id(key),
                                "The object was associated successfully",
                                str(update_comment),
                            )
                        except Exception as e:
                            logging.error(
                                'failed to generate an audit event with exception="{}"'.format(
                                    str(e)
                                )
                            )

                    # render
                    if action == "associate":
                        response = {
                            "action": "success",
                            "response": 'the following objects="{}" where successfully associated with the object_group_name="{}"'.format(
                                object_list, object_group_name
                            ),
                            "record": collection.data.query(
                                query=json.dumps(query_string)
                            ),
                        }

                    elif action == "unassociate":
                        response = {
                            "action": "success",
                            "response": 'the following objects="{}" where successfully unassociated from the object_group_name="{}"'.format(
                                object_list, object_group_name
                            ),
                            "record": collection.data.query(
                                query=json.dumps(query_string)
                            ),
                        }

                    # log
                    logging.info(json.dumps(response, indent=4))

                    return {"payload": response, "status": 200}

                except Exception as e:
                    # register an audit event
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "associate logical group",
                            str(object_value),
                            "logical_group",
                            collection.data.query_by_id(key),
                            "The object failed to be associated",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    # render response
                    response = {
                        "action": "failure",
                        "response": 'an exception was encountered, exception="{}"'.format(
                            str(e)
                        ),
                    }
                    logging.error(json.dumps(response))
                    return {"payload": response, "status": 500}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'general exception, an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}
