from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_splk_dhm.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import random
import sys
import time
from datetime import datetime
import re
import hashlib
import urllib.parse
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces

# import Splunk libs
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerSplkOutliersEngine_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkOutliersEngine_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_outliers_engine(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_outliers_engine/write",
            "resource_group_desc": "Endpoints related to the management of the Machine Learning Outliers detection (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add a new ML model to a given entity
    def post_outliers_add_model(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None
        model_json = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]
                model_json = resp_dict["model_json"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "resource_desc": "Add a new Machine Learning outliers model",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/splk_outliers_engine/write/outliers_add_model" body="{\'tenant_id\':\'mytenant\',\'component\':\'dsm\',\'object\':\'netscreen:netscreen:firewall\',\'model_json\':\'{\\\\\\"kpi_metric\\\\\\":\\\\\\"splk.feeds.perc95_eventcount_5m\\\\\\",\\\\\\"kpi_span\\\\\\":\\\\\\"10m\\\\\\",\\\\\\"method_calculation\\\\\\":\\\\\\"avg\\\\\\",\\\\\\"density_lowerthreshold\\\\\\":\\\\\\"0.005\\\\\\",\\\\\\"density_upperthreshold\\\\\\":\\\\\\"0.005\\\\\\",\\\\\\"alert_lower_breached\\\\\\":\\\\\\"1\\\\\\",\\\\\\"alert_upper_breached\\\\\\":\\\\\\"1\\\\\\",\\\\\\"period_calculation\\\\\\":\\\\\\"-30d\\\\\\",\\\\\\"time_factor\\\\\\":\\\\\\"%25H\\\\\\",\\\\\\"perc_min_lowerbound_deviation\\\\\\":\\\\\\"5.0\\\\\\",\\\\\\"perc_min_upperbound_deviation\\\\\\":\\\\\\"5.0\\\\\\"}\'}"',
                "describe": "This endpoint adds a new ML model to a given entity, it requires a POST call with the following options:",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "conponent": "The component category",
                        "object": "entity name",
                        "model_json": "The new ML model JSON definition",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Try to load the model JSON definition, if loading fails, stop here
        if not isinstance(model_json, dict):
            try:
                new_model_definition = json.loads(model_json)
                logging.info(
                    f'Successfully loaded model_json="{json.dumps(new_model_definition, indent=4)}"'
                )
            except Exception as e:
                msg = f'Failed to load the model_json="{model_json}" as a properly formatted JSON object with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}
        else:
            new_model_definition = model_json

        # records summary
        records = []

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": f"splk-{component}",
                        "object": object_value,
                    }
                ]
            }

            entity_rules = collection.data.query(query=json.dumps(query_string))[0]
            key = entity_rules.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            # Load as a dict
            try:
                entities_outliers = json.loads(entity_rules.get("entities_outliers"))
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # log debug
            logging.debug(
                f'entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            # For each expected key, try to retrieve the value
            try:
                kpi_metric = new_model_definition.get("kpi_metric")
                kpi_span = new_model_definition.get("kpi_span")
                method_calculation = new_model_definition.get("method_calculation")
                density_lowerthreshold = new_model_definition.get(
                    "density_lowerthreshold"
                )
                density_upperthreshold = new_model_definition.get(
                    "density_upperthreshold"
                )
                alert_lower_breached = new_model_definition.get("alert_lower_breached")
                alert_upper_breached = new_model_definition.get("alert_upper_breached")
                period_calculation = new_model_definition.get("period_calculation")
                time_factor = urllib.parse.unquote(
                    new_model_definition.get("time_factor")
                )
                auto_correct = new_model_definition.get("auto_correct")
                perc_min_lowerbound_deviation = new_model_definition.get(
                    "perc_min_lowerbound_deviation"
                )
                perc_min_upperbound_deviation = new_model_definition.get(
                    "perc_min_upperbound_deviation"
                )

            except Exception as e:
                msg = f'Failed to retrieve expected key from model_json with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # Set the model identifier
            model = f"model_{random.getrandbits(48)}"

            # Set the final model definition
            final_model_definition = {
                "is_disabled": 0,
                "kpi_metric": kpi_metric,
                "kpi_span": kpi_span,
                "method_calculation": method_calculation,
                "density_lowerthreshold": density_lowerthreshold,
                "density_upperthreshold": density_upperthreshold,
                "alert_lower_breached": alert_lower_breached,
                "alert_upper_breached": alert_upper_breached,
                "period_calculation": period_calculation,
                "time_factor": time_factor,
                "auto_correct": auto_correct,
                "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                "period_exclusions": [],
                "ml_model_gen_search": "pending",
                "ml_model_render_search": "pending",
                "ml_model_summary_search": "pending",
                "rules_access_search": "pending",
                "ml_model_filename": "pending",
                "ml_model_filesize": "pending",
                "ml_model_lookup_share": "pending",
                "ml_model_lookup_owner": "pending",
                "ml_model_lookup_perm_write": "pending",
                "ml_model_lookup_perm_read": "pending",
                "last_exec": "pending",
            }

            # Add the new model to the dict
            entities_outliers[model] = final_model_definition

            # log debug
            logging.debug(
                f'final model_dict="{json.dumps(entities_outliers, indent=4)}"'
            )

            try:
                # Update the record
                entity_rules["entities_outliers"] = json.dumps(
                    entities_outliers, indent=4
                )
                entity_rules["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(entity_rules))

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "action": "add",
                    "result": "success",
                    "model_definition": {
                        "kpi_metric": kpi_metric,
                        "kpi_span": kpi_span,
                        "method_calculation": method_calculation,
                        "density_lowerthreshold": density_lowerthreshold,
                        "density_upperthreshold": density_upperthreshold,
                        "alert_lower_breached": alert_lower_breached,
                        "alert_upper_breached": alert_upper_breached,
                        "period_calculation": period_calculation,
                        "time_factor": time_factor,
                        "auto_correct": auto_correct,
                        "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                        "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                    },
                    "message": f'the model="{model}" was successfully added to the outliers rules',
                }
                records.append(result)

            except Exception as e:
                logging.error(
                    f'failed to add the new model to the dictionnary with exception="{str(e)}"'
                )

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "action": "add",
                    "result": "failure",
                    "model_definition": {
                        "kpi_metric": kpi_metric,
                        "kpi_span": kpi_span,
                        "method_calculation": method_calculation,
                        "density_lowerthreshold": density_lowerthreshold,
                        "density_upperthreshold": density_upperthreshold,
                        "alert_lower_breached": alert_lower_breached,
                        "alert_upper_breached": alert_upper_breached,
                        "period_calculation": period_calculation,
                        "time_factor": time_factor,
                        "auto_correct": auto_correct,
                        "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                        "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                    },
                    "exception": f'failed to add the new model to the dictionnary with exception="{str(e)}"',
                }
                records.append(result)

            # log debug
            logging.debug(
                f'final dict, entity_rules="{json.dumps(entities_outliers, indent=4)}"'
            )

            # render HTTP status and summary

            req_summary = {
                "process_count": processed_count,
                "success_count": succcess_count,
                "failures_count": failures_count,
                "records": records,
            }

            if processed_count > 0 and processed_count == succcess_count:
                # log
                logging.info(
                    f'ML model was added successfully, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "add ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(req_summary, indent=1)),
                        "ML model was added successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 200}

            else:
                # log
                logging.error(
                    f'ML model could not be added, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "add ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(req_summary, indent=1)),
                        "ML model could not be added",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 500}

        else:
            msg = "No record found for this object"
            logging.error(msg)
            return {"payload": msg, "status": 404}

    # Add a new ML model to a given entity
    def post_outliers_cim_add_model(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        object_value = None
        entity_value = None
        cim_field = None
        model_json = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                entity_value = resp_dict["entity"]
                cim_field = resp_dict["cim_field"]
                model_json = resp_dict["model_json"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint adds a new ML model to a given entity, it requires a POST call with the following options:",
                "resource_desc": "Add a new Machine Learning outliers model",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/splk_outliers_engine/write/outliers_cim_add_model" body="{\'tenant_id\':\'cim-demo\',\'object\':\'auth001\',\'entity\': \'global\',\'cim_field\': \'signature\',\'model_json\':\'{\\\\\\"entity\\\\\\":\\\\\\"global\\\\\\",\\\\\\"cim_field\\\\\\":\\\\\\"signature\\\\\\",\\\\\\"kpi_metric\\\\\\":\\\\\\"splk.cim.total_count\\\\\\",\\\\\\"kpi_span\\\\\\":\\\\\\"10m\\\\\\",\\\\\\"method_calculation\\\\\\":\\\\\\"avg\\\\\\",\\\\\\"density_lowerthreshold\\\\\\":\\\\\\"0.005\\\\\\",\\\\\\"density_upperthreshold\\\\\\":\\\\\\"0.005\\\\\\",\\\\\\"alert_lower_breached\\\\\\":\\\\\\"1\\\\\\",\\\\\\"alert_upper_breached\\\\\\":\\\\\\"1\\\\\\",\\\\\\"period_calculation\\\\\\":\\\\\\"-30d\\\\\\",\\\\\\"time_factor\\\\\\":\\\\\\"%25H\\\\\\",\\\\\\"perc_min_lowerbound_deviation\\\\\\":\\\\\\"5.0\\\\\\",\\\\\\"perc_min_upperbound_deviation\\\\\\":\\\\\\"5.0\\\\\\"}\'}"',
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "entity name",
                        "entity": "The cim entity, if the cim tracker is not using a custom break by logic, the cim entity is global",
                        "cim_field": "The cim field",
                        "model_json": "The new ML model JSON definition",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = f"kv_trackme_cim_outliers_entity_rules_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Try to load the model JSON definition, if loading fails, stop here
        if not isinstance(model_json, dict):
            try:
                new_model_definition = json.loads(model_json)
                logging.info(
                    f'Successfully loaded model_json="{json.dumps(new_model_definition, indent=4)}"'
                )
            except Exception as e:
                msg = f'Failed to load the model_json="{model_json}" as a properly formatted JSON object with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}
        else:
            new_model_definition = model_json

        # records summary
        records = []

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": "splk-cim",
                        "object": object_value,
                        "cim_field": cim_field,
                        "entity": entity_value,
                    }
                ]
            }

            entity_rules = collection.data.query(query=json.dumps(query_string))[0]
            key = entity_rules.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            # Load as a dict
            try:
                entities_outliers = json.loads(entity_rules.get("entities_outliers"))
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # log debug
            logging.debug(
                f'entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            # For each expected key, try to retrieve the value
            try:
                kpi_metric = new_model_definition.get("kpi_metric")
                kpi_span = new_model_definition.get("kpi_span")
                method_calculation = new_model_definition.get("method_calculation")
                density_lowerthreshold = new_model_definition.get(
                    "density_lowerthreshold"
                )
                density_upperthreshold = new_model_definition.get(
                    "density_upperthreshold"
                )
                alert_lower_breached = new_model_definition.get("alert_lower_breached")
                alert_upper_breached = new_model_definition.get("alert_upper_breached")
                period_calculation = new_model_definition.get("period_calculation")
                time_factor = urllib.parse.unquote(
                    new_model_definition.get("time_factor")
                )
                auto_correct = new_model_definition.get("auto_correct")
                perc_min_lowerbound_deviation = new_model_definition.get(
                    "perc_min_lowerbound_deviation"
                )
                perc_min_upperbound_deviation = new_model_definition.get(
                    "perc_min_upperbound_deviation"
                )

            except Exception as e:
                msg = f'Failed to retrieve expected key from model_json with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # Set the model identifier
            model = f"model_{random.getrandbits(48)}"

            # Set the final model definition
            final_model_definition = {
                "is_disabled": 0,
                "entity": entity_value,
                "cim_field": cim_field,
                "kpi_metric": kpi_metric,
                "kpi_span": kpi_span,
                "method_calculation": method_calculation,
                "density_lowerthreshold": density_lowerthreshold,
                "density_upperthreshold": density_upperthreshold,
                "alert_lower_breached": alert_lower_breached,
                "alert_upper_breached": alert_upper_breached,
                "period_calculation": period_calculation,
                "time_factor": time_factor,
                "auto_correct": auto_correct,
                "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                "period_exclusions": [],
                "ml_model_gen_search": "pending",
                "ml_model_render_search": "pending",
                "ml_model_summary_search": "pending",
                "rules_access_search": "pending",
                "ml_model_filename": "pending",
                "ml_model_filesize": "pending",
                "ml_model_lookup_share": "pending",
                "ml_model_lookup_owner": "pending",
                "ml_model_lookup_perm_write": "pending",
                "ml_model_lookup_perm_read": "pending",
                "last_exec": "pending",
            }

            # Add the new model to the dict
            entities_outliers[model] = final_model_definition

            # log debug
            logging.debug(
                f'final model_dict="{json.dumps(entities_outliers, indent=4)}"'
            )

            try:
                # Update the record
                entity_rules["entities_outliers"] = json.dumps(
                    entities_outliers, indent=4
                )
                entity_rules["mtime"] = time.time()
                collection.data.update(str(key), json.dumps(entity_rules))

                # increment counter
                processed_count += 1
                succcess_count += 1
                failures_count += 0

                # append for summary
                result = {
                    "tenant_id": tenant_id,
                    "object_category": "splk-cim",
                    "object": object_value,
                    "entity": entity_value,
                    "cim_field": cim_field,
                    "action": "add",
                    "result": "success",
                    "model_definition": {
                        "entity": entity_value,
                        "cim_field": cim_field,
                        "kpi_metric": kpi_metric,
                        "kpi_span": kpi_span,
                        "method_calculation": method_calculation,
                        "density_lowerthreshold": density_lowerthreshold,
                        "density_upperthreshold": density_upperthreshold,
                        "alert_lower_breached": alert_lower_breached,
                        "alert_upper_breached": alert_upper_breached,
                        "period_calculation": period_calculation,
                        "time_factor": time_factor,
                        "auto_correct": auto_correct,
                        "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                        "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                    },
                    "message": f'the model="{model}" was successfully added to the outliers rules',
                }
                records.append(result)

            except Exception as e:
                logging.error(
                    f'failed to add the new model to the dictionnary with exception="{str(e)}"'
                )

                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                result = {
                    "tenant_id": tenant_id,
                    "object_category": "splk-cim",
                    "object": object_value,
                    "entity": entity_value,
                    "cim_field": cim_field,
                    "action": "add",
                    "result": "failure",
                    "model_definition": {
                        "entity": entity_value,
                        "cim_field": cim_field,
                        "kpi_metric": kpi_metric,
                        "kpi_span": kpi_span,
                        "method_calculation": method_calculation,
                        "density_lowerthreshold": density_lowerthreshold,
                        "density_upperthreshold": density_upperthreshold,
                        "alert_lower_breached": alert_lower_breached,
                        "alert_upper_breached": alert_upper_breached,
                        "period_calculation": period_calculation,
                        "time_factor": time_factor,
                        "auto_correct": auto_correct,
                        "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                        "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                    },
                    "exception": f'failed to add the new model to the dictionnary with exception="{str(e)}"',
                }
                records.append(result)

            # log debug
            logging.debug(
                f'final dict, entity_rules="{json.dumps(entities_outliers, indent=4)}"'
            )

            # render HTTP status and summary

            req_summary = {
                "process_count": processed_count,
                "success_count": succcess_count,
                "failures_count": failures_count,
                "records": records,
            }

            if processed_count > 0 and processed_count == succcess_count:
                # log
                logging.info(
                    f'ML model was added successfully, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "add ML models",
                        str(object_value),
                        "splk-cim",
                        str(json.dumps(req_summary, indent=1)),
                        "ML model was added successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 200}

            else:
                # log
                logging.error(
                    f'ML model could not be added, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "add ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(req_summary, indent=1)),
                        "ML model could not be added",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 500}

        else:
            msg = f'Could not find a valid record for the tenant_id="{tenant_id}", object="{object_value}", entity="{entity_value}", cim_field="{cim_field}"'
            logging.error(msg)
            return {"payload": msg, "status": 404}

    # delete one or more ML models
    def post_outliers_delete_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None
        models_list = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]
                models_list = resp_dict["models_list"]
                # Handle as a CSV list of keys
                models_list = models_list.split(",")

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint deletes or more existing ML models, it requires a POST call with the following options:",
                "resource_desc": "Delete a Machine Learning outliers model",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_delete_models\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'object': 'netscreen:netscreen:firewall', 'models_list': 'model_xxxxxxxxxx'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name",
                        "models_list": "Comma separated list of models identifiers to be deleted",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": f"splk-{component}",
                        "object": object_value,
                    }
                ]
            }

            entity_rules = collection.data.query(query=json.dumps(query_string))[0]
            key = entity_rules.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            logging.debug(entity_rules)

            # Load as a dict
            try:
                entities_outliers = json.loads(entity_rules.get("entities_outliers"))
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # log debug
            logging.debug(
                f'entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            # loop through the models
            for model in models_list:
                try:
                    model_dict = entities_outliers[model]

                    # log debug
                    logging.debug(f'model_dict="{json.dumps(model_dict, indent=4)}"')

                    # delete from the dict
                    del entities_outliers[model]

                    # if the last model from the rules was deleted, Ml is not ready any longer
                    if not len(entities_outliers) > 2:
                        last_exec = "pending"
                    else:
                        last_exec = entity_rules.get("last_exec")

                    # Update the record
                    entity_rules["entities_outliers"] = json.dumps(
                        entities_outliers, indent=4
                    )
                    entity_rules["mtime"] = time.time()
                    entity_rules["last_exec"] = last_exec
                    collection.data.update(str(key), json.dumps(entity_rules))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1
                    failures_count += 0

                    # append for summary
                    result = {
                        "tenant_id": tenant_id,
                        "object_category": f"splk-{component}",
                        "object": object_value,
                        "action": "delete",
                        "result": "success",
                        "message": f'the model="{model}" was successfully deleted from the outliers rules',
                    }
                    records.append(result)

                except Exception as e:
                    logging.error(
                        f'model="{model}" could not be found in the entity_rules="{json.dumps(entities_outliers, indent=4)}"'
                    )

                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    result = {
                        "tenant_id": tenant_id,
                        "object_category": f"splk-{component}",
                        "object": object_value,
                        "action": "delete",
                        "result": "failure",
                        "exception": f'the model="{model}" could not be found in the outliers rules',
                    }
                    records.append(result)

            # log debug
            logging.debug(
                f'final dict, entity_rules="{json.dumps(entities_outliers, indent=4)}"'
            )

            # render HTTP status and summary

            req_summary = {
                "process_count": processed_count,
                "success_count": succcess_count,
                "failures_count": failures_count,
                "records": records,
            }

            if processed_count > 0 and processed_count == succcess_count:
                # log
                logging.info(
                    f'ML models were deleted successfully, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "delete ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(req_summary, indent=1)),
                        "ML models were deleted successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 200}

            else:
                # log
                logging.error(
                    f'ML models could not be deleted, summary="{json.dumps(req_summary, indent=4)}"'
                )

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(req_summary, indent=1)),
                        "ML models could not be deleted",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": req_summary, "status": 500}

        else:
            msg = f'No record found for this object, query_string="{json.dumps(query_string, indent=2)}", collection="{collection_name}"'
            logging.error(msg)
            return {"payload": {"response": msg}, "status": 404}

    # delete one or more ML models for splk-cim
    def post_outliers_cim_delete_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        object_value = None
        json_dict = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                json_dict = json.loads(resp_dict["json_dict"])

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint deletes or more existing ML models, it requires a POST call with the following options:",
                "resource_desc": "Delete a Machine Learning outliers model",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/splk_outliers_engine/write/outliers_cim_delete_models" body="{\'tenant_id\':\'cim-demo\',\'object\':\'auth001\',\'json_dict\':\'[{\\\\\\"model_id\\\\\\":\\\\\\"model_150156483858859\\\\\\",\\\\\\"entity\\\\\\":\\\\\\"global\\\\\\",\\\\\\"cim_field\\\\\\":\\\\\\"app\\\\\\"},{\\\\\\"model_id\\\\\\":\\\\\\"model_129056918057674\\\\\\",\\\\\\"entity\\\\\\":\\\\\\"global\\\\\\",\\\\\\"cim_field\\\\\\":\\\\\\"app\\\\\\"}]\'}"',
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "The name of the object",
                        "json_dict": "A list of JSON dictionaries containing for each model to be deleted, the entity, cim_field and model_id",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = f"kv_trackme_cim_outliers_entity_rules_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Loop
        for p in json_dict:
            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                # Define the KV query
                query_string = {
                    "$and": [
                        {
                            "object_category": "splk-cim",
                            "object": object_value,
                            "entity": p["entity"],
                            "cim_field": p["cim_field"],
                        }
                    ]
                }

                entity_rules = collection.data.query(query=json.dumps(query_string))[0]
                key = entity_rules.get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # Load as a dict
                try:
                    entities_outliers = json.loads(
                        entity_rules.get("entities_outliers")
                    )
                except Exception as e:
                    msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                    logging.error(msg)
                    return {"payload": msg, "status": 500}

                # log debug
                logging.debug(
                    f'entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
                )

                # loop through the models
                model = p["model_id"]

                try:
                    model_dict = entities_outliers[model]

                    # log debug
                    logging.debug(f'model_dict="{json.dumps(model_dict, indent=4)}"')

                    # delete from the dict
                    del entities_outliers[model]

                    # if the last model from the rules was deleted, Ml is not ready any longer
                    if not len(entities_outliers) > 2:
                        last_exec = "pending"
                    else:
                        last_exec = entity_rules.get("last_exec")

                    # Update the record
                    entity_rules["entities_outliers"] = json.dumps(
                        entities_outliers, indent=4
                    )
                    entity_rules["mtime"] = time.time()
                    entity_rules["last_exec"] = last_exec
                    collection.data.update(str(key), json.dumps(entity_rules))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1
                    failures_count += 0

                    # append for summary
                    result = {
                        "tenant_id": tenant_id,
                        "object_category": "splk-cim",
                        "object": object_value,
                        "entity": p["entity"],
                        "cim_field": p["cim_field"],
                        "action": "delete",
                        "result": "success",
                        "message": f'the model="{model}" was successfully deleted from the outliers rules',
                    }
                    records.append(result)

                except Exception as e:
                    logging.error(
                        f'model="{model}" could not be found in the entity_rules="{json.dumps(entities_outliers, indent=4)}"'
                    )

                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    result = {
                        "tenant_id": tenant_id,
                        "object_category": "splk-cim",
                        "object": object_value,
                        "entity": p["entity"],
                        "cim_field": p["cim_field"],
                        "action": "delete",
                        "result": "failure",
                        "exception": f'the model="{model}" could not be found in the outliers rules',
                    }
                    records.append(result)

                # log debug
                logging.debug(
                    f'final dict, entity_rules="{json.dumps(entities_outliers, indent=4)}"'
                )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            # log
            logging.info(
                f'ML models were deleted successfully, summary="{json.dumps(req_summary, indent=4)}"'
            )

            # audit
            try:
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "delete ML models",
                    str(object_value),
                    "splk-dsm",
                    str(json.dumps(req_summary, indent=1)),
                    "ML models were deleted successfully",
                    str(update_comment),
                )
            except Exception as e:
                logging.error(
                    f'failed to generate an audit event with exception="{str(e)}"'
                )

            return {"payload": req_summary, "status": 200}

        else:
            # log
            logging.error(
                f'ML models could not be deleted, summary="{json.dumps(req_summary, indent=4)}"'
            )

            # audit
            try:
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "failure",
                    "delete ML models",
                    str(object_value),
                    "splk-dsm",
                    str(json.dumps(req_summary, indent=1)),
                    "ML models could not be deleted",
                    str(update_comment),
                )
            except Exception as e:
                logging.error(
                    f'failed to generate an audit event with exception="{str(e)}"'
                )

            return {"payload": req_summary, "status": 500}

    # update one or more ML models
    def post_outliers_update_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None
        outliers_rules = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]
                outliers_rules = resp_dict["outliers_rules"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint updates existing ML models rules, it requires a POST call with the following options:",
                "resource_desc": "Update a Machine Learning outliers model",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_update_models\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall','outliers_rules':'<redacted_json_dict>'}",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name",
                        "outliers_rules": "The JSON array object containing the outliers rules",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": f"splk-{component}",
                        "object": object_value,
                    }
                ]
            }

            entity_rules = collection.data.query(query=json.dumps(query_string))[0]
            key = entity_rules.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            logging.debug(entity_rules)

            # Load as a dict
            try:
                entities_outliers = json.loads(entity_rules.get("entities_outliers"))
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # log debug
            logging.debug(
                f'before update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            #
            # Process update
            #

            # load the JSON update as a dict
            logging.debug(f'outliers_rules="{outliers_rules}"')

            for outliers_rule in outliers_rules:
                # log debug
                logging.debug(f'outliers_rule_update="{outliers_rule}"')

                # get the model_id
                model_id = outliers_rule.get("model_id")

                # log debug
                logging.debug(f'Handling model_id="{model_id}"')

                # Update the main dict
                fields_list = [
                    "kpi_metric",
                    "kpi_span",
                    "method_calculation",
                    "period_calculation",
                    "time_factor",
                    "density_lowerthreshold",
                    "density_upperthreshold",
                    "alert_lower_breached",
                    "alert_upper_breached",
                    "auto_correct",
                    "perc_min_lowerbound_deviation",
                    "perc_min_upperbound_deviation",
                    "is_disabled",
                ]
                for field in fields_list:
                    if field in ("time_factor"):
                        entities_outliers[model_id][field] = urllib.parse.unquote(
                            outliers_rule.get(field)
                        )
                    else:
                        entities_outliers[model_id][field] = outliers_rule.get(field)
            # log debug
            logging.debug(
                f'after update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            # Finally, update the KVstore record
            try:
                # Update the record
                entity_rules["entities_outliers"] = json.dumps(
                    entities_outliers, indent=4
                )
                collection.data.update(str(key), json.dumps(entity_rules))

                # final
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "ML models were successfully updated",
                    "failures_count": 0,
                    "entities_outliers": entities_outliers,
                }

                # log
                logging.info(json.dumps(result_record, indent=4))

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "update ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(result_record, indent=1)),
                        "ML models were updated successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": result_record, "status": 200}

            except Exception as e:
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "Failed to update the KVstore record",
                    "failures_count": 1,
                    "exception": str(e),
                }
                logging.error(json.dumps(result_record, indent=4))
                return {"payload": result_record, "status": 500}

        else:
            msg = "No record found for this object"
            logging.error(msg)
            return {"payload": msg, "status": 404}

    # Add an exclusion period to the ML model
    def post_outliers_manage_model_period_exclusion(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                # required in all cases
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]
                model_id = resp_dict["model_id"]

                # for cim only
                cim_field = resp_dict.get("cim_field", None)
                # if component is cim, cim_field is required
                if component == "cim" and not cim_field:
                    msg = f'component="{component}" requires cim_field to be defined'
                    logging.error(msg)
                    return {"payload": msg, "status": 500}

                action = resp_dict["action"]
                # action value can be add or delete
                if not action in ("add", "delete", "show"):
                    msg = f'action="{action}" is not valid, it must be either "add", "delete" or "show"'
                    logging.error(msg)
                    return {"payload": msg, "status": 500}

                # required for delete, this can be given as a list of comma separated values, check if it is a list and otherwise turn it as a proper list
                period_exclusion_id = resp_dict.get("period_exclusion_id", None)
                if period_exclusion_id:
                    if not isinstance(period_exclusion_id, list):
                        period_exclusion_id = period_exclusion_id.split(",")

                # required for addition
                earliest = resp_dict.get("earliest", None)
                latest = resp_dict.get("latest", None)

                # if action is delete, the period_exclude_id is required
                if action == "delete":
                    if not period_exclusion_id:
                        msg = f'action="{action}" requires period_exclusion_id to be defined'
                        logging.error(msg)
                        return {"payload": msg, "status": 500}

                # if action is add, earliest and latest are required
                if action == "add":
                    if not earliest or not latest:
                        msg = f'action="{action}" requires earliest and latest to be defined'
                        logging.error(msg)
                        return {"payload": msg, "status": 500}
                    # earliest and latest can be provided as epochtime, or date string in the format %Y-%m-%dT%H:%M, check if provided
                    # as epochtime or date string, if dat string attempt to convrt to epochtime using datetime.datetime.strptime
                    try:
                        earliest = int(earliest)
                    except Exception as e:
                        try:
                            logging.debug(
                                f'trying to parse as datetime, earliest="{earliest}"'
                            )
                            earliest_dt = datetime.strptime(
                                str(earliest), "%Y-%m-%dT%H:%M"
                            )
                            earliest = int(round(float(earliest_dt.timestamp())))

                        except Exception as e:
                            msg = f'action="{action}" requires earliest to be defined as epochtime or date string in the format %Y-%m-%dT%H:%M, parsing as date failed with exception={str(e)}'
                            logging.error(msg)
                            return {"payload": msg, "status": 500}
                    try:
                        latest = int(latest)
                    except Exception as e:
                        try:
                            logging.debug(
                                f'trying to parse as datetime, latest="{latest}"'
                            )
                            latest_dt = datetime.strptime(str(latest), "%Y-%m-%dT%H:%M")
                            latest = int(round(float(latest_dt.timestamp())))
                        except Exception as e:
                            msg = f'action="{action}" requires latest to be defined as epochtime or date string in the format %Y-%m-%dT%H:%M, parsing as date failed with exception={str(e)}'
                            logging.error(msg)
                            return {"payload": msg, "status": 500}

                    # also verify that earliest is not lower than latest, the earliest epochtime cannot be before the latest epochtime
                    if not earliest < latest:
                        msg = f'action="{action}" requires earliest to be before latest, earliest="{earliest}", latest="{latest}"'
                        logging.error(msg)
                        return {"payload": msg, "status": 500}

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint adds a period of exclusion for a given ML model, it requires a POST call with the following options:",
                "resource_desc": "Add or delete an exclusion period to a given ML model",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_add_model_period_exclusion\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall', 'model_id': 'The model identifier', 'action': 'add', 'earliest': '<epoch earliest', 'latest': '<epoch latest>'}",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name",
                        "action": "add, delete, show",
                        "model_id": "The model identifier",
                        "earliest": "The earliest time to be excluded (for add), in epochtime or date string in the format %Y-%m-%dT%H:%M",
                        "latest": "The latest time to be excluded (for add), in epochtime or date string in the format %Y-%m-%dT%H:%M",
                        "period_exclusion_id": "The period exclusion identifier, can be provided as a comma separated list of values (for action delete)",
                        "cim_field": "Required for splk-cim only, the cim_field value",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": f"splk-{component}",
                        "object": object_value,
                    }
                ]
            }

            # for cim, add the cim_field to the query
            if component == "cim":
                query_string["$and"].append({"cim_field": cim_field})

            object_rules_definition = collection.data.query(
                query=json.dumps(query_string)
            )[0]
            key = object_rules_definition.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            logging.debug(object_rules_definition)

            # Load as a dict
            try:
                entities_outliers = json.loads(
                    object_rules_definition.get("entities_outliers")
                )
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)
                return {"payload": msg, "status": 500}

            # log debug
            logging.debug(
                f'before update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            #
            # Process update
            #

            # load the JSON update as a dict
            logging.debug(
                f'entities_outliers="{json.dumps(entities_outliers, indent=2)}"'
            )

            # boolean to check if model_id exists
            model_id_exists = False

            for entity_model_id in entities_outliers:
                logging.debug(f'model_id="{entity_model_id}"')

                entity_rules = entities_outliers[entity_model_id]
                logging.debug(f'entity_rules="{json.dumps(entity_rules, indent=2)}"')

                # if the model_id does not match, break
                if entity_model_id != model_id:
                    continue
                else:
                    model_id_exists = True

                # log debug
                logging.debug(f'Handling model_id="{entity_model_id}"')

                # Get the current period_exclusions record (list)
                period_exclusions = entity_rules.get("period_exclusions", [])

                # if action is add
                if action == "add":
                    # add our object to the list

                    # period_exclusion_id, generate the md5 hash of earliest:latest, use this to detect if the period was excluded already
                    period_exclusion_id = hashlib.md5(
                        f"{earliest}:{latest}".encode()
                    ).hexdigest()

                    # boolean
                    period_exclusion_id_exists = False

                    if len(period_exclusions) > 0:
                        for period_exclusion in period_exclusions:
                            if (
                                period_exclusion["period_exclusion_id"]
                                == period_exclusion_id
                            ):
                                period_exclusion_id_exists = True
                                logging.info(
                                    f'period_exclusion_id="{period_exclusion_id}" already exists, skipping'
                                )
                                continue

                    if not period_exclusion_id_exists:
                        # add the new period
                        period_exclusions.append(
                            {
                                "period_exclusion_id": period_exclusion_id,
                                "earliest": earliest,
                                "earliest_human": time.strftime(
                                    "%c", time.localtime(float(earliest))
                                ),
                                "latest": latest,
                                "latest_human": time.strftime(
                                    "%c", time.localtime(float(latest))
                                ),
                                "ctime": str(round(time.time(), 0)),
                            }
                        )

                # if action is delete, verify the period_exclusion_id exists in the list, if so delete it
                elif action == "delete":
                    # delete the period_exclusion_id from the list
                    for item in period_exclusion_id:
                        logging.debug(f'checking item: "{item}"')
                        for period_exclusion in period_exclusions:
                            if period_exclusion["period_exclusion_id"] == item.strip():
                                logging.debug(f'deleting item: "{item}"')
                                period_exclusions.remove(period_exclusion)

                # if action is show, return the period_exclusions list
                elif action == "show":
                    response_list = []
                    # loop through the period_exclusions, add each as a dict to the response_list (model_id, period_exclusion_id, earliest, latest))
                    for period_exclusion in period_exclusions:
                        response_list.append(
                            {
                                "model_id": model_id,
                                "period_exclusion_id": period_exclusion.get(
                                    "period_exclusion_id"
                                ),
                                "earliest": period_exclusion.get("earliest"),
                                "earliest_human": period_exclusion.get(
                                    "earliest_human"
                                ),
                                "latest": period_exclusion.get("latest"),
                                "latest_human": period_exclusion.get("latest_human"),
                                "ctime": period_exclusion.get("ctime"),
                                "ctime_human": time.strftime(
                                    "%c",
                                    time.localtime(
                                        float(period_exclusion.get("ctime"))
                                    ),
                                ),
                            }
                        )
                    return {
                        "payload": response_list,
                        "status": 200,
                    }

                # update the entity_rule record
                entity_rules["period_exclusions"] = period_exclusions

                # uodate the entities_outliers dict
                entities_outliers[model_id] = entity_rules

                # log
                logging.debug(
                    f'post update entities_outliers="{json.dumps(entities_outliers, indent=2)}"'
                )

            # if model_id does not exist, return a 500 payload
            if not model_id_exists:
                error_msg = (
                    f'The model_id="{model_id}" does not exist in the outliers rules'
                )
                logging.error(error_msg)
                return {"payload": error_msg, "status": 500}

            # log debug
            logging.debug(
                f'after update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
            )

            # update the object record
            object_rules_definition["entities_outliers"] = json.dumps(
                entities_outliers, indent=2
            )

            # Finally, update the KVstore record
            try:
                # Update the record
                collection.data.update(str(key), json.dumps(object_rules_definition))

                # final
                if action == "add":
                    msg = "Exclusion period for ML model was successfully added"
                    msg_audit_title = "add ML model exclusion period"
                elif action == "delete":
                    msg = "Exclusion period for ML model was successfully deleted"
                    msg_audit_title = "delete ML model exclusion period"

                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": msg,
                    "failures_count": 0,
                    "entities_outliers": entities_outliers,
                }

                # log
                logging.info(json.dumps(result_record, indent=4))

                # audit

                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        msg_audit_title,
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(result_record, indent=1)),
                        "ML models were updated successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": result_record, "status": 200}

            except Exception as e:
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "Failed to update the KVstore record",
                    "failures_count": 1,
                    "exception": str(e),
                }
                logging.error(json.dumps(result_record, indent=4))
                return {"payload": result_record, "status": 500}

        else:
            msg = "No record found for this object"
            logging.error(msg)
            return {"payload": msg, "status": 404}

    # update one or more ML models, spl-cim has multiple levels of objects (object / entity / cim_field)
    def post_outliers_cim_update_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        object_value = None
        outliers_rules = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                outliers_rules = resp_dict["outliers_rules"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint updates existing ML models rules, it requires a POST call with the following options:",
                "resource_desc": "Update a Machine Learning outliers model",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_cim_update_models\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall','outliers_rules':'<redacted_json_dict>'}",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "entity object name",
                        "outliers_rules": "The JSON array object containing the outliers rules",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = f"kv_trackme_cim_outliers_entity_rules_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # records summary
        records = []

        # Loop through the dict
        logging.debug(
            f'submitted outliers_rules="{json.dumps(outliers_rules, indent=2)}"'
        )

        #
        # loop in changes
        #

        for p in outliers_rules:
            model_id = p["model_id"]

            # Get the current KVsotre record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            try:
                # Define the KV query
                query_string = {
                    "$and": [
                        {
                            "object_category": "splk-cim",
                            "object": object_value,
                            "entity": p["entity"],
                            "cim_field": p["cim_field"],
                        }
                    ]
                }

                entity_rules = collection.data.query(query=json.dumps(query_string))[0]
                key = entity_rules.get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                logging.debug(entity_rules)

                # Load as a dict
                try:
                    entities_outliers = json.loads(
                        entity_rules.get("entities_outliers")
                    )
                except Exception as e:
                    msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                    logging.error(msg)
                    return {"payload": msg, "status": 500}

                # log debug
                logging.debug(
                    f'before update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
                )

                #
                # Process update
                #

                # load the JSON update as a dict
                logging.debug(
                    f'outliers_rules="{json.dumps(outliers_rules, indent=2)}"'
                )

                # update that model

                # log debug
                logging.debug(f'Handling model_id="{p["model_id"]}"')

                # Update the main dict
                fields_list = [
                    "kpi_metric",
                    "kpi_span",
                    "method_calculation",
                    "period_calculation",
                    "time_factor",
                    "density_lowerthreshold",
                    "density_upperthreshold",
                    "alert_lower_breached",
                    "alert_upper_breached",
                    "auto_correct",
                    "perc_min_lowerbound_deviation",
                    "perc_min_upperbound_deviation",
                    "is_disabled",
                ]
                for field in fields_list:
                    if field in ("time_factor"):
                        entities_outliers[model_id][field] = urllib.parse.unquote(
                            p[field]
                        )
                    else:
                        entities_outliers[model_id][field] = p[field]

                # log debug
                logging.debug(
                    f'after update, entities_outliers="{json.dumps(entities_outliers, indent=4)}"'
                )

                # Finally, update the KVstore record
                try:
                    # Update the record
                    entity_rules["entities_outliers"] = json.dumps(
                        entities_outliers, indent=4
                    )
                    collection.data.update(str(key), json.dumps(entity_rules))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1
                    failures_count += 0

                    # append for summary
                    result = {
                        "object": object_value,
                        "entity": p["entity"],
                        "cim_field": p["cim_field"],
                        "action": "update",
                        "result": "success",
                        "message": f'tenant_id="{tenant_id}", The entity was successfully updated',
                    }
                    records.append(result)

                    # log
                    logging.info(json.dumps(records, indent=4))

                    # audit
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "update ML models",
                            str(object_value),
                            "splk-dsm",
                            str(json.dumps(records, indent=1)),
                            "ML models were updated successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    msg = f'tenant_id="{tenant_id}", failed to update the entity, object="{object_value}", exception="{str(e)}"'
                    result = {
                        "object": object_value,
                        "entity": p["entity"],
                        "cim_field": p["cim_field"],
                        "action": "update",
                        "result": "failure",
                        "exception": msg,
                    }
                    records.append(result)
                    logging.error(msg)

                    # audit
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "update ML models",
                            str(object_value),
                            "splk-dsm",
                            str(json.dumps(records, indent=1)),
                            msg,
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            f'failed to generate an audit event with exception="{str(e)}"'
                        )

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                msg = f'tenant_id="{tenant_id}", failed to update the entity, object="{object_value}", exception="ressource not found"'
                result = {
                    "object": object_value,
                    "entity": p["entity"],
                    "cim_field": p["cim_field"],
                    "action": "update",
                    "result": "failure",
                    "exception": f'tenant_id="{tenant_id}", failed to update the entity, object="{object_value}", exception="ressource not found"',
                }
                records.append(result)
                logging.error(msg)

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "update ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(records, indent=1)),
                        msg,
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            logging.info(
                f'ML models bulk edit was successfull, results="{json.dumps(req_summary, indent=1)}"'
            )
            return {"payload": req_summary, "status": 200}

        else:
            logging.error(
                f'ML models bulk edit has failed, results="{json.dumps(req_summary, indent=1)}"'
            )
            return {"payload": req_summary, "status": 500}

    # reset ML models
    def post_outliers_reset_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint resets ML models rules, it requires a POST call with the following options:",
                "resource_desc": "Reset all ML outliers models for a given entity",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_reset_models\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # entity rules collection
        collection_entity_rules_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection_entity_rules = service.kvstore[collection_entity_rules_name]

        # data rules collection
        collection_entity_data_name = (
            f"kv_trackme_{component}_outliers_entity_data_tenant_{tenant_id}"
        )
        collection_entity_data = service.kvstore[collection_entity_data_name]

        # Define the KV query
        query_string = {
            "$and": [
                {
                    "object_category": f"splk-{component}",
                    "object": object_value,
                }
            ]
        }

        # get the entity_rules record
        try:
            kvrecord_entity_rules = collection_entity_rules.data.query(
                query=json.dumps(query_string)
            )[0]
            kvrecord_entity_rules_key = kvrecord_entity_rules.get("_key")
        except Exception as e:
            kvrecord_entity_rules_key = None

        # get the entity_data record
        try:
            kvrecord_entity_data = collection_entity_data.data.query(
                query=json.dumps(query_string)
            )[0]
            kvrecord_entity_data_key = kvrecord_entity_data.get("_key")
        except Exception as e:
            kvrecord_entity_data_key = None

        #
        # proceed
        #

        # first, reset the entity_data record, if any
        kvrecord_entity_data_deleted = False
        if kvrecord_entity_data_key:
            try:
                collection_entity_data.data.delete(
                    json.dumps({"_key": kvrecord_entity_data_key})
                )
                logging.info(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_rules_name}", deleted key="{kvrecord_entity_data_key}"'
                )
                kvrecord_entity_data_deleted = True
            except Exception as e:
                logging.error(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_rules_name}", failed to deleted key="{kvrecord_entity_data_key}" with exception="{str(e)}"'
                )

        # secondly, purge and re-generate ML models
        kvrecord_entity_rules_deleted = False
        if kvrecord_entity_rules_key:
            #
            # Process reset
            #

            try:
                collection_entity_rules.data.delete(
                    json.dumps({"_key": kvrecord_entity_rules_key})
                )
                logging.info(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_data_name}", deleted key="{kvrecord_entity_rules_key}"'
                )
                kvrecord_entity_rules_deleted = True
            except Exception as e:
                logging.error(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_data_name}", failed to deleted key="{kvrecord_entity_rules_key}" with exception="{str(e)}"'
                )
                kvrecord_entity_rules_deleted = False

                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "Failed to delete the KVstore record this entity, cannot proceed to reset",
                    "failures_count": 1,
                    "exception": str(e),
                }
                logging.error(json.dumps(result_record, indent=4))
                return {"payload": result_record, "status": 500}

        else:
            logging.info(
                f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", no KVstore record was found to be deleted'
            )

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }

        # Search query reset
        searchquery_reset = remove_leading_spaces(
            f"""\
            | makeresults | head 1
            | eval tenant_id="{tenant_id}", object_category="splk-{component}", object="{object_value}", key="{kvrecord_entity_rules_key}"
            | trackmesplkoutlierssetrules tenant_id="{tenant_id}" component="{component}"
            """
        )

        # log debug
        logging.debug(
            f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", searchquery_reset="{searchquery_reset}"'
        )

        # automatic retry
        max_failures_count = 24
        sleep_time = 5
        search_successful = False
        current_failures_count = 0
        while current_failures_count < max_failures_count and not search_successful:
            reset_results = []
            try:
                # spawn the search and get the results
                searchresults_reset = service.jobs.oneshot(
                    searchquery_reset, **kwargs_search
                )
                reader = results.JSONResultsReader(searchresults_reset)
                for item in reader:
                    if isinstance(item, dict):
                        reset_results.append(item)

                # final
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "ML models were successfully reset",
                    "failures_count": 0,
                    "reset_results": reset_results,
                    "searchquery_reset": searchquery_reset,
                    "entity_rules_deleted": kvrecord_entity_rules_deleted,
                    "entity_data_deleted": kvrecord_entity_data_deleted,
                }

                # log
                logging.info(json.dumps(result_record, indent=4))

                # break while
                search_successful = True

                # audit
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "reset ML models",
                    str(object_value),
                    "splk-dsm",
                    str(json.dumps(result_record, indent=1)),
                    "ML models were reset successfully",
                    str(update_comment),
                )

                return {"payload": result_record, "status": 200}

            except Exception as e:
                if re.search(
                    "The maximum number of concurrent historical searches on this instance has been reached",
                    str(e),
                ):
                    current_failures_count += 1
                    search_successful = False
                    if current_failures_count >= max_failures_count:
                        # permanent failure
                        result_record = {
                            "tenant_id": tenant_id,
                            "object_category": f"splk-{component}",
                            "object": object_value,
                            "results": "Failed to reset ML models for this entity",
                            "failures_count": 1,
                            "entity_rules_deleted": kvrecord_entity_rules_deleted,
                            "entity_data_deleted": kvrecord_entity_data_deleted,
                            "exception": str(e),
                        }
                        logging.error(json.dumps(result_record, indent=4))
                        return {"payload": result_record, "status": 500}

                    # temporary failure
                    else:
                        logging.warn(
                            f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", temporary search failure, exception="{str(e)}"'
                        )
                        time.sleep(sleep_time)

                else:
                    # permanent failure
                    result_record = {
                        "tenant_id": tenant_id,
                        "object_category": f"splk-{component}",
                        "object": object_value,
                        "results": "Failed to reset ML models for this entity",
                        "failures_count": 1,
                        "entity_rules_deleted": kvrecord_entity_rules_deleted,
                        "entity_data_deleted": kvrecord_entity_data_deleted,
                        "exception": str(e),
                    }
                    logging.error(json.dumps(result_record, indent=4))
                    return {"payload": result_record, "status": 500}

    # reset and train ML models for cim
    def post_outliers_cim_reset_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        object_value = None
        component = "cim"

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint resets ML models rules, it requires a POST call with the following options:",
                "resource_desc": "Reset all ML outliers models for a given entity",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_cim_reset_models\" body=\"{'tenant_id':'cim-demo','object':'auth001'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "entity name",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # entity rules collection
        collection_entity_rules_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection_entity_rules = service.kvstore[collection_entity_rules_name]

        # data rules collection
        collection_entity_data_name = (
            f"kv_trackme_{component}_outliers_entity_data_tenant_{tenant_id}"
        )
        collection_entity_data = service.kvstore[collection_entity_data_name]

        # Define the KV query
        query_string = {
            "$and": [
                {
                    "object_category": f"splk-{component}",
                    "object": object_value,
                }
            ]
        }

        # get the entity_rules record
        try:
            kvrecord_entity_rules = collection_entity_rules.data.query(
                query=json.dumps(query_string)
            )[0]
            kvrecord_entity_rules_key = kvrecord_entity_rules.get("_key")
        except Exception as e:
            kvrecord_entity_rules_key = None

        # get the entity_data record
        try:
            kvrecord_entity_data = collection_entity_data.data.query(
                query=json.dumps(query_string)
            )[0]
            kvrecord_entity_data_key = kvrecord_entity_data.get("_key")
        except Exception as e:
            kvrecord_entity_data_key = None

        #
        # proceed
        #

        # first, reset the entity_data record, if any
        kvrecord_entity_data_deleted = False
        if kvrecord_entity_data_key:
            try:
                collection_entity_data.data.delete(
                    json.dumps({"_key": kvrecord_entity_data_key})
                )
                logging.info(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_rules_name}", deleted key="{kvrecord_entity_data_key}"'
                )
                kvrecord_entity_data_deleted = True
            except Exception as e:
                logging.error(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_rules_name}", failed to deleted key="{kvrecord_entity_data_key}" with exception="{str(e)}"'
                )

        # secondly, purge and re-generate ML models
        kvrecord_entity_rules_deleted = False
        if kvrecord_entity_rules_key:
            #
            # Process reset
            #

            try:
                collection_entity_rules.data.delete(
                    json.dumps({"_key": kvrecord_entity_rules_key})
                )
                logging.info(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_data_name}", deleted key="{kvrecord_entity_rules_key}"'
                )
                kvrecord_entity_rules_deleted = True
            except Exception as e:
                logging.error(
                    f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", collection="{collection_entity_data_name}", failed to deleted key="{kvrecord_entity_rules_key}" with exception="{str(e)}"'
                )
                kvrecord_entity_rules_deleted = False

                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "Failed to delete the KVstore record this entity, cannot proceed to reset",
                    "failures_count": 1,
                    "exception": str(e),
                }
                logging.error(json.dumps(result_record, indent=4))
                return {"payload": result_record, "status": 500}

        else:
            logging.info(
                f'tenant_id="{tenant_id}", component="splk-cim", object="{object_value}", no KVstore record was found to be deleted'
            )

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }

        #
        # reset models
        #

        # Search query reset
        searchquery_reset = f'| trackmesplkoutlierscimsetrules tenant_id="{tenant_id}" object="{object_value}'

        # log debug
        logging.debug(
            f'tenant_id="{tenant_id}", component="splk-cim", object="{object_value}", searchquery_reset="{searchquery_reset}"'
        )

        reset_results = []
        try:
            # spawn the search and get the results
            searchresults_reset = service.jobs.oneshot(
                searchquery_reset, **kwargs_search
            )
            reader = results.JSONResultsReader(searchresults_reset)
            for item in reader:
                if isinstance(item, dict):
                    reset_results.append(item)

            # final
            reset_result_record = {
                "tenant_id": tenant_id,
                "object_category": "splk-cim",
                "object": object_value,
                "results": "ML models were successfully reset",
                "failures_count": 0,
                "reset_results": reset_results,
                "searchquery_reset": searchquery_reset,
                "entity_rules_deleted": kvrecord_entity_rules_deleted,
                "entity_data_deleted": kvrecord_entity_data_deleted,
            }

            # log
            logging.info(json.dumps(reset_result_record, indent=4))

            # audit
            try:
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "reset ML models",
                    str(object_value),
                    "splk-dsm",
                    str(json.dumps(reset_result_record, indent=1)),
                    "ML models were reset successfully",
                    str(update_comment),
                )
            except Exception as e:
                logging.error(
                    f'failed to generate an audit event with exception="{str(e)}"'
                )

        except Exception as e:
            reset_result_record = {
                "tenant_id": tenant_id,
                "object_category": "splk-cim",
                "object": object_value,
                "results": "Failed to reset ML models for this entity",
                "failures_count": 1,
                "entity_rules_deleted": kvrecord_entity_rules_deleted,
                "entity_data_deleted": kvrecord_entity_data_deleted,
                "exception": str(e),
            }
            logging.error(json.dumps(reset_result_record, indent=4))

            return {"payload": reset_result_record, "status": 500}

        #
        # Train models
        #

        # Search query train
        searchquery_train = f'| trackmesplkoutlierscimtrain tenant_id="{tenant_id}" component="cim" object="{object_value}" cim_field=*'

        # log debug
        logging.debug(
            f'tenant_id="{tenant_id}", component="cim", object="{object_value}", searchquery_train="{searchquery_train}"'
        )

        train_results = []

        try:
            # spawn the search and get the results
            searchresults_train = service.jobs.oneshot(
                searchquery_train, **kwargs_search
            )
            reader = results.JSONResultsReader(searchresults_train)
            for item in reader:
                if isinstance(item, dict):
                    train_results.append(item)

            # final
            train_results_record = {
                "tenant_id": tenant_id,
                "object_category": "splk-cim",
                "object": object_value,
                "results": "ML models were successfully trained",
                "failures_count": 0,
                "train_results": train_results,
                "searchquery_train": searchquery_train,
            }

            # log
            logging.info(json.dumps(train_results_record, indent=4))

            # audit
            try:
                trackme_audit_event(
                    request_info.system_authtoken,
                    request_info.server_rest_uri,
                    tenant_id,
                    request_info.user,
                    "success",
                    "train ML models",
                    str(object_value),
                    "splk-dsm",
                    str(json.dumps(train_results_record, indent=1)),
                    "ML models were trained successfully",
                    str(update_comment),
                )
            except Exception as e:
                logging.error(
                    f'failed to generate an audit event with exception="{str(e)}"'
                )

        except Exception as e:
            train_results_record = {
                "tenant_id": tenant_id,
                "object_category": "splk-cim",
                "object": object_value,
                "results": "Failed to train ML models for this entity",
                "failures_count": 1,
                "exception": str(e),
            }
            logging.error(json.dumps(train_results_record, indent=4))
            return {"payload": train_results_record, "status": 500}

        # return if successful
        return {
            "payload": {
                "tenant_id": tenant_id,
                "object": object_value,
                "result": "success",
                "failures_count": 0,
                "results": "ML models were successfully reset and trained for this entity",
                "reset_results": reset_result_record,
                "train_results": train_results_record,
            },
            "status": 200,
        }

    # train ML models
    def post_outliers_train_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint trains ML models rules, it requires a POST call with the following options:",
                "resource_desc": "Train all ML outliers models for a given entity",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_train_models\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Data collection
        collection_name = (
            f"kv_trackme_{component}_outliers_entity_rules_tenant_{tenant_id}"
        )
        collection = service.kvstore[collection_name]

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            # Define the KV query
            query_string = {
                "$and": [
                    {
                        "object_category": f"splk-{component}",
                        "object": object_value,
                    }
                ]
            }

            entity_rules = collection.data.query(query=json.dumps(query_string))[0]
            key = entity_rules.get("_key")

        except Exception as e:
            key = None

        # Render result
        if key:
            # Define the SPL query
            kwargs_search = {
                "app": "trackme",
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }
            searchquery = f'| trackmesplkoutlierstrain tenant_id="{tenant_id}" component="{component}" object="{object_value}"'
            logging.debug(
                f'tenant_id="{tenant_id}", component="{component}", object="{object_value}", searchquery="{searchquery}"'
            )

            query_results = []
            try:
                # spawn the search and get the results
                searchresults = service.jobs.oneshot(searchquery, **kwargs_search)
                reader = results.JSONResultsReader(searchresults)
                for item in reader:
                    if isinstance(item, dict):
                        query_results.append(item)

                # final
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "ML models were successfully trained",
                    "failures_count": 0,
                    "query_results": query_results,
                }

                # log
                logging.info(json.dumps(result_record, indent=4))

                # audit
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "train ML models",
                        str(object_value),
                        "splk-dsm",
                        str(json.dumps(result_record, indent=1)),
                        "ML models were train successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        f'failed to generate an audit event with exception="{str(e)}"'
                    )

                return {"payload": result_record, "status": 200}

            except Exception as e:
                result_record = {
                    "tenant_id": tenant_id,
                    "object_category": f"splk-{component}",
                    "object": object_value,
                    "results": "Failed to train ML models for this entity",
                    "failures_count": 1,
                    "exception": str(e),
                }
                logging.error(json.dumps(result_record, indent=4))
                return {"payload": result_record, "status": 500}

    # Force mlmonitor execution for a given object
    def post_outliers_mlmonitor_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        component = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                component = resp_dict["component"]
                object_value = resp_dict["object"]
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint runs the execution of the ML monitor backend process, it requires a POST call with the following options:",
                "resource_desc": "Runs Machine Learning Outliers monitor process for a given entity",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_mlmonitor_models\" body=\"{'tenant_id':'mytenant','component':'dsm','object':'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "the component, valid options for this endpoint are: dsm|dhm|flx",
                        "object": "the object identifier for the entity to be executed",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = f'| trackmesplkoutlierstrackerhelper tenant_id="{tenant_id}" component="{component}" object="{object_value}" force_run="True"'

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.oneshot(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            if len(query_results) > 0:
                return {
                    "payload": {
                        "action": "success",
                        "upstream_query": searchquery,
                        "query_results": query_results,
                    },
                    "status": 200,
                }
            else:
                return {
                    "payload": {
                        "action": "failure",
                        "upstream_query": searchquery,
                        "query_results": "The search was executed successfully, but no results were returned.",
                    },
                    "status": 500,
                }

        except Exception as e:
            response = {
                "action": "failure",
                "response": f'an exception was encountered, exception="{str(e)}"',
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Force mlmonitor execution for a given object
    def post_outliers_cim_mlmonitor_models(self, request_info, **kwargs):
        describe = False
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint runs the execution of the ML monitor backend process for a CIM entity, it requires a POST call with the following options:",
                "resource_desc": "Runs Machine Learning Outliers monitor process for a given CIM entity",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/write/outliers_cim_mlmonitor_models\" body=\"{'tenant_id':'mytenant', 'object':'network_traffic:pan_firewall'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "the object identifier for the entity to be executed",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = f'| trackmesplkoutlierscimtrackerhelper tenant_id="{tenant_id}" object="{object_value}" force_run="True"'

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.oneshot(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            if len(query_results) > 0:
                return {
                    "payload": {
                        "action": "success",
                        "upstream_query": searchquery,
                        "query_results": query_results,
                    },
                    "status": 200,
                }
            else:
                return {
                    "payload": {
                        "action": "failure",
                        "upstream_query": searchquery,
                        "query_results": "The search was executed successfully, but no results were returned.",
                    },
                    "status": 500,
                }

        except Exception as e:
            response = {
                "action": "failure",
                "response": f'an exception was encountered, exception="{str(e)}"',
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}
