from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_splk_dhm.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel

# import Splunk libs
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerSplkOutliersEngine_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkOutliersEngine_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_outliers_engine(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_outliers_engine",
            "resource_group_desc": "Endpoints related to the management of the Machine Learning Outliers detection (read only operations)",
        }

        return {"payload": response, "status": 200}

    # get Machine Learning models and rules
    def post_outliers_get_rules(self, request_info, **kwargs):
        """
        | trackme mode=post url="/services/trackme/v2/splk_outliers_engine/outliers_get_rules" body="{'tenant_id': 'mytenant', 'component': 'dsm', 'object': '*'}"
        """

        describe = False
        tenant_id = None
        component = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                component = resp_dict["component"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint retrieves the Machine Learning outliers models and rules, it requires a POST call with the following options:",
                "resource_desc": "Get Machine Learning outliers models and rules",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/outliers_get_rules\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "component": "The component category",
                        "object": "entity name, use a * to match all entities",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = (
            '| trackmesplkoutliersgetrules tenant_id="'
            + str(tenant_id)
            + '" component="'
            + str(component)
            + '" object="'
            + str(object_value)
            + '"'
        )

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.oneshot(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            return {"payload": query_results, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # get Machine Learning models and rules
    def post_outliers_cim_get_rules(self, request_info, **kwargs):
        """
        | trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/outliers_cim_get_rules\" body=\"{'tenant_id': 'mytenant', 'object': '*', 'entity': '*', 'cim_field':'*'}\"
        """

        describe = False
        tenant_id = None
        object_value = None
        entity_name = None
        cim_field = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]
                entity_name = resp_dict["entity"]
                cim_field = resp_dict["cim_field"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint retrieves the Machine Learning outliers models and rules, it requires a POST call with the following options:",
                "resource_desc": "Get Machine Learning outliers models and rules",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_outliers_engine/outliers_cim_get_rules\" body=\"{'tenant_id': 'mytenant', 'object': '*', 'entity': '*', 'cim_field':'*'}\"",
                "options": [
                    {
                        "tenant_id": "tenant identifier",
                        "object": "object name, use a * to match all objects",
                        "entity": "cim entity name, use a * to match all entities",
                        "cim_field": "cim field, use a * to match all fields",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }
        searchquery = (
            '| trackmesplkoutlierscimgetrules tenant_id="'
            + str(tenant_id)
            + '" component="cim" '
            + 'object="'
            + str(object_value)
            + '", entity="'
            + str(entity_name)
            + '", cim_field="'
            + str(cim_field)
            + '"'
        )

        query_results = []
        try:
            # spawn the search and get the results
            searchresults = service.jobs.oneshot(searchquery, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)
            return {"payload": query_results, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}
