from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_smart_status.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
import time

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# TrackMe REST handler
import trackme_rest_handler

# TrackMe libs
from trackme_libs import trackme_getloglevel

# TrackMe SmartStatus libs
from trackme_libs_smartstatus import (
    smartstatus_flipping_correlation,
    smartstatus_return_future_tolerance,
    smartstatus_investigations_uc_future,
    smartstatus_investigations_uc_latency,
    smartstatus_investigations_uc_delay,
    smartstatus_investigations_uc_hosts_dcount,
    smartstatus_investigations_uc_ml_outliers,
    smartstatus_investigations_uc_events_format_recognition,
    smartstatus_investigations_uc_wlk_execution_errors,
    smartstatus_investigations_uc_wlk_skipping,
    smartstatus_investigations_uc_wlk_orphan,
    smartstatus_investigations_uc_wlk_delayed,
)

# TrackMe splk-feeds libs
from trackme_libs_splk_feeds import (
    splk_dsm_return_entity_info,
    splk_dsm_return_elastic_info,
)

# Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkSmartStatus_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkSmartStatus_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_smart_status(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_smart_status",
            "resource_group_desc": "Endpoints related to the Smart Status features (automated investigations)",
        }

        return {"payload": response, "status": 200}

    # Get smart status for splk-dsm
    def post_ds_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/ds_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'netscreen:netscreen:firewall'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_dsm_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-dsm", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        # Get future_tolerance
        future_tolerance = smartstatus_return_future_tolerance(request_info)

        # Get entity info
        elastic_info = splk_dsm_return_elastic_info(
            request_info.session_key,
            request_info.server_rest_port,
            tenant_id,
            object_value,
        )
        entity_info = splk_dsm_return_entity_info(record)

        # Check if is an Elastic Source
        is_elastic = elastic_info.get("is_elastic")
        if is_elastic == 1:
            smartstatus_response["elastic_info"] = elastic_info
        else:
            smartstatus_response["entity_info"] = entity_info

        # log debug
        logging.debug(
            'function ds_smart_status, elastic_info="{}"'.format(
                json.dumps(elastic_info, indent=2)
            )
        )
        logging.debug(
            'function ds_smart_status, entity_info="{}"'.format(
                json.dumps(entity_info, indent=2)
            )
        )

        # Get entity account and search_mode
        account = None
        search_mode = None

        if is_elastic == 1:
            account = elastic_info.get("account")
            search_mode = elastic_info.get("search_mode")
        else:
            account = entity_info.get("account")
            search_mode = entity_info.get("search_mode")

        # if is elastic, update the elastic info to the entity info dict
        if is_elastic == 1:
            entity_info["account"] = account
            entity_info["search_mode"] = search_mode
            entity_info["is_elastic"] = 1
            entity_info["search_constraint"] = elastic_info.get("search_constraint")
            entity_info["elastic_search_mode"] = elastic_info.get("elastic_search_mode")
            entity_info.pop("breakby_key", None)
            entity_info.pop("breakby_value", None)
            entity_info.pop("breakby_statement", None)
        else:
            entity_info["is_elastic"] = 0

        # log debug
        logging.debug(
            'function ds_smart_status, entity_info="{}"'.format(
                json.dumps(entity_info, indent=2)
            )
        )

        # add to response
        smartstatus_response["account"] = account
        smartstatus_response["search_mode"] = search_mode

        #################################
        # Greem: happiness is in da house
        #################################

        # if there are anomalies, no need to continue further
        if anomaly_reason == "none":
            smartstatus_response["_time"] = time.time()
            smartstatus_response["tenant_id"] = tenant_id
            smartstatus_response["object_category"] = "splk-dsm"
            smartstatus_response["object"] = object_value
            smartstatus_response["alias"] = record.get("alias")
            smartstatus_response["object_state"] = record.get("object_state")
            smartstatus_response["anomaly_reason"] = anomaly_reason
            smartstatus_response["flipping_correlations"] = flipping_correlation
            smartstatus_response[
                "smart_result"
            ] = "The entity is currently in a normal state, therefore further investigations are not required at this stage."
            smartstatus_response["smart_code"] = 0

            return {"payload": smartstatus_response, "status": 200}

        ################################################
        # use case investigations: data is in the future
        ################################################

        if anomaly_reason == "future_over_tolerance":
            logging.info(
                'performing investigations use case: data in the future, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case search investigates and extracts main information related to the detection of data in the future over the tolerance"

            try:
                uc_results = smartstatus_investigations_uc_future(
                    request_info, future_tolerance, record, entity_info
                )
                # append to response

                if entity_info.get("search_mode") in ("tstats", "raw", "from"):
                    use_case_desc = "The use case search investigates and extracts the top 100 hosts and 100 sources generating data in the future over the tolerance"
                    search1_results_label = "results_hosts_future"
                    search1_info_label = "search_info_hosts_future"
                    search2_results_label = "results_sources_future"
                    search2_info_label = "search_info_sources_future"
                    search3_results_label = "results_sample_events_future"
                    search3_info_label = "search_info_sample_events_future"

                elif entity_info.get("search_mode") in ("mstats"):
                    use_case_desc = "The use case search investigates and extracts the top 100 hosts and 100 metric categories generating metrics in the future over the tolerance"
                    search1_results_label = "results_hosts_future"
                    search1_info_label = "search_info_hosts_future"
                    search2_results_label = "results_metric_categories_future"
                    search2_info_label = "search_info_metric_categories_future"
                    search3_results_label = "results_sample_metrics_future"
                    search3_info_label = "search_info_sample_metrics_future"

                smartstatus_response["smart_investigations_uc_future"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    search2_results_label: uc_results.get("results_2"),
                    search2_info_label: uc_results.get("search_info_2"),
                    search3_results_label: uc_results.get("results_3"),
                    search3_info_label: uc_results.get("search_info_3"),
                    "smart_function": "smartstatus_investigations_uc_future",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_future, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smart_investigations_uc_future"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_future CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_future",
                    "exception": str(e),
                }

        #########################################
        # use case investigations: entity latency
        #########################################

        if anomaly_reason == "lag_threshold_breached":
            logging.info(
                'performing investigations use case: latency, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and generates main indexing latency statistics for the entity"

            try:
                uc_results = smartstatus_investigations_uc_latency(
                    request_info, record, entity_info
                )
                # append to response

                if entity_info.get("search_mode") in ("tstats", "raw", "from"):
                    use_case_desc = "The use case investigates and generates main indexing latency statistics for the entity"
                    search1_results_label = "results_latency"
                    search1_info_label = "search_info_latency"
                    search2_results_label = "results_latency_overtime"
                    search2_results_label_sparkline = (
                        "results_latency_overtime_sparkline"
                    )
                    search2_info_label = "search_info_latency_overtime"
                    search3_results_label = "results_sample_events"
                    search3_info_label = "search_info_sample_events"

                smartstatus_response["smartstatus_investigations_uc_latency"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    search2_results_label: uc_results.get("results_2"),
                    search2_results_label_sparkline: uc_results.get(
                        "results_2_sparkline"
                    ),
                    search2_info_label: uc_results.get("search_info_2"),
                    search3_results_label: uc_results.get("results_3"),
                    search3_info_label: uc_results.get("search_info_3"),
                    "smart_function": "smartstatus_investigations_uc_latency",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_latency, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smart_investigations_uc_latency"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_latency CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_latency",
                    "exception": str(e),
                }

        #######################################
        # use case investigations: entity delay
        #######################################

        if anomaly_reason == "delay_threshold_breached":
            logging.info(
                'performing investigations use case: delay, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and generates delay related statistics for the entity"

            try:
                uc_results = smartstatus_investigations_uc_delay(
                    request_info, record, entity_info
                )
                # append to response

                if entity_info.get("search_mode") in ("tstats", "raw", "from"):
                    use_case_desc = "The use case investigates data flow interruption (delay breached) for the entity"
                    search1_results_label = "results_delay"
                    search1_info_label = "search_info_delay"
                    search2_results_label = "results_delay_metric_overtime"
                    search2_results_label_sparkline = (
                        "results_delay_metric_overtime_sparkline"
                    )
                    search2_info_label = "search_info_delay_metric_overtime"
                    search3_results_label = "results_flip"
                    search3_info_label = "search_info_flip"

                smartstatus_response["smartstatus_investigations_uc_delay"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    search2_results_label: uc_results.get("results_2"),
                    search2_results_label_sparkline: uc_results.get(
                        "results_2_sparkline"
                    ),
                    search2_info_label: uc_results.get("search_info_2"),
                    search3_results_label: uc_results.get("results_3"),
                    search3_info_label: uc_results.get("search_info_3"),
                    "smart_function": "smartstatus_investigations_uc_delay",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_delay, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smart_investigations_uc_delay"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_delay CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_delay",
                    "exception": str(e),
                }

        ########################################################
        # use case investigations: distinct host count threshold
        ########################################################

        if anomaly_reason == "min_hosts_dcount":
            logging.info(
                'performing investigations use case: distinct host count threshold, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and generates statistics related to the minimal distinct count of hosts for the entity"

            try:
                uc_results = smartstatus_investigations_uc_hosts_dcount(
                    request_info, record, entity_info
                )
                # append to response

                if entity_info.get("search_mode") in ("tstats", "raw", "from"):
                    use_case_desc = "The use case investigates minimal hosts distinct count threshold breached for the entity"
                    search1_results_label = "results_dcount_hosts"
                    search1_info_label = "search_info_dcount_hosts"
                    search2_results_label = "results_dcount_hosts_metric_overtime"
                    search2_results_label_sparkline = (
                        "results_dcount_hosts_metric_overtime_sparkline"
                    )
                    search2_info_label = "search_info_dcount_hosts_metric_overtime"
                    search3_results_label = "results_flip"
                    search3_info_label = "search_info_flip"

                smartstatus_response["smartstatus_investigations_uc_hosts_dcount"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    search2_results_label: uc_results.get("results_2"),
                    search2_results_label_sparkline: uc_results.get(
                        "results_2_sparkline"
                    ),
                    search2_info_label: uc_results.get("search_info_2"),
                    search3_results_label: uc_results.get("results_3"),
                    search3_info_label: uc_results.get("search_info_3"),
                    "smart_function": "smartstatus_investigations_uc_hosts_dcount",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_hosts_dcount, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smartstatus_investigations_uc_hosts_dcount"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_hosts_dcount CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_hosts_dcount",
                    "exception": str(e),
                }

        ##############################################################
        # use case investigations: Machine Learning Outliers Detection
        ##############################################################

        if anomaly_reason == "ml_outliers_detection":
            logging.info(
                'performing investigations use case: Machine Learning Outliers, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates ML Outliers for this object by executing the ML monitor process for enabled models"

            try:
                uc_results = smartstatus_investigations_uc_ml_outliers(
                    request_info, record, entity_info
                )
                # append to response

                if entity_info.get("search_mode") in ("tstats", "raw", "from"):
                    use_case_desc = (
                        "The use case investigates ML Outliers detection for the entity"
                    )
                    search1_results_label = "results_ml_outliers"
                    search1_info_label = "search_info_ml_outliers"

                smartstatus_response["smartstatus_investigations_uc_ml_outliers"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_ml_outliers",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_ml_outliers, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smartstatus_investigations_uc_ml_outliers"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_ml_outliers CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_ml_outliers",
                    "exception": str(e),
                }

        ####################################################################
        # use case investigations: Data sampling & events format recognition
        ####################################################################

        if anomaly_reason == "data_sampling_anomaly":
            logging.info(
                'performing investigations use case: Data sampling & events format recognition, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates Data sampling for the entity"

            try:
                uc_results = smartstatus_investigations_uc_events_format_recognition(
                    request_info, record, entity_info
                )

                # append to response
                use_case_desc = "The use case investigates Data sampling for the entity"
                search1_results_label = "results_data_sampling"
                search1_info_label = "search_info_data_sampling"

                smartstatus_response[
                    "smartstatus_investigations_uc_events_format_recognition"
                ] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_events_format_recognition",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_events_format_recognition, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response[
                    "smartstatus_investigations_uc_events_format_recognition"
                ] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_events_format_recognition CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_events_format_recognition",
                    "exception": str(e),
                }

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-dsm"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        return {"payload": smartstatus_response, "status": 200}

    # Get smart status for splk-dhm
    def post_dh_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/dh_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'remote|account:lab|firewall.pan.amer.design.node1'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_dhm_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-dsm", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-dhm"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        return {"payload": smartstatus_response, "status": 200}

    # Get smart status for splk-mhm
    def post_mh_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/mh_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'key:env|splunk'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_mhm_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-dsm", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-mhm"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        return {"payload": smartstatus_response, "status": 200}

    # Get smart status for splk-cim
    def post_cim_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/cim_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'auth001'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_cim_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-dsm", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-cim"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        return {"payload": smartstatus_response, "status": 200}

    # Get smart status for splk-flx
    def post_flx_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/flx_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'Okta:Splunk_TA_okta_identity_cloud:okta_logs'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_flx_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-flx", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-flx"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        return {"payload": smartstatus_response, "status": 200}

    # Get smart status for splk-wlk
    def post_wlk_smart_status(self, request_info, **kwargs):
        # define
        tenant_id = None
        object_value = None

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                tenant_id = resp_dict["tenant_id"]
                object_value = resp_dict["object"]

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoints runs the smart status for a given data source, it requires a POST call with the following options:",
                "resource_desc": "Run Smart Status for a given entity",
                "resource_spl_example": "| trackme mode=\"post\" url=\"/services/trackme/v2/splk_smart_status/wlk_smart_status\" body=\"{'tenant_id': 'mytenant', 'object': 'mysearch'}\"",
                "options": [
                    {
                        "tenant_id": "the tenant identifier",
                        "object": "name of the data source",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Define the KV query
        query_string = {
            "object": object_value,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Get the object record
        try:
            collection_name = "kv_" + "trackme_wlk_tenant_" + str(tenant_id)
            collection = service.kvstore[collection_name]

            records = collection.data.query(query=json.dumps(query_string))
            record = records[0]
            key = record.get("_key")

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'object="{}" not found'.format(object_value),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 404}

        #
        # object is known, proceed
        #

        # set the initial empty smartstatus_response
        smartstatus_response = {}

        # set the initial smart_code value
        smart_code = 0

        # get anomaly_reason
        anomaly_reason = record.get("anomaly_reason")

        if anomaly_reason in ("skipping_searches_detected"):
            smart_code = 1
        elif anomaly_reason in ("orphan_search_detected"):
            smart_code = 2
        elif anomaly_reason in ("execution_errors_detected"):
            smart_code = 3
        elif anomaly_reason in ("anomaly_outliers_detected"):
            smart_code = 4
        elif anomaly_reason in ("execution_delayed"):
            smart_code = 5
        elif anomaly_reason in ("status_not_met"):
            smart_code = 6
        elif anomaly_reason in ("out_of_monitoring_window"):
            smart_code = 7

        # retrieve flipping correlation
        try:
            flipping_correlation = smartstatus_flipping_correlation(
                request_info, tenant_id, "splk-wlk", object_value
            )
        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered with smartstatus_flipping_correlation, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

        # Increment if needed
        if flipping_correlation.get("flipping_correlation_status") == 1:
            smart_code += 1

        ##########
        # response
        ##########

        smartstatus_response["_time"] = time.time()
        smartstatus_response["tenant_id"] = tenant_id
        smartstatus_response["object_category"] = "splk-wlk"
        smartstatus_response["object"] = object_value
        smartstatus_response["alias"] = record.get("alias")
        smartstatus_response["object_state"] = record.get("object_state")
        smartstatus_response["anomaly_reason"] = anomaly_reason
        smartstatus_response["flipping_correlations"] = flipping_correlation
        smartstatus_response["smart_code"] = smart_code

        #
        # UC execution errors
        #

        if anomaly_reason == "execution_errors_detected":
            logging.info(
                'performing investigations use case: execution_errors_detected, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and retrieves Splunk scheduler execution errors for the entity"

            try:
                uc_results = smartstatus_investigations_uc_wlk_execution_errors(
                    request_info, tenant_id, record
                )
                # append to response

                search1_results_label = "results_execution_errors_detected"
                search1_info_label = "search_info_execution_errors_detected"

                smartstatus_response[
                    "smartstatus_investigations_uc_wlk_execution_errors"
                ] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_wlk_execution_errors",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_wlk_execution_errors, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response[
                    "smartstatus_investigations_uc_wlk_execution_errors"
                ] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_wlk_execution_errors CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_wlk_execution_errors",
                    "exception": str(e),
                }

        if anomaly_reason == "skipping_searches_detected":
            logging.info(
                'performing investigations use case: skipping_searches_detected, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and retrieves Splunk scheduler skipping events for the entity"

            try:
                uc_results = smartstatus_investigations_uc_wlk_skipping(
                    request_info, tenant_id, record
                )
                # append to response

                search1_results_label = "results_skipping_searches_detected"
                search1_info_label = "search_info_skipping_searches_detected"

                smartstatus_response["smartstatus_investigations_uc_wlk_skipping"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_wlk_skipping",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_wlk_skipping, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smartstatus_investigations_uc_wlk_skipping"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_wlk_skipping CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_wlk_skipping",
                    "exception": str(e),
                }

        if anomaly_reason == "orphan_search_detected":
            logging.info(
                'performing investigations use case: orphan_search_detected, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and retrieves the orphan status for the entity"

            try:
                uc_results = smartstatus_investigations_uc_wlk_orphan(
                    request_info, tenant_id, record
                )
                # append to response

                search1_results_label = "results_orphan_search_detected"
                search1_info_label = "search_info_orphan_search_detected"

                smartstatus_response["smartstatus_investigations_uc_wlk_orphan"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_wlk_orphan",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_wlk_orphan, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smartstatus_investigations_uc_wlk_orphan"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_wlk_orphan CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_wlk_orphan",
                    "exception": str(e),
                }

        if anomaly_reason == "execution_delayed":
            logging.info(
                'performing investigations use case: execution_delayed, tenant_id="{}", object="{}"'.format(
                    tenant_id, object_value
                )
            )
            use_case_desc = "The use case investigates and retrieves metadata and last execution for delayed entities"

            try:
                uc_results = smartstatus_investigations_uc_wlk_delayed(
                    request_info, tenant_id, record
                )
                # append to response

                search1_results_label = "results_execution_delayed"
                search1_info_label = "search_info_execution_delayed"

                smartstatus_response["smartstatus_investigations_uc_wlk_delayed"] = {
                    "action": "success",
                    "use_case_desc": use_case_desc,
                    search1_results_label: uc_results.get("results_1"),
                    search1_info_label: uc_results.get("search_info_1"),
                    "smart_function": "smartstatus_investigations_uc_wlk_delayed",
                }

            except Exception as e:
                logging.error(
                    'smartstatus, tenant_id="{}", object="{}", exception encountered while running uc function smartstatus_investigations_uc_wlk_delayed, exception="{}"'.format(
                        tenant_id, object_value, str(e)
                    )
                )
                smartstatus_response["smartstatus_investigations_uc_wlk_delayed"] = {
                    "action": "failure",
                    "use_case_desc": use_case_desc,
                    "response": "an exception was encountered while attempting to run the use case investigation, review REST Api logs: index=_internal sourcetype=trackme:rest_api smartstatus_investigations_uc_wlk_delayed CASE(ERROR)",
                    "smart_function": "smartstatus_investigations_uc_wlk_delayed",
                    "exception": str(e),
                }

        return {"payload": smartstatus_response, "status": 200}
