from __future__ import absolute_import, division, print_function, unicode_literals

__name__ = "trackme_rest_handler_vtenant.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import logging
from logging.handlers import RotatingFileHandler
import os, sys
import json
from collections import OrderedDict
import time

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_reqinfo

# import Splunk SDK client
import splunklib.client as client
import splunklib.results as results


class TrackMeHandlerVtenants_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerVtenants_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_vtenants(self, request_info, **kwargs):
        response = {
            "resource_group_name": "vtenants/write",
            "resource_group_desc": "Endpoints related to the management of TrackMe Virtual Tenants (power operations)",
        }

        return {"payload": response, "status": 200}

    # Runs a tracker with elevated privileges and as a the system user rather the requester
    def post_run_tenant_tracker(self, request_info, **kwargs):
        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            if not describe:
                tenant_id = resp_dict["tenant_id"]
                report = resp_dict["report"]
                earliest = resp_dict["earliest"]
                latest = resp_dict["latest"]

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint is designed to execute a Splunk query as the system user, especially to run trackers via the UI with privileges elevation and running on behalf of the system rather than the requester, it requires a POST call with optional data:",
                "resource_desc": "Run TrackMe trackers as the system user",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/vtenants/write/run_tenant_tracker\" body=\"{'tenant_id':'mytenant', 'report': 'mytracker', 'earliest': '-5m', 'latest': 'now'}\"",
                "options": [
                    {
                        "tenant_id": "Tenant identifier",
                        "report": "The name of the TrackMe report, for security reasons its execution will be refused if it is known in TrackMe knowledge objects",
                        "earliest": "The Splunk earliest time quantifier",
                        "latest": "The Splunk latest time quantifier",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # performance counter
        start = time.time()

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service - must run as system
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = "kv_trackme_virtual_tenants"
        collection = service.kvstore[collection_name]

        try:
            vtenant_record = collection.data.query(
                query=json.dumps({"tenant_id": tenant_id})
            )[0]
            vtenant_key = vtenant_record.get("_key")
            logging.debug(
                f'tenant_id="{tenant_id}", vtenant_key="{vtenant_key}", vtenant_report="{json.dumps(vtenant_record)}"'
            )
        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}" cannot be found'
            logging.error(error_msg)
            return {
                "payload": {
                    "response": error_msg,
                },
                "status": 500,
            }

        # get savedsearch definition
        try:
            savedsearch = service.saved_searches[report]
            savedsearch_search = savedsearch.content["search"]
            logging.debug(
                f'tenant_id="{tenant_id}", report="{report}", definition="{savedsearch_search}"'
            )

        except Exception as e:
            error_msg = f'report="{report}" cannot be found'
            logging.error(error_msg)
            return {
                "payload": {
                    "response": error_msg,
                },
                "status": 500,
            }

        # Define the SPL query
        kwargs_search = {
            "app": "trackme",
            "earliest_time": earliest,
            "latest_time": latest,
            "search_mode": "normal",
            "preview": False,
            "time_format": "%s",
            "count": 0,
            "output_mode": "json",
        }

        query_results = []
        search = f'| savedsearch "{report}"'
        try:
            logging.info(
                f'tenant_id={tenant_id}, executing report="{report}", earliest="{earliest}", latest="{latest}", requester="{request_info.user}"'
            )
            # spawn the search and get the results
            searchresults = service.jobs.export(search, **kwargs_search)
            reader = results.JSONResultsReader(searchresults)
            for item in reader:
                if isinstance(item, dict):
                    query_results.append(item)

            run_time = time.time() - start
            logging.info(
                f'tenant_id={tenant_id}, terminated report="{report}", earliest="{earliest}", latest="{latest}", requester="{request_info.user}", run_time="{run_time}"'
            )

            return {"payload": query_results, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}
