#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import re
import sys
import time
from ast import literal_eval

# Third-party libraries
import requests
import urllib3

# Logging handlers
from logging.handlers import RotatingFileHandler

# Disable insecure request warnings for urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# set splunkhome
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_ack_tracker.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append libs
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# Import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)

# import trackme libs
from trackme_libs import trackme_audit_event, trackme_reqinfo


@Configuration(distributed=False)
class TrackMeAckTracker(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** Optional, the tenant identifier""",
        require=False,
        default=None,
    )

    action = Option(
        doc="""
        **Syntax:** **action=****
        **Description:** Optional, the action to be performed, valid options are: ack_expired | force_expire_all_ack""",
        require=False,
        default="ack_expired",
        validate=validators.Match("action", r"^(ack_expired|force_expire_all_ack)$"),
    )

    def generate(self, **kwargs):
        # Start performance counter
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # retrieve options
        trackme_ack_remove_when_green = bool(
            int(
                reqinfo["trackme_conf"]["trackme_general"][
                    "trackme_ack_remove_when_green"
                ]
            )
        )

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # get current user
        username = self._metadata.searchinfo.username

        # get user info
        users = self.service.users

        # Get roles for the current user
        username_roles = []
        for user in users:
            if user.name == username:
                username_roles = user.roles
        logging.info('username="{}", roles="{}"'.format(username, username_roles))

        # Data collection
        collection_name = "kv_trackme_virtual_tenants"
        collection = self.service.kvstore[collection_name]

        # Define the KV query search string
        query_string = {"tenant_status": "enabled"}

        if self.tenant_id and self.tenant_id != "*":
            query_string = {
                "$and": [
                    {
                        "tenant_id": self.tenant_id,
                        "tenant_status": "enabled",
                    }
                ]
            }

        # log and yield
        msg = "Starting Ack expiration maintenance job"
        logging.info(msg)
        yield {"_time": str(time.time()), "_raw": {"event": msg}}

        # Get the records
        enabled_tenants = []
        try:
            records = collection.data.query(query=json.dumps(query_string))

            # Loop through the records
            for record in records:
                enabled_tenants.append(record["tenant_id"])

            # For each enabled tenant, access the Ack KVstore, loop through the records, verify the Ack expiration and update as needed
            for tenant_id in enabled_tenants:
                # log and yield
                msg = (
                    'Starting Ack expiration maintenance job for tenant_id="{}"'.format(
                        tenant_id
                    )
                )
                logging.info(msg)
                yield {
                    "_time": str(time.time()),
                    "_raw": {"tenant_id": tenant_id, "event": msg},
                }

                # Ack collection
                ack_collection_name = "kv_trackme_common_alerts_ack_tenant_" + str(
                    tenant_id
                )
                ack_collection = self.service.kvstore[ack_collection_name]

                # Retrieve any active Ack
                query_string = {
                    "ack_state": "active",
                }

                # Get
                ack_records = ack_collection.data.query(query=json.dumps(query_string))

                # Log a summary of operation
                response_summary = []

                # Loop through the records
                for ack_record in ack_records:
                    # debug logging
                    logging.debug(
                        'tenant_id="{}", ack_record="{}"'.format(
                            tenant_id, json.dumps(ack_record, indent=1)
                        )
                    )

                    # record key
                    key = ack_record["_key"]

                    # retrieve the ack_expiration value
                    ack_expiration = float(ack_record["ack_expiration"])
                    logging.debug(
                        'tenant_id="{}", key="{}", ack_expiration="{}", ack_record="{}"'.format(
                            tenant_id,
                            key,
                            str(ack_expiration),
                            json.dumps(ack_record, indent=1),
                        )
                    )

                    #
                    # attempt to retrieve the entity state
                    #

                    object_category = ack_record.get("object_category")

                    component_mapping = {
                        "splk-dsm": "dsm",
                        "splk-dhm": "dhm",
                        "splk-mhm": "mhm",
                        "splk-flx": "flx",
                        "splk-cim": "cim",
                        "splk-wlk": "wlk",
                    }

                    component = component_mapping.get(object_category, "none")
                    component_collection_name = (
                        f"kv_trackme_{component}_tenant_{tenant_id}"
                    )
                    component_collection = self.service.kvstore[
                        component_collection_name
                    ]

                    # Define the KV query
                    query_string = {
                        "object": ack_record["object"],
                    }

                    try:
                        entity_kvrecord = component_collection.data.query(
                            query=json.dumps(query_string)
                        )[0]
                        entity_key = entity_kvrecord.get("_key")
                        object_state = entity_kvrecord.get("object_state")
                    except Exception as e:
                        entity_key = None
                        object_state = None

                    entity_has_returned_to_green = False
                    if entity_key:
                        if not object_state:
                            logging.info(
                                'tenant_id="{}", object_category="{}", object="{}", no object state information could be retrieved, this can be expected under specific circumstances'.format(
                                    tenant_id, object_category, ack_record["object"]
                                )
                            )
                            entity_has_returned_to_green = False
                        else:
                            logging.debug(
                                f'tenant_id="{tenant_id}", object_category="{object_category}", object="{ack_record["object"]}", object_state="{object_state}"'
                            )
                            if object_state == "green":
                                entity_has_returned_to_green = True
                                logging.info(
                                    'tenant_id="{}", object_category="{}", object="{}", object_state="{}", entity_has_returned_to_green="{}"'.format(
                                        tenant_id,
                                        object_category,
                                        ack_record["object"],
                                        object_state,
                                        entity_has_returned_to_green,
                                    )
                                )
                            else:
                                entity_has_returned_to_green = False
                    else:
                        entity_has_returned_to_green = False

                    # current time
                    current_epoch = float(time.time())

                    # get the ack_type
                    try:
                        ack_type = ack_record.get("ack_type")
                    except Exception as e:
                        ack_type = "unsticky"
                    logging.debug(
                        'tenant_id="{}", object_category="{}", object="{}", object_state="{}", ack_type="{}"'.format(
                            tenant_id,
                            object_category,
                            ack_record["object"],
                            object_state,
                            ack_type,
                        )
                    )

                    if (
                        (trackme_ack_remove_when_green or ack_type == "sticky")
                        and entity_has_returned_to_green
                        and self.action == "ack_expired"
                    ):
                        if ack_type == "sticky" and not current_epoch > ack_expiration:
                            msg = 'Ack with key="{}" is active in sticky mode, it will not be purged before its expiration'.format(
                                str(key)
                            )
                            logging.info(msg)
                            yield {
                                "_time": str(time.time()),
                                "_raw": {"tenant_id": tenant_id, "event": msg},
                            }

                        else:
                            if ack_type == "sticky":
                                msg = 'Ack with key="{}" is active and sticky, the entity has returned to green and the ack has now expired'.format(
                                    str(key)
                                )
                            else:
                                msg = 'Ack with key="{}" is active and unsticky, the entity has returned to green, forcing its expiration now'.format(
                                    str(key)
                                )

                            logging.info(msg)
                            yield {
                                "_time": str(time.time()),
                                "_raw": {"tenant_id": tenant_id, "event": msg},
                            }

                            try:
                                ack_collection.data.update(
                                    str(key),
                                    json.dumps(
                                        {
                                            "object": ack_record["object"],
                                            "object_category": ack_record[
                                                "object_category"
                                            ],
                                            "ack_state": "inactive",
                                            "ack_mtime": str(current_epoch),
                                            "ack_expiration": "N/A",
                                            "ack_comment": ack_record["ack_comment"],
                                        }
                                    ),
                                )

                                # log and yield
                                msg = 'Ack with key="{}" for object="{}" object_category="{}" was updated successfully'.format(
                                    key,
                                    ack_record["object"],
                                    ack_record["object_category"],
                                )
                                logging.info(msg)
                                yield {
                                    "_time": str(time.time()),
                                    "_raw": {
                                        "tenant_id": tenant_id,
                                        "event": msg,
                                        "record": ack_collection.data.query_by_id(key),
                                    },
                                }
                                # debug only
                                logging.debug(
                                    'tenant_id="{}", updated_record=\{}"'.format(
                                        tenant_id, ack_collection.data.query_by_id(key)
                                    )
                                )

                                # Record an audit change
                                trackme_audit_event(
                                    session_key,
                                    reqinfo["server_rest_uri"],
                                    tenant_id,
                                    str(username),
                                    "success",
                                    "disable ack",
                                    ack_record["object"],
                                    ack_record["object_category"],
                                    ack_collection.data.query(
                                        query=json.dumps({"_key": key})
                                    )[0],
                                    "The entity has returned to green state or the Ack has expired now.",
                                    "Auto Ack management",
                                )

                                # Add to response
                                response_summary.append(
                                    {
                                        "object": ack_record["object"],
                                        "object_category": ack_record[
                                            "object_category"
                                        ],
                                        "action_description": "The entity has returned to green state or the Ack has expired now.",
                                        "ack_attributes": ack_collection.data.query(
                                            query=json.dumps({"_key": key})
                                        )[0],
                                    }
                                )

                            except Exception as e:
                                msg = 'tenant_id="{}", Failed to update the ack_record="{}" with exception="{}"'.format(
                                    tenant_id, json.dumps(ack_record, indent=1), str(e)
                                )
                                logging.error(msg)
                                yield {
                                    "_time": str(time.time()),
                                    "_raw": {"tenant_id": tenant_id, "event": msg},
                                }

                    elif self.action == "ack_expired":
                        # handle expiration
                        if current_epoch > ack_expiration:
                            msg = 'Ack with key="{}" is expired since {} seconds, updating the record now'.format(
                                str(key), str(round(current_epoch - ack_expiration, 2))
                            )
                            logging.info(msg)
                            yield {
                                "_time": str(time.time()),
                                "_raw": {"tenant_id": tenant_id, "event": msg},
                            }

                            try:
                                ack_collection.data.update(
                                    str(key),
                                    json.dumps(
                                        {
                                            "object": ack_record["object"],
                                            "object_category": ack_record[
                                                "object_category"
                                            ],
                                            "ack_state": "inactive",
                                            "ack_mtime": str(current_epoch),
                                            "ack_expiration": "N/A",
                                            "ack_comment": ack_record["ack_comment"],
                                        }
                                    ),
                                )

                                # log and yield
                                msg = 'Ack with key="{}" for object="{}" object_category="{}" was updated successfully'.format(
                                    key,
                                    ack_record["object"],
                                    ack_record["object_category"],
                                )
                                logging.info(msg)
                                yield {
                                    "_time": str(time.time()),
                                    "_raw": {
                                        "tenant_id": tenant_id,
                                        "event": msg,
                                        "record": ack_collection.data.query_by_id(key),
                                    },
                                }
                                # debug only
                                logging.debug(
                                    'tenant_id="{}", updated_record=\{}"'.format(
                                        tenant_id, ack_collection.data.query_by_id(key)
                                    )
                                )

                                # Record an audit change
                                trackme_audit_event(
                                    session_key,
                                    reqinfo["server_rest_uri"],
                                    tenant_id,
                                    str(username),
                                    "success",
                                    "disable ack",
                                    ack_record["object"],
                                    ack_record["object_category"],
                                    ack_collection.data.query(
                                        query=json.dumps({"_key": key})
                                    )[0],
                                    "The Ack is now expired.",
                                    "Auto Ack management",
                                )

                                # add to response
                                response_summary.append(
                                    {
                                        "object": ack_record["object"],
                                        "object_category": ack_record[
                                            "object_category"
                                        ],
                                        "action_description": "The Ack is now expired.",
                                        "ack_attributes": ack_collection.data.query(
                                            query=json.dumps({"_key": key})
                                        )[0],
                                    }
                                )

                            except Exception as e:
                                msg = 'tenant_id="{}", Failed to update the expired ack_record="{}" with exception="{}"'.format(
                                    tenant_id, json.dumps(ack_record, indent=1), str(e)
                                )
                                logging.error(msg)
                                yield {
                                    "_time": str(time.time()),
                                    "_raw": {"tenant_id": tenant_id, "event": msg},
                                }

                    elif self.action == "force_expire_all_ack":
                        msg = 'Ack with key="{}" is active, forcing its expiration now'.format(
                            str(key)
                        )
                        logging.info(msg)
                        yield {
                            "_time": str(time.time()),
                            "_raw": {"tenant_id": tenant_id, "event": msg},
                        }

                        try:
                            ack_collection.data.update(
                                str(key),
                                json.dumps(
                                    {
                                        "object": ack_record["object"],
                                        "object_category": ack_record[
                                            "object_category"
                                        ],
                                        "ack_state": "inactive",
                                        "ack_mtime": str(current_epoch),
                                        "ack_expiration": "N/A",
                                        "ack_comment": ack_record["ack_comment"],
                                    }
                                ),
                            )

                            # log and yield
                            msg = 'Ack with key="{}" for object="{}" object_category="{}" was updated successfully'.format(
                                key, ack_record["object"], ack_record["object_category"]
                            )
                            logging.info(msg)
                            yield {
                                "_time": str(time.time()),
                                "_raw": {
                                    "tenant_id": tenant_id,
                                    "event": msg,
                                    "record": ack_collection.data.query_by_id(key),
                                },
                            }
                            # debug only
                            logging.debug(
                                'tenant_id="{}", updated_record=\{}"'.format(
                                    tenant_id, ack_collection.data.query_by_id(key)
                                )
                            )

                            # Record an audit change
                            trackme_audit_event(
                                session_key,
                                reqinfo["server_rest_uri"],
                                tenant_id,
                                str(username),
                                "success",
                                "disable ack",
                                ack_record["object"],
                                ack_record["object_category"],
                                ack_collection.data.query(
                                    query=json.dumps({"_key": key})
                                )[0],
                                "Force all Ack expired has been requested.",
                                "Auto Ack management",
                            )

                            # add to response
                            response_summary.append(
                                {
                                    "object": ack_record["object"],
                                    "object_category": ack_record["object_category"],
                                    "action_description": "Force all Ack expired has been requested.",
                                    "ack_attributes": ack_collection.data.query(
                                        query=json.dumps({"_key": key})
                                    )[0],
                                }
                            )

                        except Exception as e:
                            msg = 'tenant_id="{}", Failed to update the ack_record="{}" with exception="{}"'.format(
                                tenant_id, json.dumps(ack_record, indent=1), str(e)
                            )
                            logging.error(msg)
                            yield {
                                "_time": str(time.time()),
                                "_raw": {"tenant_id": tenant_id, "event": msg},
                            }

                    else:
                        # log and yield
                        msg = 'The Ack record with key="{}" for object="{}" object_category="{}" will expire in {} seconds'.format(
                            key,
                            ack_record["object"],
                            ack_record["object_category"],
                            str(round(ack_expiration - current_epoch, 2)),
                        )
                        logging.info(msg)
                        yield {
                            "_time": str(time.time()),
                            "_raw": {"tenant_id": tenant_id, "event": msg},
                        }

                # log and yield
                msg = 'Ending Ack expiration maintenance job for tenant_id="{}"'.format(
                    tenant_id
                )
                logging.info(msg)
                yield {
                    "_time": str(time.time()),
                    "_raw": {
                        "tenant_id": tenant_id,
                        "event": msg,
                        "no_ack_action": len(response_summary),
                        "response_summary": response_summary,
                    },
                }

            # log and yield
            duration = round(time.time() - start, 3)
            msg = 'Ending Ack expiration maintenance job, duration_sec="{}"'.format(
                str(duration)
            )
            logging.info(msg)
            yield {
                "_time": str(time.time()),
                "_raw": {"event": msg, "duration_sec": str(duration)},
            }

        except Exception as e:
            # log and yield
            msg = 'failed to run the Ack maintenance job with exception="{}"'.format(
                str(e)
            )
            logging.error(msg)
            yield {"_time": str(time.time()), "_raw": {"event": msg}}


dispatch(TrackMeAckTracker, sys.argv, sys.stdin, sys.stdout, __name__)
