#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import json
import time

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_dataset_gen.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)

# Import trackme libs
from trackme_libs import trackme_reqinfo


@Configuration(distributed=False)
class TrackMeDataSetGen(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
        validate=validators.Match("tenant_id", r".*"),
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** The tracker component name to be executed.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(?:dsm|dhm|mhm|cim)$"),
    )

    regex = Option(
        doc="""
        **Syntax:** **regex=****
        **Description:** matches dataset expressions not regex based, or regex based,\
        valid options are: true | false""",
        require=True,
        default=None,
        validate=validators.Match("regex", r"^(?:true|false)$"),
    )

    addprefix = Option(
        doc="""
        **Syntax:** **addprefix=****
        **Description:** add prefix to field names,\
        valid options are: true | false""",
        require=False,
        default="false",
        validate=validators.Match("addprefix", r"^(?:true|false)$"),
    )

    target = Option(
        doc="""
        **Syntax:** **target=****
        **Description:** The target, valid options are: allowlist | blocklist""",
        require=True,
        default=None,
        validate=validators.Match("target", r"^(?:allowlist|blocklist)$"),
    )

    fields = Option(
        doc="""
        **Syntax:** **fields=****
        **Description:** The list of fields, separated by a comma""",
        require=False,
        default=None,
        validate=validators.Match("fields", r".*"),
    )

    store_cache = Option(
        doc="""
        **Syntax:** **store_cache=****
        **Description:** Store the result in KVstore cache, for faster later usage""",
        require=False,
        default="false",
        validate=validators.Match("store_cache", r"^(?:true|false)$"),
    )

    def generate(self, **kwargs):
        if self:
            # performance counter
            start = time.time()

            # Get request info and set logging level
            reqinfo = trackme_reqinfo(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
            )
            log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

            # component suffix
            if self.component == "dsm":
                subcomponent = "dsm"
            elif self.component == "dhm":
                subcomponent = "dhm"
            elif self.component == "mhm":
                subcomponent = "mhm"

            # Data collection
            collection_name = (
                "kv_trackme_"
                + str(subcomponent)
                + "_allowlist_tenant_"
                + str(self.tenant_id)
            )

            try:
                # get data
                collection = self.service.kvstore[collection_name]

                # prefix: depending on the component, a prefix is added automatically by TrackMe
                prefix = None
                if subcomponent in ("dsm", "dhm"):
                    prefix = "data_"
                elif subcomponent in ("mhm"):
                    prefix = "metric_"

                #
                # indextime filtering
                #

                if self.regex == "false":
                    #
                    # allowlist
                    #

                    allowlist_final = []
                    # these are static fields
                    if self.fields:
                        object_types = self.fields.split(",")
                    else:
                        object_types = ["index"]

                    for object_type in object_types:
                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "object_category": object_type,
                                    "action": "allow",
                                    "is_rex": "false",
                                }
                            ]
                        }

                        # get records
                        allowlist_records = collection.data.query(
                            query=json.dumps(query_string)
                        )

                        # create a list
                        indexes_list = []
                        for record in allowlist_records:
                            if not record["object"] in indexes_list:
                                indexes_list.append(record["object"])

                        # Add the prefix if needed, except for object never has one
                        if self.addprefix == "true" and object_type != "object":
                            object_type = prefix + object_type

                        # form the search
                        allowlist_search_str = []
                        if indexes_list:
                            for index in indexes_list:
                                # _internal is a specific use case where we filter on the splunkd sourcetype too
                                if index == "_internal":
                                    if self.addprefix == "true":
                                        allowlist_search_str.append(
                                            "("
                                            + object_type
                                            + '="'
                                            + index
                                            + '" '
                                            + prefix
                                            + 'sourcetype="splunkd")'
                                        )
                                    else:
                                        allowlist_search_str.append(
                                            "("
                                            + object_type
                                            + '="'
                                            + index
                                            + '" sourcetype="splunkd")'
                                        )
                                else:
                                    allowlist_search_str.append(
                                        "(" + object_type + '="' + index + '")'
                                    )
                        else:
                            allowlist_search_str.append("(" + object_type + "=*)")

                        # final
                        allowlist_str = "(" + " OR ".join(allowlist_search_str) + ")"
                        allowlist_final.append(allowlist_str)

                    allowlist_final_str = "( " + " OR ".join(allowlist_final) + " )"

                    #
                    # blocklist
                    #

                    blocklist_final = []
                    # these are static fields
                    if self.fields:
                        object_types = self.fields.split(",")
                    else:
                        if subcomponent == "dsm":
                            object_types = ["index", "sourcetype", "object"]
                        elif subcomponent == "dhm":
                            object_types = ["index", "sourcetype", "object"]
                        elif subcomponent == "mhm":
                            object_types = ["index", "object", "metric_category"]

                    for object_type in object_types:
                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "object_category": object_type,
                                    "action": "block",
                                    "is_rex": "false",
                                }
                            ]
                        }

                        # get records
                        blocklist_records = collection.data.query(
                            query=json.dumps(query_string)
                        )

                        # create a list
                        indexes_list = []
                        for record in blocklist_records:
                            if not record["object"] in indexes_list:
                                indexes_list.append(record["object"])

                        # Add the prefix if needed, except for object and metric_category
                        if (
                            self.addprefix == "true"
                            and object_type != "object"
                            and object_type != "metric_category"
                        ):
                            object_type = prefix + object_type

                        # form the search
                        blocklist_search_str = []
                        if indexes_list:
                            for index in indexes_list:
                                blocklist_search_str.append(
                                    object_type + '!="' + index + '"'
                                )
                        else:
                            blocklist_search_str.append(object_type + '!=""')

                        # final
                        blocklist_str = "(" + " AND ".join(blocklist_search_str) + ")"
                        blocklist_final.append(blocklist_str)

                    blocklist_final_str = "( " + " AND ".join(blocklist_final) + " )"

                #
                # searchtime filtering
                #

                elif self.regex == "true":
                    #
                    # allowlist
                    #

                    allowlist_final = []
                    if self.fields:
                        object_types = self.fields.split(",")
                    else:
                        object_types = ["index"]

                    for object_type in object_types:
                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "object_category": object_type,
                                    "action": "allow",
                                    "is_rex": "true",
                                }
                            ]
                        }

                        # get records
                        allowlist_records = collection.data.query(
                            query=json.dumps(query_string)
                        )

                        # create a list
                        indexes_list = []
                        for record in allowlist_records:
                            if not record["object"] in indexes_list:
                                indexes_list.append(record["object"])

                        # Add the prefix if needed, except for object never has one
                        if self.addprefix == "true" and object_type != "object":
                            object_type = prefix + object_type

                        # form the search
                        allowlist_search_str = []
                        if indexes_list:
                            for index in indexes_list:
                                allowlist_search_str.append(
                                    "match(" + object_type + ', "' + index + '")'
                                )
                        else:
                            allowlist_search_str.append(
                                "match(" + object_type + ', ".*")'
                            )

                        # final
                        allowlist_str = "(" + " OR ".join(allowlist_search_str) + ")"
                        allowlist_final.append(allowlist_str)

                    allowlist_final_str = "( " + " OR ".join(allowlist_final) + " )"

                    #
                    # blocklist
                    #

                    blocklist_final = []
                    object_type = []

                    if self.fields:
                        object_types = self.fields.split(",")
                    else:
                        if subcomponent == "dsm":
                            object_types = ["index", "sourcetype", "object"]
                        elif subcomponent == "dhm":
                            object_types = ["index", "sourcetype", "object"]
                        elif subcomponent == "mhm":
                            object_types = ["index", "object", "metric_category"]

                    for object_type in object_types:
                        # regex

                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "object_category": object_type,
                                    "action": "block",
                                    "is_rex": "true",
                                }
                            ]
                        }

                        # get records
                        blocklist_records = collection.data.query(
                            query=json.dumps(query_string)
                        )

                        # create a list
                        indexes_list = []
                        for record in blocklist_records:
                            if not record["object"] in indexes_list:
                                indexes_list.append(record["object"])

                        # Add the prefix if needed, except for object and metric_category
                        if (
                            self.addprefix == "true"
                            and object_type != "object"
                            and object_type != "metric_category"
                        ):
                            object_type = prefix + object_type

                        # form the search
                        blocklist_search_str = []
                        if indexes_list:
                            for index in indexes_list:
                                blocklist_search_str.append(
                                    "NOT match(" + object_type + ', "' + index + '")'
                                )
                        else:
                            blocklist_search_str.append(
                                "NOT match(" + object_type + ', "^(?![\s\S])")'
                            )

                        # final
                        blocklist_str = "(" + " AND ".join(blocklist_search_str) + ")"
                        blocklist_final.append(blocklist_str)

                    blocklist_final_str = "( " + " AND ".join(blocklist_final) + " )"

                #
                # Final
                #

                # final search string
                if self.target == "allowlist":
                    search_str = allowlist_final_str
                elif self.target == "blocklist":
                    search_str = blocklist_final_str

                # finally, if store_cache is set to true, store the result in its own KVstore record for faster later on usage
                if self.store_cache == "true":
                    logging.debug(
                        'tenant_id="{}", component="{}", fields="{}", storing the result in KVstore cache for faster later on usage'.format(
                            self.tenant_id, self.component, self.fields
                        )
                    )

                    # Cache collection
                    collection_name = "kv_trackme_common_datagen_cache_tenant_" + str(
                        self.tenant_id
                    )
                    collection = self.service.kvstore[collection_name]

                    # Define the KV query
                    if not self.fields:
                        query_string = {
                            "$and": [
                                {
                                    "component": subcomponent,
                                    "target": self.target,
                                    "addprefix": self.addprefix,
                                    "regex": self.regex,
                                }
                            ]
                        }
                    else:
                        query_string = {
                            "$and": [
                                {
                                    "component": subcomponent,
                                    "target": self.target,
                                    "addprefix": self.addprefix,
                                    "regex": self.regex,
                                    "fields": self.fields,
                                }
                            ]
                        }
                    logging.debug(
                        'query_string="{}"'.format(json.dumps(query_string, indent=2))
                    )

                    # check if the record exists
                    try:
                        record = collection.data.query(query=json.dumps(query_string))
                        key = record[0].get("_key")

                    except Exception as e:
                        key = None

                    # if the record does exist, create it, if it does, check if it is up to date
                    if not key:
                        try:
                            # Insert the record
                            if not self.fields:
                                collection.data.insert(
                                    json.dumps(
                                        {
                                            "component": subcomponent,
                                            "target": self.target,
                                            "addprefix": self.addprefix,
                                            "regex": self.regex,
                                            "search_str": search_str,
                                        }
                                    )
                                )
                            else:
                                collection.data.insert(
                                    json.dumps(
                                        {
                                            "component": subcomponent,
                                            "target": self.target,
                                            "addprefix": self.addprefix,
                                            "regex": self.regex,
                                            "fields": self.fields,
                                            "search_str": search_str,
                                        }
                                    )
                                )

                        except Exception as e:
                            logging.error(
                                'tenant_id="{}", component="{}", failed to insert the KVstore record with exception="{}"'.format(
                                    self.tenant_id, self.component, str(e)
                                )
                            )
                    else:
                        if str(record[0]["search_str"]) != str(search_str):
                            try:
                                # Update the record
                                if not self.fields:
                                    collection.data.update(
                                        str(key),
                                        json.dumps(
                                            {
                                                "component": subcomponent,
                                                "target": self.target,
                                                "addprefix": self.addprefix,
                                                "regex": self.regex,
                                                "search_str": search_str,
                                            }
                                        ),
                                    )
                                else:
                                    collection.data.update(
                                        str(key),
                                        json.dumps(
                                            {
                                                "component": subcomponent,
                                                "target": self.target,
                                                "addprefix": self.addprefix,
                                                "regex": self.regex,
                                                "fields": self.fields,
                                                "search_str": search_str,
                                            }
                                        ),
                                    )
                            except Exception as e:
                                logging.error(
                                    'tenant_id="{}", component="{}", failed to update the KVstore record with exception="{}"'.format(
                                        self.tenant_id, self.component, str(e)
                                    )
                                )

                # yield
                yield {
                    "search_str": search_str,
                }

                # performance counter
                logging.info(
                    'tenant_id="{}", component="{}", status="success", TrackmeDataSetGen terminated, enable debug to get full generated search string, run_time="{}"'.format(
                        self.tenant_id, self.component, round(time.time() - start, 3)
                    )
                )
                logging.debug(
                    'tenant_id="{}", component="{}", search_str="{}"'.format(
                        self.tenant_id, self.component, search_str
                    )
                )

            except Exception as e:
                logging.error(
                    'tenant_id="{}", component="{}", status="failure", An exception was encountered, exception="{}"'.format(
                        self.tenant_id, self.component, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", component="{}", status="failure", An exception was encountered, exception="{}"'.format(
                        self.tenant_id, self.component, str(e)
                    )
                )


dispatch(TrackMeDataSetGen, sys.argv, sys.stdin, sys.stdout, __name__)
