#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library
import os
import sys
import json
import ast
import re
import time

# Logging
import logging
from logging.handlers import RotatingFileHandler

# Networking
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_decision_maker.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk
from splunklib.searchcommands import (
    dispatch,
    StreamingCommand,
    Configuration,
    Option,
    validators,
)

# Import trackme libs
from trackme_libs import trackme_reqinfo, trackme_vtenant_account


@Configuration(distributed=False)
class TrackMeDecisionMaker(StreamingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** Specify the TrackMe component.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(dsm|dhm|mhm|wlk|flx|cim)$"),
    )

    """
    This function ensures that records have the same list of fields to allow Splunk to automatically extract these fields
    If a given result does not have a given field, it will be added to the record as an empty value    
    """

    def generate_fields(self, records):
        all_keys = set()
        for record in records:
            all_keys.update(record.keys())

        for record in records:
            for key in all_keys:
                if key not in record:
                    record[key] = ""
            yield record

    """
    Define the function convert_seconds_to_duration
    behaviour: converts seconds to duration, duration is a string from as [D+]HH:MM:SS
    The first segment represents the number of days, the second the number of hours, third the number of minutes, and the fourth the number of seconds.
    """

    def convert_seconds_to_duration(self, seconds):
        try:
            original_seconds = int(seconds)
        except ValueError:
            return "Invalid input"

        # Check if the original seconds were negative
        is_negative = original_seconds < 0
        seconds = abs(original_seconds)

        # Calculate days, hours, minutes, and seconds
        days = seconds // (24 * 3600)
        seconds = seconds % (24 * 3600)
        hours = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        seconds %= 60

        # Format the duration string
        if days > 0:
            duration = f"{days}+{hours:02d}:{minutes:02d}:{seconds:02d}"
        else:
            duration = f"{hours:02d}:{minutes:02d}:{seconds:02d}"

        # Add "-" if the original seconds were negative
        if is_negative:
            duration = "-" + duration

        return duration

    """
    Define the function convert_epoch_to_datetime
    """

    def convert_epoch_to_datetime(self, epoch):
        # convert epoch to float
        try:
            epoch = float(epoch)
            # convert epoch to datetime
            datetime = time.strftime("%d %b %Y %H:%M", time.localtime(epoch))
            return datetime
        except:
            epoch = 0
            return epoch

    """
    Create a new function called get_monitoring_wdays which returns True or False
    arguments: monitoring_wdays, monitoring_hours_ranges
    original SPL logic:
    
    ``` apply wdays rules ```\
    | eval current_wday_no=strftime(now(), "%w")\
    | eval manual_wdays_regex=data_monitoring_wdays\
    | rex field=manual_wdays_regex mode=sed "s/manual:(\d)/\1/g"\
    | eval manual_wdays_regex = "^(" . replace(manual_wdays_regex, ",", "|") . ")$"\
    | eval isUnderMonitoringDays=case(\
    match(data_monitoring_wdays, "^(auto|manual):all_days"), "true",\
    match(data_monitoring_wdays, "^(auto|manual):monday-to-friday") AND match(current_wday_no, "^(1|2|3|4|5)"), "true",\
    match(data_monitoring_wdays, "^(auto|manual):monday-to-saturday") AND match(current_wday_no, "^(1|2|3|4|5|6)"), "true",\
    match(data_monitoring_wdays, "^(manual):\d") AND match(current_wday_no, manual_hours_regex), "true",\
    0=0, "false"\
    )\
    ``` apply hours ranges rules ```\
    | eval current_hour_range=strftime(now(), "%H")\
    | eval manual_hours_regex=data_monitoring_hours_ranges\
    | rex field=manual_hours_regex mode=sed "s/manual:(\d)/\1/g"\
    | eval manual_hours_regex = "^(" . replace(manual_hours_regex, ",", "|") . ")$"\
    | eval isUnderMonitoringHours=case(\
    match(data_monitoring_hours_ranges, "^(auto|manual):all_ranges"), "true",\
    match(data_monitoring_hours_ranges, "^(auto|manual):08h-to-20h") AND match(current_hour_range, "^(8|9|10|11|12|13|14|15|16|17|18|19)$"), "true",\
    match(data_monitoring_hours_ranges, "^(manual):\d") AND match(current_hour_range, manual_hours_regex), "true",\
    0=0, "false"\
    )\    
    """

    def get_monitoring_wdays(self, monitoring_wdays, monitoring_hours_ranges):
        # get current wday
        current_wday_no = int(time.strftime("%w"))

        # get current hour range
        current_hour_range = int(time.strftime("%H"))

        # get manual wdays
        manual_wdays_pattern = monitoring_wdays.replace("manual:", "").replace(",", "|")
        manual_wdays_regex = re.compile(f"^{manual_wdays_pattern}$")

        # get manual hours ranges
        manual_hours_pattern = monitoring_hours_ranges.replace("manual:", "").replace(
            ",", "|"
        )
        manual_hours_regex = re.compile(f"^({manual_hours_pattern})$")

        # apply wdays rules
        if monitoring_wdays == "auto:all_days" or monitoring_wdays == "manual:all_days":
            isUnderMonitoringDays = True
        elif (
            monitoring_wdays == "auto:monday-to-friday"
            or monitoring_wdays == "manual:monday-to-friday"
        ) and current_wday_no in [
            1,
            2,
            3,
            4,
            5,
        ]:
            isUnderMonitoringDays = True
        elif (
            monitoring_wdays == "auto:monday-to-saturday"
            or monitoring_wdays == "manual:monday-to-saturday"
        ) and current_wday_no in [
            1,
            2,
            3,
            4,
            5,
            6,
        ]:
            isUnderMonitoringDays = True
        elif monitoring_wdays.startswith("manual:") and manual_wdays_regex.match(
            str(current_wday_no)
        ):
            isUnderMonitoringDays = True
        else:
            isUnderMonitoringDays = False

        # apply hours ranges rules
        if (
            monitoring_hours_ranges == "auto:all_ranges"
            or monitoring_hours_ranges == "manual:all_ranges"
        ):
            isUnderMonitoringHours = True
        elif (
            monitoring_hours_ranges == "auto:08h-to-20h"
            or monitoring_hours_ranges == "manual:08h-to-20h"
        ) and current_hour_range in range(8, 19):
            isUnderMonitoringHours = True
        elif monitoring_hours_ranges.startswith("manual:") and manual_hours_regex.match(
            str(current_hour_range)
        ):
            isUnderMonitoringHours = True
        else:
            isUnderMonitoringHours = False

        # generate a field isUnderMonitoringMsg (string) as a summary of the monitoring status
        if isUnderMonitoringDays is True and isUnderMonitoringHours is True:
            isUnderMonitoringMsg = f'''"This entity is currently under monitoring for days {monitoring_wdays} and hours {monitoring_hours_ranges}"'''

        elif isUnderMonitoringDays is False and isUnderMonitoringHours is True:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring days conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        elif isUnderMonitoringDays is True and isUnderMonitoringHours is False:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring hours conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        else:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring days and hours conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        # return
        return isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg

    """
    Create a function called get_outliers_status:
    - arguments: isOutlier, tenant_outliers_set_state
    - tenant_outliers_set_state is 0 or 1, 0 means it is disabled at tenant level, 1 means it is enabled at tenant level
    - original SPL logic:
        eval tenant_pref_outliers_set_state = [ | trackme url="/services/trackme/v2/vtenants/vtenants_accounts" mode=post body="{'tenant_id': '$tenant_id$'}" | spath | table outliers_set_state | return $outliers_set_state ]\
        | eval isOutlier=case(isOutlier=1 AND tenant_pref_outliers_set_state=1, 1, isOutlier=1 AND tenant_pref_outliers_set_state=0, 2, 1=1, 0)\
    - behaviour: alter isOutlier, 0=not outlier, 1=outlier, 2=outlier but disabled at tenant level
    """

    def get_outliers_status(self, isOutlier, tenant_outliers_set_state):
        if tenant_outliers_set_state == 1:
            if isOutlier == 1:
                isOutlier = 1
            else:
                isOutlier = 0
        else:
            if isOutlier == 1:
                isOutlier = 2
            else:
                isOutlier = 0

        return isOutlier

    """
    Create a function called get_data_sampling_status:
    - arguments: isAnomaly, tenant_data_sampling_set_state
    - tenant_data_sampling_set_state is 0 or 1, 0 means it is disabled at tenant level, 1 means it is enabled at tenant level
    - original SPL logic:
        ``` Verify Anomaly Outliers Detection ``` \
        | eval tenant_pref_data_sampling_set_state = [ | trackme url="/services/trackme/v2/vtenants/vtenants_accounts" mode=post body="{'tenant_id': '$tenant_id$'}" | spath | table data_sampling_set_state | return $data_sampling_set_state ]\
        | eval isAnomaly=case(isAnomaly=1 AND tenant_pref_data_sampling_set_state=1, 1, isAnomaly=2 AND tenant_pref_data_sampling_set_state=0, 2, 1=1, 0)\
    - behaviour: alter isAnomaly, 0=not anomaly, 1=anomaly, 2=anomaly but disabled at tenant level
    """

    def get_data_sampling_status(self, isAnomaly, tenant_data_sampling_set_state):
        if tenant_data_sampling_set_state == 1:
            if isAnomaly == 1:
                isAnomaly = 1
            else:
                isAnomaly = 0
        else:
            if isAnomaly == 1:
                isAnomaly = 2
            else:
                isAnomaly = 0

        return isAnomaly

    """
    Create a function called get_future_status:
    - arguments: future_tolerance (expressed in seconds), system_future_tolerance (expressed in seconds), data_last_lag_seen (expressed in seconds)
    - behaviour: returns a boolean, True if data_last_lag_seen is lower than future_tolerance, False otherwise. If future_tolerance is 0, rely on system_future_tolerance
    """

    def get_future_status(
        self,
        future_tolerance,
        system_future_tolerance,
        data_last_lag_seen,
        data_last_ingestion_lag_seen,
        data_last_time_seen,
        data_last_ingest,
    ):
        isFuture = False
        isFutureMsg = ""
        if future_tolerance == 0:
            future_tolerance = system_future_tolerance

        # convert all to int
        try:
            future_tolerance = int(round(float(future_tolerance), 0))
            data_last_lag_seen = int(round(float(data_last_lag_seen), 0))
            data_last_ingestion_lag_seen = int(
                round(float(data_last_ingestion_lag_seen), 0)
            )
        except:
            pass

        logging.debug(
            f"data_last_lag_seen={data_last_lag_seen}, system_future_tolerance={system_future_tolerance}, future_tolerance={future_tolerance}"
        )

        if float(data_last_lag_seen) < float(future_tolerance) or float(
            data_last_ingestion_lag_seen
        ) < float(future_tolerance):
            isFuture = True

            # convert data_last_lag_seen to duration
            data_last_lag_seen_duration = self.convert_seconds_to_duration(
                data_last_lag_seen
            )

            # convert data_last_ingestion_lag_seen to duration
            data_last_ingestion_lag_seen_duration = self.convert_seconds_to_duration(
                data_last_ingestion_lag_seen
            )

            # convert data_last_time_seen to %c
            data_last_time_seen_datetime = self.convert_epoch_to_datetime(
                data_last_time_seen
            )

            # convert data_last_ingest to %c
            data_last_ingest_datetime = self.convert_epoch_to_datetime(data_last_ingest)

            isFutureMsg = f"""detected data indexed in the future which is most likely due to timestamping misconfiguration, timezone or time synchronization issue. Event delay is {data_last_lag_seen} seconds (duration: {data_last_lag_seen_duration}), Event latency is {data_last_ingestion_lag_seen} seconds (duration: {data_last_ingestion_lag_seen_duration}), this is beyond current tolerance threshold of {future_tolerance} seconds, latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event ingested for this entity: {data_last_ingest_datetime}. Review and fix the root cause, or adapt the future tolerance at the system level or for this entity especially."""

        else:
            isFuture = False

        return isFuture, isFutureMsg, future_tolerance

    """
    Create a function called get_future_metrics_status:
    - arguments: future_tolerance (expressed in seconds), system_future_tolerance (expressed in seconds), data_last_lag_seen (expressed in seconds)
    - behaviour: returns a boolean, True if metric_last_time_seen is lower than system_future_tolerance, False otherwise.
    """

    def get_future_metrics_status(
        self,
        system_future_tolerance,
        metric_last_time_seen,
    ):
        isFuture = False
        isFutureMsg = ""

        logging.debug(
            f"metric_last_time_seen={metric_last_time_seen}, system_future_tolerance={system_future_tolerance}"
        )

        if float(metric_last_time_seen) < float(system_future_tolerance):
            isFuture = True
            # convert data_last_lag_seen to duration
            metric_last_time_seen_duration = self.convert_seconds_to_duration(
                metric_last_time_seen
            )
            # convert data_last_time_seen to %c
            metric_last_time_seen_datetime = self.convert_epoch_to_datetime(
                metric_last_time_seen
            )

            isFutureMsg = f"""detected metrics indexed in the future which is most likely due to timestamping misconfiguration, timezone or time synchronization issue. Metric delay is {metric_last_time_seen} seconds (duration: {metric_last_time_seen_duration}) which is beyond tolerance threshold of {system_future_tolerance}, latest event available (_time) for this entity: {metric_last_time_seen_datetime}"""
        else:
            isFuture = False

        return isFuture, isFutureMsg

    """
    Create a function call get_is_under_dcount_host:
    - arguments: min_dcount_host, min_dcount_threshold
    - returns: isUnderDcountHost (boolean), isUnderDcountHostMsg (string)
    - behaviour: returns a boolean, True if min_dcount_threshold is a numerical and is lower than min_dcount_host, False otherwise
    """

    def get_is_under_dcount_host(
        self, min_dcount_host, min_dcount_threshold, min_dcount_field
    ):
        if (
            isinstance(min_dcount_host, float)
            and min_dcount_threshold < min_dcount_host
        ):
            isUnderDcountHost = True
            isUnderDcountHostMsg = f"""Monitoring conditions are not met due to low number of hosts. Number of hosts is {min_dcount_threshold} based on the metric {min_dcount_field} which is lower than the minimum required number of hosts of {min_dcount_host}"""
        else:
            isUnderDcountHost = False
            isUnderDcountHostMsg = ""

        return isUnderDcountHost, isUnderDcountHostMsg

    """
    Create a function called get_logical_group_status:
    - arguments: object_group_key, object_group_name, object_group_min_green_percent, object_group_green_percent
    - behaviour: returns booleans: isUnderLogicalGroup, LogicalGroupStateInAlert 
      isUnderLogicalGroup: True if object_group_key is defined
      LogicalGroupStateInAlert: True if object_group_key is defined and object_group_min_green_percent is lower than object_group_green_percent, False otherwise
    """

    def get_logical_group_status(
        self,
        object_group_key,
        object_group_name,
        object_group_min_green_percent,
        object_group_green_percent,
        object_group_members_count,
        object_group_members_red,
    ):
        if object_group_key != "":
            isUnderLogicalGroup = True
            if object_group_min_green_percent < object_group_green_percent:
                LogicalGroupStateInAlert = True
                LogicalGroupMsg = f"""Logical Group {object_group_name} with key="{object_group_key}" is in alert state. The minimum green percent is {object_group_min_green_percent} which is lower than the minimal green percent of {object_group_green_percent}, object_group_members_count={object_group_members_count}, object_group_members_red={object_group_members_red}"""

            else:
                LogicalGroupStateInAlert = False
                LogicalGroupMsg = f"""Logical Group {object_group_name} with key="{object_group_key}" is in normal state. The minimum green percent is {object_group_min_green_percent} which is higher or equal to the minimal green percent of {object_group_green_percent}, object_group_members_count={object_group_members_count}, object_group_members_red={object_group_members_red}"""

        else:
            isUnderLogicalGroup = False
            LogicalGroupStateInAlert = False
            LogicalGroupMsg = ""

        return isUnderLogicalGroup, LogicalGroupStateInAlert, LogicalGroupMsg

    """
    Create a function called get_dsm_latency_status:
    - arguments: data_last_ingestion_lag_seen, data_max_lag_allowed, data_lag_alert_kpis, data_last_ingest, data_last_time_seen
    - returns: isUnderLatencyAlert (boolean), isUnderLatencyMessage (string)
    - behaviour: 
      isUnderLatencyAlert: If data_lag_alert_kpis is "all_kpis" or "lag_ingestion_kpi" and if data_last_ingestion_lag_seen is higher than data_max_lag_allowed, then isUnderLatencyAlert is True
      isUnderLatencyMessage: If isUnderLatencyAlert is True:
        "Monitoring conditions are not met due to latency issues. Ingestion latency is $data_last_ingestion_lag_seen$ seconds (duration: <conversion of data_last_ingestion_lag_seen to duration), which is higher than the maximum allowed latency of $data_max_lag_allowed$ seconds (duration: <conversion of data_max_lag_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
      isUnderLatencyMessage: If isUnderLatencyAlert is False:
        "Monitoring conditions are not met for ingest latency are met. Ingestion latency is $data_last_ingestion_lag_seen$ seconds (duration: <conversion of data_last_ingestion_lag_seen to duration), which is lower than the maximum allowed latency of $data_max_lag_allowed$ seconds (duration: <conversion of data_max_lag_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
    """

    def get_dsm_latency_status(
        self,
        data_last_ingestion_lag_seen,
        data_max_lag_allowed,
        data_lag_alert_kpis,
        data_last_ingest,
        data_last_time_seen,
    ):
        # convert data_last_ingestion_lag_seen to float
        try:
            data_last_ingestion_lag_seen = float(data_last_ingestion_lag_seen)
        except:
            data_last_ingestion_lag_seen = 0

        # convert data_max_lag_allowed to float
        try:
            data_max_lag_allowed = float(data_max_lag_allowed)
        except:
            data_max_lag_allowed = 0

        # convert data_last_ingest to float
        try:
            data_last_ingest = float(data_last_ingest)
        except:
            data_last_ingest = 0

        # convert data_last_time_seen to float
        try:
            data_last_time_seen = float(data_last_time_seen)
        except:
            data_last_time_seen = 0

        # convert data_last_ingestion_lag_seen to duration
        data_last_ingestion_lag_seen_duration = self.convert_seconds_to_duration(
            data_last_ingestion_lag_seen
        )

        # convert data_max_lag_allowed to duration
        data_max_lag_allowed_duration = self.convert_seconds_to_duration(
            data_max_lag_allowed
        )

        # convert data_last_ingest to %c
        data_last_ingest_datetime = self.convert_epoch_to_datetime(data_last_ingest)

        # convert data_last_time_seen to %c
        data_last_time_seen_datetime = self.convert_epoch_to_datetime(
            data_last_time_seen
        )

        # define isUnderLatencyAlert
        if data_lag_alert_kpis in ["all_kpis", "lag_ingestion_kpi"]:
            if float(data_last_ingestion_lag_seen) > float(data_max_lag_allowed):
                isUnderLatencyAlert = True
            else:
                isUnderLatencyAlert = False
        else:
            isUnderLatencyAlert = False

        # define isUnderLatencyMessage
        if isUnderLatencyAlert:
            isUnderLatencyMessage = f"""Monitoring conditions are not met due to latency issues. Ingestion latency is approximately {round(float(data_last_ingestion_lag_seen), 3)} seconds (duration: {data_last_ingestion_lag_seen_duration}), which is higher than the maximum allowed latency of {round(float(data_max_lag_allowed), 0)} seconds (duration: {data_max_lag_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event indexed (_indextime) for this entity: {data_last_ingest_datetime}"""
        else:
            isUnderLatencyMessage = f"""monitoring conditions for ingest latency are met. Ingestion latency is approximately {round(float(data_last_ingestion_lag_seen), 3)} seconds (duration: {data_last_ingestion_lag_seen_duration}), which is lower than the maximum allowed latency of {round(float(data_max_lag_allowed), 0)} seconds (duration: {data_max_lag_allowed_duration}), latest event indexed (_indextime) for this entity: {data_last_ingest_datetime}"""

        # return
        return isUnderLatencyAlert, isUnderLatencyMessage

    """
    Create a function called get_dsm_delay_status:
    - arguments: data_last_lag_seen, data_max_delay_allowed, data_lag_alert_kpis, data_last_ingest, data_last_time_seen
    - returns: isUnderDelayAlert (boolean), isUnderDelayMessage (string)
    - behaviour:
        isUnderDelayAlert: If data_lag_alert_kpis is "all_kpis" or "lag_event_kpi" and if data_last_lag_seen is higher than data_max_delay_allowed, then isUnderDelayAlert is True
        isUnderDelayMessage: If isUnderDelayAlert is True:
            "Monitoring conditions are not met due to delay issues. Event delay is $data_last_lag_seen$ seconds (duration: <conversion of data_last_lag_seen to duration), which is higher than the maximum allowed delay of $data_max_delay_allowed$ seconds (duration: <conversion of data_max_delay_allowed to duration>), latest event available (_time) for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
        isUnderDelayMessage: If isUnderDelayAlert is False:
            "monitoring conditions for event delay are met. Event delay is $data_last_lag_seen$ seconds (duration: <conversion of data_last_lag_seen to duration), which is lower than the maximum allowed delay of $data_max_delay_allowed$ seconds (duration: <conversion of data_max_delay_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"            
    """

    def get_dsm_delay_status(
        self,
        data_last_lag_seen,
        data_max_delay_allowed,
        data_lag_alert_kpis,
        data_last_ingest,
        data_last_time_seen,
    ):
        # convert data_last_lag_seen to float
        try:
            data_last_lag_seen = float(data_last_lag_seen)
        except:
            data_last_lag_seen = 0

        # convert data_max_delay_allowed to float
        try:
            data_max_delay_allowed = float(data_max_delay_allowed)
        except:
            data_max_delay_allowed = 0

        # convert data_last_ingest to float
        try:
            data_last_ingest = float(data_last_ingest)
        except:
            data_last_ingest = 0

        # convert data_last_time_seen to float
        try:
            data_last_time_seen = float(data_last_time_seen)
        except:
            data_last_time_seen = 0

        # convert data_last_lag_seen to duration
        data_last_lag_seen_duration = self.convert_seconds_to_duration(
            data_last_lag_seen
        )

        # convert data_max_delay_allowed to duration
        data_max_delay_allowed_duration = self.convert_seconds_to_duration(
            data_max_delay_allowed
        )

        # convert data_last_ingest to %c
        data_last_ingest_datetime = self.convert_epoch_to_datetime(data_last_ingest)

        # convert data_last_time_seen to %c
        data_last_time_seen_datetime = self.convert_epoch_to_datetime(
            data_last_time_seen
        )

        # define isUnderDelayAlert
        if data_lag_alert_kpis in ["all_kpis", "lag_event_kpi"]:
            if float(data_last_lag_seen) > float(data_max_delay_allowed):
                isUnderDelayAlert = True
            else:
                isUnderDelayAlert = False
        else:
            isUnderDelayAlert = False

        # define isUnderDelayMessage
        if isUnderDelayAlert:
            isUnderDelayMessage = f"""Monitoring conditions are not met due to delay issues. Event delay is {round(float(data_last_lag_seen), 3)} seconds (duration: {data_last_lag_seen_duration}), which is higher than the maximum allowed delay of {int(round(float(data_max_delay_allowed), 0))} seconds (duration: {data_max_delay_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event ingested (_indextime) for this entity: {data_last_ingest_datetime}"""
        else:
            isUnderDelayMessage = f"""monitoring conditions for event delay are met. Event delay is {round(float(data_last_lag_seen), 3)} seconds (duration: {data_last_lag_seen_duration}), which is lower than the maximum allowed delay of {int(round(float(data_max_delay_allowed), 0))} seconds (duration: {data_max_delay_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}"""

        # return
        return isUnderDelayAlert, isUnderDelayMessage

    """
    Create a function called set_dsm_status:
    - arguments: record, isOutlier, isAnomaly, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderDcountHost, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isAnomaly is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                isUnderDcountHost is False
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_dsm_status(
        self,
        record,
        isOutlier,
        isAnomaly,
        isFuture,
        isFutureMsg,
        isUnderMonitoringDays,
        isUnderMonitoringHours,
        isUnderMonitoringMsg,
        isUnderDcountHost,
        isUnderDcountHostMsg,
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
        isUnderLatencyAlert,
        isUnderLatencyMessage,
        isUnderDelayAlert,
        isUnderDelayMessage,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # define object_state
        if (
            (isOutlier == 0 or isOutlier == 2)
            and (isAnomaly == 0 or isAnomaly == 2)
            and isUnderDcountHost is False
            and isUnderLatencyAlert is False
            and isUnderDelayAlert is False
        ):
            object_state = "green"
        else:
            object_state = "red"

        # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
        if object_state == "red" and isUnderLogicalGroup is True:
            if LogicalGroupStateInAlert is False:
                object_state = "blue"

        # if object_state is not red or blue but isFuture is True, then object_state is orange
        if object_state not in ["red", "blue"]:
            if isFuture is True:
                object_state = "orange"

        # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
        if object_state == "red":
            if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message.append(isUnderDelayMessage)
            status_message.append(isUnderLatencyMessage)
            anomaly_reason.append("none")

        else:
            if isOutlier == 1:
                outlier_reasons = record.get("isOutlierReason", [])
                if isinstance(outlier_reasons, list):
                    # Join the list elements into a single string
                    outlier_reasons_str = " | ".join(outlier_reasons)
                    status_message.append(outlier_reasons_str)
                else:
                    # If it's not a list, append it directly
                    status_message.append(outlier_reasons)
                anomaly_reason.append("ml_outliers_detection")

            if isAnomaly == 1:
                status_message.append(
                    "anomalies detected in the data sampling and format recognition, review the data sampling screen to investigate. This alert means that trackMe detected an issue in the format of the events compared to the format that was previously identified for this source"
                )
                anomaly_reason.append("data_sampling_anomaly")

            if isFuture is True:
                status_message.append(isFutureMsg)
                anomaly_reason.append("future_over_tolerance")

            if isUnderMonitoringDays is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_days")

            if isUnderMonitoringHours is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_hours")

            if isUnderDcountHost is True:
                status_message.append(isUnderDcountHostMsg)
                anomaly_reason.append("min_hosts_dcount")

            if isUnderLatencyAlert is True:
                status_message.append(isUnderLatencyMessage)
                anomaly_reason.append("lag_threshold_breached")

            if isUnderDelayAlert is True:
                status_message.append(isUnderDelayMessage)
                anomaly_reason.append("delay_threshold_breached")

            # logical group
            if isUnderLogicalGroup is True:
                status_message.append(LogicalGroupMsg)
                anomaly_reason.append("in_logical_group")

        # form status_message_json
        status_message_json["status_message"] = status_message
        status_message_json["anomaly_reason"] = anomaly_reason

        # return
        return object_state, status_message, status_message_json, anomaly_reason

    """
    Create a function called set_dhm_status:
    - arguments: record, isOutlier, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_dhm_status(
        self,
        record,
        isOutlier,
        isFuture,
        isFutureMsg,
        isUnderMonitoringDays,
        isUnderMonitoringHours,
        isUnderMonitoringMsg,
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
        isUnderLatencyAlert,
        isUnderLatencyMessage,
        isUnderDelayAlert,
        isUnderDelayMessage,
        default_splk_dhm_alerting_policy,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # define object_state
        if (
            (isOutlier == 0 or isOutlier == 2)
            and isUnderLatencyAlert is False
            and isUnderDelayAlert is False
        ):
            object_state = "green"
        else:
            object_state = "red"

        """
        Convert the following SPL to python:
        | eval object_state=case(\
        splk_dhm_alerting_policy="global_policy" AND default_splk_dhm_alerting_policy="track_per_host", object_state,\
        splk_dhm_alerting_policy="global_policy" AND default_splk_dhm_alerting_policy="track_per_sourcetype", if(match(splk_dhm_st_summary, "\'state\':\s\'red\'"), "red", object_state),\
        splk_dhm_alerting_policy="track_per_host", object_state,\
        splk_dhm_alerting_policy="track_per_sourcetype", if(match(splk_dhm_st_summary, "\'state\':\s\'red\'"), "red", object_state)\
        )\
        """

        splk_dhm_alerting_policy = record.get(
            "splk_dhm_alerting_policy", "global_policy"
        )
        if not len(splk_dhm_alerting_policy) > 0:
            splk_dhm_alerting_policy = "global_policy"
        splk_dhm_st_summary = record.get("splk_dhm_st_summary")

        # get the entity global max delay allowed
        global_max_delay_allowed = int(
            round(float(record.get("data_max_delay_allowed", 0)), 0)
        )

        # get the entity global max lag allowed
        global_max_lag_allowed = int(
            round(float(record.get("data_max_lag_allowed", 0)), 0)
        )

        # get the entity global last delay seen (data_last_lag_seen)
        global_last_event_lag = int(
            round(float(record.get("data_last_lag_seen", 0)), 0)
        )

        # get the entity global last lag seen (data_last_ingestion_lag_seen)
        global_last_ingest_lag = int(
            round(float(record.get("data_last_ingestion_lag_seen", 0)), 0)
        )

        # Convert splk_dhm_st_summary to a list if it is a string
        if isinstance(splk_dhm_st_summary, str):
            splk_dhm_st_summary = [splk_dhm_st_summary]

        # counters
        count_red = 0
        count_green = 0
        sourcetypes_red_list = []

        # splk_dhm_st_summary can actually be a list
        if isinstance(splk_dhm_st_summary, list):
            for item_str in splk_dhm_st_summary:
                dict_str = "{" + item_str + "}"
                dict_loaded = False
                dict_loading_error = []

                # try both options
                try:
                    new_dict = ast.literal_eval(dict_str)
                    dict_loaded = True
                except Exception as e:
                    dict_loaded = False
                    dict_loading_error.append(str(e))

                if not dict_loaded:
                    try:
                        new_dict = json.loads(item_str)
                        dict_loaded = True
                    except Exception as e:
                        dict_loaded = False
                        dict_loading_error.append(str(e))

                if dict_loaded:
                    # Iterate through the inner dictionaries
                    for inner_dict in new_dict.values():
                        if inner_dict.get("state") == "red":
                            count_red += 1

                            max_lag_allowed = float(inner_dict.get("max_lag_allowed"))
                            max_delay_allowed = float(
                                inner_dict.get("max_delay_allowed")
                            )
                            last_ingest_lag = float(inner_dict.get("last_ingest_lag"))
                            last_event_lag = float(inner_dict.get("last_event_lag"))

                            if (
                                last_ingest_lag > max_lag_allowed
                                and "lag_threshold_breached" not in anomaly_reason
                            ):
                                anomaly_reason.append("lag_threshold_breached")
                                sourcetypes_red_list.append(
                                    f'(idx: {inner_dict.get("idx")}, st: {inner_dict.get("st")}, anomaly_reason: lag_threshold_breached)'
                                )
                            if (
                                last_event_lag > max_delay_allowed
                                and "delay_threshold_breached" not in anomaly_reason
                            ):
                                anomaly_reason.append("delay_threshold_breached")
                                sourcetypes_red_list.append(
                                    f'(idx: {inner_dict.get("idx")}, st: {inner_dict.get("st")}, anomaly_reason: delay_threshold_breached)'
                                )

                        elif inner_dict.get("state") == "green":
                            count_green += 1

                else:
                    logging.error(
                        f"Error in processing item_str: {item_str}. Error: {dict_loading_error}"
                    )

            logging.debug(
                f'object="{record.get("object")}", count_red={count_red}, count_green={count_green}'
            )

            # turn sourcetypes_red_list into a pipe separated string
            sourcetypes_red_list = "|".join(sourcetypes_red_list)

            # Decision making based on the counts of red and green states
            if splk_dhm_alerting_policy == "global_policy":
                if default_splk_dhm_alerting_policy == "track_per_host":
                    # Use object_state as it is
                    pass
                elif default_splk_dhm_alerting_policy == "track_per_sourcetype":
                    if count_red > 0:
                        object_state = "red"
                        status_message.append(
                            f"One or more sourcetypes are in alert for this entity, and policy is set to track_per_sourcetype, sourcetypes in alert: {sourcetypes_red_list}"
                        )
                    else:
                        # Use object_state as it is
                        pass
            elif splk_dhm_alerting_policy == "track_per_host":
                # Use object_state as it is
                pass
            elif splk_dhm_alerting_policy == "track_per_sourcetype":
                if count_red > 0:
                    object_state = "red"
                    status_message.append(
                        f"One or more sourcetypes are in alert for this entity, and policy is set to track_per_sourcetype, sourcetypes in alert: {sourcetypes_red_list}"
                    )
                else:
                    # Use object_state as it is
                    pass
            else:
                # Use object_state as it is
                pass

        # if all sourcetypes are in alert, object_state is red or orange depending on the global max delay entity values
        if (
            count_green == 0
            and (global_last_event_lag >= global_max_delay_allowed)
            and (global_last_ingest_lag >= global_max_lag_allowed)
        ):
            object_state = "red"
            status_message.append(
                f"all sourcetypes are in alert for this entity, global entity max delay allowed is breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_event_lag)}), global entity max lag allowed is breached (max_delay_allowed: {global_max_lag_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_lag_allowed)}), last_event_lag: {global_last_ingest_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_ingest_lag)})"
            )

        elif (
            count_green == 0
            and (global_last_event_lag < global_max_delay_allowed)
            and (global_last_ingest_lag >= global_max_lag_allowed)
        ):
            object_state = "red"
            status_message.append(
                f"all sourcetypes are in alert for this entity, global entity max delay allowed is not breached but max lag allowed is breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_event_lag)})"
            )

        elif (
            count_green == 0
            and (global_last_event_lag >= global_max_delay_allowed)
            and (global_last_ingest_lag < global_max_lag_allowed)
        ):
            object_state = "red"
            status_message.append(
                f"all sourcetypes are in alert for this entity, global entity max delay allowed is breached but max lag allowed is not breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_event_lag)})"
            )

        elif (
            count_green == 0
            and (global_last_event_lag < global_max_delay_allowed)
            and (global_last_ingest_lag < global_max_lag_allowed)
        ):
            object_state = "orange"
            status_message.append(
                f"all sourcetypes are in alert for this entity, however global entity max delay allowed and max lag allowed are not breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_event_lag)})"
            )

        elif count_green == 0:
            object_state = "red"
            status_message.append(
                f"all sourcetypes are in alert for this entity, however global entity level max delay allowed and max lag allowed could not be determined, verify TrackMe logs for more information (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {self.convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {self.convert_seconds_to_duration(global_last_event_lag)})"
            )

        # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
        if object_state == "red" and isUnderLogicalGroup is True:
            if LogicalGroupStateInAlert is False:
                object_state = "blue"

        # if object_state is not red or blue but isFuture is True, then object_state is orange
        if object_state not in ["red", "blue"]:
            if isFuture is True:
                object_state = "orange"

        # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
        if object_state == "red":
            if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message.append(isUnderDelayMessage)
            status_message.append(isUnderLatencyMessage)
            anomaly_reason.append("none")

        else:
            if isOutlier == 1:
                outlier_reasons = record.get("isOutlierReason", [])
                if isinstance(outlier_reasons, list):
                    # Join the list elements into a single string
                    outlier_reasons_str = " | ".join(outlier_reasons)
                    status_message.append(outlier_reasons_str)
                else:
                    # If it's not a list, append it directly
                    status_message.append(outlier_reasons)
                anomaly_reason.append("ml_outliers_detection")

            if isFuture is True:
                status_message.append(isFutureMsg)
                anomaly_reason.append("future_over_tolerance")

            if isUnderMonitoringDays is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_days")

            if isUnderMonitoringHours is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_hours")

            if isUnderLatencyAlert is True:
                status_message.append(isUnderLatencyMessage)
                anomaly_reason.append("lag_threshold_breached")

            if isUnderDelayAlert is True:
                status_message.append(isUnderDelayMessage)
                anomaly_reason.append("delay_threshold_breached")

            # logical group
            if isUnderLogicalGroup is True:
                status_message.append(LogicalGroupMsg)
                anomaly_reason.append("in_logical_group")

        # form status_message_json
        status_message_json["status_message"] = status_message
        # deduplicate anomaly_reason
        anomaly_reason = list(set(anomaly_reason))
        status_message_json["anomaly_reason"] = anomaly_reason

        # return
        logging.debug(
            f'set_dhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
        )
        return (
            object_state,
            status_message,
            status_message_json,
            anomaly_reason,
            splk_dhm_alerting_policy,
        )

    """
    Create a function called set_mhm_status:
    - arguments: record, isFuture, isFutureMsg, isUnderLogicalGroup, LogicalGroupStateInAlert
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                all metric caterogies are green
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_mhm_status(
        self,
        record,
        isFuture,
        isFutureMsg,
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # define object_state
        object_state = "green"

        metric_details = record.get("metric_details")

        # Convert metric_details to a list if it is a string
        if isinstance(metric_details, str):
            metric_details = [metric_details]

        # counters
        count_red = 0
        count_green = 0
        metrics_red_list = []

        # splk_dhm_st_summary can actually be a list
        if isinstance(metric_details, list):
            for item_str in metric_details:
                try:
                    new_dict = ast.literal_eval(item_str)

                    # Iterate through the inner dictionaries
                    for inner_dict in new_dict.values():
                        if inner_dict.get("state") == "red":
                            count_red += 1
                            anomaly_reason.append("delay_threshold_breached")
                            metrics_red_list.append(
                                f'(idx: {inner_dict.get("idx")}, metrics: {inner_dict.get("metric_category")}, anomaly_reason: delay_threshold_breached)'
                            )

                        elif inner_dict.get("state") == "green":
                            count_green += 1

                except Exception as e:
                    logging.error(
                        f"Error in processing item_str: {item_str}. Error: {str(e)}"
                    )

            logging.debug(
                f'object="{record.get("object")}", count_red={count_red}, count_green={count_green}'
            )

            # turn metrics_red_list into a pipe separated string
            metrics_red_list = "|".join(metrics_red_list)

            # Decision making based on the counts of red and green states
            if count_red > 0:
                object_state = "red"
                status_message.append(
                    f"One or more metric categories are in alert for this entity, metrics in alert: {metrics_red_list}"
                )
            else:
                # Use object_state as it is
                pass

        # if all metrics are in alert, then object_state is red
        if count_green == 0:
            object_state = "red"
            status_message.append("all metric categories are in alert for this entity")

        # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
        if object_state == "red" and isUnderLogicalGroup is True:
            if LogicalGroupStateInAlert is False:
                object_state = "blue"

        # if object_state is not red or blue but isFuture is True, then object_state is orange
        if object_state not in ["red", "blue"]:
            if isFuture is True:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message.append(
                "All metric categories are in normal state for this entity"
            )
            anomaly_reason.append("none")

        else:
            if isFuture is True:
                status_message.append(isFutureMsg)
                anomaly_reason.append("future_over_tolerance")

            # logical group
            if isUnderLogicalGroup is True:
                status_message.append(LogicalGroupMsg)
                anomaly_reason.append("in_logical_group")

        # form status_message_json
        status_message_json["status_message"] = status_message
        status_message_json["anomaly_reason"] = anomaly_reason

        # return
        logging.debug(
            f'set_mhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
        )
        return object_state, status_message, status_message_json, anomaly_reason

    """
    Create a function called set_flx_status:
    - arguments: record, isOutlier, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_flx_status(
        self,
        record,
        isOutlier,
        isUnderMonitoringDays,
        isUnderMonitoringHours,
        isUnderMonitoringMsg,
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # for flx, attempt to retrieve extra_attributes, if present attempt to load as an object
        extra_attributes = record.get("extra_attributes", {})
        if isinstance(extra_attributes, str):
            if len(extra_attributes) > 0:
                try:
                    extra_attributes = json.loads(extra_attributes)
                except Exception as e:
                    logging.error(
                        f"Error in processing extra_attributes: {extra_attributes}. Error: {str(e)}"
                    )
            else:
                extra_attributes = {}

        # for flx, first retrieve object_state which is defined upstream
        object_state = record.get("object_state", "green")

        # status and status_description are used to compose the anomaly_reason
        status = record.get("status", "unknown")
        status_description = record.get("status_description", "unknown")

        # Verify isOutlier
        if isOutlier == 1:
            object_state = "red"
        else:
            pass

        # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
        if object_state == "red":
            if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message_str = f"The entity status is complying with monitoring rules (status: {status}, status_description: {status_description})"
            status_message.append(status_message_str)
            anomaly_reason.append("none")

        else:
            # in flx, the status is managed upstream
            status_message_str = f"The entity status is not complying with monitoring rules (status: {status}, status_description: {status_description})"
            status_message.append(status_message_str)
            anomaly_reason.append("status_not_met")

            # Other statements
            if isOutlier == 1:
                outlier_reasons = record.get("isOutlierReason", [])
                if isinstance(outlier_reasons, list):
                    # Join the list elements into a single string
                    outlier_reasons_str = " | ".join(outlier_reasons)
                    status_message.append(outlier_reasons_str)
                else:
                    # If it's not a list, append it directly
                    status_message.append(outlier_reasons)
                anomaly_reason.append("ml_outliers_detection")

            if isUnderMonitoringDays is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_days")

            if isUnderMonitoringHours is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_hours")

            # logical group
            if isUnderLogicalGroup is True:
                status_message.append(LogicalGroupMsg)
                anomaly_reason.append("in_logical_group")

        # form status_message_json
        status_message_json["status_message"] = status_message
        status_message_json["anomaly_reason"] = anomaly_reason
        if extra_attributes:
            status_message_json["extra_attributes"] = extra_attributes

        # return
        logging.debug(
            f'set_dhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", status_message_json="{status_message_json}", anomaly_reason="{anomaly_reason}"'
        )
        return object_state, status_message, status_message_json, anomaly_reason

    """
    Create a function called set_wlk_status:
    - arguments: record, isOutlier, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_wlk_status(
        self,
        record,
        isOutlier,
        isUnderMonitoringDays,
        isUnderMonitoringHours,
        isUnderMonitoringMsg,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # for wlk, first retrieve object_state which is defined upstream
        object_state = record.get("object_state", "green")

        # status and status_description are used to compose the anomaly_reason
        status = record.get("status", "unknown")
        status_description = record.get("status_description", "unknown")

        # Verify isOutlier
        if isOutlier == 1:
            object_state = "red"
        else:
            pass

        # for wlk, get various functional fields used for the anomaly_reason and status_message definition
        # skipping KPis: skipped_pct, skipped_pct_last_60m, skipped_pct_last_4h, skipped_pct_last_24h
        try:
            skipped_pct = float(record.get("skipped_pct", 0))
        except Exception as e:
            skipped_pct = 0

        try:
            skipped_pct_last_60m = float(record.get("skipped_pct_last_60m", 0))
        except Exception as e:
            skipped_pct_last_60m = 0

        try:
            skipped_pct_last_4h = float(record.get("skipped_pct_last_4h", 0))
        except Exception as e:
            skipped_pct_last_4h = 0

        try:
            skipped_pct_last_24h = float(record.get("skipped_pct_last_24h", 0))
        except Exception as e:
            skipped_pct_last_24h = 0

        # similarly, load:
        # count_errors, count_errors_last_60m, count_errors_last_4h, count_errors_last_24h
        try:
            count_errors = int(record.get("count_errors", 0))
        except Exception as e:
            count_errors = 0

        try:
            count_errors_last_60m = int(record.get("count_errors_last_60m", 0))
        except Exception as e:
            count_errors_last_60m = 0

        try:
            count_errors_last_4h = int(record.get("count_errors_last_4h", 0))
        except Exception as e:
            count_errors_last_4h = 0

        try:
            count_errors_last_24h = int(record.get("count_errors_last_24h", 0))
        except Exception as e:
            count_errors_last_24h = 0

        # retrieve last_seen (epochtime) and cron_exec_sequence_sec (value in seconds)
        try:
            last_seen = int(record.get("last_seen", 0))
        except Exception as e:
            last_seen = 0

        # get last_seen_datetime
        if last_seen > 0:
            last_seen_datetime = self.convert_epoch_to_datetime(last_seen)
        else:
            last_seen_datetime = "unknown"

        try:
            cron_exec_sequence_sec = int(record.get("cron_exec_sequence_sec", 0))
        except Exception as e:
            cron_exec_sequence_sec = 0

        # calculate isDelayed (0 or 1)
        # if now()-last_seen)>(cron_exec_sequence_sec+3600, isDelayed is 1
        now = time.time()
        if (now - last_seen) > (cron_exec_sequence_sec + 3600):
            isDelayed = 1
        else:
            isDelayed = 0

        # calculate the current delay in seconds
        current_delay = now - last_seen

        # get the current delay durection
        current_delay_duration = self.convert_seconds_to_duration(current_delay)

        # retrieve orphan boolean (0 or 1) and load as an integer, as well as orphan_last_check (human readable date)
        try:
            orphan = int(record.get("orphan", 0))
        except Exception as e:
            orphan = 0

        orphan_last_check = record.get("orphan_last_check", "unknown")

        # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
        if object_state == "red":
            if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message_str = f"The entity status is complying with monitoring rules (status: {status}, status_description: {status_description})"
            status_message.append(status_message_str)
            anomaly_reason.append("none")

        else:
            # Other statements

            # ML Outlier
            if isOutlier == 1:
                outlier_reasons = record.get("isOutlierReason", [])
                if isinstance(outlier_reasons, list):
                    # Join the list elements into a single string
                    outlier_reasons_str = " | ".join(outlier_reasons)
                    status_message.append(outlier_reasons_str)
                else:
                    # If it's not a list, append it directly
                    status_message.append(outlier_reasons)
                anomaly_reason.append("ml_outliers_detection")

            # Skipping
            if (
                skipped_pct > 0
                or skipped_pct_last_60m > 0
                or skipped_pct_last_4h > 0
                or skipped_pct_last_24h > 0
            ):
                status_message.append(
                    f"skipping searches were detected, review and address performance issues for this search or finetune its scheduling plan to clear this alert. (skipped_pct_last_60m: {skipped_pct_last_60m}, skipped_pct_last_4h: {skipped_pct_last_4h}, skipped_pct_last_24h: {skipped_pct_last_24h})"
                )
                anomaly_reason.append("skipping_searches_detected")

            # Errors count
            if (
                count_errors > 0
                or count_errors_last_60m > 0
                or count_errors_last_4h > 0
                or count_errors_last_24h > 0
            ):
                status_message.append(
                    f"execution errors were detected, review and address these errors to clear this alert. (count_errors_last_60m: {count_errors_last_60m}, count_errors_last_4h: {count_errors_last_4h}, count_errors_last_24h: {count_errors_last_24h})"
                )
                anomaly_reason.append("execution_errors_detected")

            # orphan
            if orphan == 1:
                status_message.append(
                    f"the search was detected as an orphan search which means the user owning the search is not currently a valid user (orphan: {orphan}, time check: {orphan_last_check}"
                )
                anomaly_reason.append("orphan_search_detected")

            # delayed
            if isDelayed == 1 and cron_exec_sequence_sec > 0:
                status_message.append(
                    f"the search was detected as delayed, this means the search is not running as expected (isDelayed: {isDelayed}, last_seen: {last_seen_datetime}, cron_exec_sequence_sec: {cron_exec_sequence_sec}, current delay: {current_delay_duration} duration)"
                )
                anomaly_reason.append("execution_delayed")

            if isUnderMonitoringDays is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_days")

            if isUnderMonitoringHours is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_hours")

            # if we failed to identify the reason
            if len(status_message) == 0:
                status_message_str = f"The entity status is not complying with monitoring rules (status: {status}, status_description: {status_description})"
                status_message.append(status_message_str)
                anomaly_reason.append("status_not_met")

        # form status_message_json
        status_message_json["status_message"] = status_message
        status_message_json["anomaly_reason"] = anomaly_reason

        # return
        logging.debug(
            f'set_wlk_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
        )
        return object_state, status_message, status_message_json, anomaly_reason

    """
    Create a function called set_cim_status:
    - arguments: record, isOutlier
    - returns: 
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                cim compliance is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    def set_cim_status(
        self,
        record,
        isOutlier,
        isUnderMonitoringDays,
        isUnderMonitoringHours,
        isUnderMonitoringMsg,
    ):
        # init status_message and anomaly_reason
        status_message = []
        anomaly_reason = []

        # init status_message_json
        status_message_json = {}

        # for cim, attempt to retrieve cim_tracking_rules and cim_tracking_results, if present attempt to load as an object
        cim_tracking_rules = record.get("cim_tracking_rules", {})
        if isinstance(cim_tracking_rules, str):
            try:
                cim_tracking_rules = json.loads(cim_tracking_rules)
            except Exception as e:
                logging.error(
                    f"Error in processing cim_tracking_rules: {cim_tracking_rules}. Error: {str(e)}"
                )

        cim_tracking_results = record.get("cim_tracking_results", {})
        if isinstance(cim_tracking_results, str):
            if cim_tracking_results == "unknown":
                cim_tracking_results = {}
            else:
                try:
                    cim_tracking_results = json.loads(cim_tracking_results)
                except Exception as e:
                    logging.error(
                        f"Error in processing cim_tracking_results: {cim_tracking_results}. Error: {str(e)}"
                    )

        # for cim, first retrieve object_state which is defined upstream
        object_state = record.get("object_state", "red")
        if object_state == "unknown":  # if returned unknown, then it's red
            object_state = "red"

        # Verify isOutlier
        if isOutlier == 1:
            object_state = "red"
        else:
            pass

        # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
        if object_state == "red":
            if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
                object_state = "orange"

        # define anomaly_reason
        if object_state == "green":
            status_message_str = (
                f"all compliancy checks on the CIM entity are currently successful."
            )
            status_message.append(status_message_str)
            anomaly_reason.append("none")

        else:
            # in cim, the anomaly_reason is part of the cim_tracking_results

            if len(cim_tracking_results) > 0:
                try:
                    anomaly_reason_str = cim_tracking_results.get(
                        "anomaly_reason", None
                    )

                    if anomaly_reason_str:
                        """
                        if anomaly_reason_str contains:
                        - "compliance_failed:", anomaly_reason is "compliance_failed", status_message is:
                        "CIM compliancy checks have failed for one or more CIM fields: [ " . cim_fields_red . " ], review the compliancy check results and rules, idenfity and address the root causes to fix the entity state."
                        - "no_results:", anomaly_reason is "no_results", status_message is:
                        "Alert: CIM compliancy checks have failed since the main search did not return any results, review the entity definition and its schedule."
                        """

                        if anomaly_reason_str.startswith("compliance_failed:"):
                            anomaly_reason_str = "compliance_failed"
                            cim_fields_red = cim_tracking_results.get(
                                "cim_fields_red", None
                            )
                            if cim_fields_red:
                                status_message_str = f"CIM compliancy checks have failed for one or more CIM fields: {cim_fields_red}, review the compliancy check results and rules, idenfity and address the root causes to fix the entity state."
                            else:
                                status_message_str = f"CIM compliancy checks have failed for compliance issue reason but we failed to extract the cim_fields_red!"

                        elif anomaly_reason_str.startswith("no_results:"):
                            anomaly_reason_str = "no_results"
                            status_message_str = f"CIM compliancy checks have failed since the main search did not return any results, review the entity definition and its schedule."

                        anomaly_reason.append(anomaly_reason_str)
                        status_message.append(status_message_str)

                    else:
                        anomaly_reason_str = "unknown"
                        status_message_str = (
                            "CIM compliancy checks have failed for an unknown reason!"
                        )
                        anomaly_reason.append(anomaly_reason_str)
                        status_message.append(status_message_str)

                except Exception as e:
                    logging.error(
                        f'failed to process the record="{json.dumps(record, indent=2)}", exception="{str(e)}"'
                    )
                    anomaly_reason_str = "unknown"
                    status_message_str = (
                        "CIM compliancy checks have failed for an unknown reason!"
                    )
                    anomaly_reason.append(anomaly_reason_str)
                    status_message.append(status_message_str)

            else:
                anomaly_reason_str = "unknown"
                status_message_str = "CIM compliancy checks have not been performed yet, or no results were resulting since the CIM entity was created, run the tracker now and review the results again."
                anomaly_reason.append(anomaly_reason_str)
                status_message.append(status_message_str)

            # Other statements
            if isOutlier == 1:
                outlier_reasons = record.get("isOutlierReason", [])
                if isinstance(outlier_reasons, list):
                    # Join the list elements into a single string
                    outlier_reasons_str = " | ".join(outlier_reasons)
                    status_message.append(outlier_reasons_str)
                else:
                    # If it's not a list, append it directly
                    status_message.append(outlier_reasons)
                anomaly_reason.append("ml_outliers_detection")

            if isUnderMonitoringDays is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_days")

            if isUnderMonitoringHours is False:
                status_message.append(isUnderMonitoringMsg)
                anomaly_reason.append("out_of_monitoring_hours")

        # form status_message_json
        status_message_json["status_message"] = status_message
        status_message_json["anomaly_reason"] = anomaly_reason

        # return
        logging.debug(
            f'set_cim_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", status_message_json="{status_message_json}", anomaly_reason="{anomaly_reason}"'
        )
        return object_state, status_message, status_message_json, anomaly_reason

    """
    Stream function
    """

    def stream(self, records):
        # Start performance counter
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # Get virtual tenant account
        vtenant_conf = trackme_vtenant_account(
            self._metadata.searchinfo.session_key,
            self._metadata.searchinfo.splunkd_uri,
            self.tenant_id,
        )

        #
        # System level settings
        #

        system_future_tolerance = float(
            reqinfo["trackme_conf"]["trackme_general"]["trackme_future_tolerance"]
        )

        # dhm specific

        if self.component == "dhm":
            macro_name = (
                f"trackme_dhm_default_splk_dhm_alert_policy_tenant_{self.tenant_id}"
            )
            macro_current = self.service.confs["macros"][macro_name]
            default_splk_dhm_alerting_policy = macro_current.content.get("definition")
            # remove double quotes from default_splk_dhm_alerting_policy
            default_splk_dhm_alerting_policy = default_splk_dhm_alerting_policy.replace(
                '"', ""
            )

            logging.debug(
                f'default_splk_dhm_alerting_policy="{default_splk_dhm_alerting_policy}"'
            )

        #
        # Virtual tenant account settings
        #

        # outliers tenant level settings
        tenant_outliers_set_state = int(vtenant_conf["outliers_set_state"])
        tenant_data_sampling_set_state = int(vtenant_conf["data_sampling_set_state"])

        # Process records
        processed_records = []
        records_count = 0
        for record in records:
            records_count += 1
            try:
                new_record = {}

                # get object_value and key
                object_value = record.get("object", None)
                logging.debug(
                    f'object="{object_value}", record="{json.dumps(record, indent=2)}"'
                )

                # The value for key is normally in the field keyid, but in some cases it is in the field key or _key
                # use keyid, key, _key in that order of preference
                key_value = record.get("keyid", None)
                if key_value is None:
                    key_value = record.get("key", None)
                if key_value is None:
                    key_value = record.get("_key", None)

                #
                # splk-dsm
                #

                # get record fields depending on the component
                if self.component == "dsm":
                    # get outliers and data sampling
                    isOutlier = int(record.get("isOutlier", 0))
                    isAnomaly = int(record.get("isAnomaly", 0))
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", isAnomaly="{isAnomaly}"'
                    )

                    # get future_tolerance
                    future_tolerance = record.get("future_tolerance", 0)
                    try:
                        future_tolerance = float(future_tolerance)
                    except:
                        future_tolerance = 0

                    # get data_lag_alert_kpis
                    data_lag_alert_kpis = record.get("data_lag_alert_kpis", "all_kpis")

                    # get actual primary KPI values
                    data_last_ingestion_lag_seen = record.get(
                        "data_last_ingestion_lag_seen", 0
                    )
                    data_last_lag_seen = record.get("data_last_lag_seen", 0)

                    # get per entity thresholds
                    data_max_lag_allowed = float(record.get("data_max_lag_allowed", 0))
                    data_max_delay_allowed = float(
                        record.get("data_max_delay_allowed", 0)
                    )
                    min_dcount_threshold = record.get("min_dcount_threshold", 0)
                    try:
                        min_dcount_threshold = float(min_dcount_threshold)
                    except:
                        min_dcount_threshold = 0

                    # get dcount host related information
                    min_dcount_host = record.get("min_dcount_host", "any")
                    try:
                        min_dcount_host = float(min_dcount_host)
                    except:
                        pass
                    min_dcount_field = record.get("min_dcount_field", None)

                    # get monitoring wdays and hours ranges
                    data_monitoring_wdays = record.get(
                        "data_monitoring_wdays", "auto:all_days"
                    )
                    data_monitoring_hours_ranges = record.get(
                        "data_monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # get logical group information: object_group_key, object_group_name, object_group_min_green_percent, object_group_green_percent, object_group_members_count, object_group_members_red
                    object_group_key = record.get("object_group_key", "")
                    object_group_name = record.get("object_group_name", "")
                    object_group_min_green_percent = record.get(
                        "object_group_min_green_percent", 0
                    )
                    try:
                        object_group_min_green_percent = float(
                            object_group_min_green_percent
                        )
                    except:
                        object_group_min_green_percent = 0
                    object_group_green_percent = record.get(
                        "object_group_green_percent", 0
                    )
                    try:
                        object_group_green_percent = float(object_group_green_percent)
                    except:
                        object_group_green_percent = 0

                    object_group_members_count = record.get(
                        "object_group_members_count", 0
                    )
                    object_group_members_red = record.get("object_group_members_red", 0)

                    # get data_last_ingest, data_last_time_seen, data_last_time_seen_idx (epochtime)
                    data_last_ingest = record.get("data_last_ingest", 0)
                    try:
                        data_last_ingest = float(data_last_ingest)
                    except:
                        pass
                    data_last_time_seen = record.get("data_last_time_seen", 0)
                    try:
                        data_last_time_seen = float(data_last_time_seen)
                    except:
                        pass
                    data_last_time_seen_idx = record.get("data_last_time_seen_idx", 0)
                    try:
                        data_last_time_seen_idx = float(data_last_time_seen_idx)
                    except:
                        pass

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = self.get_monitoring_wdays(
                        data_monitoring_wdays, data_monitoring_hours_ranges
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", data_monitoring_wdays="{data_monitoring_wdays}", data_monitoring_hours_ranges="{data_monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = self.get_outliers_status(
                        isOutlier, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # call get_data_sampling_status and define isAnomaly
                    isAnomaly = self.get_data_sampling_status(
                        isAnomaly, tenant_data_sampling_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isAnomaly="{isAnomaly}", tenant_data_sampling_set_state="{tenant_data_sampling_set_state}"'
                    )

                    # call get_future_status and define isFuture
                    (
                        isFuture,
                        isFutureMsg,
                        merged_future_tolerance,
                    ) = self.get_future_status(
                        future_tolerance,
                        system_future_tolerance,
                        data_last_lag_seen,
                        data_last_ingestion_lag_seen,
                        data_last_time_seen,
                        data_last_ingest,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", future_tolerance="{future_tolerance}", system_future_tolerance="{system_future_tolerance}", merged_future_tolerance="{merged_future_tolerance}", data_last_lag_seen="{data_last_lag_seen}", isFutureMsg="{isFutureMsg}"'
                    )

                    # call get_is_under_dcount_host and define isUnderDcountHost
                    (
                        isUnderDcountHost,
                        isUnderDcountHostMsg,
                    ) = self.get_is_under_dcount_host(
                        min_dcount_host, min_dcount_threshold, min_dcount_field
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDcountHost="{isUnderDcountHost}", isUnderDcountHostMsg="{isUnderDcountHostMsg}", min_dcount_host="{min_dcount_host}", min_dcount_threshold="{min_dcount_threshold}"'
                    )

                    # call get_logical_group_status and define isUnderLogicalGroup and LogicalGroupStateInAlert
                    (
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    ) = self.get_logical_group_status(
                        object_group_key,
                        object_group_name,
                        object_group_min_green_percent,
                        object_group_green_percent,
                        object_group_members_count,
                        object_group_members_red,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}", object_group_key="{object_group_key}", object_group_name="{object_group_name}", object_group_min_green_percent="{object_group_min_green_percent}", object_group_green_percent="{object_group_green_percent}", object_group_members_count="{object_group_members_count}", object_group_members_red="{object_group_members_red}"'
                    )

                    # call get_dsm_latency_status and define isUnderLatencyAlert and isUnderLatencyMessage
                    (
                        isUnderLatencyAlert,
                        isUnderLatencyMessage,
                    ) = self.get_dsm_latency_status(
                        data_last_ingestion_lag_seen,
                        data_max_lag_allowed,
                        data_lag_alert_kpis,
                        data_last_ingest,
                        data_last_time_seen,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLatencyAlert="{isUnderLatencyAlert}", isUnderLatencyMessage="{isUnderLatencyMessage}", data_last_ingestion_lag_seen="{data_last_ingestion_lag_seen}", data_max_lag_allowed="{data_max_lag_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                    )

                    # call get_dsm_delay_status and define isUnderDelayAlert and isUnderDelayMessage
                    (
                        isUnderDelayAlert,
                        isUnderDelayMessage,
                    ) = self.get_dsm_delay_status(
                        data_last_lag_seen,
                        data_max_delay_allowed,
                        data_lag_alert_kpis,
                        data_last_ingest,
                        data_last_time_seen,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDelayAlert="{isUnderDelayAlert}", isUnderDelayMessage="{isUnderDelayMessage}", data_last_lag_seen="{data_last_lag_seen}", data_max_delay_allowed="{data_max_delay_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                    )

                    # call set_dsm_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = self.set_dsm_status(
                        record,
                        isOutlier,
                        isAnomaly,
                        isFuture,
                        isFutureMsg,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                        isUnderDcountHost,
                        isUnderDcountHostMsg,
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                        isUnderLatencyAlert,
                        isUnderLatencyMessage,
                        isUnderDelayAlert,
                        isUnderDelayMessage,
                    )
                    logging.debug(
                        f'set_dsm_status, tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # future tolerance
                    new_record["future_tolerance"] = int(
                        round(merged_future_tolerance, 0)
                    )

                    # convert data_last_time_seen to last_time from epoch
                    last_time = self.convert_epoch_to_datetime(data_last_time_seen)
                    new_record["last_time"] = last_time

                    # convert data_last_ingest to last_ingest from epoch
                    last_ingest = self.convert_epoch_to_datetime(data_last_ingest)
                    new_record["last_ingest"] = last_ingest

                    # convert data_last_time_seen_idx to last_time_idx from epoch
                    last_time_idx = self.convert_epoch_to_datetime(data_last_time_seen)
                    new_record["last_time_idx"] = last_time_idx

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

                #
                # splk-dhm
                #

                elif self.component == "dhm":
                    # get splk_dhm_st_summary
                    splk_dhm_st_summary = record.get("splk_dhm_st_summary", None)
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", splk_dhm_st_summary="{splk_dhm_st_summary}"'
                    )

                    # get outliers and data sampling
                    isOutlier = int(record.get("isOutlier", 0))
                    isAnomaly = int(record.get("isAnomaly", 0))
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", isAnomaly="{isAnomaly}"'
                    )

                    # get future_tolerance
                    future_tolerance = record.get("future_tolerance", 0)
                    try:
                        future_tolerance = float(future_tolerance)
                    except:
                        future_tolerance = 0

                    # get data_lag_alert_kpis
                    data_lag_alert_kpis = record.get("data_lag_alert_kpis", "all_kpis")

                    # get actual primary KPI values
                    data_last_ingestion_lag_seen = record.get(
                        "data_last_ingestion_lag_seen", 0
                    )
                    data_last_lag_seen = record.get("data_last_lag_seen", 0)

                    # get per entity thresholds
                    data_max_lag_allowed = float(record.get("data_max_lag_allowed", 0))
                    data_max_delay_allowed = float(
                        record.get("data_max_delay_allowed", 0)
                    )

                    # get monitoring wdays and hours ranges
                    data_monitoring_wdays = record.get(
                        "data_monitoring_wdays", "auto:all_days"
                    )
                    data_monitoring_hours_ranges = record.get(
                        "data_monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # get logical group information: object_group_key, object_group_name, object_group_min_green_percent, object_group_green_percent, object_group_members_count, object_group_members_red
                    object_group_key = record.get("object_group_key", "")
                    object_group_name = record.get("object_group_name", "")
                    object_group_min_green_percent = record.get(
                        "object_group_min_green_percent", 0
                    )
                    try:
                        object_group_min_green_percent = float(
                            object_group_min_green_percent
                        )
                    except:
                        object_group_min_green_percent = 0
                    object_group_green_percent = record.get(
                        "object_group_green_percent", 0
                    )
                    try:
                        object_group_green_percent = float(object_group_green_percent)
                    except:
                        object_group_green_percent = 0

                    object_group_members_count = record.get(
                        "object_group_members_count", 0
                    )
                    object_group_members_red = record.get("object_group_members_red", 0)

                    # get data_last_ingest, data_last_time_seen, data_last_time_seen_idx (epochtime)
                    data_last_ingest = record.get("data_last_ingest", 0)
                    try:
                        data_last_ingest = float(data_last_ingest)
                    except:
                        pass
                    data_last_time_seen = record.get("data_last_time_seen", 0)
                    try:
                        data_last_time_seen = float(data_last_time_seen)
                    except:
                        pass
                    data_last_time_seen_idx = record.get("data_last_time_seen_idx", 0)
                    try:
                        data_last_time_seen_idx = float(data_last_time_seen_idx)
                    except:
                        pass

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = self.get_monitoring_wdays(
                        data_monitoring_wdays, data_monitoring_hours_ranges
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", data_monitoring_wdays="{data_monitoring_wdays}", data_monitoring_hours_ranges="{data_monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = self.get_outliers_status(
                        isOutlier, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # call get_future_status and define isFuture
                    (
                        isFuture,
                        isFutureMsg,
                        merged_future_tolerance,
                    ) = self.get_future_status(
                        future_tolerance,
                        system_future_tolerance,
                        data_last_lag_seen,
                        data_last_ingestion_lag_seen,
                        data_last_time_seen,
                        data_last_ingest,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", future_tolerance="{future_tolerance}", system_future_tolerance="{system_future_tolerance}", merged_future_tolerance="{merged_future_tolerance}", data_last_lag_seen="{data_last_lag_seen}", isFutureMsg="{isFutureMsg}"'
                    )

                    # call get_logical_group_status and define isUnderLogicalGroup and LogicalGroupStateInAlert
                    (
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    ) = self.get_logical_group_status(
                        object_group_key,
                        object_group_name,
                        object_group_min_green_percent,
                        object_group_green_percent,
                        object_group_members_count,
                        object_group_members_red,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}", object_group_key="{object_group_key}", object_group_name="{object_group_name}", object_group_min_green_percent="{object_group_min_green_percent}", object_group_green_percent="{object_group_green_percent}", object_group_members_count="{object_group_members_count}", object_group_members_red="{object_group_members_red}"'
                    )

                    # call get_dsm_latency_status and define isUnderLatencyAlert and isUnderLatencyMessage
                    (
                        isUnderLatencyAlert,
                        isUnderLatencyMessage,
                    ) = self.get_dsm_latency_status(
                        data_last_ingestion_lag_seen,
                        data_max_lag_allowed,
                        data_lag_alert_kpis,
                        data_last_ingest,
                        data_last_time_seen,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLatencyAlert="{isUnderLatencyAlert}", isUnderLatencyMessage="{isUnderLatencyMessage}", data_last_ingestion_lag_seen="{data_last_ingestion_lag_seen}", data_max_lag_allowed="{data_max_lag_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                    )

                    # call get_dsm_delay_status and define isUnderDelayAlert and isUnderDelayMessage
                    (
                        isUnderDelayAlert,
                        isUnderDelayMessage,
                    ) = self.get_dsm_delay_status(
                        data_last_lag_seen,
                        data_max_delay_allowed,
                        data_lag_alert_kpis,
                        data_last_ingest,
                        data_last_time_seen,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDelayAlert="{isUnderDelayAlert}", isUnderDelayMessage="{isUnderDelayMessage}", data_last_lag_seen="{data_last_lag_seen}", data_max_delay_allowed="{data_max_delay_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                    )

                    # call set_dhm_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                        splk_dhm_alerting_policy,
                    ) = self.set_dhm_status(
                        record,
                        isOutlier,
                        isFuture,
                        isFutureMsg,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                        isUnderLatencyAlert,
                        isUnderLatencyMessage,
                        isUnderDelayAlert,
                        isUnderDelayMessage,
                        default_splk_dhm_alerting_policy,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # future tolerance
                    new_record["future_tolerance"] = int(
                        round(merged_future_tolerance, 0)
                    )

                    # specific for dhm
                    new_record["splk_dhm_alerting_policy"] = splk_dhm_alerting_policy

                    # convert data_last_time_seen to last_time from epoch
                    last_time = self.convert_epoch_to_datetime(data_last_time_seen)
                    new_record["last_time"] = last_time

                    # convert data_last_ingest to last_ingest from epoch
                    last_ingest = self.convert_epoch_to_datetime(data_last_ingest)
                    new_record["last_ingest"] = last_ingest

                    # convert data_last_time_seen_idx to last_time_idx from epoch
                    last_time_idx = self.convert_epoch_to_datetime(data_last_time_seen)
                    new_record["last_time_idx"] = last_time_idx

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

                #
                # splk-mhm
                #

                elif self.component == "mhm":
                    # get metric_details
                    metric_details = record.get("metric_details", None)
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", metric_details="{metric_details}"'
                    )

                    # get logical group information: object_group_key, object_group_name, object_group_min_green_percent, object_group_green_percent, object_group_members_count, object_group_members_red
                    object_group_key = record.get("object_group_key", "")
                    object_group_name = record.get("object_group_name", "")
                    object_group_min_green_percent = record.get(
                        "object_group_min_green_percent", 0
                    )
                    try:
                        object_group_min_green_percent = float(
                            object_group_min_green_percent
                        )
                    except:
                        object_group_min_green_percent = 0
                    object_group_green_percent = record.get(
                        "object_group_green_percent", 0
                    )
                    try:
                        object_group_green_percent = float(object_group_green_percent)
                    except:
                        object_group_green_percent = 0

                    object_group_members_count = record.get(
                        "object_group_members_count", 0
                    )
                    object_group_members_red = record.get("object_group_members_red", 0)

                    # get metric_last_time_seen (epochtime)
                    metric_last_time_seen = record.get("metric_last_time_seen", 0)
                    try:
                        metric_last_time_seen = float(metric_last_time_seen)
                    except:
                        pass

                    # call get_future_metrics_status and define isFuture
                    isFuture, isFutureMsg = self.get_future_metrics_status(
                        system_future_tolerance,
                        metric_last_time_seen,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", system_future_tolerance="{system_future_tolerance}", metric_last_time_seen="{metric_last_time_seen}", isFutureMsg="{isFutureMsg}"'
                    )

                    # call get_logical_group_status and define isUnderLogicalGroup and LogicalGroupStateInAlert
                    (
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    ) = self.get_logical_group_status(
                        object_group_key,
                        object_group_name,
                        object_group_min_green_percent,
                        object_group_green_percent,
                        object_group_members_count,
                        object_group_members_red,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}", object_group_key="{object_group_key}", object_group_name="{object_group_name}", object_group_min_green_percent="{object_group_min_green_percent}", object_group_green_percent="{object_group_green_percent}", object_group_members_count="{object_group_members_count}", object_group_members_red="{object_group_members_red}"'
                    )

                    # call set_mhm_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = self.set_mhm_status(
                        record,
                        isFuture,
                        isFutureMsg,
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # convert metric_last_time_seen to last_time from epoch
                    last_time = self.convert_epoch_to_datetime(metric_last_time_seen)
                    new_record["last_time"] = last_time

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

                #
                # splk-flx
                #

                # get record fields depending on the component
                elif self.component == "flx":
                    # get outliers
                    isOutlier = int(record.get("isOutlier", 0))
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                    )

                    # get monitoring wdays and hours ranges
                    monitoring_wdays = record.get("monitoring_wdays", "auto:all_days")
                    monitoring_hours_ranges = record.get(
                        "monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # get logical group information: object_group_key, object_group_name, object_group_min_green_percent, object_group_green_percent, object_group_members_count, object_group_members_red
                    object_group_key = record.get("object_group_key", "")
                    object_group_name = record.get("object_group_name", "")
                    object_group_min_green_percent = record.get(
                        "object_group_min_green_percent", 0
                    )
                    try:
                        object_group_min_green_percent = float(
                            object_group_min_green_percent
                        )
                    except:
                        object_group_min_green_percent = 0
                    object_group_green_percent = record.get(
                        "object_group_green_percent", 0
                    )
                    try:
                        object_group_green_percent = float(object_group_green_percent)
                    except:
                        object_group_green_percent = 0

                    object_group_members_count = record.get(
                        "object_group_members_count", 0
                    )
                    object_group_members_red = record.get("object_group_members_red", 0)

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = self.get_monitoring_wdays(
                        monitoring_wdays, monitoring_hours_ranges
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = self.get_outliers_status(
                        isOutlier, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # call get_logical_group_status and define isUnderLogicalGroup and LogicalGroupStateInAlert
                    (
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    ) = self.get_logical_group_status(
                        object_group_key,
                        object_group_name,
                        object_group_min_green_percent,
                        object_group_green_percent,
                        object_group_members_count,
                        object_group_members_red,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}", object_group_key="{object_group_key}", object_group_name="{object_group_name}", object_group_min_green_percent="{object_group_min_green_percent}", object_group_green_percent="{object_group_green_percent}", object_group_members_count="{object_group_members_count}", object_group_members_red="{object_group_members_red}"'
                    )

                    # call set_flx_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = self.set_flx_status(
                        record,
                        isOutlier,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                        isUnderLogicalGroup,
                        LogicalGroupStateInAlert,
                        LogicalGroupMsg,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

                #
                # splk-wlk
                #

                # get record fields depending on the component
                elif self.component == "wlk":
                    # get outliers
                    isOutlier = int(record.get("isOutlier", 0))
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                    )

                    # get monitoring wdays and hours ranges
                    monitoring_wdays = record.get("monitoring_wdays", "auto:all_days")
                    monitoring_hours_ranges = record.get(
                        "monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = self.get_monitoring_wdays(
                        monitoring_wdays, monitoring_hours_ranges
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = self.get_outliers_status(
                        isOutlier, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # call set_wlk_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = self.set_wlk_status(
                        record,
                        isOutlier,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

                #
                # splk-cim
                #

                # get record fields depending on the component
                elif self.component == "cim":
                    # get outliers
                    isOutlier = int(record.get("isOutlier", 0))
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = self.get_outliers_status(
                        isOutlier, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # get monitoring wdays and hours ranges
                    monitoring_wdays = record.get("monitoring_wdays", "auto:all_days")
                    monitoring_hours_ranges = record.get(
                        "monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = self.get_monitoring_wdays(
                        monitoring_wdays, monitoring_hours_ranges
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call set_cim_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = self.set_cim_status(
                        record,
                        isOutlier,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record[
                        "latest_flip_time_human"
                    ] = self.convert_epoch_to_datetime(latest_flip_time_human)

            #
            # End per component processing
            #

            except Exception as e:
                logging.error(
                    f'tenant_id="{self.tenant_id}", component="{self.component}", Error processing record, record="{json.dumps(record, indent=2)}", exception="{str(e)}"'
                )
                continue  # Proceed with next record

            #
            # End per component processing
            #

            # add all key value pairs from the original record to new_record if not present already
            for key, value in record.items():
                if key not in new_record:
                    new_record[key] = value

            # add new_record to processed_records
            processed_records.append(new_record)

        #
        # Render
        #

        for yield_record in self.generate_fields(processed_records):
            # logging
            logging.debug(f'yield_record="{json.dumps(yield_record, indent=2)}"')

            # yield record
            yield yield_record

        # performance counter
        logging.info(
            f'trackmedecisionmaker has terminated, tenant_id="{self.tenant_id}", component="{self.component}", upstream_records="{records_count}", processed_records="{len(processed_records)}", run_time="{round(time.time() - start, 3)}"'
        )


dispatch(TrackMeDecisionMaker, sys.argv, sys.stdin, sys.stdout, __name__)
