#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library
import os
import sys
import time
import json
import hashlib
from datetime import datetime

# External libraries
import urllib3

# Disable urllib3 warnings
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Configure logging
import logging
from logging.handlers import RotatingFileHandler

splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    os.path.join(splunkhome, "var", "log", "splunk", "trackme_persistentfields.log"),
    mode="a",
    maxBytes=10_000_000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    StreamingCommand,
    Configuration,
    Option,
    validators,
)

# Import trackme libs
from trackme_libs import trackme_reqinfo
from trackme_libs_utils import decode_unicode


@Configuration(distributed=False)
class TrackMePersistentHandler(StreamingCommand):
    collection = Option(
        doc="""
        **Syntax:** **collection=****
        **Description:** Specify the collection.""",
        require=True,
        default="None",
        validate=validators.Match("collection", r"^.*$"),
    )

    key = Option(
        doc="""
        **Syntax:** **key=****
        **Description:** Specify the key.""",
        require=True,
        default="None",
        validate=validators.Match("key", r"^.*$"),
    )

    update_collection = Option(
        doc="""
        **Syntax:** **update_collection=****
        **Description:** Enables or disables updating and inserting innto the collection, this replaces the need from calling outputlookup.""",
        require=False,
        default=False,
        validate=validators.Match("key", r"^(True|False)$"),
    )

    def stream(self, records):
        # performance counter
        start_time = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # connect to the KVstore
        target_collection = f"kv_{self.collection}"
        collection = self.service.kvstore[target_collection]

        # set the component
        persistent_fields = []

        if self.collection.startswith("trackme_dsm_"):
            persistent_fields = [
                "priority",
                "alias",
                "data_max_lag_allowed",
                "data_max_delay_allowed",
                "future_tolerance",
                "data_lag_alert_kpis",
                "monitored_state",
                "data_monitoring_wdays",
                "data_monitoring_hours_ranges",
                "data_override_lagging_class",
                "allow_adaptive_delay",
                "min_dcount_host",
                "min_dcount_field",
                "tags_manual",
            ]
        elif self.collection.startswith("trackme_dhm_"):
            persistent_fields = [
                "priority",
                "alias",
                "data_max_lag_allowed",
                "data_max_delay_allowed",
                "future_tolerance",
                "data_lag_alert_kpis",
                "monitored_state",
                "data_monitoring_wdays",
                "data_monitoring_hours_ranges",
                "data_override_lagging_class",
                "allow_adaptive_delay",
                "min_dcount_host",
                "tags",
                "splk_dhm_st_summary_full",
                "splk_dhm_st_summary_compact",
            ]
        elif self.collection.startswith("trackme_mhm_"):
            persistent_fields = [
                "priority",
                "alias",
                "metric_max_lag_allowed",
                "monitored_state",
                "metric_monitoring_wdays",
                "metric_override_lagging_class",
                "metric_details_full",
                "metric_details_compact",
            ]
        elif self.collection.startswith("trackme_flx_"):
            persistent_fields = [
                "priority",
                "alias",
                "monitored_state",
                "data_monitoring_wdays",
                "data_monitoring_hours_ranges",
                "data_override_lagging_class",
                "tags",
            ]
        elif self.collection.startswith("trackme_cim_"):
            persistent_fields = [
                "priority",
                "alias",
                "monitored_state",
                "cim_tracking_rules",
            ]
        elif self.collection.startswith("trackme_wlk_"):
            persistent_fields = [
                "priority",
                "alias",
                "monitored_state",
                "monitoring_wdays",
                "monitoring_hours_ranges",
                "tags",
            ]

        # get all records
        get_collection_start = time.time()
        collection_records = []
        collection_records_keys = set()

        end = False
        skip_tracker = 0
        while end == False:
            process_collection_records = collection.data.query(skip=skip_tracker)
            if len(process_collection_records) != 0:
                for item in process_collection_records:
                    if item.get("_key") not in collection_records_keys:
                        collection_records.append(item)
                        collection_records_keys.add(item.get("_key"))
                skip_tracker += 5000
            else:
                end = True

        logging.info(
            f'context="perf", get collection records, no_records="{len(collection_records)}", run_time="{round((time.time() - get_collection_start), 3)}", collection="{self.collection}"'
        )

        # create a list of all keys for fast verification
        collection_keys = []

        # create a dictionary of all keys with their persistent fields
        collection_dict = {}

        for collection_record in collection_records:
            collection_keys.append(collection_record.get("_key"))
            collection_subrecord = {}

            # add time of update of this record
            try:
                collection_subrecord["mtime"] = float(collection_record.get("mtime"))
            except Exception as e:
                collection_subrecord["mtime"] = float(time.time())

            # loop and add persistent fields
            for field in persistent_fields:
                try:
                    collection_subrecord[field] = collection_record.get(field)
                except Exception as e:
                    logging.debug(f'field="{field}" was not found in the record')

            # add to our dict
            collection_dict[collection_record.get("_key")] = collection_subrecord

        #
        # Define Meta
        #

        final_records = []

        # Loop in the results
        for record in records:
            record_is_new = record.get(self.key) not in collection_keys

            # get time, if any
            time_event = None
            try:
                time_event = record["_time"]
            except Exception as e:
                time_event = time.time()

            # always set an object_md5 and add _key in the record
            record_object_value = decode_unicode(record["object"])
            record_alias = decode_unicode(record["alias"])
            logging.debug(
                f'object="{record["object"]}", decoded_object="{record_object_value}", alias="{record["alias"]}", decoded_alias="{record_alias}"'
            )

            # add the _key in the record if there is none
            if not record.get("_key"):
                object_md5 = hashlib.md5(
                    record_object_value.encode("utf-8")
                ).hexdigest()
                record["_key"] = object_md5
                logging.debug(
                    f'adding _key="{object_md5}" to record for object="{record_object_value}", alias="{record_alias}"'
                )

            # handle unicode for object and alias
            record["object"] = record_object_value
            record["alias"] = record_alias

            # get tracker_runtime, if any
            tracker_runtime = record.get("tracker_runtime", None)
            if tracker_runtime:
                try:
                    tracker_runtime = float(tracker_runtime)
                except Exception as e:
                    tracker_runtime = time.time()
            else:
                tracker_runtime = time.time()

            # detect conflict update
            conflict_update = False
            if not record_is_new:
                # attempt to retrieve and convert mtime value, if fails for any reason, set conflict_update to False
                try:
                    mtime = float(collection_dict[record.get(self.key)].get("mtime"))
                    if mtime > float(tracker_runtime):
                        conflict_update = True
                        logging.info(
                            f'record key="{record.get(self.key)}", conflict update detected, preserving persistent fields="{persistent_fields}", record="{json.dumps(record, indent=2)}"'
                        )
                    else:
                        conflict_update = False

                except Exception as e:
                    logging.error(
                        f'failed to extract and convert mtime="{mtime}", tracker_runtime="{tracker_runtime}", exception message="{str(e)}"'
                    )
                    conflict_update = False

            # create a summary record
            summary_record = {}

            # Add _time first
            summary_record["_time"] = float(time_event)

            # loop through the dict
            for k in record:
                logging.debug(f'field="{k}", value="{record[k]}"')

                # Exclude the event time, add existing fields
                if k != "_time":
                    #
                    # handle persistent field
                    #

                    if not record_is_new:
                        # if field is in persistent list of fields
                        if k in persistent_fields:
                            # preserve persistent fields if conflict update is detected
                            if conflict_update:
                                summary_record[k] = collection_dict[
                                    record.get(self.key)
                                ].get(k)
                            else:
                                summary_record[k] = record[k]

                        # normal field
                        else:
                            summary_record[k] = record[k]

                    else:
                        # record is new, no need to consider it
                        summary_record[k] = record[k]

            # insert and update the collection if requested
            final_records.append(record)

        # batch update/insert
        batch_update_collection_start = time.time()

        # process by chunk
        chunks = [final_records[i : i + 500] for i in range(0, len(final_records), 500)]
        for chunk in chunks:
            try:
                collection.data.batch_save(*chunk)
            except Exception as e:
                logging.error(f'KVstore batch failed with exception="{str(e)}"')

            # yield
            for record in chunk:
                yield record

        # perf counter for the batch operation
        logging.info(
            f'context="perf", batch KVstore update terminated, no_records="{len(final_records)}", run_time="{round((time.time() - batch_update_collection_start), 3)}", collection="{self.collection}"'
        )

        # perf counter for the entire call
        logging.info(
            f'trackmepersistentfields has terminated, collection="{self.collection}", key="{self.key}", update_collection="{self.update_collection}", run_time="{round((time.time() - start_time), 3)}"'
        )


dispatch(TrackMePersistentHandler, sys.argv, sys.stdin, sys.stdout, __name__)
