#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import re
import json

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_sampling_executor.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# import trackme libs
from trackme_libs import trackme_reqinfo, trackme_register_tenant_object_summary


@Configuration(distributed=False)
class DataSamplingExecutor(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    mode = Option(
        doc="""
        **Syntax:** **mode=****
        **Description:** The data sampling executor mode, valid options are: run_sampling | get_samples""",
        require=False,
        default="run_sampling",
        validate=validators.Match("mode", r"^(run_sampling|get_samples)$"),
    )

    object = Option(
        doc="""
        **Syntax:** **mode=****
        **Description:** The object target, only used if mode is get_samples""",
        require=False,
        default="*",
        validate=validators.Match("object", r"^.*$"),
    )

    earliest = Option(
        doc="""
        **Syntax:** **earliest=****
        **Description:** The earliest time quantifier.""",
        require=False,
        default="-24h",
    )

    latest = Option(
        doc="""
        **Syntax:** **latest=****
        **Description:** The latest time quantifier.""",
        require=False,
        default="now",
    )

    max_runtime = Option(
        doc="""
        **Syntax:** **max_runtime=****
        **Description:** The max runtime for the job in seconds, defaults to 15 minutes less 120 seconds of margin.""",
        require=False,
        default="900",
        validate=validators.Match("max_runtime", r"^\d*$"),
    )

    get_samples_max_count = Option(
        doc="""
        **Syntax:** **get_samples_max_count=****
        **Description:** The max number of events to be sampled in get sample mode, default to 10k events.""",
        require=False,
        default="10000",
        validate=validators.Match("get_samples_max_count", r"^\d*$"),
    )

    def generate(self, **kwargs):
        if self:
            # set max failed re-attempt
            max_failures_count = 24
            sleep_time = 5

            # performance counter
            mainstart = time.time()

            # Get request info and set logging level
            reqinfo = trackme_reqinfo(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
            )
            log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

            # get the default number of events to be sampled per object, at discovery and run
            splk_data_sampling_default_sample_record_at_discovery = reqinfo[
                "trackme_conf"
            ]["splk_data_sampling"][
                "splk_data_sampling_default_sample_record_at_discovery"
            ]
            splk_data_sampling_default_sample_record_at_run = reqinfo["trackme_conf"][
                "splk_data_sampling"
            ]["splk_data_sampling_default_sample_record_at_run"]

            # counter
            count = 0

            # Get the session key
            session_key = self._metadata.searchinfo.session_key

            # Get splunkd port
            splunkd_port = reqinfo["server_rest_port"]

            # build header and target
            header = "Splunk " + str(session_key)

            # Data collection
            collection_name = "kv_trackme_dsm_data_sampling_tenant_" + str(
                self.tenant_id
            )
            collection = self.service.kvstore[collection_name]

            # Get data, purge any record older than 48 hours

            # note: do not allow crap entities, sometimes users get very fancy, entities with double quotes in the value
            # are not expected nor desirable, do not allow crap to get in

            search = (
                "| inputlookup trackme_dsm_tenant_"
                + str(self.tenant_id)
                + ' where monitored_state="enabled"'
                + "\n| eval key=_key"
                + "\n| `trackme_exclude_badentities`"
                + '\n| search object="'
                + self.object
                + '"'
                + '\n| where data_last_time_seen>relative_time(now(), "-24h")'
                + "\n| lookup trackme_dsm_data_sampling_tenant_"
                + str(self.tenant_id)
                + " object OUTPUT data_sample_feature"
                + '\n| eval data_sample_feature=if(isnull(data_sample_feature), "enabled", data_sample_feature) | where data_sample_feature!="disabled"'
                + "\n| sort limit=0 data_sample_lastrun"
                + "\n| fields object, key"
            )

            kwargs_oneshot = {
                "earliest_time": self.earliest,
                "latest_time": self.latest,
                "count": 0,
                "output_mode": "json",
            }

            logging.debug('search="{}"'.format(search))

            # available modes
            # - run_sampling: run the full sampling process, as expected per schedule
            # - get_samples: get samples for simulation or inline search purposes

            if self.mode == "run_sampling":
                # report name for logging purposes
                report_name = "trackme_dsm_data_sampling_tracker_tenant_" + str(
                    self.tenant_id
                )

                # run the main report, every result is a Splunk search to be executed on its own thread
                objects_list = []

                # automatic retry
                search_successful = False
                current_failures_count = 0
                while (
                    current_failures_count < max_failures_count
                    and not search_successful
                ):
                    try:
                        oneshotsearch_results = self.service.jobs.oneshot(
                            search, **kwargs_oneshot
                        )
                        reader = results.JSONResultsReader(oneshotsearch_results)

                        for item in reader:
                            if isinstance(item, dict):
                                logging.debug('search_results="{}"'.format(item))
                                # append to the list of searches
                                objects_list.append(
                                    {
                                        "object": item.get("object"),
                                        "key": item.get("key"),
                                    }
                                )

                        # break while
                        search_successful = True
                        break

                    except Exception as e:
                        if re.search(
                            "The maximum number of concurrent historical searches on this instance has been reached",
                            str(e),
                        ):
                            current_failures_count += 1
                            search_successful = False
                            if current_failures_count >= max_failures_count:
                                # Call the component register
                                trackme_register_tenant_object_summary(
                                    session_key,
                                    self._metadata.searchinfo.splunkd_uri,
                                    self.tenant_id,
                                    "splk-dsm",
                                    "trackme_dsm_data_sampling_tracker_tenant_"
                                    + str(self.tenant_id),
                                    "failure",
                                    time.time(),
                                    str(time.time() - mainstart),
                                    str(e),
                                    "-24h",
                                    "now",
                                )
                                msg = 'tenant_id="{}", component="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                    self.tenant_id,
                                    "splk-dsm",
                                    current_failures_count,
                                    max_failures_count,
                                    str(e),
                                )
                                logging.error(msg)
                                raise Exception(msg)
                            else:
                                msg = 'tenant_id="{}", component="{}", temporary search failure with exception="{}"'.format(
                                    self.tenant_id, "splk-dsm", str(e)
                                )
                                logging.warn(msg)
                                time.sleep(sleep_time)
                        else:
                            # Call the component register
                            trackme_register_tenant_object_summary(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                "splk-dsm",
                                "trackme_dsm_data_sampling_tracker_tenant_"
                                + str(self.tenant_id),
                                "failure",
                                time.time(),
                                str(time.time() - mainstart),
                                str(e),
                                "-24h",
                                "now",
                            )
                            msg = 'tenant_id="{}", component="{}", search failed with exception="{}"'.format(
                                self.tenant_id, "splk-dsm", str(e)
                            )
                            logging.error(msg)
                            raise Exception(msg)

                # loop
                logging.debug(
                    'objects_list="{}"'.format(json.dumps(objects_list, indent=2))
                )
                for object_dict in objects_list:
                    object_value = object_dict.get("object")
                    object_key = object_dict.get("key")

                    # break if reaching the max run time less 120 seconds of margin
                    if (time.time() - int(mainstart)) - 120 >= int(self.max_runtime):
                        logging.info(
                            'tenant_id="{}" max_runtime="{}" is about to be reached, current_runtime="{}", job will be terminated now'.format(
                                self.tenant_id,
                                str(self.max_runtime),
                                str(time.time() - int(mainstart)),
                            )
                        )
                        break

                    # check if we have a KVrecord already for this object
                    query_string = {
                        "$and": [
                            {
                                "object": object_value,
                            }
                        ]
                    }
                    try:
                        # try get to get the key
                        kvrecord = collection.data.query(
                            query=(json.dumps(query_string))
                        )[0]
                        key = kvrecord.get("_key")
                    except Exception as e:
                        key = None

                    # if we have a KVrecord, check if a custom value is set for the number of records to be sampled, otherwise fallback to default
                    data_sampling_nr = None
                    if key:
                        try:
                            data_sampling_nr = kvrecord.get("data_sampling_nr")
                        except Exception as e:
                            data_sampling_nr = None

                    if not data_sampling_nr and key:
                        data_sampling_nr = (
                            splk_data_sampling_default_sample_record_at_discovery
                        )
                    elif key:
                        data_sampling_nr = (
                            splk_data_sampling_default_sample_record_at_run
                        )
                    else:
                        data_sampling_nr = (
                            splk_data_sampling_default_sample_record_at_discovery
                        )

                    logging.debug('data_sampling_nr="{}"'.format(data_sampling_nr))

                    # init
                    search = None

                    # get the entity info
                    try:
                        json_data = {
                            "tenant_id": self.tenant_id,
                            "object": object_value,
                        }
                        target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dsm/ds_entity_info"

                        response = requests.post(
                            target_url,
                            headers={
                                "Authorization": header,
                                "Content-Type": "application/json",
                            },
                            verify=False,
                            data=json.dumps(json_data),
                            timeout=300,
                        )
                        search = json.loads(response.text).get(
                            "splk_dsm_sampling_search"
                        )

                        if search and search != "N/A":
                            # replace the number of records to be sampled
                            if re.search("search", "splunkremote "):
                                search = search.replace(
                                    "head 1000", "head " + str(data_sampling_nr)
                                )
                            else:
                                search = search + " | head " + str(data_sampling_nr)

                            # add the key
                            search = (
                                search
                                + '\n| eval key="'
                                + str(object_key)
                                + '", object="'
                                + str(object_value)
                                + '"'
                                + "\n| rename _raw as raw_sample, sourcetype as data_sourcetype"
                            )

                        else:
                            search = None

                    except Exception as e:
                        logging.error(
                            'object="{}", could not retrieve entity info data sampling search, exception="{}"'.format(
                                object_value, str(e)
                            )
                        )

                    # run the main report, every result is a Splunk search to be executed on its own thread
                    if search:
                        # proceed
                        logging.info(
                            'tenant_id="{}" Executing data sampling resulting search="{}"'.format(
                                self.tenant_id, search
                            )
                        )

                        # automatic retry
                        search_successful = False
                        current_failures_count = 0
                        while (
                            current_failures_count < max_failures_count
                            and not search_successful
                        ):
                            # start
                            start = time.time()
                            try:
                                search_results = self.service.jobs.oneshot(
                                    search, **kwargs_oneshot
                                )
                                count += 1

                                # Get the results and display them using the JSONResultsReader
                                reader = results.JSONResultsReader(search_results)
                                for item in reader:
                                    if isinstance(item, dict):
                                        logging.debug(
                                            'search_results="{}"'.format(item)
                                        )
                                        data = {
                                            "_time": time.time(),
                                            "_raw": item,
                                            "key": item.get("key"),
                                            "object": item.get("object"),
                                            "raw_sample": item.get("raw_sample"),
                                            "data_sourcetype": item.get(
                                                "data_sourcetype"
                                            ),
                                        }
                                        yield data

                                logging.info(
                                    'tenant_id="{}" search successfully executed in {} seconds'.format(
                                        self.tenant_id, round(time.time() - start, 3)
                                    )
                                )
                                search_successful = True

                            except Exception as e:
                                if re.search(
                                    "The maximum number of concurrent historical searches on this instance has been reached",
                                    str(e),
                                ):
                                    current_failures_count += 1
                                    search_successful = False
                                    if current_failures_count >= max_failures_count:
                                        msg = 'tenant_id="{}" permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                            self.tenant_id,
                                            current_failures_count,
                                            max_failures_count,
                                            str(e),
                                        )
                                        logging.error(msg)
                                    else:
                                        msg = 'tenant_id="{}" temporary search failure with exception="{}"'.format(
                                            self.tenant_id, str(e)
                                        )
                                        logging.warn(msg)
                                        time.sleep(sleep_time)
                                else:
                                    # Call the component register
                                    msg = 'tenant_id="{}" search failed with exception="{}"'.format(
                                        self.tenant_id, str(e)
                                    )
                                    logging.error(msg)
                                    raise Exception(msg)

                            # break while
                            break

                # end
                logging.info(
                    'tenant_id="{}" data sampling job successfully executed, status="success", run_time="{}", report_name="{}", entities_count="{}"'.format(
                        self.tenant_id,
                        round(time.time() - mainstart, 3),
                        str(report_name),
                        str(count),
                    )
                )

                # Call the component register
                trackme_register_tenant_object_summary(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    "splk-dsm",
                    "trackme_dsm_data_sampling_tracker_tenant_" + str(self.tenant_id),
                    "success",
                    time.time(),
                    str(time.time() - mainstart),
                    "The report was executed successfully",
                    "-24h",
                    "now",
                )

            elif self.mode == "get_samples":
                # check if we have a KVrecord already for this object
                query_string = {
                    "$and": [
                        {
                            "object": self.object,
                        }
                    ]
                }
                try:
                    # try get to get the key
                    kvrecord = collection.data.query(query=(json.dumps(query_string)))[
                        0
                    ]
                    key = kvrecord.get("_key")
                except Exception as e:
                    key = None

                # run the main report, every result is a Splunk search to be executed on its own thread
                if not key:
                    raise Exception("this entity was not found in the collection")

                # init
                search = None

                # get the entity info
                try:
                    json_data = {
                        "tenant_id": self.tenant_id,
                        "object": self.object,
                    }
                    target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_dsm/ds_entity_info"
                    response = requests.post(
                        target_url,
                        headers={
                            "Authorization": header,
                            "Content-Type": "application/json",
                        },
                        verify=False,
                        data=json.dumps(json_data),
                        timeout=300,
                    )
                    search = json.loads(response.text).get("splk_dsm_sampling_search")

                except Exception as e:
                    raise Exception(
                        'could not retrieve entity info data sampling search, exception="{}"'.format(
                            str(e)
                        )
                    )

                # run the main report, every result is a Splunk search to be executed on its own thread
                if not search:
                    raise Exception(
                        "could not retrieve entity info data sampling search, this entity was not found"
                    )

                else:
                    # replace the number of records to be sampled
                    if not re.search("search", "splunkremote "):
                        search = search + " | head " + str(self.get_samples_max_count)

                    # add the key
                    search = (
                        search
                        + '\n| eval key="'
                        + str(key)
                        + '", object="'
                        + str(self.object)
                        + '"'
                        + "\n| rename _raw as raw_sample, sourcetype as data_sourcetype"
                    )

                    try:
                        suboneshotsearch_results = self.service.jobs.oneshot(
                            search, **kwargs_oneshot
                        )
                        count += 1

                        # Get the results and display them using the JSONResultsReader
                        subreader = results.JSONResultsReader(suboneshotsearch_results)
                        for item in subreader:
                            if isinstance(item, dict):
                                logging.debug('search_results="{}"'.format(item))

                                data = {
                                    "_time": time.time(),
                                    "key": item.get("key"),
                                    "object": item.get("object"),
                                    "_raw": item.get("raw_sample"),
                                    "data_sourcetype": item.get("data_sourcetype"),
                                }
                                yield data

                    except Exception as e:
                        logging.error(
                            'tenant_id="{}" search failed with exception="{}"'.format(
                                self.tenant_id, str(e)
                            )
                        )


dispatch(DataSamplingExecutor, sys.argv, sys.stdin, sys.stdout, __name__)
