#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = ["Guilhem Marchand"]
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library
import os
import sys
import time
import json

# External libraries
import urllib3

# Disable urllib3 warnings
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Configure logging
import logging
from logging.handlers import RotatingFileHandler

# set splunkhome
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splkflx_inactive_inspector.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append libs
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs (after lib appended)
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)

# import trackme libs (after lib appended)
from trackme_libs import (
    trackme_reqinfo,
    trackme_audit_event,
    trackme_audit_flip,
    trackme_register_tenant_object_summary,
    trackme_gen_state,
)


@Configuration(distributed=False)
class SplkFlxGetReportsDef(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    register_component = Option(
        doc="""
        **Syntax:** **register_component=****
        **Description:** If the search is invoked by a tracker, register_component can be called to capture and register any execution exception.""",
        require=False,
        default=True,
    )

    report = Option(
        doc="""
        **Syntax:** **report=****
        **Description:** If register_component is set, a value for report is required.""",
        require=False,
        default=None,
        validate=validators.Match("report", r"^.*$"),
    )

    max_sec_since_inactivity_before_status_update = Option(
        doc="""
        **Syntax:** **max_sec_since_inactivity_before_status_update=****
        **Description:** value for max_sec_since_inactivity_before_status_update is required.""",
        require=False,
        default="3600",
        validate=validators.Match(
            "max_sec_since_inactivity_before_status_update", r"^\d*$"
        ),
    )

    max_days_since_inactivity_before_purge = Option(
        doc="""
        **Syntax:** **max_days_since_inactivity_before_purge=****
        **Description:** value for max_days_since_inactivity_before_purge is required.""",
        require=False,
        default="7",
        validate=validators.Match(
            "max_days_since_inactivity_before_purge", r"^(\d|\.)*$"
        ),
    )

    # main
    def generate(self, **kwargs):
        if self:
            # start perf duration counter
            start = time.time()

            # Get request info and set logging level
            reqinfo = trackme_reqinfo(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
            )
            log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

            # Get configuration and define metadata
            trackme_summary_idx = reqinfo["trackme_conf"]["index_settings"][
                "trackme_summary_idx"
            ]

            # Data collection
            data_collection_name = "kv_trackme_flx_tenant_%s" % (self.tenant_id)
            data_collection = self.service.kvstore[data_collection_name]

            # convert the max day in sec (status update)
            max_sec_record_age_before_status_update = float(
                self.max_sec_since_inactivity_before_status_update
            )

            # convert the max day in sec (purge)
            max_sec_record_age_before_purge = (
                float(self.max_days_since_inactivity_before_purge) * 86400
            )

            # end of get configuration

            #
            # loop through the KVstore records
            #

            records = data_collection.data.query()

            # records to be processed counter
            count = 0

            # capture exceptions
            errors_count = 0
            errors_list = []

            # Loop in the results
            for record in records:
                # object
                object_name = record.get("object")

                # get the age in seconds since the latest execution
                sec_since_last_execution = time.time() - float(
                    record.get("tracker_runtime")
                )

                # get status
                record_status = int(record.get("status"))

                # get current_state
                record_current_state = record.get("current_state")

                # attempt to retrieve max_sec_inactive on a per record basic
                max_sec_inactive = record.get("max_sec_inactive", None)
                if not max_sec_inactive:
                    max_sec_inactive = max_sec_record_age_before_status_update
                else:
                    max_sec_inactive = float(max_sec_inactive)

                #
                # update inactive entities status
                #

                # if the requested conditions are met
                if (
                    float(sec_since_last_execution) > max_sec_inactive
                    and record_status != 2
                ):
                    count += 1

                    # Update the record
                    record["object_state"] = "red"
                    record["current_state"] = "red"
                    record["status"] = 2
                    record[
                        "status_message"
                    ] = f"This entity has been inactive for more than {sec_since_last_execution} seconds and was not actively managed by any tracker, its status was updated automatically by the inactive entities tracker"
                    record[
                        "status_description"
                    ] = f"The entity status is red due to inactivity"
                    record["anomaly_reason"] = "inactive"
                    flip_previous_state = record.get("latest_flip_state")
                    record["latest_flip_state"] = "red"
                    record["latest_flip_time"] = time.time()

                    try:
                        data_collection.data.update(
                            str(record.get("_key")), json.dumps(record)
                        )
                        logging.info(
                            'successfully updated record="{}"'.format(
                                json.dumps(record, indent=2)
                            )
                        )

                        # flip
                        new_latest_flip_state = "red"
                        new_latest_flip_time = time.time()
                        new_flip_time_human = time.strftime(
                            "%d/%m/%Y %H:%M:%S", time.localtime(new_latest_flip_time)
                        )
                        flip_result = f'{new_flip_time_human}, object="{object_name}" has flipped from previous_state="{flip_previous_state}" to state="{new_latest_flip_state}" with anomaly_reason="inactive"'

                        try:
                            trackme_audit_flip(
                                self._metadata.searchinfo.session_key,
                                reqinfo["server_rest_uri"],
                                str(self.tenant_id),
                                object_name,
                                "splk-flx",
                                "red",
                                flip_previous_state,
                                str(new_latest_flip_state),
                                str(new_latest_flip_time),
                                str(flip_result),
                            )
                        except Exception as e:
                            logging.error(
                                'tenant_id="{}", object="{}", failed to generate a flipping state event with exception="{}"'.format(
                                    self.tenant_id, object_name, str(e)
                                )
                            )

                        # audit
                        trackme_audit_event(
                            self._metadata.searchinfo.session_key,
                            reqinfo["server_rest_uri"],
                            self.tenant_id,
                            "system",
                            "success",
                            "update status inactive entity",
                            object_name,
                            "splk-flx",
                            str(json.dumps(record, indent=1)),
                            "Entity was updated successfully",
                            "automated management of inactive entities",
                        )

                        # sunnary state
                        try:
                            trackme_gen_state(
                                index=trackme_summary_idx,
                                sourcetype="trackme:state",
                                source=f"current_state_tracking:splk-flx:{self.tenant_id}",
                                event=json.dumps(record),
                            )
                            logging.debug(
                                f'TrackMe summary event created successfully, record="{json.dumps(record, indent=1)}"'
                            )
                        except Exception as e:
                            logging.error(
                                f'TrackMe summary event creation failure, record="{json.dumps(record, indent=1)}", exception="{str(e)}"'
                            )

                        yield_record = {
                            "_time": time.time(),
                            "_raw": {
                                "response": "processing with record update, _key={}, object={}".format(
                                    record.get("_key"), object_name
                                ),
                                "action": "success",
                                "record": record,
                            },
                        }

                        yield yield_record

                    except Exception as e:
                        error_msg = f'An exception was encountered while attempting to update the KVstore record, exception="{str(e)}", record="{json.dumps(record, indent=2)}"'
                        errors_count += 1
                        errors_list.append(error_msg)
                        yield_record = {
                            "_time": time.time(),
                            "_raw": {
                                "response": "processing with record deletion, _key={}, object={}".format(
                                    record.get("_key"), object_name
                                ),
                                "action": "failure",
                                "exception": str(e),
                                "record": record,
                            },
                        }

                        yield yield_record

                # if the entity is already red but inactive, gen a summary event
                elif float(sec_since_last_execution) > max_sec_inactive:
                    if record_current_state != "red":
                        try:
                            record["current_state"] = "red"
                            data_collection.data.update(
                                str(record.get("_key")), json.dumps(record)
                            )
                            logging.info(
                                'successfully updated record="{}"'.format(
                                    json.dumps(record, indent=2)
                                )
                            )

                        except Exception as e:
                            error_msg = f'An exception was encountered while attempting to update the KVstore record, exception="{str(e)}", record="{json.dumps(record, indent=2)}"'
                            errors_count += 1
                            errors_list.append(error_msg)
                            yield_record = {
                                "_time": time.time(),
                                "_raw": {
                                    "response": "processing with record update, _key={}, object={}".format(
                                        record.get("_key"), object_name
                                    ),
                                    "action": "failure",
                                    "exception": str(e),
                                    "record": record,
                                },
                            }

                            yield yield_record

                    # sunnary state
                    try:
                        trackme_gen_state(
                            index=trackme_summary_idx,
                            sourcetype="trackme:state",
                            source=f"current_state_tracking:splk-flx:{self.tenant_id}",
                            event=json.dumps(record),
                        )
                        logging.debug(
                            f'TrackMe summary event created successfully, record="{json.dumps(record, indent=1)}"'
                        )
                    except Exception as e:
                        logging.error(
                            f'TrackMe summary event creation failure, record="{json.dumps(record, indent=1)}", exception="{str(e)}"'
                        )

                #
                # purge inactive entities
                #

                # if the requested conditions are met
                if (
                    float(sec_since_last_execution) > max_sec_record_age_before_purge
                    and record_status != 2
                ):
                    count += 1

                    # Remove the record
                    try:
                        data_collection.data.delete(
                            json.dumps({"_key": record.get("_key")})
                        )
                        logging.info(
                            'successfully deleted record="{}"'.format(
                                json.dumps(record, indent=2)
                            )
                        )
                        trackme_audit_event(
                            self._metadata.searchinfo.session_key,
                            reqinfo["server_rest_uri"],
                            self.tenant_id,
                            "system",
                            "success",
                            "delete inactive entity",
                            object_name,
                            "splk-flx",
                            str(json.dumps(record, indent=1)),
                            "Entity was deleted successfully",
                            "automated deletion of inactive entities",
                        )

                        yield_record = {
                            "_time": time.time(),
                            "_raw": {
                                "response": "processing with record deletion, _key={}, object={}".format(
                                    record.get("_key"), object_name
                                ),
                                "action": "success",
                                "record": record,
                            },
                        }

                        yield yield_record

                    except Exception as e:
                        error_msg = f'An exception was encountered while attempting to delete the KVstore record, exception="{str(e)}", record="{json.dumps(record, indent=2)}"'
                        errors_count += 1
                        errors_list.append(error_msg)
                        yield_record = {
                            "_time": time.time(),
                            "_raw": {
                                "response": "processing with record deletion, _key={}, object={}".format(
                                    record.get("_key"), object_name
                                ),
                                "action": "failure",
                                "exception": str(e),
                                "record": record,
                            },
                        }

                        yield yield_record

            if not count > 0:
                yield_record = {
                    "_time": time.time(),
                    "_raw": {
                        "response": "There are not records to be processed at the moment, nothing to do.",
                        "action": "success",
                    },
                }

                yield yield_record

            # capture the job status
            if errors_count > 0:
                if self.register_component and self.tenant_id and self.report:
                    try:
                        trackme_register_tenant_object_summary(
                            self._metadata.searchinfo.session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            "splk-flx",
                            self.report,
                            "failure",
                            time.time(),
                            round(time.time() - start, 3),
                            errors_list,
                            "-5m",
                            "now",
                        )
                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", component="splk-flx", Failed to call trackme_register_tenant_object_summary with exception="{str(e)}"'
                        )
            else:
                if self.register_component and self.tenant_id and self.report:
                    try:
                        trackme_register_tenant_object_summary(
                            self._metadata.searchinfo.session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            "splk-flx",
                            self.report,
                            "success",
                            time.time(),
                            round(time.time() - start, 3),
                            "the job was executed successfully",
                            "-5m",
                            "now",
                        )
                    except Exception as e:
                        logging.error(
                            f'tenant_id="{self.tenant_id}", component="splk-flx", Failed to call trackme_register_tenant_object_summary with exception="{str(e)}"'
                        )

            if self.report:
                logging.info(
                    'trackmesplkflxinactiveinspector has terminated, report="{}", run_time="{}"'.format(
                        self.report, round(time.time() - start, 3)
                    )
                )
            else:
                logging.info(
                    'trackmesplkflxinactiveinspector has terminated, run_time="{}"'.format(
                        round(time.time() - start, 3)
                    )
                )


dispatch(SplkFlxGetReportsDef, sys.argv, sys.stdin, sys.stdout, __name__)
