#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import json
import re
import hashlib
import ast
import time

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splk_flx_parse.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# Import TrackMe splk-flx libs
from trackme_libs_splk_flx import trackme_flx_gen_metrics

# Import trackme libs
from trackme_libs import trackme_reqinfo

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    StreamingCommand,
    Configuration,
    Option,
    validators,
)


@Configuration(distributed=False)
class TrackMeSplkFlxParse(StreamingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    context = Option(
        doc="""
        **Syntax:** **context=****
        **Description:** The context is used for simulation purposes, defaults to live.""",
        require=False,
        default="live",
        validate=validators.Match("context", r"^(live|simulation)$"),
    )

    def get_tenant_metric_idx(self):
        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": "Splunk %s" % self._metadata.searchinfo.session_key,
            "Content-Type": "application/json",
        }

        # get the index conf for this tenant
        url = "%s/services/trackme/v2/vtenants/tenant_idx_settings" % (
            self._metadata.searchinfo.splunkd_uri
        )
        data = {"tenant_id": self.tenant_id, "idx_stanza": "trackme_metric_idx"}

        # Retrieve and set the tenant idx, if any failure, logs and use the global index
        try:
            response = requests.post(
                url,
                headers=header,
                data=json.dumps(data, indent=1),
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                error_msg = f'failed to retrieve the tenant metric index, response.status_code="{response.status_code}", response.text="{response.text}"'
                logging.error(error_msg)
                raise Exception(error_msg)
            else:
                response_data = json.loads(json.dumps(response.json(), indent=1))
                tenant_trackme_metric_idx = response_data["trackme_metric_idx"]
        except Exception as e:
            error_msg = (
                f'failed to retrieve the tenant metric index, exception="{str(e)}"'
            )
            logging.error(error_msg)
            raise Exception(error_msg)

        return tenant_trackme_metric_idx

    def stream(self, records):
        # performance counter
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # log info
        logging.info(
            f'tenant_id="{self.tenant_id}", context="{self.context}", TrackMeSplkFlxParse is starting'
        )

        # get metric index
        metric_index = self.get_tenant_metric_idx()

        # counter
        count = 0

        # Loop in the results
        for record in records:
            # increment
            count += 1

            yield_record = {}
            flx_record = {}

            #
            # field: _time
            #

            try:
                flx_time = record.get("_time")
                flx_record["_time"] = flx_time
            except Exception as e:
                flx_time = None

            if not flx_time:
                flx_time = time.time()
                flx_record["_time"] = flx_time

            #
            # field: object
            # mandatory: try to get the object field
            #

            try:
                flx_object = record.get("object")
            except Exception as e:
                flx_object = None
            if not flx_object:
                logging.error(
                    f'The field object is mandatory and should be rendered as a part of the search results, the object field could not be found in result="{json.dumps(record, indent=2)}"'
                )
                raise Exception(
                    "The field object is mandatory and should be rendered as a part of the search results, the object field could not be found in search results"
                )

            #
            # field: alias
            # optional: if provided in the upstream results, use it, otherwise alias will be set explicitly equal to the short version of the object
            #

            flx_alias = record.get("alias", flx_object)
            flx_record["alias"] = flx_alias

            #
            # field: priority
            # optional: if provided in the upstream results, use it, otherwise priority will be set automatically
            #

            flx_priority = record.get("priority", None)
            if flx_priority:
                flx_record["priority"] = flx_priority

            #
            # field: tracker_name
            # optional
            #

            try:
                flx_tracker_name = record.get("tracker_name")
                flx_record["tracker_name"] = flx_tracker_name
            except Exception as e:
                flx_tracker_name = None

            #
            # field: group
            # optional
            #

            try:
                flx_group = record.get("group")
            except Exception as e:
                flx_group = None

            #
            # manage group and object
            #

            # group: if not specified in the search logic, set it equal to tracker_name
            if not flx_group:
                flx_group = str(flx_tracker_name)
            # add to the flx_record
            flx_record["group"] = flx_group

            # object: concatenate group and object if the object does not contain the group name yet
            regex = "^%s\:" % (str(flx_group))
            if not re.search(regex, flx_object):
                flx_object = str(flx_group) + ":" + str(flx_object)

            # add to the flx_record
            flx_record["object"] = flx_object

            # convert the flx_object into an md5 sum, we use the same value when storing into the KVstore
            flx_md5 = hashlib.md5(flx_object.encode("utf-8")).hexdigest()

            #
            # field: status_description
            # try to get the status field
            #

            try:
                flx_status = record.get("status")
                flx_record["status"] = flx_status
            except Exception as e:
                flx_status = None
            if not flx_status:
                logging.error(
                    f'The field status is mandatory and should be rendered as a part of the search results, the status field could not be found in result="{json.dumps(record, indent=2)}"'
                )
                raise Exception(
                    "The field status is mandatory and should be rendered as a part of the search results, the status field could not be found in search results"
                )

            # the status should be an integer
            try:
                flx_status = int(flx_status)
            except Exception as e:
                raise Exception(
                    'The field status is not an integer, value="{}"'.format(flx_status)
                )

            #
            # field: status_description
            # try to get the status_description field
            #

            try:
                flx_status_description = record.get("status_description")
                flx_record["status_description"] = flx_status_description
            except Exception as e:
                flx_status_description = None

            #
            # field: object_description
            # try to get the object_description field
            #

            try:
                flx_object_description = record.get("object_description")
                flx_record["object_description"] = flx_object_description
            except Exception as e:
                flx_object_description = None

            #
            # field: metrics
            # try to get and parse the metrics
            #

            try:
                flx_metrics = record.get("metrics")
                flx_record["metrics"] = flx_metrics
            except Exception as e:
                flx_metrics = None

            # Extract or generate metrics
            flx_metrics_parsed = False
            flx_metrics_parsed_msg = None

            if flx_metrics:
                # store the exception, if any
                flx_metrics_parsed_exception = None

                # attempt json.loads
                try:
                    flx_metrics = json.loads(record.get("metrics"))
                    flx_metrics_parsed = True
                    flx_metrics_parsed_msg = (
                        "Metrics JSON were submitted and successfully parsed"
                    )
                except Exception as e:
                    flx_metrics_parsed = False
                    flx_metrics_parsed_exception = str(e)

                # attempt ast (if json is submitted with single quotes)
                if not flx_metrics_parsed:
                    try:
                        flx_metrics = ast.literal_eval(record.get("metrics"))
                        flx_metrics_parsed = True
                        flx_metrics_parsed_msg = (
                            "Metrics JSON were submitted and successfully parsed"
                        )
                    except Exception as e:
                        flx_metrics_parsed = False
                        flx_metrics_parsed_exception = str(e)

                if flx_metrics and not flx_metrics_parsed:
                    logging.error(
                        f'failed to load the submitted metrics as a JSON object, verify it is properly JSON encoded using single or double quotes delimitors, exception="{flx_metrics_parsed_exception}"'
                    )
                    flx_metrics = {
                        "status": flx_status,
                    }
                    flx_metrics_parsed_msg = f'Metrics JSON were submitted but could not be parsed properly, verify the JSON syntax, properties should be enquoted with single or double quotes, exception="{flx_metrics_parsed_exception}"'
                else:
                    # add the status
                    flx_metrics["status"] = flx_status

            # if no metrics, include the status only
            else:
                flx_metrics = {
                    "status": flx_status,
                }
                flx_metrics_parsed_msg = (
                    "There were no metrics provided, will include the status only"
                )

            #
            # field: outliers_metrics
            # try to get and parse the ML outliers metrics
            #

            try:
                flx_outliers_metrics = record.get("outliers_metrics")
                flx_record["outliers_metrics"] = flx_outliers_metrics
            except Exception as e:
                flx_outliers_metrics = None

            # Extract or generate metrics
            flx_outliers_metrics_parsed = False
            flx_outliers_metrics_parsed_msg = None

            if flx_outliers_metrics:
                # store the exception, if any
                flx_outliers_metrics_parsed_exception = None

                # attempt json.loads
                try:
                    flx_outliers_metrics = json.loads(record.get("outliers_metrics"))
                    flx_outliers_metrics_parsed = True

                    # test extracting values
                    for flx_outliers_metric in flx_outliers_metrics:
                        kpi_name = "splk.flx." + flx_outliers_metric
                        kpi_dict = flx_outliers_metrics[flx_outliers_metric]
                        value_alert_lower_breached = kpi_dict.get(
                            "alert_lower_breached"
                        )
                        value_alert_upper_breached = kpi_dict.get(
                            "alert_upper_breached"
                        )
                        logging.debug(
                            f'Extracting outliers_metrics, kpi_name="{kpi_name}", alert_lower_breached="{value_alert_lower_breached}", alert_upper_breached="{value_alert_upper_breached}"'
                        )

                    flx_outliers_metrics_parsed_msg = (
                        "Outliers Metrics JSON were submitted and successfully parsed"
                    )
                except Exception as e:
                    flx_outliers_metrics_parsed = False
                    flx_outliers_metrics_parsed_exception = str(e)

                # attempt ast (if json is submitted with single quotes)
                if not flx_outliers_metrics_parsed:
                    try:
                        flx_outliers_metrics = ast.literal_eval(
                            record.get("outliers_metrics")
                        )
                        flx_outliers_metrics_parsed = True

                        # test extracting values
                        for flx_outliers_metric in flx_outliers_metrics:
                            kpi_name = "splk.flx." + flx_outliers_metric
                            kpi_dict = flx_outliers_metrics[flx_outliers_metric]
                            value_alert_lower_breached = kpi_dict.get(
                                "alert_lower_breached"
                            )
                            value_alert_upper_breached = kpi_dict.get(
                                "alert_upper_breached"
                            )
                            logging.debug(
                                f'Extracting outliers_metrics, kpi_name="{kpi_name}", alert_lower_breached="{value_alert_lower_breached}", alert_upper_breached="{value_alert_upper_breached}"'
                            )

                        flx_outliers_metrics_parsed_msg = "Outliers Metrics JSON were submitted and successfully parsed"
                    except Exception as e:
                        flx_outliers_metrics_parsed = False
                        flx_outliers_metrics_parsed_exception = str(e)

                if flx_outliers_metrics and not flx_outliers_metrics_parsed:
                    logging.error(
                        "failed to load the submitted outliers metrics as a JSON object, verify it is properly JSON encoded using single or double quotes delimitors"
                    )
                    flx_outliers_metrics = flx_outliers_metrics
                    flx_outliers_metrics_parsed_msg = f'Outliers Metrics JSON were submitted but could not be parsed properly, verify the JSON syntax, properties should be enquoted with single or double quotes, exception="{flx_outliers_metrics_parsed_exception}"'

            # if no metrics, include the message only only
            else:
                flx_outliers_metrics = {}
                flx_outliers_metrics_parsed_msg = (
                    "There were no outliers metrics provided"
                )

            #
            # Generate and index metrics
            # if not in simulation mode, ingest metrics now
            #

            if self.context == "live":
                try:
                    trackme_flx_gen_metrics(
                        self.tenant_id,
                        flx_object,
                        flx_md5,
                        metric_index,
                        json.dumps(flx_metrics),
                    )
                except Exception as e:
                    error_msg = f'tenant_id="{self.tenant_id}", object="{flx_object}", object_id="{flx_md5}", failed to stream events to metrics with exception="{str(e)}"'.format(
                        self.tenant_id, flx_object, flx_md5, str(e)
                    )
                    logging.error(error_msg)
                    raise Exception(error_msg)

            # get a _raw, if any, otherwise build
            try:
                raw = record.get("_raw")
                if raw:
                    flx_record["_raw"] = raw
            except Exception as e:
                raw = {}
                for k in record:
                    raw[k] = record[k]

            # finally
            yield_record[flx_object] = flx_record

            # add metrics
            flx_record["metrics"] = flx_metrics

            # add outliers_metrics
            flx_record["outliers_metrics"] = flx_outliers_metrics

            # extra_attibutes, this is optional
            flx_extra_attributes = record.get("extra_attributes", None)
            flx_record["extra_attributes"] = flx_extra_attributes

            # max_sec_inactive, this is optional
            flx_max_sec_inactive = record.get("max_sec_inactive", None)
            flx_record["max_sec_inactive"] = flx_max_sec_inactive

            # if in simulation, add an informational message related to the metrics management
            if self.context == "simulation":
                # add metrics_message
                flx_record["metrics_message"] = flx_metrics_parsed_msg

                # add outliers_metrics_message
                flx_record["outliers_metrics_message"] = flx_outliers_metrics_parsed_msg

                # yield
                yield_record = {
                    "_time": flx_time,
                    "_raw": flx_record,
                    "group": flx_group,
                    "object": flx_object,
                    "alias": flx_alias,
                    "object_category": "splk-flx",
                    "object_description": flx_object_description,
                    "status": flx_status,
                    "status_description": flx_status_description,
                    "metrics": flx_metrics,
                    "metrics_message": flx_metrics_parsed_msg,
                    "outliers_metrics": flx_outliers_metrics,
                    "outliers_metrics_message": flx_outliers_metrics_parsed_msg,
                    "extra_attributes": flx_extra_attributes,
                    "max_sec_inactive": flx_max_sec_inactive,
                }

                # conditional additions to the final result
                if flx_priority:
                    yield_record["priority"] = flx_priority

                yield yield_record

            else:
                # yield
                yield_record = {
                    "_time": flx_time,
                    "_raw": flx_record,
                    "group": flx_group,
                    "object": flx_object,
                    "alias": flx_alias,
                    "object_category": "splk-flx",
                    "object_description": flx_object_description,
                    "status": flx_status,
                    "status_description": flx_status_description,
                    "metrics": flx_metrics,
                    "outliers_metrics": flx_outliers_metrics,
                    "extra_attributes": flx_extra_attributes,
                    "max_sec_inactive": flx_max_sec_inactive,
                }

                # conditional additions to the final result
                if flx_priority:
                    yield_record["priority"] = flx_priority

                yield yield_record

            # log info
            logging.debug(
                f'tenant_id="{self.tenant_id}", context="{self.context}", processed result="{json.dumps(flx_record, indent=2)}"'
            )

        # log info
        run_time = run_time = round(time.time() - start, 3)
        logging.info(
            f'tenant_id="{self.tenant_id}", context="{self.context}", TrackMeSplkFlxParse has terminated successfully, turn debug mode on for more details, results_count="{count}", run_time={run_time}'
        )


dispatch(TrackMeSplkFlxParse, sys.argv, sys.stdin, sys.stdout, __name__)
