#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import re

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splk_outliers_cim_tracker_helper.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk lib
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# import trackme libs
from trackme_libs import trackme_reqinfo, trackme_register_tenant_object_summary


@Configuration(distributed=False)
class SplkOutliersTracker(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    object = Option(
        doc="""
        **Syntax:** **object=****
        **Description:** Optional, The value for object.""",
        require=False,
        default="*",
        validate=validators.Match("object", r"^.*$"),
    )

    max_runtime = Option(
        doc="""
        **Syntax:** **max_runtime=****
        **Description:** Optional, The max value in seconds for the total runtime of the job, defaults to 900 (15 min) which is substracted by 120 sec of margin. Once the job reaches this, it gets terminated""",
        require=False,
        default="900",
        validate=validators.Match("object", r"^\d*$"),
    )

    force_run = Option(
        doc="""
        **Syntax:** **force_run=****
        **Description:** Optional, force run monitor, if set to True we will not honour the minimal time betwen two monitor execution.""",
        require=False,
        default="False",
        validate=validators.Match("force_run", r"^(?:True|False)$"),
    )

    def generate(self, **kwargs):
        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5

        # performance counter
        mainstart = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # Get app level config
        splk_outliers_time_monitor_mlmodels_default = reqinfo["trackme_conf"][
            "splk_outliers_detection"
        ]["splk_outliers_time_monitor_mlmodels_default"]

        # counter
        count = 0

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # Outliers rules storage collection
        collection_rules_name = (
            "kv_trackme_cim_outliers_entity_rules" + "_tenant_" + str(self.tenant_id)
        )
        collection_rule = self.service.kvstore[collection_rules_name]

        # Outliers data storage collection
        collection_data_name = (
            "kv_trackme_cim_outliers_entity_data" + "_tenant_" + str(self.tenant_id)
        )
        collection_data = self.service.kvstore[collection_data_name]

        # Get data
        kwargs_oneshot = {
            "earliest_time": "-5m",
            "latest_time": "now",
            "output_mode": "json",
            "count": 0,
        }

        #
        # RUN
        #

        # report name for logging purposes
        report_name = "trackme_cim_outliers_mlmonitor_tracker_tenant_" + str(
            self.tenant_id
        )

        # Define the search providing the list of entities which models need to be monitored
        if self.force_run == "False":
            search_query = (
                '| `get_splk_outliers_cim_entities("'
                + self.tenant_id
                + '", "'
                + self.object
                + '")`'
                + "\n | lookup trackme_cim_outliers_entity_data_tenant_"
                + self.tenant_id
                + " entity, object, cim_field OUTPUT mtime as last_exec_monitor"
                + "\n | eval duration_since_last=if(isnum(last_exec_monitor), now()-last_exec_monitor, 0)"
                + "\n | where duration_since_last=0 OR duration_since_last>="
                + str(splk_outliers_time_monitor_mlmodels_default)
                + "\n | sort - duration_since_last"
            )

        else:
            search_query = (
                '| `get_splk_outliers_cim_entities("'
                + self.tenant_id
                + '", "'
                + self.object
                + '")`'
                + "\n | lookup trackme_cim_outliers_entity_data_tenant_"
                + self.tenant_id
                + " entity, object, cim_field OUTPUT mtime as last_exec_monitor"
                + "\n | eval duration_since_last=if(isnum(last_exec_monitor), now()-last_exec_monitor, 0)"
                + "\n | sort - duration_since_last"
            )

        logging.debug('search_query="{}"'.format(search_query))

        # automatic retry
        search_successful = False
        current_failures_count = 0
        while current_failures_count < max_failures_count and not search_successful:
            # run the main report, every result is a Splunk search to be executed on its own thread
            try:
                oneshotsearch_results = self.service.jobs.oneshot(
                    search_query, **kwargs_oneshot
                )
                reader = results.JSONResultsReader(oneshotsearch_results)

                # loop through the results, and train models per entity

                # store processed entities in a list
                processed_entities = []

                for item in reader:
                    logging.debug('search_results="{}"'.format(item))

                    if isinstance(item, dict):
                        # break if reaching the max run time less 120 seconds of margin
                        if (time.time() - int(mainstart)) - 120 >= int(
                            self.max_runtime
                        ):
                            logging.info(
                                'tenant_id="{}" max_runtime="{}" for ML models is about to be reached with current_runtime="{}", job will be terminated now'.format(
                                    self.tenant_id,
                                    int(self.max_runtime),
                                    time.time() - int(mainstart),
                                )
                            )
                            break

                        # run the resulting search
                        object_value = item.get("object")
                        model_ids = item.get("model_id")
                        cim_field = item.get("cim_field")
                        entity = item.get("entity")

                        # log debug
                        logging.debug(
                            'object="{}", cim_field="{}", model_ids="{}"'.format(
                                object_value, cim_field, model_ids
                            )
                        )

                        # set the global_isOutlier
                        global_isOutlier = 0
                        global_models_in_anomaly = []
                        global_isOutlierReason = []

                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "object_category": "splk-cim",
                                    "object": object_value,
                                    "entity": entity,
                                    "cim_field": cim_field,
                                }
                            ]
                        }

                        # Get the current record
                        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
                        key = None

                        # for cim: we expect multiple records potentially
                        try:
                            records_outliers_rules = collection_rule.data.query(
                                query=json.dumps(query_string)
                            )

                        except Exception as e:
                            records_outliers_rules = None

                        # if no records, log a warning message and break
                        if not records_outliers_rules:
                            msg = 'tenant_id="{}", object="{}" outliers rules record cannot be found or are not yet available for this entity.'.format(
                                self.tenant_id, self.object
                            )
                            logging.warn(msg)
                            break

                        logging.debug(
                            'record_outliers_rules="{}"'.format(
                                json.dumps(records_outliers_rules, indent=4)
                            )
                        )

                        # turn into a list
                        model_ids = model_ids.split(",")

                        # loop through the records
                        for record_outliers_rules in records_outliers_rules:
                            # get the key
                            key = record_outliers_rules.get("_key")

                            # log debug
                            logging.debug(
                                'record_outliers_rules="{}"'.format(
                                    record_outliers_rules
                                )
                            )

                            # Get the JSON outliers rules object
                            entities_outliers = record_outliers_rules.get(
                                "entities_outliers"
                            )

                            # Load as a dict
                            try:
                                entities_outliers = json.loads(
                                    record_outliers_rules.get("entities_outliers")
                                )
                            except Exception as e:
                                msg = 'Failed to load entities_outliers with exception="{}"'.format(
                                    str(e)
                                )

                            # log debug
                            logging.debug(
                                'entities_outliers="{}"'.format(entities_outliers)
                            )

                            # Load the general enablement
                            try:
                                outliers_is_disabled = int(
                                    record_outliers_rules.get("is_disabled")
                                )
                                logging.debug(
                                    'is_disabled="{}"'.format(outliers_is_disabled)
                                )

                            except Exception as e:
                                msg = 'Failed to extract one or more expected settings from the entity, is this record corrupted? Exception="{}"'
                                logging.error(msg)
                                outliers_is_disabled = 1

                            # process the entity if general outliers is enabled
                            if outliers_is_disabled == 0:
                                # Process all ML models per entity
                                processed_entity_models = {}

                                # process
                                if model_ids:
                                    # loop through each model per entity
                                    for model_id in model_ids:
                                        # model configuration
                                        model_config = entities_outliers[model_id]
                                        logging.debug(
                                            'configuration for model_id="{}" config="{}"'.format(
                                                model_id,
                                                json.dumps(model_config, indent=4),
                                            )
                                        )

                                        # If the model is enabled
                                        if int(model_config["is_disabled"]) == 0:
                                            # Get conf from the model
                                            alert_lower_breached = int(
                                                model_config["alert_lower_breached"]
                                            )
                                            alert_upper_breached = int(
                                                model_config["alert_upper_breached"]
                                            )

                                            # get the kpi metric name and value
                                            kpi_metric_name = model_config["kpi_metric"]
                                            logging.debug(
                                                'kpi_metric_name="{}"'.format(
                                                    kpi_metric_name
                                                )
                                            )

                                            # Set the initial state for that model
                                            isOutlier = 0
                                            isOutlierReason = "None"

                                            # Set the search
                                            model_render_search = (
                                                '| trackmesplkoutlierscimrender tenant_id="'
                                                + self.tenant_id
                                                + '" component="cim" object="'
                                                + object_value
                                                + '" earliest="-24h" latest="now" model_id="'
                                                + model_id
                                                + '" entity="'
                                                + entity
                                                + '" cim_field="'
                                                + record_outliers_rules.get("cim_field")
                                                + '" | table _time, *, LowerBound, UpperBound | sort 0 - _time | head 1'
                                            )

                                            logging.info(
                                                'tenant_id="{}", object="{}", model_id="{}\, Executing resulting search="{}"'.format(
                                                    self.tenant_id,
                                                    object_value,
                                                    model_id,
                                                    model_render_search,
                                                )
                                            )

                                            # set kwargs
                                            kwargs_oneshot = {
                                                "earliest_time": "-24h",
                                                "latest_time": "now",
                                                "search_mode": "normal",
                                                "preview": False,
                                                "count": 0,
                                                "output_mode": "json",
                                            }

                                            # Performance timer
                                            start = time.time()

                                            # Run the search
                                            search_results = None
                                            try:
                                                # use the export API which is not limited in number of results
                                                search_results = (
                                                    self.service.jobs.export(
                                                        model_render_search,
                                                        **kwargs_oneshot,
                                                    )
                                                )
                                                reader = results.JSONResultsReader(
                                                    search_results
                                                )

                                                # loop through the reader results
                                                for item in reader:
                                                    if isinstance(item, dict):
                                                        search_results = item

                                                        # raw results logged only in debug
                                                        logging.debug(
                                                            'search_results="{}"'.format(
                                                                search_results
                                                            )
                                                        )

                                                        # Inspect results
                                                        time_outlier = search_results[
                                                            "_time"
                                                        ]

                                                        # try to get the LowerBound and UpperBound
                                                        try:
                                                            LowerBound = search_results[
                                                                "LowerBound"
                                                            ]
                                                        except Exception as e:
                                                            LowerBound = None
                                                            logging.warning(
                                                                'Could not retrieve a LowerBound value from item="{}"'.format(
                                                                    item
                                                                )
                                                            )

                                                        try:
                                                            UpperBound = search_results[
                                                                "UpperBound"
                                                            ]
                                                        except Exception as e:
                                                            UpperBound = None
                                                            logging.warning(
                                                                'Could not retrieve a UpperBound value from item="{}"'.format(
                                                                    item
                                                                )
                                                            )

                                                        try:
                                                            kpi_metric_value = (
                                                                search_results[
                                                                    kpi_metric_name
                                                                ]
                                                            )
                                                            logging.debug(
                                                                'kpi_metric_value="{}"'.format(
                                                                    kpi_metric_value
                                                                )
                                                            )
                                                        except Exception as e:
                                                            kpi_metric_value = None
                                                            logging.warning(
                                                                'Could not retrieve the kpi_metric_value from item="{}"'.format(
                                                                    item
                                                                )
                                                            )

                                                        # Define the outliers status
                                                        if (
                                                            kpi_metric_value
                                                            and LowerBound
                                                            and UpperBound
                                                        ):
                                                            if int(
                                                                alert_lower_breached
                                                            ) == 1 and float(
                                                                kpi_metric_value
                                                            ) < float(
                                                                LowerBound
                                                            ):
                                                                isOutlier = 1
                                                                pct_decrease = (
                                                                    (
                                                                        float(
                                                                            LowerBound
                                                                        )
                                                                        - float(
                                                                            kpi_metric_value
                                                                        )
                                                                    )
                                                                    / float(LowerBound)
                                                                ) * 100
                                                                isOutlierReason = f'Outliers ML for kpi="{kpi_metric_name}", model_id="{model_id}", LowerBound="{round(float(LowerBound), 3)}" breached with kpi_metric_value="{round(float(kpi_metric_value), 3)}" at time="{time_outlier}", pct_decrease="{round(float(pct_decrease), 2)}"'

                                                            elif int(
                                                                alert_upper_breached
                                                            ) == 1 and float(
                                                                kpi_metric_value
                                                            ) > float(
                                                                UpperBound
                                                            ):
                                                                isOutlier = 1
                                                                pct_increase = (
                                                                    (
                                                                        float(
                                                                            kpi_metric_value
                                                                        )
                                                                        - float(
                                                                            UpperBound
                                                                        )
                                                                    )
                                                                    / float(UpperBound)
                                                                ) * 100
                                                                isOutlierReason = f'Outliers ML for kpi="{kpi_metric_name}", model_id="{model_id}", UpperBound="{round(float(UpperBound), 3)}" breached with kpi_metric_value="{round(float(kpi_metric_value), 3)}" at time="{time_outlier}", pct_increase="{round(float(pct_increase), 2)}"'

                                                            else:
                                                                isOutlier = 0
                                                                isOutlierReason = "No outliers anomalies were detected"

                                                            # impact the global_isOutlier accordingly
                                                            if isOutlier == 1:
                                                                global_isOutlier = 1
                                                                global_models_in_anomaly.append(
                                                                    model_id
                                                                )
                                                                global_isOutlierReason.append(
                                                                    isOutlierReason
                                                                )

                                            except Exception as e:
                                                error_msg = 'tenant_id="{}", object="{}", model_id="{}", search has failed with the following exception="{}", search="{}"'.format(
                                                    self.tenant_id,
                                                    object_value,
                                                    model_id,
                                                    str(e),
                                                    model_render_search,
                                                )
                                                logging.error(error_msg)

                                            # Performance timer
                                            model_search_runtime = float(
                                                time.time()
                                            ) - float(start)

                                            # try loading search results and define a message if did not produce any results
                                            try:
                                                summary_search_results = {
                                                    "time": search_results["_time"],
                                                    "_raw": json.loads(
                                                        search_results["_raw"]
                                                    ),
                                                    "model_render_search": model_render_search,
                                                }
                                            except Exception as e:
                                                summary_search_results = "Outliers search did not produce any results"

                                            # Insert the summary record
                                            model_summary = {
                                                "isOutlier": isOutlier,
                                                "isOutlierReason": isOutlierReason,
                                                "alert_lower_breached": alert_lower_breached,
                                                "alert_upper_breached": alert_upper_breached,
                                                "summary_search_results": summary_search_results,
                                                "model_render_search": model_render_search,
                                                "search_run_time": str(
                                                    model_search_runtime
                                                ),
                                                "time_exec": str(time.time()),
                                            }
                                            processed_entity_models[
                                                model_id
                                            ] = model_summary

                                            # log info
                                            logging.info(
                                                'tenant_id="{}", object="{}", model_summary="{}"'.format(
                                                    self.tenant_id,
                                                    object_value,
                                                    json.dumps(model_summary, indent=4),
                                                )
                                            )

                                    # summary record for that entity
                                    processed_entity_record = {
                                        "entity": object_value,
                                        "cim_field": cim_field,
                                        "processed_model_ids": processed_entity_models,
                                    }

                                    # append
                                    processed_entities.append(processed_entity_record)

                                    # increment the entity counter
                                    count += 1

                                    #
                                    # Finally, update the outliers data KVstore
                                    #

                                    # Define the KV query
                                    query_string = {
                                        "$and": [
                                            {
                                                "object_category": "splk-cim",
                                                "object": object_value,
                                                "entity": entity,
                                                "cim_field": cim_field,
                                            }
                                        ]
                                    }

                                    # Get the current record
                                    # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
                                    key = None

                                    try:
                                        record_outliers_data = (
                                            collection_data.data.query(
                                                query=json.dumps(query_string)
                                            )[0]
                                        )
                                        key = record_outliers_data.get("_key")

                                    except Exception as e:
                                        key = None

                                    if not key:
                                        # new record
                                        try:
                                            collection_data.data.insert(
                                                json.dumps(
                                                    {
                                                        "entity": str(entity),
                                                        "object": str(object_value),
                                                        "object_category": "splk-cim",
                                                        "cim_field": record_outliers_rules.get(
                                                            "cim_field"
                                                        ),
                                                        "mtime": str(time.time()),
                                                        "isOutlier": global_isOutlier,
                                                        "isOutlierReason": global_isOutlierReason,
                                                        "models_in_anomaly": global_models_in_anomaly,
                                                        "models_summary": json.dumps(
                                                            processed_entity_models,
                                                            indent=4,
                                                        ),
                                                    }
                                                )
                                            )

                                        except Exception as e:
                                            logging.error(
                                                'tenant_id="{}", object="{}", failed to insert a new KVstore record with exception="{}"'.format(
                                                    self.tenant_id, object_value, str(e)
                                                )
                                            )

                                    else:
                                        try:
                                            # update existing record
                                            collection_data.data.update(
                                                str(key),
                                                json.dumps(
                                                    {
                                                        "entity": str(entity),
                                                        "object": str(object_value),
                                                        "object_category": "splk-cim",
                                                        "cim_field": record_outliers_rules.get(
                                                            "cim_field"
                                                        ),
                                                        "mtime": str(time.time()),
                                                        "isOutlier": global_isOutlier,
                                                        "isOutlierReason": global_isOutlierReason,
                                                        "models_in_anomaly": global_models_in_anomaly,
                                                        "models_summary": json.dumps(
                                                            processed_entity_models,
                                                            indent=4,
                                                        ),
                                                    }
                                                ),
                                            )

                                        except Exception as e:
                                            logging.error(
                                                'tenant_id="{}", object="{}", failed to update a KVstore record with exception="{}"'.format(
                                                    self.tenant_id, object_value, str(e)
                                                )
                                            )

                # end
                if int(count) > 0:
                    logging.info(
                        'tenant_id="{}" outliers tracker job successfully executed, status="success", run_time="{}", report="{}", entities_count="{}"'.format(
                            self.tenant_id,
                            round(time.time() - mainstart, 3),
                            str(report_name),
                            str(count),
                        )
                    )
                    # yield
                    results_dict = {
                        "tenant_id": self.tenant_id,
                        "action": "success",
                        "results": "outliers tracker job successfully executed",
                        "run_time": time.time() - mainstart,
                        "entities_count": str(count),
                        "processed_entities": processed_entities,
                        "upstream_search_query": search_query,
                    }
                    yield {"_time": time.time(), "_raw": results_dict}

                else:
                    logging.info(
                        'tenant_id="{}" outliers tracker job successfully executed but there were no entities to be tracked at this time, status="success", run_time="{}", report="{}", entities_count="{}"'.format(
                            self.tenant_id,
                            round(time.time() - mainstart, 3),
                            str(report_name),
                            str(count),
                        )
                    )
                    # yield
                    results_dict = {
                        "tenant_id": self.tenant_id,
                        "action": "success",
                        "results": "outliers tracker job successfully executed but there were no entities to be tracked at this time",
                        "run_time": time.time() - mainstart,
                        "entities_count": str(count),
                        "upstream_search_query": search_query,
                    }
                    yield {"_time": time.time(), "_raw": results_dict}

                # Call the component register
                trackme_register_tenant_object_summary(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    "splk-cim",
                    "trackme_cim_outliers_mlmonitor_tracker_tenant_"
                    + str(self.tenant_id),
                    "success",
                    time.time(),
                    round(time.time() - mainstart, 3),
                    "The report was executed successfully",
                    "-24h",
                    "now",
                )

                # break while
                search_successful = True
                break

            except Exception as e:
                if re.search(
                    "The maximum number of concurrent historical searches on this instance has been reached",
                    str(e),
                ):
                    current_failures_count += 1
                    search_successful = False
                    if current_failures_count >= max_failures_count:
                        trackme_register_tenant_object_summary(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            "splk-cim",
                            "trackme_cim_outliers_mlmonitor_tracker_tenant_"
                            + str(self.tenant_id),
                            "failure",
                            time.time(),
                            round(time.time() - mainstart, 3),
                            str(e),
                            "-24h",
                            "now",
                        )
                        msg = 'tenant_id="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                            self.tenant_id,
                            current_failures_count,
                            max_failures_count,
                            str(e),
                        )
                        logging.error(msg)
                        raise Exception(msg)
                    else:
                        msg = 'tenant_id="{}", temporary search failure with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                        logging.warn(msg)
                        time.sleep(sleep_time)

                else:
                    trackme_register_tenant_object_summary(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        "splk-cim",
                        "trackme_cim_outliers_mlmonitor_tracker_tenant_"
                        + str(self.tenant_id),
                        "failure",
                        time.time(),
                        round(time.time() - mainstart, 3),
                        str(e),
                        "-24h",
                        "now",
                    )
                    msg = (
                        'tenant_id="{}", main search failed with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                    )
                    logging.error(msg)
                    raise Exception(msg)


dispatch(SplkOutliersTracker, sys.argv, sys.stdin, sys.stdout, __name__)
