#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import re

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splk_outliers_cim_train.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# Import trackme libs
from trackme_libs import trackme_reqinfo

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces


@Configuration(distributed=False)
class SplkOutliersTrain(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The value for tenant_id.""",
        require=True,
        validate=validators.Match("tenant_id", r"^.*$"),
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** The component category.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(?:cim)$"),
    )

    object = Option(
        doc="""
        **Syntax:** **object=****
        **Description:** The value for object.""",
        require=True,
        validate=validators.Match("object", r"^.*$"),
    )

    cim_field = Option(
        doc="""
        **Syntax:** **cim_field=****
        **Description:** The value for cim_field.""",
        require=True,
        validate=validators.Match("cim_field", r"^.*$"),
    )

    model_id = Option(
        doc="""
        **Syntax:** **model_id=****
        **Description:** The Machine Learning model ID to be trained, optional and defaults to all models defined for the entity.""",
        require=False,
        validate=validators.Match("object", r"^.*$"),
    )

    def generate(self, **kwargs):
        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5

        # track run_time
        global_start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # TrackMe sharing level
        trackme_default_sharing = reqinfo["trackme_conf"]["trackme_general"][
            "trackme_default_sharing"
        ]

        # get current user
        username = self._metadata.searchinfo.username

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": "Splunk %s" % session_key,
            "Content-Type": "application/json",
        }

        # Get global index conf
        trackme_metric_idx = reqinfo["trackme_conf"]["index_settings"][
            "trackme_metric_idx"
        ]

        # Get the value for splk_outliers_min_days_history
        splk_outliers_min_days_history = reqinfo["trackme_conf"][
            "splk_outliers_detection"
        ]["splk_outliers_min_days_history"]

        # set the tenant_trackme_metric_idx
        tenant_trackme_metric_idx = None

        # get the index conf for this tenant
        url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/tenant_idx_settings"
        data = {"tenant_id": self.tenant_id, "idx_stanza": "trackme_metric_idx"}

        # Retrieve and set the tenant idx, if any failure, logs and use the global index
        try:
            response = requests.post(
                url,
                headers=header,
                data=json.dumps(data, indent=1),
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                tenant_trackme_metric_idx = trackme_metric_idx
            else:
                try:
                    response_data = json.loads(json.dumps(response.json(), indent=1))
                    tenant_trackme_metric_idx = response_data["trackme_metric_idx"]
                except Exception as e:
                    tenant_trackme_metric_idx = response_data["trackme_metric_idx"]

        except Exception as e:
            tenant_trackme_metric_idx = trackme_metric_idx

        # Component collection
        collection_name = (
            "kv_trackme_" + self.component + "_tenant_" + str(self.tenant_id)
        )
        collection = self.service.kvstore[collection_name]

        # Outliers rules storage collection
        collection_rules_name = (
            "kv_trackme_"
            + self.component
            + "_outliers_entity_rules"
            + "_tenant_"
            + str(self.tenant_id)
        )
        collection_rule = self.service.kvstore[collection_rules_name]

        # Vtenants storage collection
        vtenants_collection_name = "kv_trackme_virtual_tenants"
        vtenants_collection = self.service.kvstore[vtenants_collection_name]

        #
        # First, get the full vtenant definition
        #

        # Define the KV query search string
        query_string = {
            "tenant_id": self.tenant_id,
        }

        # get
        try:
            vtenant_record = vtenants_collection.data.query(
                query=json.dumps(query_string)
            )
        except Exception as e:
            error_msg = 'tenant_id="{}", Could not retrieve the virtual tenant definition in the KVstore storage="{}", exception="{}"'.format(
                self.tenant_id, vtenants_collection_name, str(e)
            )
            logging.error(error_msg)
            raise Exception(error_msg)

        # required for RBAC purposes

        tenant_owner = str(vtenant_record[0].get("tenant_owner"))
        tenant_roles_admin = str(vtenant_record[0].get("tenant_roles_admin"))
        tenant_roles_power = str(vtenant_record[0].get("tenant_roles_power"))
        tenant_roles_user = str(vtenant_record[0].get("tenant_roles_user"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        #
        # Get the Outliers rules
        #

        # Define the KV query
        if self.cim_field and self.cim_field != "*":
            query_string_filter = {
                "object_category": "splk-" + self.component,
                "object": self.object,
                "cim_field": self.cim_field,
            }
        else:
            query_string_filter = {
                "object_category": "splk-" + self.component,
                "object": self.object,
            }

        query_string = {"$and": [query_string_filter]}

        # get records
        try:
            records_outliers_rules = collection_rule.data.query(
                query=json.dumps(query_string)
            )
        except Exception as e:
            records_outliers_rules = None

        if not records_outliers_rules:
            msg = 'tenant_id="{}", component="{}", object="{}" outliers rules record cannot be found or are not yet available for this entity.'.format(
                self.tenant_id, self.component, self.object
            )
            logging.error(msg)
            raise Exception(msg)

        else:
            # process records
            for record_outliers_rules in records_outliers_rules:
                # get key
                key = record_outliers_rules.get("_key")

                # log debug
                logging.debug(
                    'record_outliers_rules="{}"'.format(record_outliers_rules)
                )

                # Get the JSON outliers rules object
                entities_outliers = record_outliers_rules.get("entities_outliers")

                # Load as a dict
                try:
                    entities_outliers = json.loads(
                        record_outliers_rules.get("entities_outliers")
                    )
                except Exception as e:
                    msg = 'Failed to load entities_outliers with exception="{}"'.format(
                        str(e)
                    )

                # log debug
                logging.debug('entities_outliers="{}"'.format(entities_outliers))

                # Load the account and the general enablement
                try:
                    ds_account = record_outliers_rules.get("ds_account")
                    outliers_is_disabled = int(record_outliers_rules.get("is_disabled"))
                    logging.debug(
                        'ds_account="{}", is_disabled="{}"'.format(
                            ds_account, outliers_is_disabled
                        )
                    )

                except Exception as e:
                    msg = 'Failed to extract one or more expected settings from the entity, is this record corrupted? Exception="{}"'
                    logging.error(msg)
                    raise Exception(msg)

                # Load the entity (breakby)
                try:
                    entity_name = record_outliers_rules.get("entity")

                except Exception as e:
                    msg = 'Failed to extract one or more expected settings from the entity, is this record corrupted? Exception="{}"'
                    logging.error(msg)
                    raise Exception(msg)

                # Only proceed is enabled

                # proceed
                if outliers_is_disabled == 1:
                    yield {
                        "_time": time.time(),
                        "_raw": "Outliers detection are disabled at the global level for this entity, nothing to do.",
                        "response": "Outliers detection are disabled at the global level for this entity, nothing to do.",
                    }

                elif outliers_is_disabled == 0:
                    #
                    # Establish confidence level
                    #

                    # Establish low as the basis level
                    ml_confidence = "low"
                    ml_confidence_reason = (
                        "ML has insufficient historical metrics to proceed (pending)"
                    )
                    ml_metrics_duration = "unknown"

                    # set kwargs
                    kwargs_confidence = {
                        "earliest_time": "-90d",
                        "latest_time": "now",
                        "output_mode": "json",
                        "count": 0,
                    }

                    # define the gen search
                    ml_confidence_search = remove_leading_spaces(
                        f"""\
                        | mstats latest(trackme.splk.cim.*) as * where index={tenant_trackme_metric_idx} tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" by object span=1d
                        | stats min(_time) as first_time by object
                        | eval metrics_duration=now()-first_time
                        | eval confidence=if(metrics_duration<({splk_outliers_min_days_history}*86400), "low", "normal")
                        | eval metrics_duration=tostring(metrics_duration, "duration")
                        | head 1
                        """
                    )
                    logging.info(
                        f"tenant_id={self.tenant_id}, component={self.component}, object={self.object}, ML confidence inspection, ml_confidence_search={ml_confidence_search}"
                    )

                    # automatic retry
                    search_successful = False
                    current_failures_count = 0
                    while (
                        current_failures_count < max_failures_count
                        and not search_successful
                    ):
                        # track the search runtime
                        start = time.time()

                        # proceed
                        try:
                            ml_confidence_results = self.service.jobs.export(
                                ml_confidence_search, **kwargs_confidence
                            )
                            reader = results.JSONResultsReader(ml_confidence_results)

                            for item in reader:
                                if isinstance(item, dict):
                                    # log
                                    logging.info(
                                        f'tenant_id={self.tenant_id}, object={self.object}, ML confidence inspection, ml_confidence={item.get("confidence")}, metrics_duration={item.get("metrics_duration")}, ml_confidence_search={ml_confidence_search}'
                                    )
                                    ml_confidence = item.get("confidence", "low")
                                    ml_metrics_duration = item.get(
                                        "metrics_duration", "unknown"
                                    )

                            # break while
                            search_successful = True
                            break

                        except Exception as e:
                            if re.search(
                                "The maximum number of concurrent historical searches on this instance has been reached",
                                str(e),
                            ):
                                current_failures_count += 1
                                search_successful = False
                                if current_failures_count >= max_failures_count:
                                    logging.error(msg)
                                    raise Exception(msg)
                                else:
                                    msg = 'tenant_id="{}", component="{}", object="{}", temporary search failure with exception="{}", run_time="{}"'.format(
                                        self.tenant_id,
                                        self.component,
                                        self.object,
                                        str(e),
                                        str(time.time() - start),
                                    )
                                    logging.warn(msg)
                                    time.sleep(sleep_time)
                            else:
                                msg = 'tenant_id="{}", component="{}", object="{}", Machine Learning model training search failed with exception="{}", run_time="{}"'.format(
                                    self.tenant_id,
                                    self.component,
                                    self.object,
                                    str(e),
                                    str(time.time() - start),
                                )
                                logging.error(msg)
                                raise Exception(msg)

                    # set the ml_confidence_reason
                    if ml_confidence == "low":
                        ml_confidence_reason = f"ML has insufficient historical metrics to proceed (metrics_duration={ml_metrics_duration}, required={splk_outliers_min_days_history}days)"
                    elif ml_confidence == "normal":
                        ml_confidence_reason = f"ML has sufficient historical metrics to proceed (metrics_duration={ml_metrics_duration}, required={splk_outliers_min_days_history}days)"

                    #
                    # ML inspection
                    #

                    # use a counter to condition final actions
                    process_counter = 0

                    # set a list for error reporting purposes of available modesl
                    entity_outliers_models = []

                    # Loop through outliers entities
                    for entity_outlier in entities_outliers:
                        # Add to the list
                        entity_outliers_models.append(entity_outlier)

                        # Extract as a dict
                        entity_outlier_dict = entities_outliers[entity_outlier]

                        logging.debug(
                            'entity_outlier_dict="{}"'.format(entity_outlier_dict)
                        )

                        # Define if entity is enabled
                        process_entity = False

                        try:
                            is_disabed = int(entity_outlier_dict["is_disabled"])
                            if is_disabed == 0:
                                process_entity = True
                            else:
                                process_entity = False

                        except Exception as e:
                            msg = 'tenant_id="{}", component="{}", object="{}", Failed to extract the entity enablement status, is this record corrupted? Exception="{}"'.format(
                                self.tenant_id,
                                self.component,
                                self.object,
                                entity_outlier,
                                str(e),
                            )
                            logging.error(msg)
                            raise Exception(msg)

                        # Extract the outliers entity settings
                        if process_entity:
                            # Process all outliers entities, unless filtering is set at the custom command level
                            process_entity_outliers = False

                            # decide
                            if self.model_id:
                                if self.model_id == entity_outlier:
                                    process_entity_outliers = True
                                else:
                                    process_entity_outliers = False
                            else:
                                process_entity_outliers = True

                            # log debug
                            logging.debug(
                                'process_entity_outliers="{}"'.format(
                                    process_entity_outliers
                                )
                            )

                            # if process
                            if process_entity_outliers:
                                # increment counter
                                process_counter += 1

                                # log
                                logging.info(
                                    'tenant_id="{}", component="{}", object="{}", Processing Machine Learning outliers training for entity_outlier="{}", rules="{}"'.format(
                                        self.tenant_id,
                                        self.component,
                                        self.object,
                                        entity_outlier,
                                        json.dumps(entity_outlier_dict, indent=4),
                                    )
                                )

                                try:
                                    cim_breakby = entity_outlier_dict["cim_breakby"]
                                    entity_name = entity_outlier_dict["entity"]
                                    cim_field = entity_outlier_dict["cim_field"]
                                    kpi_metric = entity_outlier_dict["kpi_metric"]
                                    kpi_span = entity_outlier_dict["kpi_span"]
                                    method_calculation = entity_outlier_dict[
                                        "method_calculation"
                                    ]
                                    density_lowerthreshold = entity_outlier_dict[
                                        "density_lowerthreshold"
                                    ]
                                    density_upperthreshold = entity_outlier_dict[
                                        "density_upperthreshold"
                                    ]
                                    alert_lower_breached = entity_outlier_dict[
                                        "alert_lower_breached"
                                    ]
                                    alert_upper_breached = entity_outlier_dict[
                                        "alert_upper_breached"
                                    ]
                                    period_calculation = entity_outlier_dict[
                                        "period_calculation"
                                    ]
                                    time_factor = entity_outlier_dict["time_factor"]
                                    perc_min_lowerbound_deviation = entity_outlier_dict[
                                        "perc_min_lowerbound_deviation"
                                    ]
                                    perc_min_upperbound_deviation = entity_outlier_dict[
                                        "perc_min_upperbound_deviation"
                                    ]
                                    period_exclusions = entity_outlier_dict.get(
                                        "period_exclusions", []
                                    )
                                    # ensure period_exclusions is a list, otherwise set it to an empty list
                                    if not isinstance(period_exclusions, list):
                                        period_exclusions = []

                                    rules_summary = {
                                        "is_disabled": is_disabed,
                                        "cim_breakby": cim_breakby,
                                        "entity_name": entity_name,
                                        "cim_field": cim_field,
                                        "kpi_metric": kpi_metric,
                                        "kpi_span": kpi_span,
                                        "method_calculation": method_calculation,
                                        "density_lowerthreshold": density_lowerthreshold,
                                        "density_upperthreshold": density_upperthreshold,
                                        "period_calculation": period_calculation,
                                        "time_factor": time_factor,
                                        "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                                        "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                                        "alert_lower_breached": alert_lower_breached,
                                        "alert_upper_breached": alert_upper_breached,
                                        "period_exclusions": period_exclusions,
                                    }

                                    logging.info(
                                        'Processing outliers entity="{}", rules_summary="{}"'.format(
                                            entity_outlier, rules_summary
                                        )
                                    )

                                except Exception as e:
                                    msg = 'tenant_id="{}", component="{}", object="{}", entity_outlier="{}", failed to extract one or more expected settings from the entity, is this record corrupted? Exception="{}"'.format(
                                        self.tenant_id,
                                        self.component,
                                        self.object,
                                        entity_outlier,
                                        str(e),
                                    )
                                    logging.error(msg)
                                    raise Exception(msg)

                                # Define the Splunk searches
                                ml_model_gen_search = None
                                ml_model_render_search = None

                                # if no breakby, override
                                if cim_breakby == "none":
                                    cim_breakby = "cim_entity_zone"

                                # Set the densityFunction threshold parameters
                                if (
                                    float(density_lowerthreshold) > 0
                                    and float(density_upperthreshold) > 0
                                ):
                                    density_threshold_str = (
                                        "lower_threshold="
                                        + str(density_lowerthreshold)
                                        + " upper_threshold="
                                        + str(density_upperthreshold)
                                    )
                                else:
                                    density_threshold_str = (
                                        "lower_threshold=0.005 upper_threshold=0.005"
                                    )
                                    error_msg = (
                                        "densityFunction threshold parameters are incorrects for this entity, "
                                        + "lower_threshold and upper_threshold must both be a positive value, "
                                        + "will be using using factory value."
                                    )
                                    logging.error(
                                        'tenant_id="{}", compoent="{}", object="{}", {}'.format(
                                            self.tenant_id,
                                            self.component,
                                            self.tenant_id,
                                            error_msg,
                                        )
                                    )

                                # Construct the where NOT conditions, and also verifies if the period_exclusions are valid
                                where_conditions = ""
                                if period_exclusions:
                                    for period in period_exclusions:
                                        logging.debug(f"period_exclusion: {period}")

                                        # get the period_latest
                                        period_latest = period["latest"]

                                        # period_calculation is a time relative expression to now, such as -30d for the past 30 days from now, so we need to convert it to a timestamp
                                        # extract the first two digits after the minus sign which corresponds to the number of days, then convert to seconds, and apply against the current time
                                        period_calculation_no_days = (
                                            int(period_calculation[1:3]) * 86400
                                        )
                                        period_calculation_timestamp = (
                                            int(time.time())
                                            - period_calculation_no_days
                                        )

                                        # if the period_earliest and period_latest are not valid, then we need to skip this period_exclusion
                                        if period_latest < period_calculation_timestamp:
                                            logging.info(
                                                f"tenant_id={self.tenant_id}, object={self.object}, model_id={entity_outlier} rejecting period exclusion as it is now out of the model period calculation: {json.dumps(period, indent=4)}"
                                            )

                                            # delete the period_exclusion from the list
                                            period_exclusions.remove(period)

                                            # update the entity_outlier_dict
                                            entity_outlier_dict[
                                                "period_exclusions"
                                            ] = period_exclusions

                                        else:
                                            logging.info(
                                                f"tenant_id={self.tenant_id}, object={self.object}, model_id={entity_outlier} accepting period exclusion: {json.dumps(period, indent=4)}"
                                            )
                                            where_conditions += f'``` period_exclusions for this ML model: ```\n| where NOT (_time>{period["earliest"]} AND _time<{period["latest"]})\n'

                                else:
                                    where_conditions = (
                                        "``` no period_exclusions for this ML model ```"
                                    )

                                # set the lookup name
                                ml_model_lookup_name = (
                                    "__mlspl_" + str(entity_outlier) + ".mlmodel"
                                )

                                # retrieve the current size, if the size is not equal to pending, then the lookup model needs to be deleted first
                                ml_model_filesize = entity_outlier_dict[
                                    "ml_model_filesize"
                                ]

                                if ml_model_filesize != "pending":
                                    # Attempt to delete the current ml model
                                    rest_url = f'{self._metadata.searchinfo.splunkd_uri}"/servicesNS/{username}/trackme/data/lookup-table-files/{ml_model_lookup_name}'

                                    logging.info(
                                        'tenant_id="{}", attempting to delete Machine Learning lookup_name="{}"'.format(
                                            self.tenant_id, ml_model_lookup_name
                                        )
                                    )
                                    try:
                                        response = requests.delete(
                                            rest_url,
                                            headers=header,
                                            verify=False,
                                            timeout=300,
                                        )
                                        if response.status_code not in (200, 201, 204):
                                            logging.error(
                                                'tenant_id="{}", failure to delete ML lookup_name="{}", url="{}", response.status_code="{}", response.text="{}"'.format(
                                                    self.tenant_id,
                                                    ml_model_lookup_name,
                                                    rest_url,
                                                    response.status_code,
                                                    response.text,
                                                )
                                            )
                                        else:
                                            logging.info(
                                                'tenant_id="{}", action="success", deleted lookup_name="{}" successfully'.format(
                                                    self.tenant_id, ml_model_lookup_name
                                                )
                                            )

                                            # Update ml_model_filesize / ml_model_lookup_share
                                            entity_outlier_dict[
                                                "ml_model_filesize"
                                            ] = "pending"
                                            entity_outlier_dict[
                                                "ml_model_lookup_share"
                                            ] = "pending"

                                    except Exception as e:
                                        logging.error(
                                            'tenant_id="{}", failure to delete ML lookup_name="{}" with exception:"{}"'.format(
                                                self.tenant_id,
                                                ml_model_lookup_name,
                                                str(e),
                                            )
                                        )
                                        return {
                                            "payload": "Warn: exception encountered: "
                                            + str(e),
                                            "status": 500,
                                        }

                                # define the gen search depending on if time_factor is set to none or not
                                if time_factor == "none":
                                    ml_model_gen_search = remove_leading_spaces(
                                        f"""
                                        | mstats avg(trackme.{kpi_metric}) as {kpi_metric} where index="{tenant_trackme_metric_idx}"
                                        tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" {cim_breakby}="{entity_name}" 
                                        cim_field="{cim_field}" by object span="{kpi_span}"
                                        {where_conditions}
                                        | fit DensityFunction {kpi_metric} {density_threshold_str} into {entity_outlier}
                                        | rex field=BoundaryRanges "(-Infinity:(?<LowerBound>[\\d|\\.]*))|((?<UpperBound>[\\d|\\.]*):Infinity)"
                                        | foreach LowerBound UpperBound [ eval <<FIELD>> = if(isnum('<<FIELD>>'), '<<FIELD>>', 0) ]
                                        | fields _time {kpi_metric} LowerBound UpperBound
                                        | stats count as metrics_count
                                        """
                                    )

                                else:
                                    ml_model_gen_search = remove_leading_spaces(
                                        f"""
                                        | mstats avg(trackme.{kpi_metric}) as {kpi_metric} where index="{tenant_trackme_metric_idx}"
                                        tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" {cim_breakby}="{entity_name}" 
                                        cim_field="{cim_field}" by object span="{kpi_span}"
                                        {where_conditions}
                                        | eval factor=strftime(_time, "{time_factor}")
                                        | fit DensityFunction {kpi_metric} {density_threshold_str} into {entity_outlier} by factor
                                        | rex field=BoundaryRanges "(-Infinity:(?<LowerBound>[\\d|\\.]*))|((?<UpperBound>[\\d|\\.]*):Infinity)"
                                        | foreach LowerBound UpperBound [ eval <<FIELD>> = if(isnum('<<FIELD>>'), '<<FIELD>>', 0) ]
                                        | fields _time {kpi_metric} LowerBound UpperBound
                                        | stats count as metrics_count
                                        """
                                    )

                                # log debug
                                logging.debug(
                                    'ml_model_gen_search="{}"'.format(
                                        ml_model_gen_search
                                    )
                                )

                                # define the render search depending on if time_factor is set to none or not, to be stored for further usage purposes
                                if time_factor == "none":
                                    ml_model_render_search = remove_leading_spaces(
                                        f"""
                                        | mstats avg(trackme.{kpi_metric}) as {kpi_metric} where index="{tenant_trackme_metric_idx}"
                                        tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" {cim_breakby}="{entity_name}" 
                                        cim_field="{cim_field}" by object span="{kpi_span}"
                                        | apply {entity_outlier}
                                        | rex field=BoundaryRanges "(-Infinity:(?<LowerBound>[\\d|\\.]*))|((?<UpperBound>[\\d|\\.]*):Infinity)"
                                        | foreach LowerBound UpperBound [ eval <<FIELD>> = if(isnum('<<FIELD>>'), '<<FIELD>>', 0) ]
                                        | fields _time {kpi_metric} LowerBound UpperBound
                                        """
                                    )

                                else:
                                    ml_model_render_search = remove_leading_spaces(
                                        f"""
                                        | mstats avg(trackme.{kpi_metric}) as {kpi_metric} where index="{tenant_trackme_metric_idx}"
                                        tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" {cim_breakby}="{entity_name}" 
                                        cim_field="{cim_field}" by object span="{kpi_span}"
                                        | eval factor=strftime(_time, "{time_factor}")
                                        | apply {entity_outlier}
                                        | rex field=BoundaryRanges "(-Infinity:(?<LowerBound>[\\d|\\.]*))|((?<UpperBound>[\\d|\\.]*):Infinity)"
                                        | foreach LowerBound UpperBound [ eval <<FIELD>> = if(isnum('<<FIELD>>'), '<<FIELD>>', 0) ]
                                        | fields _time {kpi_metric} LowerBound UpperBound
                                        """
                                    )

                                # set kwargs
                                kwargs_oneshot = {
                                    "earliest_time": str(period_calculation),
                                    "latest_time": "now",
                                    "output_mode": "json",
                                    "count": 0,
                                }

                                #
                                # Run
                                #

                                # automatic retry
                                search_successful = False
                                current_failures_count = 0
                                while (
                                    current_failures_count < max_failures_count
                                    and not search_successful
                                ):
                                    # track the search runtime
                                    start = time.time()

                                    # proceed
                                    try:
                                        oneshotsearch_results = (
                                            self.service.jobs.export(
                                                ml_model_gen_search, **kwargs_oneshot
                                            )
                                        )
                                        reader = results.JSONResultsReader(
                                            oneshotsearch_results
                                        )

                                        for item in reader:
                                            if isinstance(item, dict):
                                                # log
                                                logging.info(
                                                    'tenant_id="{}", component="{}", object="{}", kpi_metric="{}", entity_outlier="{}", Machine Learning model training search executed successfully, run_time="{}", results="{}"'.format(
                                                        self.tenant_id,
                                                        self.component,
                                                        self.tenant_id,
                                                        kpi_metric,
                                                        entity_outlier,
                                                        round(time.time() - start, 3),
                                                        json.dumps(item, indent=0),
                                                    )
                                                )

                                        # break while
                                        search_successful = True
                                        break

                                    except Exception as e:
                                        if re.search(
                                            "The maximum number of concurrent historical searches on this instance has been reached",
                                            str(e),
                                        ):
                                            current_failures_count += 1
                                            search_successful = False
                                            if (
                                                current_failures_count
                                                >= max_failures_count
                                            ):
                                                msg = 'tenant_id="{}", component="{}", object="{}", kpi_metric="{}", entity_outlier="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                                    self.tenant_id,
                                                    self.component,
                                                    self.object,
                                                    kpi_metric,
                                                    entity_outlier,
                                                    current_failures_count,
                                                    max_failures_count,
                                                    str(e),
                                                )
                                                logging.error(msg)
                                                raise Exception(msg)
                                            else:
                                                msg = 'tenant_id="{}", component="{}", object="{}", kpi_metric="{}", entity_outlier="{}", temporary search failure with exception="{}", run_time="{}"'.format(
                                                    self.tenant_id,
                                                    self.component,
                                                    self.object,
                                                    kpi_metric,
                                                    entity_outlier,
                                                    str(e),
                                                    round(time.time() - start, 3),
                                                )
                                                logging.warn(msg)
                                                time.sleep(sleep_time)
                                        else:
                                            msg = 'tenant_id="{}", component="{}", object="{}", kpi_metric="{}", entity_outlier="{}", Machine Learning model training search failed with exception="{}", run_time="{}"'.format(
                                                self.tenant_id,
                                                self.component,
                                                self.object,
                                                kpi_metric,
                                                entity_outlier,
                                                str(e),
                                                round(time.time() - start, 3),
                                            )
                                            logging.error(msg)
                                            raise Exception(msg)

                                # if the share level is not app, then permissions need to be updated

                                # retrieve the current share level
                                ml_model_lookup_share = entity_outlier_dict[
                                    "ml_model_lookup_share"
                                ]

                                if ml_model_lookup_share != "app":
                                    # Attempt to update the ml lookup permissions
                                    rest_url = f"{self._metadata.searchinfo.splunkd_uri}/servicesNS/{username}/trackme/data/lookup-table-files/{ml_model_lookup_name}/acl"

                                    logging.info(
                                        'tenant_id="{}", attempting to update permissions of Machine Learning lookup_name="{}"'.format(
                                            self.tenant_id, ml_model_lookup_name
                                        )
                                    )
                                    try:
                                        response = requests.post(
                                            rest_url,
                                            headers=header,
                                            data={
                                                "owner": tenant_owner,
                                                "sharing": trackme_default_sharing,
                                                "perms.write": tenant_roles_write_perms,
                                                "perms.read": tenant_roles_read_perms,
                                            },
                                            verify=False,
                                            timeout=300,
                                        )
                                        if response.status_code not in (200, 201, 204):
                                            logging.error(
                                                'tenant_id="{}", failure to update ML lookup_name="{}", url="{}", response.status_code="{}", response.text="{}"'.format(
                                                    self.tenant_id,
                                                    ml_model_lookup_name,
                                                    rest_url,
                                                    response.status_code,
                                                    response.text,
                                                )
                                            )
                                        else:
                                            logging.info(
                                                'tenant_id="{}", action="success", permissions of lookup_name="{}" were updated successfully'.format(
                                                    self.tenant_id, ml_model_lookup_name
                                                )
                                            )

                                            # Update ml_model_lookup_share
                                            entity_outlier_dict[
                                                "ml_model_lookup_share"
                                            ] = "app"

                                            # Update owner and perms
                                            entity_outlier_dict[
                                                "ml_model_lookup_owner"
                                            ] = tenant_owner
                                            entity_outlier_dict[
                                                "ml_model_lookup_perm_write"
                                            ] = tenant_roles_write_perms
                                            entity_outlier_dict[
                                                "ml_model_lookup_perm_read"
                                            ] = tenant_roles_read_perms

                                    except Exception as e:
                                        logging.error(
                                            'tenant_id="{}", failure to update ML lookup_name="{}" with exception:"{}"'.format(
                                                self.tenant_id,
                                                ml_model_lookup_name,
                                                str(e),
                                            )
                                        )
                                        return {
                                            "payload": "Warn: exception encountered: "
                                            + str(e),
                                            "status": 500,
                                        }

                                # Update last_exec
                                entity_outlier_dict["last_exec"] = str(time.time())

                                # Update ml_model_gen_search
                                entity_outlier_dict[
                                    "ml_model_gen_search"
                                ] = ml_model_gen_search

                                # Update
                                entity_outlier_dict[
                                    "ml_model_render_search"
                                ] = ml_model_render_search

                                # Update rules_access_search
                                entity_outlier_dict["rules_access_search"] = (
                                    "| inputlookup trackme_"
                                    + self.component
                                    + "_outliers_entity_rules_tenant_"
                                    + str(self.tenant_id)
                                    + ' where _key="'
                                    + str(key)
                                    + '"'
                                )

                                # Update ml_model_filename
                                entity_outlier_dict[
                                    "ml_model_filename"
                                ] = ml_model_lookup_name

                                # Update ml_model_summary_search
                                entity_outlier_dict[
                                    "ml_model_summary_search"
                                ] = "| summary " + str(entity_outlier)

                                # Update ml_model_filesize
                                try:
                                    entity_outlier_dict[
                                        "ml_model_filesize"
                                    ] = os.path.getsize(
                                        os.path.join(
                                            splunkhome,
                                            "etc",
                                            "apps",
                                            "trackme",
                                            "lookups",
                                            ml_model_lookup_name,
                                        )
                                    )
                                except Exception as e:
                                    logging.error(
                                        'tenant_id="{}", failure to retrieve the size of the ML lookup_name="{}" with exception:"{}"'.format(
                                            self.tenant_id, ml_model_lookup_name, str(e)
                                        )
                                    )

                                # Update the main dict
                                entities_outliers[entity_outlier] = entity_outlier_dict

                    if process_counter > 0:
                        # record global last exec
                        last_exec = time.time()

                        final_dict = {
                            "_key": key,
                            "entity": entity_name,
                            "ds_account": record_outliers_rules.get("ds_account"),
                            "entities_outliers": entities_outliers,
                            "is_disabled": record_outliers_rules.get("is_disabled"),
                            "mtime": record_outliers_rules.get("mtime"),
                            "object": record_outliers_rules.get("object"),
                            "object_category": record_outliers_rules.get(
                                "object_category"
                            ),
                            "cim_field": record_outliers_rules.get("cim_field"),
                            "last_exec": str(last_exec),
                        }

                        # Update the KVrecord
                        try:
                            collection_rule.data.update(
                                str(key),
                                json.dumps(
                                    {
                                        "entity": entity_name,
                                        "ds_account": record_outliers_rules.get(
                                            "ds_account"
                                        ),
                                        "entities_outliers": json.dumps(
                                            entities_outliers, indent=4
                                        ),
                                        "is_disabled": record_outliers_rules.get(
                                            "is_disabled"
                                        ),
                                        "mtime": record_outliers_rules.get("mtime"),
                                        "object": record_outliers_rules.get("object"),
                                        "object_category": record_outliers_rules.get(
                                            "object_category"
                                        ),
                                        "cim_field": record_outliers_rules.get(
                                            "cim_field"
                                        ),
                                        "last_exec": str(last_exec),
                                        "confidence": ml_confidence,
                                        "confidence_reason": ml_confidence_reason,
                                    }
                                ),
                            )

                        except Exception as e:
                            logging.error(
                                'failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                                    str(e)
                                )
                            )

                        # Update the component KVrecord
                        try:
                            kvrecord = collection.data.query(
                                query=json.dumps(
                                    {"object": record_outliers_rules.get("object")}
                                )
                            )[0]
                            kvrecordkey = kvrecord.get("_key")

                        except Exception as e:
                            logging.error(
                                'Could not retrieve the component KVstore record with key="{}", this installation seems to be corrupted!'.format(
                                    record_outliers_rules.get("object")
                                )
                            )
                            kvrecordkey = None

                        # proceed
                        if kvrecordkey:
                            kvrecord["outliers_readiness"] = "True"
                            try:
                                collection.data.update(
                                    str(kvrecordkey), json.dumps(kvrecord)
                                )
                                logging.debug(
                                    'outliers_readiness status was updated successfully for object="{}"'.format(
                                        record_outliers_rules.get("object")
                                    )
                                )
                            except Exception as e:
                                logging.error(
                                    'Failed to update the component Kvstore record with key="{}"'.format(
                                        record_outliers_rules.get("object"), str(e)
                                    )
                                )

                        # yield
                        yield {"_time": time.time(), "_raw": final_dict}

                        # log final
                        logging.info(
                            'tenant_id="{}", component="{}", object="{}", Machine Learning model training search terminated, run_time="{}"'.format(
                                self.tenant_id,
                                self.component,
                                self.tenant_id,
                                str(time.time() - global_start),
                            )
                        )

                    else:
                        if self.model_id:
                            error_msg = 'tenant_id="{}", component="{}", object="{}", The model_id="{}" was not found for this object, available models="{}"'.format(
                                self.tenant_id,
                                self.component,
                                self.object,
                                self.model_id,
                                json.dumps(entity_outliers_models),
                            )
                            logging.error(error_msg)
                            raise Exception(error_msg)
                        else:
                            error_msg = 'tenant_id="{}", component="{}", object="{}", no models could not found for this object, is the KVrecord corrupted? rules_definition="{}"'.format(
                                self.tenant_id, self.component, self.object
                            )
                            logging.error(error_msg)
                            raise Exception(error_msg)


dispatch(SplkOutliersTrain, sys.argv, sys.stdin, sys.stdout, __name__)
