#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import re

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splk_outliers_render.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# Import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.client as client
import splunklib.results as results

# Import trackme libs
from trackme_libs import trackme_reqinfo

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces


@Configuration(distributed=False)
class SplkOutliersRender(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The value for tenant_id.""",
        require=True,
        validate=validators.Match("tenant_id", r"^.*$"),
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** The component category.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(?:dsm|dhm|flx|wlk)$"),
    )

    object = Option(
        doc="""
        **Syntax:** **object=****
        **Description:** The value for object.""",
        require=True,
        validate=validators.Match("object", r"^.*$"),
    )

    model_id = Option(
        doc="""
        **Syntax:** **model_id=****
        **Description:** The Machine Learning model ID to be rendered, optional and defaults to the first model defined for the entity.""",
        require=False,
        validate=validators.Match("model_id", r"^.*$"),
    )

    mode = Option(
        doc="""
        **Syntax:** **mode=****
        **Description:** The rendering mode, live model retrieves the model definition from the KVstore, simulation from the model_def argument.""",
        require=False,
        default="live",
        validate=validators.Match("mode", r"^(live|simulation)$"),
    )

    model_json_def = Option(
        doc="""
        **Syntax:** **model_json_def=****
        **Description:** If in simulation mode, the JSON definition for the ML model.""",
        require=False,
        validate=validators.Match("mode", r"^.*$"),
    )

    earliest = Option(
        doc="""
        **Syntax:** **earliest=****
        **Description:** The earliest time for the search.""",
        require=False,
        default=None,
    )

    latest = Option(
        doc="""
        **Syntax:** **latest=****
        **Description:** The latest time for the search.""",
        require=False,
        default=None,
    )

    lowerbound_negative = Option(
        doc="""
        **Syntax:** **lowerbound_negative=****
        **Description:** Allow negative lowerBound.""",
        require=False,
        default=False,
        validate=validators.Match("mode", r"^(True|False)$"),
    )

    auto_correct = Option(
        doc="""
        **Syntax:** **auto_correct=****
        **Description:** Automatically correct lower bound and upper bound calculations notably using the min lower and upper deviation percentage.""",
        require=False,
        default=True,
        validate=validators.Match("mode", r"^(True|False)$"),
    )

    def generate(self, **kwargs):
        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5

        # track run_time
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # set earliest and latest
        if not self.earliest:
            earliest = self._metadata.searchinfo.earliest_time
        else:
            earliest = self.earliest

        if not self.latest:
            latest = self._metadata.searchinfo.latest_time
        else:
            latest = self.latest

        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": "Splunk %s" % session_key,
            "Content-Type": "application/json",
        }

        # Get global index conf
        trackme_metric_idx = reqinfo["trackme_conf"]["index_settings"][
            "trackme_metric_idx"
        ]

        # Outliers rules storage collection
        collection_rules_name = (
            "kv_trackme_"
            + self.component
            + "_outliers_entity_rules"
            + "_tenant_"
            + str(self.tenant_id)
        )
        collection_rule = self.service.kvstore[collection_rules_name]

        #
        # mode live
        #

        if self.mode == "live":
            # log debug
            logging.debug("node is live")

            #
            # Get the Outliers rules
            #

            # Define the KV query
            query_string_filter = {
                "object_category": "splk-" + self.component,
                "object": self.object,
            }

            query_string = {"$and": [query_string_filter]}

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

            key = None

            try:
                records_outliers_rules = collection_rule.data.query(
                    query=json.dumps(query_string)
                )
                record_outliers_rules = records_outliers_rules[0]
                key = record_outliers_rules.get("_key")

            except Exception as e:
                key = None

            # if no records
            if not key:
                msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}" outliers rules record cannot be found or are not yet available for this entity.'
                logging.error(msg)
                raise Exception(msg)

            # log debug
            logging.debug(f'record_outliers_rules="{record_outliers_rules}"')

            # Get the JSON outliers rules object
            entities_outliers = record_outliers_rules.get("entities_outliers")

            # Load as a dict
            try:
                entities_outliers = json.loads(
                    record_outliers_rules.get("entities_outliers")
                )
            except Exception as e:
                msg = f'Failed to load entities_outliers with exception="{str(e)}"'
                logging.error(msg)

            # log debug
            logging.debug(
                f'entities_outliers="{json.dumps(entities_outliers, indent=2)}"'
            )

            # Load the account and the general enablement
            try:
                ds_account = record_outliers_rules.get("ds_account")
                outliers_is_disabled = int(record_outliers_rules.get("is_disabled"))
                logging.debug(
                    f'ds_account="{ds_account}", is_disabled="{outliers_is_disabled}"'
                )
            except Exception as e:
                msg = f'Failed to extract one or more expected settings from the entity, is this record corrupted? Exception="{e}"'
                logging.error(msg)
                raise Exception(msg)

            #
            # Start
            #

            # Only proceed is enabled

            # proceed
            if outliers_is_disabled == 1:
                yield {
                    "_time": time.time(),
                    "_raw": "Outliers detection are disabled at the global level for this entity, nothing to do.",
                    "response": "Outliers detection are disabled at the global level for this entity, nothing to do.",
                }

            elif outliers_is_disabled == 0:
                # set a list for error reporting purposes of available modesl
                entity_outliers_models = []

                # Process render
                process_render = False

                # Loop through outliers entities
                for entity_outlier in entities_outliers:
                    # check is_disabled
                    is_disabled = int(entities_outliers[entity_outlier]["is_disabled"])

                    # log debug
                    logging.debug(
                        f'entity_outlier="{entity_outlier}", is_disabled="{is_disabled}"'
                    )

                    # Add to the list
                    if is_disabled == 0:
                        entity_outliers_models.append(entity_outlier)
                    else:
                        logging.debug(
                            f'entity_outlier="{entity_outlier}", entity is disabled, is_disabled="{is_disabled}"'
                        )

                # if all models have been disabked
                if not entity_outliers_models:
                    # bool
                    process_render = False

                    # yield
                    yield {
                        "_time": time.time(),
                        "_raw": "All models for this entity are currently disabled, nothing to do.",
                        "response": "All models for this entity are currently disabled, nothing to do.",
                    }

                elif self.model_id:
                    # check is_disabled for this model
                    is_disabled = int(entities_outliers[self.model_id]["is_disabled"])

                    # log debug
                    logging.debug(
                        f'model_id="{self.model_id}", is_disabled="{is_disabled}"'
                    )

                    if is_disabled != 0:
                        # bool
                        process_render = False

                        # yield
                        yield {
                            "_time": time.time(),
                            "_raw": "This model is currently disabled, nothing to do.",
                            "response": "This model is currently disabled, nothing to do.",
                        }

                    else:
                        # bool
                        process_render = True

                        # normalise
                        model_id = self.model_id

                else:
                    # bool
                    process_render = True

                    # normalise, select first available model
                    model_id = entity_outliers_models[0]

                # if process render
                if process_render:
                    # Extract as a dict
                    entity_outlier_dict = entities_outliers[model_id]

                    # log debug
                    logging.debug(f'entity_outlier_dict="{entity_outlier_dict}"')

                    # Extract the render search
                    ml_model_render_search = entity_outlier_dict[
                        "ml_model_render_search"
                    ]

                    # if the search is pending, rendering outliers is not ready yet
                    if ml_model_render_search == "pending":
                        error_msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{model_id}", The ML search is not yet available for rendering, please train this model first.'
                        logging.error(error_msg)
                        raise Exception(error_msg)

                    # log debug
                    logging.debug(f'ml_model_render_search="{ml_model_render_search}"')

                    # Get the perc_min_lowerbound_deviation
                    perc_min_lowerbound_deviation = float(
                        entity_outlier_dict["perc_min_lowerbound_deviation"]
                    )
                    logging.debug(
                        f'perc_min_lowerbound_deviation="{perc_min_lowerbound_deviation}"'
                    )

                    # Get the perc_min_upperbound_deviation
                    perc_min_upperbound_deviation = float(
                        entity_outlier_dict["perc_min_upperbound_deviation"]
                    )
                    logging.debug(
                        f'perc_min_upperbound_deviation="{perc_min_upperbound_deviation}"'
                    )

                    # Run the search and render outliers

                    # set kwargs
                    kwargs_oneshot = {
                        "earliest_time": earliest,
                        "latest_time": latest,
                        "search_mode": "normal",
                        "preview": False,
                        "time_format": "%s",
                        "count": 0,
                        "output_mode": "json",
                    }

                    # automatic retry
                    search_successful = False
                    current_failures_count = 0
                    while (
                        current_failures_count < max_failures_count
                        and not search_successful
                    ):
                        try:
                            # use the export API which is not limited in number of results
                            search_results = self.service.jobs.export(
                                ml_model_render_search, **kwargs_oneshot
                            )
                            reader = results.JSONResultsReader(search_results)

                            # loop through the reader results
                            for item in reader:
                                if isinstance(item, dict):
                                    search_results = item

                                    # raw results logged only in debug
                                    logging.debug(f'search_results="{search_results}"')

                                    # yield_record
                                    yield_record = {}

                                    # auto correct parameter, can come as an option to the CLI or part of the model definition
                                    auto_correct = True

                                    try:
                                        model_auto_correct = int(
                                            entity_outlier_dict["auto_correct"]
                                        )
                                        if model_auto_correct == 0:
                                            auto_correct = False
                                        elif model_auto_correct == 1:
                                            auto_correct = True
                                    except Exception as e:
                                        if self.auto_correct == "True":
                                            auto_correct = True
                                        elif self.auto_correct == "False":
                                            auto_correct = self.auto_correct

                                    # log
                                    logging.debug(f'auto_correct="{auto_correct}"')

                                    # loop through the fields, and process outliers rendering
                                    for k in search_results:
                                        # log if the lower and/or upper outliers were corrected
                                        LowerBoundWasCorrected = False
                                        LowerBoundCorrectionReason = None
                                        UpperBoundWasCorrected = False
                                        UpperBoundCorrectionReason = None

                                        # get the kpi metric name and value
                                        kpi_metric_name = entity_outlier_dict[
                                            "kpi_metric"
                                        ]
                                        kpi_metric_value = search_results[
                                            entity_outlier_dict["kpi_metric"]
                                        ]
                                        logging.debug(
                                            f'kpi_metric_name="{kpi_metric_name}", kpi_metric_value="{kpi_metric_value}"'
                                        )

                                        # calculate the perc_min_lowerbound_deviation value
                                        perc_min_lowerbound_deviation_value = (
                                            float(kpi_metric_value)
                                            * int(perc_min_lowerbound_deviation)
                                            / 100
                                        )
                                        logging.debug(
                                            f"kpi_metric_value={kpi_metric_value}, perc_min_lowerbound_deviation={perc_min_lowerbound_deviation}, perc_min_lowerbound_deviation_value={perc_min_lowerbound_deviation_value}"
                                        )

                                        # calculate the perc_min_upperbound_deviation value
                                        perc_min_upperbound_deviation_value = (
                                            float(kpi_metric_value)
                                            * int(perc_min_upperbound_deviation)
                                            / 100
                                        )
                                        logging.debug(
                                            f"kpi_metric_value={kpi_metric_value}, perc_min_upperbound_deviation={perc_min_upperbound_deviation}, perc_min_upperbound_deviation_value={perc_min_upperbound_deviation_value}"
                                        )

                                        # caclulate the corrected candidates
                                        LowerBoundMin = float(kpi_metric_value) - float(
                                            perc_min_lowerbound_deviation_value
                                        )
                                        UpperBoundMin = float(kpi_metric_value) + float(
                                            perc_min_upperbound_deviation_value
                                        )
                                        logging.debug(
                                            f'LowerBoundMin="{LowerBoundMin}", UpperBoundMin="{UpperBoundMin}"'
                                        )

                                        # try to get the LowerBound and UpperBound, if we have no results (not enough historical data), apply corrected values instead
                                        try:
                                            LowerBound = search_results["LowerBound"]
                                        except Exception as e:
                                            LowerBoundWasCorrected = True
                                            LowerBoundCorrectionReason = "No value was generated, likely due to lack of historical data"
                                            LowerBound = LowerBoundMin
                                            logging.warning(
                                                f'Could not retrieve a LowerBound value from item="{item}", likely we have not enough historical data yet, applying corrected value="{LowerBound}" instead'
                                            )

                                        try:
                                            UpperBound = search_results["UpperBound"]
                                        except Exception as e:
                                            UpperBoundWasCorrected = True
                                            UpperBoundCorrectionReason = "No value was generated, likely due to lack of historical data"
                                            UpperBound = UpperBoundMin
                                            logging.warning(
                                                f'Could not retrieve a UpperBound value from item="{item}", likely we have not enough historical data yet, applying corrected value="{UpperBound}" instead'
                                            )

                                        # apply
                                        if auto_correct:
                                            # condition for a lower outlier: generated lower threshold is greater than the kpi value
                                            # condition for an upper outlier: generated upper threshold is lower than the kpi value

                                            currentLowerBoundDeviationValue = float(
                                                LowerBound
                                            ) - float(kpi_metric_value)

                                            logging.debug(
                                                f"currentLowerBoundDeviationValue={currentLowerBoundDeviationValue}"
                                            )

                                            currentUpperBoundDeviationValue = float(
                                                kpi_metric_value
                                            ) - float(UpperBound)

                                            logging.debug(
                                                f"currentUpperBoundDeviationValue={currentUpperBoundDeviationValue}"
                                            )

                                            # for lowerBound, replace as well if equal or lower than 0 unless requested to allow this behavior
                                            if (
                                                float(LowerBound) <= 0
                                                and not self.lowerbound_negative
                                                == "True"
                                            ):
                                                LowerBoundWasCorrected = True
                                                LowerBoundCorrectionReason = f"Generated LowerBound {float(LowerBound)} is negative or equal to 0"
                                                LowerBoundOrig = LowerBound
                                                LowerBound = float(LowerBoundMin)

                                            # for upperBound, replace as well if equal or lower than 0
                                            if float(UpperBound) <= 0:
                                                UpperBoundWasCorrected = True
                                                UpperBoundCorrectionReason = f"Generated UpperBound {float(UpperBound)} is negative or equal to 0"
                                                UpperBoundOrig = UpperBound
                                                UpperBound = float(UpperBoundMin)

                                            #
                                            # lower
                                            #

                                            # if a lower outlier is said to be detected
                                            if float(LowerBound) > float(
                                                kpi_metric_value
                                            ):
                                                # the generated lower bound should be not lower than the safety margin
                                                if not float(
                                                    currentLowerBoundDeviationValue
                                                ) > float(
                                                    perc_min_lowerbound_deviation_value
                                                ):
                                                    # apply safeties instead of generated
                                                    LowerBoundWasCorrected = True
                                                    LowerBoundCorrectionReason = f"Current LowerBound deviation value {currentLowerBoundDeviationValue} is not higher than minimal deviation value {perc_min_lowerbound_deviation_value} using {perc_min_lowerbound_deviation} pct deviation"
                                                    LowerBoundOrig = LowerBound
                                                    LowerBound = float(LowerBoundMin)

                                                else:
                                                    # else accept the outlier
                                                    LowerBoundOrig = LowerBound

                                            else:
                                                LowerBoundOrig = LowerBound

                                            #
                                            # upper
                                            #

                                            # If an upper outlier is said to be detected
                                            if float(UpperBound) < float(
                                                kpi_metric_value
                                            ):
                                                # the generated upper bound should be higher than the safety margin
                                                if not float(
                                                    currentUpperBoundDeviationValue
                                                ) > float(
                                                    perc_min_upperbound_deviation_value
                                                ):
                                                    # apply safeties instead of generated
                                                    UpperBoundWasCorrected = True
                                                    UpperBoundCorrectionReason = f"Current UpperBound deviation value {currentUpperBoundDeviationValue} is not higher than minimal deviation value {perc_min_upperbound_deviation_value} using {perc_min_upperbound_deviation} pct deviation"
                                                    UpperBoundOrig = UpperBound
                                                    UpperBound = float(UpperBoundMin)

                                                # else accept the outlier
                                                else:
                                                    UpperBoundOrig = UpperBound

                                            else:
                                                UpperBoundOrig = UpperBound

                                            # lower bound and upper bound cannot be equal
                                            if float(LowerBound) == float(UpperBound):
                                                # apply safeties instead of generated
                                                LowerBoundWasCorrected = True
                                                LowerBoundCorrectionReason = f"LowerBound value {LowerBoundOrig} and UpperBound value {UpperBoundOrig} cannot be equal"
                                                LowerBoundOrig = LowerBound
                                                LowerBound = float(LowerBoundMin)

                                                # apply safeties instead of generated
                                                UpperBoundWasCorrected = True
                                                UpperBoundCorrectionReason = f"LowerBound value {LowerBoundOrig} and UpperBound value {UpperBoundOrig} cannot be equal"
                                                UpperBoundOrig = UpperBound
                                                UpperBound = float(UpperBoundMin)

                                        # do not correct anything
                                        else:
                                            LowerBoundOrig = LowerBound
                                            UpperBoundOrig = UpperBound

                                        # Add to the dict
                                        yield_record["_time"] = search_results["_time"]
                                        yield_record["LowerBound"] = LowerBound
                                        yield_record["UpperBound"] = UpperBound
                                        yield_record[kpi_metric_name] = kpi_metric_value

                                        # Add _raw
                                        yield_record["_raw"] = {
                                            "_time": search_results["_time"],
                                            "kpi_metric_name": kpi_metric_name,
                                            "kpi_metric_value": kpi_metric_value,
                                            "LowerBoundMin": LowerBoundMin,
                                            "LowerBoundOrig": LowerBoundOrig,
                                            "LowerBound": LowerBound,
                                            "UpperBoundMin": UpperBoundMin,
                                            "UpperBoundOrig": UpperBoundOrig,
                                            "UpperBound": UpperBound,
                                            "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                                            "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                                            "LowerBoundWasCorrected": LowerBoundWasCorrected,
                                            "LowerBoundCorrectionReason": LowerBoundCorrectionReason,
                                            "UpperBoundWasCorrected": UpperBoundWasCorrected,
                                            "UpperBoundCorrectionReason": UpperBoundCorrectionReason,
                                            "search_results": search_results,
                                        }

                                    # yield
                                    yield yield_record

                            # for diag purposes
                            logging.debug(
                                f'diagnostic, is_preview="{reader.is_preview}"'
                            )

                            # log
                            logging.info(
                                f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{model_id}", search was terminated successfully, duration={time.time() - start}, search="{ml_model_render_search}"'
                            )

                            # break while
                            search_successful = True
                            break

                        except Exception as e:
                            if re.search(
                                "The maximum number of concurrent historical searches on this instance has been reached",
                                str(e),
                            ):
                                current_failures_count += 1
                                search_successful = False
                                if current_failures_count >= max_failures_count:
                                    msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{model_id}", permanent search failure, current_failures_count="{current_failures_count}", max_failures_count="{max_failures_count}", exception="{e}"'
                                    logging.error(msg)
                                    raise Exception(msg)
                                else:
                                    msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{model_id}", temporary search failure with the following exception="{e}", search="{ml_model_render_search}"'
                                    logging.warn(msg)
                                    time.sleep(sleep_time)
                            else:
                                msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{model_id}", search has failed with the following exception="{e}", search="{ml_model_render_search}"'
                                logging.error(msg)
                                raise Exception(msg)

        elif self.mode == "simulation":
            # log debug
            logging.debug("mode is simulation")

            # log debug
            logging.debug(f"model_json_def={self.model_json_def}")

            # load the model definition as a dict
            try:
                model_json_def = json.loads(self.model_json_def)
                # log debug
                logging.debug(
                    f'successfully loaded model_json_def="{json.dumps(model_json_def, indent=4)}"'
                )
            except Exception as e:
                msg = f'failed to load the submitted model_json_def="{self.model_json_def}" with exception="{e}"'
                logging.error(msg)
                raise Exception(msg)

            # set the tenant_trackme_metric_idx
            tenant_trackme_metric_idx = None

            # get the index conf for this tenant
            url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/tenant_idx_settings"
            data = {"tenant_id": self.tenant_id, "idx_stanza": "trackme_metric_idx"}

            # Retrieve and set the tenant idx, if any failure, logs and use the global index
            try:
                response = requests.post(
                    url, headers=header, data=json.dumps(data, indent=1), verify=False, timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    tenant_trackme_metric_idx = trackme_metric_idx
                else:
                    try:
                        response_data = json.loads(
                            json.dumps(response.json(), indent=1)
                        )
                        tenant_trackme_metric_idx = response_data["trackme_metric_idx"]
                    except Exception as e:
                        tenant_trackme_metric_idx = response_data["trackme_metric_idx"]

            except Exception as e:
                tenant_trackme_metric_idx = trackme_metric_idx

            # auto correct parameter, can come as an option to the CLI or part of the model definition
            auto_correct = True

            try:
                model_auto_correct = int(model_json_def.get("auto_correct"))
                if model_auto_correct == 0:
                    auto_correct = False
                elif model_auto_correct == 1:
                    auto_correct = True
            except Exception as e:
                if self.auto_correct == "True":
                    auto_correct = True
                elif self.auto_correct == "False":
                    auto_correct = self.auto_correct

            # log
            logging.debug(f'auto_correct="{auto_correct}"')

            # Set the densityFunction threshold parameters
            if (
                float(model_json_def.get("density_lowerthreshold")) > 0
                and float(model_json_def.get("density_lowerthreshold")) > 0
            ):
                density_threshold_str = (
                    "lower_threshold="
                    + str(model_json_def.get("density_lowerthreshold"))
                    + " upper_threshold="
                    + str(model_json_def.get("density_upperthreshold"))
                )
            else:
                density_threshold_str = "lower_threshold=0.005 upper_threshold=0.005"
                error_msg = (
                    "densityFunction threshold parameters are incorrects for this entity, "
                    + "lower_threshold and upper_threshold must both be a positive value, "
                    + "will be using using factory value."
                )
                logging.error(
                    f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.tenant_id}", {error_msg}'
                )

            # define the simulation search
            ml_model_render_search = remove_leading_spaces(
                f"""\
            | mstats avg(trackme.{model_json_def.get("kpi_metric")}) as {model_json_def.get("kpi_metric")} where index="{tenant_trackme_metric_idx}" tenant_id="{self.tenant_id}" object_category="splk-{self.component}" object="{self.object}" by object span="{model_json_def.get("kpi_span")}"
            | eval factor=strftime(_time, "{model_json_def.get("time_factor")}")
            | fit DensityFunction {model_json_def.get("kpi_metric")} {density_threshold_str} by factor
            | rex field=BoundaryRanges "(-Infinity:(?<LowerBound>[\\d|\\.]*))|((?<UpperBound>[\\d|\\.]*):Infinity)"
            | foreach LowerBound UpperBound [ eval <<FIELD>> = if(isnum('<<FIELD>>'), '<<FIELD>>', 0) ]
            | fields _time {model_json_def.get("kpi_metric")} LowerBound UpperBound
            | where _time>={earliest}
            """
            )
            logging.debug(f'ml_model_render_search="{ml_model_render_search}"')

            # Get the perc_min_lowerbound_deviation
            perc_min_lowerbound_deviation = float(
                model_json_def.get("perc_min_lowerbound_deviation")
            )
            logging.debug(
                f'perc_min_lowerbound_deviation="{perc_min_lowerbound_deviation}"'
            )

            # Get the perc_min_upperbound_deviation
            perc_min_upperbound_deviation = float(
                model_json_def.get("perc_min_upperbound_deviation")
            )
            logging.debug(
                f'perc_min_upperbound_deviation="{perc_min_upperbound_deviation}"'
            )

            # set kwargs
            kwargs_oneshot = {
                "earliest_time": model_json_def.get("period_calculation"),
                "latest_time": latest,
                "search_mode": "normal",
                "preview": False,
                "time_format": "%s",
                "count": 0,
                "output_mode": "json",
            }

            # automatic retry
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                try:
                    # use the export API which is not limited in number of results
                    search_results = self.service.jobs.export(
                        ml_model_render_search, **kwargs_oneshot
                    )
                    reader = results.JSONResultsReader(search_results)

                    # loop through the reader results
                    for item in reader:
                        if isinstance(item, dict):
                            search_results = item

                            # raw results logged only in debug
                            logging.debug(f'search_results="{search_results}"')

                            # yield_record
                            yield_record = {}

                            # loop through the fields, and process outliers rendering
                            for k in search_results:
                                # log if the lower and/or upper outliers were corrected
                                LowerBoundWasCorrected = False
                                LowerBoundCorrectionReason = None
                                UpperBoundWasCorrected = False
                                UpperBoundCorrectionReason = None

                                # get the kpi metric name and value
                                kpi_metric_name = model_json_def.get("kpi_metric")
                                kpi_metric_value = search_results[
                                    model_json_def.get("kpi_metric")
                                ]
                                logging.debug(
                                    f'kpi_metric_name="{kpi_metric_name}", kpi_metric_value="{kpi_metric_value}"'
                                )

                                # calculate the perc_min_lowerbound_deviation value
                                perc_min_lowerbound_deviation_value = (
                                    float(kpi_metric_value)
                                    * int(perc_min_lowerbound_deviation)
                                    / 100
                                )
                                logging.debug(
                                    f"kpi_metric_value={kpi_metric_value}, perc_min_lowerbound_deviation={perc_min_lowerbound_deviation}, perc_min_lowerbound_deviation_value={perc_min_lowerbound_deviation_value}"
                                )

                                # calculate the perc_min_upperbound_deviation value
                                perc_min_upperbound_deviation_value = (
                                    float(kpi_metric_value)
                                    * int(perc_min_upperbound_deviation)
                                    / 100
                                )
                                logging.info(
                                    f"kpi_metric_value={kpi_metric_value}, perc_min_upperbound_deviation={perc_min_upperbound_deviation}, perc_min_upperbound_deviation_value={perc_min_upperbound_deviation_value}"
                                )

                                # calculate the corrected candidates
                                LowerBoundMin = float(kpi_metric_value) - float(
                                    perc_min_lowerbound_deviation_value
                                )
                                UpperBoundMin = float(kpi_metric_value) + float(
                                    perc_min_upperbound_deviation_value
                                )
                                logging.debug(
                                    f'LowerBoundMin="{LowerBoundMin}", UpperBoundMin="{UpperBoundMin}"'
                                )

                                # try to get the LowerBound and UpperBound, if we have no results (not enough historical data), apply corrected values instead
                                try:
                                    LowerBound = search_results["LowerBound"]
                                except Exception as e:
                                    LowerBoundWasCorrected = True
                                    LowerBoundCorrectionReason = "No value was generated, likely due to lack of historical data"
                                    LowerBound = LowerBoundMin
                                    logging.warning(
                                        f'Could not retrieve a LowerBound value from item="{item}", likely we have not enough historical data yet, applying corrected value="{LowerBound}" instead'
                                    )

                                try:
                                    UpperBound = search_results["UpperBound"]
                                except Exception as e:
                                    UpperBoundWasCorrected = True
                                    UpperBoundCorrectionReason = "No value was generated, likely due to lack of historical data"
                                    UpperBound = UpperBoundMin
                                    logging.warning(
                                        f'Could not retrieve a UpperBound value from item="{item}", likely we have not enough historical data yet, applying corrected value="{UpperBound}" instead'
                                    )

                                # apply
                                if auto_correct:
                                    # condition for a lower outlier: generated lower threshold is greater than the kpi value
                                    # condition for an upper outlier: generated upper threshold is lower than the kpi value

                                    currentLowerBoundDeviationValue = float(
                                        LowerBound
                                    ) - float(kpi_metric_value)

                                    logging.debug(
                                        f"currentLowerBoundDeviationValue={currentLowerBoundDeviationValue}"
                                    )

                                    currentUpperBoundDeviationValue = float(
                                        kpi_metric_value
                                    ) - float(UpperBound)

                                    logging.debug(
                                        f"currentUpperBoundDeviationValue={currentUpperBoundDeviationValue}"
                                    )

                                    # for lowerBound, replace as well if equal or lower than 0 unless requested to allow this behavior
                                    if (
                                        float(LowerBound) <= 0
                                        and not self.lowerbound_negative == "True"
                                    ):
                                        LowerBoundWasCorrected = True
                                        LowerBoundCorrectionReason = f"Generated LowerBound {float(LowerBound)} is negative or equal to 0"
                                        LowerBoundOrig = LowerBound
                                        LowerBound = float(LowerBoundMin)

                                    # for upperBound, replace as well if equal or lower than 0
                                    if float(UpperBound) <= 0:
                                        UpperBoundWasCorrected = True
                                        UpperBoundCorrectionReason = f"Generated UpperBound {float(UpperBound)} is negative or equal to 0"
                                        UpperBoundOrig = UpperBound
                                        UpperBound = float(UpperBoundMin)

                                    #
                                    # lower
                                    #

                                    # if a lower outlier is said to be detected
                                    if float(LowerBound) > float(kpi_metric_value):
                                        # the generated lower bound should be not lower than the safety margin
                                        if not float(
                                            currentLowerBoundDeviationValue
                                        ) > float(perc_min_lowerbound_deviation_value):
                                            # apply safeties instead of generated
                                            LowerBoundWasCorrected = True
                                            LowerBoundCorrectionReason = f"Current LowerBound deviation value {currentLowerBoundDeviationValue} is not higher than minimal deviation value {perc_min_lowerbound_deviation_value} using {perc_min_lowerbound_deviation} pct deviation"
                                            LowerBoundOrig = LowerBound
                                            LowerBound = float(LowerBoundMin)

                                        else:
                                            # else accept the outlier
                                            LowerBoundOrig = LowerBound

                                    else:
                                        LowerBoundOrig = LowerBound

                                    #
                                    # upper
                                    #

                                    # If an upper outlier is said to be detected
                                    if float(UpperBound) < float(kpi_metric_value):
                                        # the generated upper bound should be higher than the safety margin

                                        if not float(
                                            currentUpperBoundDeviationValue
                                        ) > float(perc_min_upperbound_deviation_value):
                                            # apply safeties instead of generated
                                            UpperBoundWasCorrected = True
                                            UpperBoundCorrectionReason = f"Current UpperBound deviation value {currentUpperBoundDeviationValue} is not higher than minimal deviation value {perc_min_upperbound_deviation_value} using {perc_min_upperbound_deviation} pct deviation"
                                            UpperBoundOrig = UpperBound
                                            UpperBound = float(UpperBoundMin)

                                        # else accept the outlier
                                        else:
                                            UpperBoundOrig = UpperBound

                                    else:
                                        UpperBoundOrig = UpperBound

                                    # lower bound and upper bound cannot be equal
                                    if float(LowerBound) == float(UpperBound):
                                        # apply safeties instead of generated
                                        LowerBoundWasCorrected = True
                                        LowerBoundCorrectionReason = f"LowerBound value {LowerBoundOrig} and UpperBound value {UpperBoundOrig} cannot be equal"
                                        LowerBoundOrig = LowerBound
                                        LowerBound = float(LowerBoundMin)

                                        # apply safeties instead of generated
                                        UpperBoundWasCorrected = True
                                        UpperBoundCorrectionReason = f"LowerBound value {LowerBoundOrig} and UpperBound value {UpperBoundOrig} cannot be equal"
                                        UpperBoundOrig = UpperBound
                                        UpperBound = float(UpperBoundMin)

                                # do not correct anything
                                else:
                                    LowerBoundOrig = LowerBound
                                    UpperBoundOrig = UpperBound

                                # Add to the dict
                                yield_record["_time"] = search_results["_time"]
                                yield_record["LowerBound"] = LowerBound
                                yield_record["UpperBound"] = UpperBound
                                yield_record[kpi_metric_name] = kpi_metric_value

                                # Add _raw
                                yield_record["_raw"] = {
                                    "_time": search_results["_time"],
                                    "kpi_metric_name": kpi_metric_name,
                                    "kpi_metric_value": kpi_metric_value,
                                    "LowerBoundMin": LowerBoundMin,
                                    "LowerBoundOrig": LowerBoundOrig,
                                    "LowerBound": LowerBound,
                                    "UpperBoundMin": UpperBoundMin,
                                    "UpperBoundOrig": UpperBoundOrig,
                                    "UpperBound": UpperBound,
                                    "perc_min_lowerbound_deviation": perc_min_lowerbound_deviation,
                                    "perc_min_upperbound_deviation": perc_min_upperbound_deviation,
                                    "LowerBoundWasCorrected": LowerBoundWasCorrected,
                                    "LowerBoundCorrectionReason": LowerBoundCorrectionReason,
                                    "UpperBoundWasCorrected": UpperBoundWasCorrected,
                                    "UpperBoundCorrectionReason": UpperBoundCorrectionReason,
                                    "search_results": search_results,
                                }

                            # yield
                            yield yield_record

                    # log
                    logging.info(
                        f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{self.model_id}", search was terminated successfully, duration={time.time() - start}, search="{ml_model_render_search}"'
                    )

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{self.model_id}", permanent search failure with the following exception="{e}", search="{ml_model_render_search}"'
                            logging.error(msg)
                            raise Exception(msg)
                        else:
                            msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{self.model_id}", temporary search failure with the following exception="{e}", search="{ml_model_render_search}"'
                            logging.warn(msg)
                            time.sleep(sleep_time)
                    else:
                        msg = f'tenant_id="{self.tenant_id}", component="{self.component}", object="{self.object}", model_id="{self.model_id}", search has failed with the following exception="{e}", search="{ml_model_render_search}"'
                        logging.error(msg)
                        raise Exception(msg)


dispatch(SplkOutliersRender, sys.argv, sys.stdin, sys.stdout, __name__)
