#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import re
import json

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_splk_outliers_train_helper.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.client as client
import splunklib.results as results

# import trackme libs
from trackme_libs import trackme_reqinfo, trackme_register_tenant_object_summary


@Configuration(distributed=False)
class SplkOutliersExecutor(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** The component category.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(?:dsm|dhm|cim|flx|wlk)$"),
    )

    earliest = Option(
        doc="""
        **Syntax:** **earliest=****
        **Description:** The earliest time quantifier.""",
        require=False,
        default="-24h",
    )

    latest = Option(
        doc="""
        **Syntax:** **latest=****
        **Description:** The latest time quantifier.""",
        require=False,
        default="now",
    )

    max_runtime_sec = Option(
        doc="""
        **Syntax:** **max_runtime_sec=****
        **Description:** The max runtime for the job in seconds, defaults to 60 minutes less 120 seconds of margin.""",
        require=False,
        default="3600",
        validate=validators.Match("max_runtime_sec", r"^\d*$"),
    )

    def generate(self, **kwargs):
        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5

        # performance counter
        mainstart = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # Get app level config
        splk_outliers_time_train_mlmodels_default = reqinfo["trackme_conf"][
            "splk_outliers_detection"
        ]["splk_outliers_time_train_mlmodels_default"]
        splk_outliers_max_runtime_train_mlmodels_default = reqinfo["trackme_conf"][
            "splk_outliers_detection"
        ]["splk_outliers_max_runtime_train_mlmodels_default"]

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # Get data
        kwargs_oneshot = {
            "earliest_time": self.earliest,
            "latest_time": self.latest,
            "output_mode": "json",
            "count": 0,
        }

        #
        # RUN
        #

        # Define the search providing the list of entities which models need to be train

        # note: do not allow crap entities, sometimes users get very fancy, entities with double quotes in the value
        # are not expected nor desirable, do not allow crap to get in

        search_query = (
            '| inputlookup trackme_%s_outliers_entity_rules_tenant_%s where object_category="splk-%s"'
            % (self.component, self.tenant_id, self.component)
            + "\n | `trackme_exclude_badentities`"
            + "\n | lookup local=t trackme_%s_tenant_%s object OUTPUT monitored_state"
            % (self.component, self.tenant_id)
            + '\n | where monitored_state="enabled"'
            + '\n | eval duration_since_last=if(last_exec!="pending", now()-last_exec, 0)'
            + "\n | where duration_since_last=0 OR duration_since_last>=%s"
            % (str(splk_outliers_time_train_mlmodels_default))
            + "\n | sort - duration_since_last"
        )

        logging.debug(
            'tenant_id="{}", component="{}", retrieving the list of entities to be trained from the upstream search="{}"'.format(
                self.tenant_id, self.component, search_query
            )
        )

        # automatic retry
        search_successful = False
        current_failures_count = 0
        while current_failures_count < max_failures_count and not search_successful:
            # run the main report, every result is a Splunk search to be executed on its own thread
            try:
                oneshotsearch_results = self.service.jobs.oneshot(
                    search_query, **kwargs_oneshot
                )
                reader = results.JSONResultsReader(oneshotsearch_results)

                # loop through the results, and train models per entity

                # Store various key information
                processed_entities = []
                failures_entities = []
                failures_entities_object_list = []
                searches_toprocess = []

                for item in reader:
                    logging.debug('search_results="{}"'.format(item))

                    if isinstance(item, dict):
                        # break if reaching the max run time less 30 seconds of margin
                        if (time.time() - int(mainstart)) - 30 >= int(
                            splk_outliers_max_runtime_train_mlmodels_default
                        ):
                            logging.info(
                                'tenant_id="{}" max_runtime="{}" for ML models was reached with current_runtime="{}", job will be terminated now'.format(
                                    self.tenant_id,
                                    str(
                                        splk_outliers_max_runtime_train_mlmodels_default
                                    ),
                                    str(mainstart),
                                )
                            )
                            break

                        # set the search depending on the component
                        if self.component in ("dsm", "dhm", "flx", "wlk"):
                            search_train = (
                                '| trackmesplkoutlierstrain tenant_id="%s" component="%s" object="%s"'
                                % (self.tenant_id, self.component, item.get("object"))
                            )
                        elif self.component in ("cim"):
                            search_train = (
                                '| trackmesplkoutlierscimtrain tenant_id="%s" component="%s" object="%s" cim_field="%s"'
                                % (
                                    self.tenant_id,
                                    self.component,
                                    item.get("object"),
                                    item.get("cim_field"),
                                )
                            )

                        # append to our list
                        searches_toprocess.append(
                            {
                                "object_category": item.get("object_category"),
                                "object": item.get("object"),
                                "search": search_train,
                            }
                        )

                        logging.debug(
                            'entity_dict="{}"'.format(
                                json.dumps(
                                    {
                                        "object_category": item.get("object_category"),
                                        "object": item.get("object"),
                                        "search": search_train,
                                    },
                                    indent=2,
                                )
                            )
                        )

                # break while
                search_successful = True
                break

            except Exception as e:
                if re.search(
                    "The maximum number of concurrent historical searches on this instance has been reached",
                    str(e),
                ):
                    current_failures_count += 1
                    search_successful = False
                    if current_failures_count >= max_failures_count:
                        # Call the component register
                        trackme_register_tenant_object_summary(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            "splk-" + self.component,
                            "trackme_"
                            + self.component
                            + "_outliers_mltrain_tracker_tenant_"
                            + str(self.tenant_id),
                            "failure",
                            time.time(),
                            round(time.time() - mainstart, 3),
                            str(e),
                            "-24h",
                            "now",
                        )
                        msg = 'tenant_id="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                            self.tenant_id,
                            current_failures_count,
                            max_failures_count,
                            str(e),
                        )
                        logging.error(msg)
                        raise Exception(msg)

                    else:
                        msg = 'tenant_id="{}", temporary search failure with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                        logging.warn(msg)
                        time.sleep(sleep_time)

                else:
                    msg = (
                        'tenant_id="{}", main search failed with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                    )
                    logging.error(msg)
                    raise Exception(msg)

        #
        # Process
        #

        logging.debug(
            'searched to be processed="{}"'.format(
                json.dumps(searches_toprocess, indent=2)
            )
        )

        # end

        # errors counter
        search_errors_count = 0

        # run
        if len(searches_toprocess) > 0:
            #
            # process searches
            #

            for search_record in searches_toprocess:
                search = search_record.get("search")
                object_category = search_record.get("object_category")
                object_value = search_record.get("object")
                results_entity = []

                # Break if reaching the max run time less 120 seconds of margin
                current_runtime = time.time() - int(mainstart)
                if current_runtime >= (int(self.max_runtime_sec) - 120):
                    logging.info(
                        f'tenant_id="{self.tenant_id}" max_runtime="{self.max_runtime_sec}" is about to be reached, current_runtime="{current_runtime}", job will be terminated now'
                    )
                    break

                logging.debug(
                    'tenant_id="{}" Executing resulting search="{}"'.format(
                        self.tenant_id, search
                    )
                )

                # automatic retry
                search_successful = False
                current_failures_count = 0
                while (
                    current_failures_count < max_failures_count
                    and not search_successful
                ):
                    start = time.time()
                    try:
                        oneshotsearch_results = self.service.jobs.oneshot(
                            search, **kwargs_oneshot
                        )

                        # Get the results and display them using the JSONResultsReader
                        reader = results.JSONResultsReader(oneshotsearch_results)
                        for item in reader:
                            if isinstance(item, dict):
                                logging.debug('search_results="{}"'.format(item))
                                results_entity.append(item)

                        # don't be too noisy
                        if reqinfo["logging_level"] == "DEBUG":
                            processed_entities.append(
                                {
                                    "object_category": object_category,
                                    "object": object_value,
                                    "results_entity": results_entity,
                                    "search": search,
                                    "runtime": str(time.time() - start),
                                }
                            )
                        else:
                            processed_entities.append(
                                {
                                    "object_category": object_category,
                                    "object": object_value,
                                    "search": search,
                                    "runtime": str(time.time() - start),
                                }
                            )

                        # only in debug
                        logging.debug(
                            'tenant_id="{}" search successfully executed in {} seconds'.format(
                                self.tenant_id, str(time.time() - start)
                            )
                        )

                        # break while
                        search_successful = True
                        break

                    except Exception as e:
                        if re.search(
                            "The maximum number of concurrent historical searches on this instance has been reached",
                            str(e),
                        ):
                            current_failures_count += 1
                            search_successful = False
                            if current_failures_count >= max_failures_count:
                                msg = 'tenant_id="{}", component="{}", search="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                    self.tenant_id,
                                    self.component,
                                    search,
                                    current_failures_count,
                                    max_failures_count,
                                    str(e),
                                )
                                logging.error(msg)
                                search_errors_count += 1
                                failures_entities.append(
                                    {
                                        "object_category": object_category,
                                        "object": object_value,
                                        "results_entity": {
                                            "action": "failure",
                                            "search": search,
                                            "exception": str(e),
                                        },
                                    }
                                )
                                failures_entities_object_list.append(
                                    {"object": object_value, "exception": str(e)}
                                )

                            else:
                                msg = 'tenant_id="{}", component="{}", search="{}", temporary search failure with exception="{}"'.format(
                                    self.tenant_id, self.component, search, str(e)
                                )
                                logging.warn(msg)
                                time.sleep(sleep_time)

                        else:
                            msg = 'tenant_id="{}", component="{}", search="{}", main search failed with exception="{}"'.format(
                                self.tenant_id, self.component, search, str(e)
                            )
                            logging.error(msg)
                            search_errors_count += 1
                            failures_entities.append(
                                {
                                    "object_category": object_category,
                                    "object": object_value,
                                    "results_entity": {
                                        "action": "failure",
                                        "search": search,
                                        "exception": str(e),
                                    },
                                }
                            )
                            failures_entities_object_list.append(
                                {"object": object_value, "exception": str(e)}
                            )
                            search_successful = False
                            break

            #
            # end process searches loop
            #

            # yield and log
            results_dict = {
                "tenant_id": self.tenant_id,
                "action": "success",
                "results": "outliers models training job successfully executed",
                "run_time": round((time.time() - mainstart), 3),
                "entities_count": len(searches_toprocess),
                "processed_entities": processed_entities,
                "failures_entities": failures_entities,
                "search_errors_count": search_errors_count,
                "upstream_search_query": search_query,
            }
            yield {"_time": time.time(), "_raw": results_dict}
            logging.info(json.dumps(results_dict, indent=2))

        else:
            # yield and log
            results_dict = {
                "tenant_id": self.tenant_id,
                "action": "success",
                "results": "outliers models training job successfully executed but there were no entities to be trained at this time",
                "run_time": round((time.time() - mainstart), 3),
                "entities_count": len(searches_toprocess),
                "upstream_search_query": search_query,
            }
            yield {"_time": time.time(), "_raw": results_dict}
            logging.info(json.dumps(results_dict, indent=2))

        # Call the component register
        report_name = "trackme_%s_outliers_mltrain_tracker_tenant_%s" % (
            self.component,
            self.tenant_id,
        )
        if search_errors_count == 0:
            trackme_register_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                "splk-%s" % (self.component),
                report_name,
                "success",
                time.time(),
                round(time.time() - mainstart, 3),
                "The report was executed successfully",
                "-24h",
                "now",
            )

        else:
            trackme_register_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                "splk-%s" % (self.component),
                report_name,
                "failure",
                time.time(),
                round(time.time() - mainstart, 3),
                failures_entities_object_list,
                "-24h",
                "now",
            )


dispatch(SplkOutliersExecutor, sys.argv, sys.stdin, sys.stdout, __name__)
