#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import re

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_tracker_executor.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# splunk home
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# import trackme libs
from trackme_libs import (
    trackme_reqinfo,
    trackme_register_tenant_object_summary,
    trackme_return_tenant_object_summary,
    dataset_update_cache,
)

# import trackme licensing libs
from trackme_libs_licensing import trackme_check_license


@Configuration(distributed=False)
class TrackMeTrackerExecutor(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** The tracker component name to be executed.""",
        require=True,
        default=None,
        validate=validators.Match(
            "mode", r"^(?:splk-dsm|splk-dhm|splk-mhm|splk-flx|splk-wlk|splk-cim)$"
        ),
    )

    report = Option(
        doc="""
        **Syntax:** **report=****
        **Description:** The tracker report to be executed.""",
        require=True,
        default=None,
    )

    args = Option(
        doc="""
        **Syntax:** **args=****
        **Description:** optional arguments to the report.""",
        require=False,
        default=None,
    )

    earliest = Option(
        doc="""
        **Syntax:** **earliest=****
        **Description:** The earliest time quantifier.""",
        require=True,
        default=None,
    )

    latest = Option(
        doc="""
        **Syntax:** **latest=****
        **Description:** The latest time quantifier.""",
        require=True,
        default=None,
    )

    def generate(self, **kwargs):
        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5

        # performance counter
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

        # Get the session key
        session_key = self._metadata.searchinfo.session_key

        # component suffix
        subcomponent = self.component.split("splk-")[1]

        # check license state
        try:
            check_license = trackme_check_license(
                reqinfo["server_rest_uri"], session_key
            )
            license_is_valid = check_license.get("license_is_valid")
            logging.debug(
                'function check_license called, response="{}"'.format(
                    json.dumps(check_license, indent=2)
                )
            )

        except Exception as e:
            license_is_valid = 0
            logging.error('function check_license exception="{}"'.format(str(e)))

        # check restricted components
        if license_is_valid != 1 and self.component in (
            "splk-flx",
            "splk-wlk",
            "splk-cim",
        ):
            logging.error(
                "The requested component is restricted to the Full and Trial edition mode, its execution cannot be accepted"
            )

            # Call the component register
            trackme_register_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                self.component,
                self.report,
                "failure",
                time.time(),
                round(time.time() - start, 3),
                "The requested component is restricted to the Full and Trial edition mode, its execution cannot be accepted",
                self.earliest,
                self.latest,
            )

            raise Exception(
                "The requested component is restricted to the Full and Trial edition mode, its execution cannot be accepted"
            )

        # logging
        logging.info(
            'tenant_id="{}", component="{}", Starting tracker, report="{}"'.format(
                self.tenant_id, self.component, self.report
            )
        )

        # run update dataset cache first
        if subcomponent in ("dsm", "dhm", "mhm"):
            logging.debug(
                'tenant_id="{}", component="{}", calling dataset_update_cache'.format(
                    self.tenant_id, self.component
                )
            )
            try:
                cache_update = dataset_update_cache(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    subcomponent,
                )
                logging.debug(
                    'calling dataset_update_cache result="{}"'.format(cache_update)
                )
            except Exception as e:
                logging.error(
                    'failed to run the update dataset cache with exception="{}"'.format(
                        str(e)
                    )
                )

        # Define the query
        search = '| savedsearch "' + str(self.report) + '"'
        if self.args:
            search = search + " " + str(self.args)

        kwargs_oneshot = {
            "earliest_time": self.earliest,
            "latest_time": self.latest,
            "output_mode": "json",
            "count": 0,
        }

        # run the main report, every result is a Splunk search to be executed on its own thread

        # local event count, for the purposes of futher decision
        local_total_count = 0

        # automatic retry
        search_successful = False
        current_failures_count = 0
        while current_failures_count < max_failures_count and not search_successful:
            try:
                oneshotsearch_results = self.service.jobs.oneshot(
                    search, **kwargs_oneshot
                )
                reader = results.JSONResultsReader(oneshotsearch_results)

                # Call the component register
                trackme_register_tenant_object_summary(
                    session_key,
                    self._metadata.searchinfo.splunkd_uri,
                    self.tenant_id,
                    self.component,
                    self.report,
                    "success",
                    time.time(),
                    round(time.time() - start, 3),
                    "The report was executed successfully",
                    self.earliest,
                    self.latest,
                )

                for item in reader:
                    if isinstance(item, dict):
                        # increment
                        local_total_count += 1

                        logging.info(
                            'tenant_id="{}", component="{}", report="{}", earliest="{}", latest="{}", status="success", run_time="{}", results="{}"'.format(
                                self.tenant_id,
                                self.component,
                                self.report,
                                self.earliest,
                                self.latest,
                                round(time.time() - start, 3),
                                json.dumps(item, indent=0),
                            )
                        )
                        results_dict = {
                            "tenant_id": self.tenant_id,
                            "component": self.component,
                            "report": self.report,
                            "earliest": self.earliest,
                            "latest": self.latest,
                            "status": "success",
                            "run_time": round((time.time() - start), 3),
                            "results": item,
                        }
                        yield {"_time": time.time(), "_raw": results_dict}

                # break while
                search_successful = True
                break

            except Exception as e:
                if re.search(
                    "The maximum number of concurrent historical searches on this instance has been reached",
                    str(e),
                ):
                    current_failures_count += 1
                    search_successful = False
                    if current_failures_count >= max_failures_count:
                        # Call the component register
                        trackme_register_tenant_object_summary(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            self.component,
                            self.report,
                            "failure",
                            time.time(),
                            round(time.time() - start, 3),
                            str(e),
                            self.earliest,
                            self.latest,
                        )
                        msg = 'tenant_id="{}", component="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                            self.tenant_id,
                            self.component,
                            current_failures_count,
                            max_failures_count,
                            str(e),
                        )
                        logging.error(msg)
                        raise Exception(msg)
                    else:
                        msg = 'tenant_id="{}", component="{}", report="{}", earliest="{}", latest="{}", temporary search failure with exception="{}"'.format(
                            self.tenant_id,
                            self.component,
                            self.report,
                            self.earliest,
                            self.latest,
                            str(e),
                        )
                        logging.warn(msg)
                        time.sleep(sleep_time)
                else:
                    # Call the component register
                    trackme_register_tenant_object_summary(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        self.component,
                        self.report,
                        "failure",
                        time.time(),
                        round(time.time() - start, 3),
                        str(e),
                        self.earliest,
                        self.latest,
                    )
                    msg = 'tenant_id="{}", component="{}", report="{}", earliest="{}", latest="{}", main search failed with exception="{}"'.format(
                        self.tenant_id,
                        self.component,
                        self.report,
                        self.earliest,
                        self.latest,
                        str(e),
                    )
                    logging.error(msg)
                    raise Exception(msg)

        # Investigate the registered component summary
        # in some cases such as a remote search failure, we would not be able to detect it at the search exec time

        # set bool
        detected_failure_register = False

        try:
            last_status_exec = trackme_return_tenant_object_summary(
                session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
                self.component,
                self.report,
            )
            logging.debug(
                'function="trackme_return_tenant_object_summary", tenant_id="{}", component="{}", report="{}", result="{}"'.format(
                    self.tenant_id, self.component, self.report, last_status_exec
                )
            )

            if last_status_exec:
                if isinstance(last_status_exec, dict):
                    if (
                        last_status_exec.get("last_status") == "failure"
                        and local_total_count == 0
                    ):
                        logging.error(
                            'tenant_id="{}", object="{}", detected search execution failure from register component, result="{}"'.format(
                                self.tenant_id,
                                self.report,
                                json.dumps(last_status_exec, indent=2),
                            )
                        )
                        detected_failure_register = True

        except Exception as e:
            logging.error(
                'could not verify the register object summary, exception="{}"'.format(
                    str(e)
                )
            )

        if detected_failure_register:
            raise Exception(
                "TrackMe has detected a failure in the search execution, consult the component register status or the logs for more information"
            )


dispatch(TrackMeTrackerExecutor, sys.argv, sys.stdin, sys.stdout, __name__)
