#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import json
import re

# Logging imports
import logging
from logging.handlers import RotatingFileHandler

# Networking imports
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_tracker_health.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
from splunklib.searchcommands import (
    dispatch,
    GeneratingCommand,
    Configuration,
    Option,
    validators,
)
import splunklib.results as results

# import trackme libs
from trackme_libs import (
    trackme_reqinfo,
    trackme_register_tenant_object_summary,
    trackme_delete_tenant_object_summary,
    trackme_idx_for_tenant,
    trackme_state_event,
)

# import trackme licensing libs
from trackme_libs_licensing import trackme_check_license

# import trackme libs
from trackme_libs import trackme_report_update_enablement

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces

# import default vtenant account settings
from collections_data import vtenant_account_default


@Configuration(distributed=False)
class HealthTracker(GeneratingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    def generate(self, **kwargs):
        if self:
            # set max failed re-attempt
            max_failures_count = 24
            sleep_time = 5

            # performance counter
            start = time.time()

            # Get request info and set logging level
            reqinfo = trackme_reqinfo(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
            )
            log.setLevel(logging.getLevelName(reqinfo["logging_level"]))

            logging.info(
                f'tenant_id="{self.tenant_id}", context="general_execution", trackmetrackerhealth is starting now.'
            )

            # get the target index
            tenant_indexes = trackme_idx_for_tenant(
                self._metadata.searchinfo.session_key,
                self._metadata.searchinfo.splunkd_uri,
                self.tenant_id,
            )

            # get trackme release
            trackme_version = None
            app_confs = self.service.confs["app"]
            for stanza in app_confs:
                if stanza.name == "id":
                    for stanzakey, stanzavalue in stanza.content.items():
                        if stanzakey == "version":
                            trackme_version = stanzavalue
            logging.info(
                'context="check_version", running trackme version="{}"'.format(
                    trackme_version
                )
            )

            # set the schema_version_required
            schema_version_required = trackme_version.replace(".", "")

            # Get the session key
            session_key = self._metadata.searchinfo.session_key

            # Add the session_key to the reqinfo
            reqinfo["session_key"] = session_key

            # Splunk header
            header = "Splunk " + str(session_key)

            # report name for logging purposes
            report_name = "trackme_health_tracker_tenant_" + str(self.tenant_id)

            # Data collection
            collection_name = "kv_trackme_virtual_tenants"
            collection = self.service.kvstore[collection_name]

            # Get the tenant KVrecord
            query_string = {
                "tenant_id": self.tenant_id,
            }
            vtenant_record = collection.data.query(query=json.dumps(query_string))[0]

            # check license state
            try:
                check_license = trackme_check_license(
                    reqinfo["server_rest_uri"], session_key
                )
                license_is_valid = check_license.get("license_is_valid")
                license_subscription_class = check_license.get(
                    "license_subscription_class"
                )
                license_active_tenants = check_license.get("license_active_tenants")
                license_active_tenants_list = check_license.get(
                    "license_active_tenants_list"
                )
                logging.debug(
                    'context="check_licensing", function check_license called, response="{}"'.format(
                        json.dumps(check_license, indent=2)
                    )
                )

            except Exception as e:
                license_is_valid = 0
                logging.error(
                    'context="check_licensing", function check_license exception="{}"'.format(
                        str(e)
                    )
                )

            # Set the query
            health_search = remove_leading_spaces(
                f"""
            | trackme mode=post url=/services/trackme/v2/configuration/get_tenant_ops_status body=\"{{'mode': 'raw', 'tenant_id': '{self.tenant_id}'}}\"
            | trackmeopsstatusexpand
            """
            )

            # logging
            logging.info(
                'context="check_trackers", tenant_id="{}", Starting tracker, report="{}", search="{}"'.format(
                    self.tenant_id, report_name, health_search
                )
            )

            ##################################################################################
            # Global system verifications: verify that the relevant scheduled jobs are enabled
            ##################################################################################

            # These jobs are not tenant specifics, however we use the health tracker to ensure that
            # these are effectively enabled when at least one tenant has been created and is active

            savedsearch_names = [
                "trackme_ack_expiration_tracker",
                "trackme_maintenance_mode_tracker",
                "trackme_backup_scheduler",
            ]

            for savedsearch_name in savedsearch_names:
                # check ack expiration tracker
                update_properties_required = False

                try:
                    mysavedsearch = self.service.saved_searches[savedsearch_name]
                    current_disabled = int(mysavedsearch["disabled"])
                    logging.debug(
                        'context="check_global_trackers_enablement", global config check, verifying savedsearch="{}", disabled="{}"'.format(
                            mysavedsearch.name, current_disabled
                        )
                    )

                    if current_disabled == 1:
                        update_properties_required = True

                except Exception as e:
                    logging.error(
                        'context="check_global_trackers_enablement", global configuration verification, could not retrieve the status for {}'.format(
                            savedsearch_name
                        )
                    )

                if update_properties_required:
                    try:
                        action = trackme_report_update_enablement(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            savedsearch_name,
                            "enable",
                        )
                        logging.info(
                            'context="check_global_trackers_enablement", global config check, enabling savedsearch="{}", result="{}"'.format(
                                savedsearch_name, action
                            )
                        )

                    except Exception as e:
                        logging.error(
                            'context="check_global_trackers_enablement", global config check, an exception was encountered while trying to enable savedsearch="{}", exception="{}"'.format(
                                savedsearch_name, str(e)
                            )
                        )

            ##################################################################################
            # Replica orchestrator
            ##################################################################################

            # This job scheduled will automatically be enabled if we detect that at least one
            # replica tracker has been created

            # Try to get the current definition
            try:
                tenant_replica_objects = vtenant_record.get("tenant_replica_objects")

                # logging debug
                logging.debug(
                    'context="replica_orchestrator", tenant_replica_objects="{}"'.format(
                        tenant_replica_objects
                    )
                )
            except Exception as e:
                tenant_replica_objects = None

            # only run if we have a proper replica object
            if tenant_replica_objects:
                savedsearch_names = [
                    "trackme_replica_executor",
                ]

                for savedsearch_name in savedsearch_names:
                    # check
                    update_properties_required = False

                    try:
                        mysavedsearch = self.service.saved_searches[savedsearch_name]
                        current_disabled = int(mysavedsearch["disabled"])
                        logging.debug(
                            'context="replica_orchestrator", replica config check, verifying savedsearch="{}", disabled="{}"'.format(
                                mysavedsearch.name, current_disabled
                            )
                        )

                        if current_disabled == 1:
                            update_properties_required = True

                    except Exception as e:
                        logging.error(
                            'context="replica_orchestrator", replica configuration verification, could not retrieve the status for {}'.format(
                                savedsearch_name
                            )
                        )

                    if update_properties_required:
                        try:
                            action = trackme_report_update_enablement(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                savedsearch_name,
                                "enable",
                            )
                            logging.info(
                                'context="replica_orchestrator", replica config check, enabling savedsearch="{}", result="{}"'.format(
                                    savedsearch_name, action
                                )
                            )

                        except Exception as e:
                            logging.error(
                                'context="replica_orchestrator", replica config check, an exception was encountered while trying to enable savedsearch="{}", exception="{}"'.format(
                                    savedsearch_name, str(e)
                                )
                            )

            ###########################################################################
            # schema update and migration: detect and migrate Virtual Tenants if needed
            ###########################################################################

            from trackme_libs_schema import (
                trackme_schema_get_version,
                trackme_schema_update_version,
                trackme_schema_upgrade_209,
                trackme_schema_upgrade_2015,
                trackme_schema_upgrade_2016,
                trackme_schema_upgrade_2020,
                trackme_schema_upgrade_2026,
                trackme_schema_upgrade_2034,
                trackme_schema_upgrade_2034_least_privileges,
                trackme_schema_upgrade_2036,
                trackme_schema_upgrade_2038,
                trackme_schema_upgrade_2043,
                trackme_schema_upgrade_2044,
                trackme_schema_upgrade_2045,
                trackme_schema_upgrade_2054,
                trackme_schema_upgrade_2064,
                trackme_schema_upgrade_2067,
                trackme_schema_upgrade_2070,
                trackme_schema_upgrade_2071,
                trackme_schema_upgrade_2072,
                trackme_schema_upgrade_2075,
                trackme_schema_upgrade_2077,
                trackme_schema_upgrade_2078,
            )

            # Define a mapping between schema versions and their upgrade functions
            schema_upgrades = [
                (209, trackme_schema_upgrade_209),
                (2015, trackme_schema_upgrade_2015),
                (2016, trackme_schema_upgrade_2016),
                (2020, trackme_schema_upgrade_2020),
                (2026, trackme_schema_upgrade_2026),
                (2034, trackme_schema_upgrade_2034),
                (2034, trackme_schema_upgrade_2034_least_privileges),
                (2036, trackme_schema_upgrade_2036),
                (2038, trackme_schema_upgrade_2038),
                (2043, trackme_schema_upgrade_2043),
                (2043, trackme_schema_upgrade_2044),
                (2045, trackme_schema_upgrade_2045),
                (2054, trackme_schema_upgrade_2054),
                (2064, trackme_schema_upgrade_2064),
                (2067, trackme_schema_upgrade_2067),
                (2070, trackme_schema_upgrade_2070),
                (2071, trackme_schema_upgrade_2071),
                (2072, trackme_schema_upgrade_2072),
                (2075, trackme_schema_upgrade_2075),
                (2077, trackme_schema_upgrade_2077),
                (2078, trackme_schema_upgrade_2078),
            ]

            # Get the current schema version
            try:
                schema_version = trackme_schema_get_version(
                    reqinfo, self.tenant_id, schema_version_required
                )

            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{self.tenant_id}", failed to call function trackme_schema_get_version, exception="{str(e)}"'
                )

            # Check and act accordingly
            if not schema_version or int(schema_version) != int(
                schema_version_required
            ):
                for version, upgrade_func in schema_upgrades:
                    if not schema_version or int(schema_version) < version:
                        logging.info(
                            f'context="schema_upgrade", tenant_id="{self.tenant_id}", detected migration required for schema version {version}, processing now.'
                        )
                        try:
                            schema_version_update = upgrade_func(
                                reqinfo, self.tenant_id
                            )
                            logging.info(
                                f'context="schema_upgrade", tenant_id="{self.tenant_id}", schema version {version} migrated successfully.'
                            )
                        except Exception as e:
                            logging.error(
                                f'context="schema_upgrade", tenant_id="{self.tenant_id}", failed to call function {upgrade_func.__name__}, exception="{str(e)}"'
                            )

                #
                # finally migrate the schema version
                #

                try:
                    schema_version_update = trackme_schema_update_version(
                        reqinfo, self.tenant_id, schema_version_required
                    )
                except Exception as e:
                    logging.error(
                        'context="schema_upgrade", tenant_id="{}", failed to call function trackme_schema_update_version, exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                    )

            else:
                logging.info(
                    f'context="schema_upgrade", tenant_id="{self.tenant_id}", schema is up to date, no action required, schema_version="{schema_version}"'
                )

            #
            #
            #

            ###########################################################################
            # verify that the tenant account exists as expected
            ###########################################################################

            logging.info(
                'context="check_vtenant_accounts", tenant_id="{}", verifying the vtenant account'.format(
                    self.tenant_id
                )
            )

            # target
            url = "%s/services/trackme/v2/vtenants/vtenants_accounts" % (
                self._metadata.searchinfo.splunkd_uri
            )

            # init
            vtenant_account = None

            # proceed
            try:
                response = requests.post(
                    url,
                    headers={
                        "Authorization": header,
                        "Content-Type": "application/json",
                    },
                    data=json.dumps({"tenant_id": self.tenant_id}),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        'context="check_vtenant_accounts", verify vtenant account has failed, was this account deleted by mistake? response.status_code="{}", response.text="{}"'.format(
                            response.status_code, response.text
                        )
                    )
                else:
                    logging.info(
                        f'context="check_vtenant_accounts", tenant_id="{self.tenant_id}", account was verified successfully'
                    )
                    vtenant_account = response.json()
            except Exception as e:
                logging.error(
                    'context="check_vtenant_accounts", verify vtenant account has failed, exception="{}"'.format(
                        str(e)
                    )
                )

            # act if needed
            if not vtenant_account:
                url = "%s/servicesNS/nobody/trackme/trackme_vtenants" % (
                    self._metadata.searchinfo.splunkd_uri
                )

                # load default vtenant config
                data = dict(vtenant_account_default)

                # add the name value
                data["name"] = self.tenant_id

                # Retrieve and set the tenant idx, if any failure, logs and use the global index
                try:
                    response = requests.post(
                        url,
                        headers={
                            "Authorization": header,
                            "Content-Type": "application/json",
                        },
                        data=data,
                        verify=False,
                        timeout=300,
                    )
                    if response.status_code not in (200, 201, 204):
                        logging.error(
                            'context="check_vtenant_accounts", create vtenant account has failed, response.status_code="{}", response.text="{}"'.format(
                                response.status_code, response.text
                            )
                        )
                    else:
                        logging.info(
                            'context="check_vtenant_accounts", create vtenant account was operated successfully, response.status_code="{}"'.format(
                                response.status_code
                            )
                        )
                except Exception as e:
                    logging.error(
                        'context="check_vtenant_accounts", create vtenant account has failed, exception="{}"'.format(
                            str(e)
                        )
                    )

            #
            #
            #

            #
            # splk-dsm - untracked entities
            #
            # context: this activity tracks and maintain state for untracked entities
            # untracked entities are entities which are entirely out of the scope of any trackers, and therefore not maintained otherwise

            if vtenant_record.get("tenant_dsm_enabled") == True:
                logging.info(
                    f'tenant_id="{self.tenant_id}", component="splk-dsm", context="untracked_entities", inspecting untracked entities now.'
                )

                # kwargs
                kwargs_oneshot = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "output_mode": "json",
                    "count": 0,
                }

                untracked_entities_count = 0

                untracked_entities_search = f"""\
                    | inputlookup trackme_dsm_tenant_{self.tenant_id} | eval key=_key

                    ``` target any entity that has not been updated since more than 15m ```
                    | eval time_sec_since_inspection=now()-tracker_runtime
                    | where ( time_sec_since_inspection>900 OR isnull(tracker_runtime) )

                    ``` called the offline abstract macro version ```
                    `trackme_dsm_tracker_abstract({self.tenant_id})`

                    ``` collects latest collection state into the summary index ```
                    | `trackme_collect_state("current_state_tracking:splk-dsm:{self.tenant_id}", "object", "{self.tenant_id}")`

                    ``` output flipping change status if changes ```
                    | trackmesplkgetflipping tenant_id="{self.tenant_id}"

                    ``` update the KVstore collection ```
                    | `trackme_outputlookup(trackme_dsm_tenant_{self.tenant_id}, key)`

                    ``` update the delay metric only ```
                    | `trackme_mcollect(object, splk-dsm, "metric_name:trackme.splk.feeds.lag_event_sec=data_last_lag_seen", "tenant_id, object_category, object", "{self.tenant_id}")`

                    ``` summarize job ```
                    | stats count as report_entities_count, values(object) as objects by tenant_id
                """

                # run the main report, every result is a Splunk search to be executed on its own thread

                # automatic retry
                search_successful = False
                current_failures_count = 0
                while (
                    current_failures_count < max_failures_count
                    and not search_successful
                ):
                    try:
                        oneshotsearch_results = self.service.jobs.export(
                            remove_leading_spaces(untracked_entities_search),
                            **kwargs_oneshot,
                        )
                        reader = results.JSONResultsReader(oneshotsearch_results)

                        for item in reader:
                            if isinstance(item, dict):
                                untracked_entities_count += 1
                                logging.info(
                                    f'tenant_id="{self.tenant_id}", component="splk-dsm", context="untracked_entities", results="{item}"'
                                )

                        if untracked_entities_count == 0:
                            logging.info(
                                f'tenant_id="{self.tenant_id}", component="splk-dsm", context="untracked_entities", there are no untracked entities currently.'
                            )

                        # break while
                        search_successful = True
                        break

                    except Exception as e:
                        if re.search(
                            "The maximum number of concurrent historical searches on this instance has been reached",
                            str(e),
                        ):
                            current_failures_count += 1
                            search_successful = False
                            if current_failures_count >= max_failures_count:
                                # Call the component register
                                trackme_register_tenant_object_summary(
                                    session_key,
                                    self._metadata.searchinfo.splunkd_uri,
                                    self.tenant_id,
                                    "all",
                                    report_name,
                                    "failure",
                                    time.time(),
                                    str(time.time() - start),
                                    str(e),
                                    "-5m",
                                    "now",
                                )
                                msg = 'tenant_id="{}", component="splk-dsm", context="untracked_entities", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                    self.tenant_id,
                                    current_failures_count,
                                    max_failures_count,
                                    str(e),
                                )
                                logging.error(msg)
                                raise Exception(msg)
                            else:
                                msg = 'tenant_id="{}", component="splk-dsm", context="untracked_entities", temporary search failure with exception="{}"'.format(
                                    self.tenant_id, str(e)
                                )
                                logging.warn(msg)
                                time.sleep(sleep_time)
                        else:
                            # Call the component register
                            trackme_register_tenant_object_summary(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                "all",
                                report_name,
                                "failure",
                                time.time(),
                                str(time.time() - start),
                                str(e),
                                "-5m",
                                "now",
                            )
                            msg = 'context="untracked_entities", tenant_id="{}", main search failed with exception="{}"'.format(
                                self.tenant_id, str(e)
                            )
                            logging.error(msg)
                            raise Exception(msg)

            #
            # splk-dhm - untracked entities
            #
            # context: this activity tracks and maintain state for untracked entities
            # untracked entities are entities which are entirely out of the scope of any trackers, and therefore not maintained otherwise

            if vtenant_record.get("tenant_dhm_enabled") == True:
                logging.info(
                    f'tenant_id="{self.tenant_id}", component="splk-dhm", context="untracked_entities", inspecting untracked entities now.'
                )

                # kwargs
                kwargs_oneshot = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "output_mode": "json",
                    "count": 0,
                }

                untracked_entities_count = 0

                untracked_entities_search = f"""\
                    | inputlookup trackme_dhm_tenant_{self.tenant_id} | eval key=_key

                    ``` target any entity that has not been updated since more than 15m ```
                    | eval time_sec_since_inspection=now()-tracker_runtime
                    | where ( time_sec_since_inspection>900 OR isnull(tracker_runtime) )

                    ``` called the offline abstract macro version ```
                    `trackme_dhm_tracker_abstract({self.tenant_id})`

                    ``` collects latest collection state into the summary index ```
                    | `trackme_collect_state("current_state_tracking:splk-dhm:{self.tenant_id}", "object", "{self.tenant_id}")`

                    ``` output flipping change status if changes ```
                    | trackmesplkgetflipping tenant_id="{self.tenant_id}"

                    ``` update the KVstore collection ```
                    | `trackme_outputlookup(trackme_dhm_tenant_{self.tenant_id}, key)`

                    ``` update the delay metric only ```
                    | `trackme_mcollect(object, splk-dhm, "metric_name:trackme.splk.feeds.lag_event_sec=data_last_lag_seen", "tenant_id, object_category, object", "{self.tenant_id}")`

                    ``` summarize job ```
                    | stats count as report_entities_count, values(object) as objects by tenant_id
                """

                # run the main report, every result is a Splunk search to be executed on its own thread

                # automatic retry
                search_successful = False
                current_failures_count = 0
                while (
                    current_failures_count < max_failures_count
                    and not search_successful
                ):
                    try:
                        oneshotsearch_results = self.service.jobs.export(
                            remove_leading_spaces(untracked_entities_search),
                            **kwargs_oneshot,
                        )
                        reader = results.JSONResultsReader(oneshotsearch_results)

                        for item in reader:
                            if isinstance(item, dict):
                                untracked_entities_count += 1
                                logging.info(
                                    f'tenant_id="{self.tenant_id}", component="splk-dhm", context="untracked_entities", results="{item}"'
                                )

                        if untracked_entities_count == 0:
                            logging.info(
                                f'tenant_id="{self.tenant_id}", component="splk-dhm", context="untracked_entities", there are no untracked entities currently.'
                            )

                        # break while
                        search_successful = True
                        break

                    except Exception as e:
                        if re.search(
                            "The maximum number of concurrent historical searches on this instance has been reached",
                            str(e),
                        ):
                            current_failures_count += 1
                            search_successful = False
                            if current_failures_count >= max_failures_count:
                                # Call the component register
                                trackme_register_tenant_object_summary(
                                    session_key,
                                    self._metadata.searchinfo.splunkd_uri,
                                    self.tenant_id,
                                    "all",
                                    report_name,
                                    "failure",
                                    time.time(),
                                    str(time.time() - start),
                                    str(e),
                                    "-5m",
                                    "now",
                                )
                                msg = 'tenant_id="{}", component="splk-dhm", context="untracked_entities", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                    self.tenant_id,
                                    current_failures_count,
                                    max_failures_count,
                                    str(e),
                                )
                                logging.error(msg)
                                raise Exception(msg)
                            else:
                                msg = 'tenant_id="{}", component="splk-dhm", context="untracked_entities", temporary search failure with exception="{}"'.format(
                                    self.tenant_id, str(e)
                                )
                                logging.warn(msg)
                                time.sleep(sleep_time)
                        else:
                            # Call the component register
                            trackme_register_tenant_object_summary(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                "all",
                                report_name,
                                "failure",
                                time.time(),
                                str(time.time() - start),
                                str(e),
                                "-5m",
                                "now",
                            )
                            msg = 'context="untracked_entities", tenant_id="{}", main search failed with exception="{}"'.format(
                                self.tenant_id, str(e)
                            )
                            logging.error(msg)
                            raise Exception(msg)

            #
            # splk-wlk - duplicated entities
            #
            # context: this activity tracks for duplicated entities in the Workload component
            # under some rare circumstances, the Splunk scheduler logs may lack the user context, althrough we implement several safeties
            # if this happens, we need to verify and purge any duplicated entity with the system user context instead of the proper user context

            if (
                vtenant_record.get("tenant_wlk_enabled") == True
                and vtenant_record.get("tenant_replica") == False
            ):
                logging.info(
                    f'tenant_id="{self.tenant_id}", component="splk-wlk", context="duplicated_entities", inspecting entities now.'
                )

                # kwargs
                kwargs_oneshot = {
                    "earliest_time": "-5m",
                    "latest_time": "now",
                    "output_mode": "json",
                    "count": 0,
                }

                duplicated_entities_count = 0
                duplicated_entities_list = []

                duplicated_entities_search = remove_leading_spaces(
                    f"""\
                    | inputlookup trackme_wlk_tenant_{self.tenant_id} | eval keyid=_key
                    | fields keyid, account, app, user, savedsearch_name, object, last_seen
                    | eventstats count as dcount by account, app, savedsearch_name
                    | where dcount>1
                    | sort - 0 savedsearch_name, last_seen
                    | dedup savedsearch_name
                """
                )

                # run the main report, every result is a Splunk search to be executed on its own thread

                # automatic retry
                search_successful = False
                current_failures_count = 0
                while (
                    current_failures_count < max_failures_count
                    and not search_successful
                ):
                    try:
                        oneshotsearch_results = self.service.jobs.export(
                            remove_leading_spaces(duplicated_entities_search),
                            **kwargs_oneshot,
                        )
                        reader = results.JSONResultsReader(oneshotsearch_results)

                        for item in reader:
                            if isinstance(item, dict):
                                duplicated_entities_count += 1
                                duplicated_entities_list.append(item.get("keyid"))
                                logging.info(
                                    f'tenant_id="{self.tenant_id}", component="splk-wlk", context="duplicated_entities", results="{item}"'
                                )

                        if duplicated_entities_count == 0:
                            logging.info(
                                f'tenant_id="{self.tenant_id}", component="splk-wlk", context="duplicated_entities", there are no duplicated entities currently.'
                            )

                        # break while
                        search_successful = True
                        break

                    except Exception as e:
                        if re.search(
                            "The maximum number of concurrent historical searches on this instance has been reached",
                            str(e),
                        ):
                            current_failures_count += 1
                            search_successful = False
                            if current_failures_count >= max_failures_count:
                                # Call the component register
                                trackme_register_tenant_object_summary(
                                    session_key,
                                    self._metadata.searchinfo.splunkd_uri,
                                    self.tenant_id,
                                    "all",
                                    report_name,
                                    "failure",
                                    time.time(),
                                    str(time.time() - start),
                                    str(e),
                                    "-5m",
                                    "now",
                                )
                                msg = 'tenant_id="{}", component="splk-wlk", context="duplicated_entities", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                    self.tenant_id,
                                    current_failures_count,
                                    max_failures_count,
                                    str(e),
                                )
                                logging.error(msg)
                                raise Exception(msg)
                            else:
                                msg = 'tenant_id="{}", component="splk-wlk", context="duplicated_entities", temporary search failure with exception="{}"'.format(
                                    self.tenant_id, str(e)
                                )
                                logging.warn(msg)
                                time.sleep(sleep_time)
                        else:
                            # Call the component register
                            trackme_register_tenant_object_summary(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                "all",
                                report_name,
                                "failure",
                                time.time(),
                                str(time.time() - start),
                                str(e),
                                "-5m",
                                "now",
                            )
                            msg = 'context="duplicated_entities", tenant_id="{}", main search failed with exception="{}"'.format(
                                self.tenant_id, str(e)
                            )
                            logging.error(msg)
                            raise Exception(msg)

                # process if needed
                if duplicated_entities_count > 0:
                    # target
                    target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/splk_wlk/write/wlk_delete"

                    # data
                    # turn duplicated_entities_list into a comma separated string
                    duplicated_entities_list = ",".join(duplicated_entities_list)
                    post_data = {
                        "tenant_id": self.tenant_id,
                        "keys_list": duplicated_entities_list,
                        "deletion_type": "temporary",
                        "update_comment": "One or more duplicated entities were detected by the health tracker, this condition can happen when Splunk scheduler logs lack the user context, automated purge of these entities.",
                    }

                    try:
                        response = requests.post(
                            target_url,
                            headers={
                                "Authorization": header,
                                "Content-Type": "application/json",
                            },
                            data=json.dumps(post_data),
                            verify=False,
                            timeout=300,
                        )
                        msg = 'context="duplicated_entities", tenant_id="{}", duplicated entities purge successful, results="{}"'.format(
                            self.tenant_id, json.dumps(response.text, indent=2)
                        )
                        logging.info(msg)

                    except Exception as e:
                        msg = 'context="duplicated_entities", tenant_id="{}", duplicated entities purge failed with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                        logging.info(msg)

            #
            # Verify and synchronize trackers attribute knowledge
            #

            def process_sync_tracker_component(component):
                logging.info(
                    f'tenant_id="{self.tenant_id}", component="splk-{component}", context="synchronize_trackers_attr", verify and synchronize trackers knowledge attributes.'
                )

                # get macros
                macros = self.service.confs["macros"]

                # Data collection
                tracker_collection_name = (
                    f"kv_trackme_{component}_hybrid_trackers_tenant_{self.tenant_id}"
                )
                tracker_collection = self.service.kvstore[tracker_collection_name]

                # Get the tenant KVrecord
                tracker_records = tracker_collection.data.query()

                for tracker_record in tracker_records:
                    record_knowledge_objects = json.loads(
                        tracker_record.get("knowledge_objects")
                    )
                    record_key = tracker_record.get("_key")

                    logging.debug(
                        f'context="synchronize_trackers_attr", record_knowledge_objects={record_knowledge_objects}'
                    )

                    tracker_root_constraint_macro = record_knowledge_objects[
                        "properties"
                    ][0].get("root_constraint_macro")
                    tracker_root_constraint = record_knowledge_objects["properties"][
                        0
                    ].get("root_constraint")

                    try:
                        macro_definition = macros[
                            tracker_root_constraint_macro
                        ].content["definition"]
                    except Exception as e:
                        logging.info(
                            f'tenant_id="{self.tenant_id}", component="splk-{component}", context="synchronize_trackers_attr", failed to retrieve the definition for the macro={tracker_root_constraint_macro}, exception={str(e)}'
                        )
                        macro_definition = None

                    if macro_definition:
                        if macro_definition != tracker_root_constraint:
                            logging.info(
                                f'tenant_id="{self.tenant_id}", component="splk-{component}", context="synchronize_trackers_attr", updating macro with new definition={macro_definition}'
                            )
                            # update the dict
                            record_knowledge_objects["properties"][0][
                                "root_constraint"
                            ] = macro_definition
                            # update the record
                            tracker_record["knowledge_objects"] = json.dumps(
                                record_knowledge_objects, indent=2
                            )

                            try:
                                tracker_collection.data.update(
                                    str(record_key), json.dumps(tracker_record)
                                )
                                logging.info(
                                    f'tenant_id="{self.tenant_id}", component="splk-{component}", context="synchronize_trackers_attr", tracker record was successfully updated.'
                                )
                            except Exception as e:
                                logging.error(
                                    f'tenant_id="{self.tenant_id}", component="splk-{component}", context="synchronize_trackers_attr", tracker record failed to be updated, exception={str(e)}'
                                )

            # Main logic
            components = ["dsm", "dhm", "mhm"]
            for component in components:
                if vtenant_record.get("tenant_" + component + "_enabled") == True:
                    process_sync_tracker_component(component)

            #
            #
            #

            # kwargs
            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # run the main report, every result is a Splunk search to be executed on its own thread

            # automatic retry
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                try:
                    oneshotsearch_results = self.service.jobs.oneshot(
                        health_search, **kwargs_oneshot
                    )
                    reader = results.JSONResultsReader(oneshotsearch_results)

                    # Call the component register
                    trackme_register_tenant_object_summary(
                        session_key,
                        self._metadata.searchinfo.splunkd_uri,
                        self.tenant_id,
                        "all",
                        report_name,
                        "success",
                        time.time(),
                        str(time.time() - start),
                        "The report was executed successfully",
                        "-5m",
                        "now",
                    )

                    for item in reader:
                        if isinstance(item, dict):
                            # verify the knowledge object - if for some reason it is not existing anymore, we should remove it
                            # and not take it into account any longer

                            # process
                            savedsearch_definition = None
                            report_name = item.get("report")
                            try:
                                savedsearch = self.service.saved_searches[report_name]
                                savedsearch_definition = savedsearch.content["search"]
                            except Exception as e:
                                savedsearch_definition = None
                                logging.warning(
                                    f'context="check_trackers", the report="{report_name}" does not exist anymore, somehow it was removed without TrackMe being aware of it, will get rid of this now.'
                                )

                            if not savedsearch_definition:
                                # extract component
                                component = report_name.split("_")[1]

                                # purge
                                try:
                                    delete_register_summary = (
                                        trackme_delete_tenant_object_summary(
                                            self._metadata.searchinfo.session_key,
                                            self._metadata.searchinfo.splunkd_uri,
                                            self.tenant_id,
                                            f"splk-{component}",
                                            report_name,
                                        )
                                    )
                                    logging.info(
                                        f'context="check_trackers", knowledge for the report="{report_name}" was purged successfully, response="{delete_register_summary}"'
                                    )
                                except Exception as e:
                                    logging.error(
                                        f'context="check_trackers", exception encountered while calling function trackme_delete_tenant_object_summary, exception="{str(e)}"'
                                    )

                            else:
                                yield {
                                    "_time": time.time(),
                                    "_raw": item,
                                    "component": item.get("component"),
                                    "earliest": item.get("earliest"),
                                    "last_duration": item.get("last_duration"),
                                    "last_exec": item.get("last_exec"),
                                    "last_result": item.get("last_result"),
                                    "last_status": item.get("last_status"),
                                    "latest": item.get("latest"),
                                    "report": report_name,
                                    "tenant_id": item.get("tenant_id"),
                                }

                                # index the audit record
                                try:
                                    trackme_state_event(
                                        session_key=self._metadata.searchinfo.session_key,
                                        splunkd_uri=self._metadata.searchinfo.splunkd_uri,
                                        tenant_id=self.tenant_id,
                                        index=tenant_indexes["trackme_audit_idx"],
                                        sourcetype="trackme:health",
                                        source=f"trackme:health:{self.tenant_id}",
                                        record={
                                            "component": item.get("component"),
                                            "earliest": item.get("earliest"),
                                            "last_duration": item.get("last_duration"),
                                            "last_exec": item.get("last_exec"),
                                            "last_result": item.get("last_result"),
                                            "last_status": item.get("last_status"),
                                            "latest": item.get("latest"),
                                            "report": report_name,
                                            "tenant_id": item.get("tenant_id"),
                                        },
                                    )
                                except Exception as e:
                                    error_msg = f'context="check_trackers", exception encountered while calling function trackme_state_event, exception="{str(e)}"'
                                    logging.error(error_msg)
                                    raise Exception(error_msg)

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            # Call the component register
                            trackme_register_tenant_object_summary(
                                session_key,
                                self._metadata.searchinfo.splunkd_uri,
                                self.tenant_id,
                                "all",
                                report_name,
                                "failure",
                                time.time(),
                                str(time.time() - start),
                                str(e),
                                "-5m",
                                "now",
                            )
                            msg = 'context="check_trackers", tenant_id="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                self.tenant_id,
                                current_failures_count,
                                max_failures_count,
                                str(e),
                            )
                            logging.error(msg)
                            raise Exception(msg)
                        else:
                            msg = 'context="check_trackers", tenant_id="{}", temporary search failure with exception="{}"'.format(
                                self.tenant_id, str(e)
                            )
                            logging.warn(msg)
                            time.sleep(sleep_time)
                    else:
                        # Call the component register
                        trackme_register_tenant_object_summary(
                            session_key,
                            self._metadata.searchinfo.splunkd_uri,
                            self.tenant_id,
                            "all",
                            report_name,
                            "failure",
                            time.time(),
                            str(time.time() - start),
                            str(e),
                            "-5m",
                            "now",
                        )
                        msg = 'context="check_trackers", tenant_id="{}", main search failed with exception="{}"'.format(
                            self.tenant_id, str(e)
                        )
                        logging.error(msg)
                        raise Exception(msg)

            #
            # licensing restriction
            #

            # if the component is a restricted component and the product is not registered, it should be disabled now
            if license_is_valid != 1 and (
                vtenant_record.get("tenant_cim_enabled") == 1
                or vtenant_record.get("tenant_flx_enabled") == 1
                or vtenant_record.get("tenant_wlk_enabled") == 1
            ):
                logging.info(
                    'context="check_licensing", tenant_id="{}", due to licensing restrictions, this tenant will be automatically disabled, the tenant is running a restricted component while this instance is not registered'.format(
                        self.tenant_id
                    )
                )

                # target
                target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

                # data
                post_data = {
                    "tenant_id": self.tenant_id,
                    "update_comment": "Auto disabling this tenant due to licensing limitation, the tenant is running a restricted component while the product is not currently registered",
                    "force": "true",
                }

                try:
                    response = requests.post(
                        target_url,
                        headers={
                            "Authorization": header,
                            "Content-Type": "application/json",
                        },
                        data=json.dumps(post_data),
                        verify=False,
                        timeout=300,
                    )
                    return json.loads(response.text)

                except Exception as e:
                    raise Exception(
                        'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{}"'.format(
                            str(e)
                        )
                    )

            elif (
                license_is_valid != 1
                and license_active_tenants > 2
                and self.tenant_id not in license_active_tenants_list[0:2]
            ):
                logging.info(
                    'context="check_licensing", tenant_id="{}", due to licensing restrictions, this tenant will be automatically disabled, the tenant is running a restricted component while this instance is not registered'.format(
                        self.tenant_id
                    )
                )

                # target
                target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

                # data
                post_data = {
                    "tenant_id": self.tenant_id,
                    "update_comment": "Auto disabling this tenant due to licensing limitation, this deployment has reached the maximum number of tenants allowed ({}), only the following tenants can be used: {}".format(
                        license_active_tenants, license_active_tenants_list[0:2]
                    ),
                    "force": "true",
                }

                try:
                    response = requests.post(
                        target_url,
                        headers={
                            "Authorization": header,
                            "Content-Type": "application/json",
                        },
                        data=json.dumps(post_data),
                        verify=False,
                        timeout=300,
                    )
                    return json.loads(response.text)

                except Exception as e:
                    raise Exception(
                        'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{}"'.format(
                            str(e)
                        )
                    )

            elif (
                license_is_valid == 1
                and license_subscription_class == "enterprise"
                and license_active_tenants > 6
                and self.tenant_id not in license_active_tenants_list[0:6]
            ):
                logging.info(
                    'context="check_licensing", tenant_id="{}", due to licensing restrictions, this tenant will be automatically disabled, the tenant is over the maximum number of allowed tenants in Enterprise Edition'.format(
                        self.tenant_id
                    )
                )

                # target
                target_url = f"{self._metadata.searchinfo.splunkd_uri}/services/trackme/v2/vtenants/admin/disable_tenant"

                # data
                post_data = {
                    "tenant_id": self.tenant_id,
                    "update_comment": "Auto disabling this tenant due to licensing limitation, this deployment has reached the maximum number of tenants allowed ({}), only the following tenants can be used: {}".format(
                        license_active_tenants, license_active_tenants_list[0:6]
                    ),
                    "force": "true",
                }

                try:
                    response = requests.post(
                        target_url,
                        headers={
                            "Authorization": header,
                            "Content-Type": "application/json",
                        },
                        data=json.dumps(post_data),
                        verify=False,
                        timeout=300,
                    )
                    return json.loads(response.text)

                except Exception as e:
                    raise Exception(
                        'An exception was encountered while attempting to disable the tenant due to licensing restrictions, exception="{}"'.format(
                            str(e)
                        )
                    )

        else:
            # Raise
            raise Exception("Error: bad request")

        logging.info(
            f'tenant_id="{self.tenant_id}", context="general_execution", trackmetrackerhealth has terminated, run_time={round(time.time() - start, 3)}'
        )


dispatch(HealthTracker, sys.argv, sys.stdin, sys.stdout, __name__)
