#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import re
import json
import random
import time
import logging
from logging.handlers import RotatingFileHandler

# Networking and URL handling imports
import requests
from requests.structures import CaseInsensitiveDict
from urllib.parse import urlencode
import urllib.parse
import urllib3

# Disable insecure request warnings for urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# appebd lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
import splunklib.client as client
import splunklib.results as results

# import trackme libs
from trackme_libs_licensing import trackme_check_license

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


# cd context manager
class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


class JSONFormatter(logging.Formatter):
    def format(self, record):
        log_record = {
            "time": self.formatTime(record, self.datefmt),
        }

        # Unwanted attributes specific to the logging module
        unwanted_attrs = set(
            [
                "name",
                "msg",
                "args",
                "levelname",
                "levelno",
                "pathname",
                "filename",
                "module",
                "exc_info",
                "exc_text",
                "stack_info",
                "lineno",
                "funcName",
                "created",
                "msecs",
                "relativeCreated",
                "thread",
                "threadName",
                "processName",
                "process",
            ]
        )

        # Add any extra attributes present in the record
        for key, value in record.__dict__.items():
            if (
                key not in log_record
                and not key.startswith("_")
                and key not in unwanted_attrs
            ):
                log_record[key] = value

        return json.dumps(log_record)


def trackme_reqinfo(session_key, splunkd_uri):
    """
    Retrieve request info & settings.
    """

    # Ensure splunkd_uri starts with "https://"
    if not splunkd_uri.startswith("https://"):
        splunkd_uri = f"https://{splunkd_uri}"

    # Build header and target URL
    headers = CaseInsensitiveDict()
    headers["Authorization"] = f"Splunk {session_key}"
    headers["Content-Type"] = "application/json"
    target_url = f"{splunkd_uri}/services/trackme/v2/configuration/request_info"

    # Create a requests session for better performance
    session = requests.Session()
    session.headers.update(headers)

    try:
        # Use a context manager to handle the request
        with session.get(target_url, verify=False) as response:
            if response.ok:
                logging.debug(f'Success retrieving conf, data="{response}"')
                response_json = response.json()
                return response_json
            else:
                error_message = f'Failed to retrieve conf, status_code={response.status_code}, response_text="{response.text}"'
                logging.error(error_message)
                raise Exception(error_message)

    except Exception as e:
        error_message = f'Failed to retrieve conf, exception="{str(e)}"'
        logging.error(error_message)
        raise Exception(error_message)


def trackme_getloglevel(system_authtoken, splunkd_port):
    """
    Simply get and return the loglevel with elevated privileges to avoid code duplication
    """

    # Get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=splunkd_port,
        token=system_authtoken,
        timeout=300,
    )

    # set loglevel
    loglevel = "INFO"
    conf_file = "trackme_settings"
    confs = service.confs[str(conf_file)]
    for stanza in confs:
        if stanza.name == "logging":
            for stanzakey, stanzavalue in stanza.content.items():
                if stanzakey == "loglevel":
                    loglevel = stanzavalue

    return loglevel


def trackme_vtenant_account(session_key, splunkd_uri, tenant_id):
    """
    Retrieve vtenant specific settings.
    """

    # Ensure splunkd_uri starts with "https://"
    if not splunkd_uri.startswith("https://"):
        splunkd_uri = f"https://{splunkd_uri}"

    # Build header and target URL
    headers = CaseInsensitiveDict()
    headers["Authorization"] = f"Splunk {session_key}"
    headers["Content-Type"] = "application/json"
    target_url = f"{splunkd_uri}/services/trackme/v2/vtenants/vtenants_accounts"

    # Create a requests session for better performance
    session = requests.Session()
    session.headers.update(headers)

    try:
        # Use a context manager to handle the request
        with session.post(
            target_url, data=json.dumps({"tenant_id": tenant_id}), verify=False
        ) as response:
            if response.ok:
                logging.debug(f'Success retrieving conf, data="{response}"')
                response_json = response.json()
                return response_json
            else:
                error_message = f'Failed to retrieve conf, status_code={response.status_code}, response_text="{response.text}"'
                logging.error(error_message)
                raise Exception(error_message)

    except Exception as e:
        error_message = f'Failed to retrieve conf, exception="{str(e)}"'
        logging.error(error_message)
        raise Exception(error_message)


def trackme_idx_for_tenant(session_key, splunkd_uri, tenant_id):
    """
    Retrieve request info & settings.
    """

    # Ensure splunkd_uri starts with "https://"
    if not splunkd_uri.startswith("https://"):
        splunkd_uri = f"https://{splunkd_uri}"

    # Build header and target URL
    headers = CaseInsensitiveDict()
    headers["Authorization"] = f"Splunk {session_key}"
    headers["Content-Type"] = "application/json"
    target_url = f"{splunkd_uri}/services/trackme/v2/vtenants/tenant_idx_settings"

    # Create a requests session for better performance
    session = requests.Session()
    session.headers.update(headers)

    try:
        # Use a context manager to handle the request
        with session.post(
            target_url, data=json.dumps({"tenant_id": tenant_id}), verify=False
        ) as response:
            if response.ok:
                logging.debug(f'Success retrieving conf, data="{response}"')
                response_json = response.json()
                return response_json
            else:
                error_message = f'Failed to retrieve conf, status_code={response.status_code}, response_text="{response.text}"'
                logging.error(error_message)
                raise Exception(error_message)

    except Exception as e:
        error_message = f'Failed to retrieve conf, exception="{str(e)}"'
        logging.error(error_message)
        raise Exception(error_message)


def trackme_gen_state(index, source, sourcetype, event):
    try:
        filehandler = RotatingFileHandler(
            "%s/var/log/splunk/trackme_state_events.log" % splunkhome,
            mode="a",
            maxBytes=100000000,
            backupCount=1,
        )
        formatter = JSONFormatter()
        logging.Formatter.converter = time.gmtime
        filehandler.setFormatter(formatter)
        log = logging.getLogger()  # root logger
        orig_filehandler = None
        for hdlr in log.handlers[:]:  # remove the existing file handlers
            if isinstance(hdlr, logging.FileHandler):
                orig_filehandler = hdlr
                log.removeHandler(hdlr)
        log.addHandler(filehandler)  # set the new handler

        logging.info(
            "TrackMe State Events",
            extra={
                "target_index": index,
                "target_sourcetype": sourcetype,
                "target_source": source,
                "event": event,
            },
        )

        if orig_filehandler:
            log.removeHandler(filehandler)
            log.addHandler(orig_filehandler)

    except Exception as e:
        raise Exception(str(e))


#
# remote account connectivity
#


def is_reachable(session, url):
    try:
        response = session.get(url, timeout=10, verify=False)
        return True if response.status_code == 200 else False
    except requests.exceptions.RequestException:
        return False


def select_url(session, splunk_url):
    # Transform the splunk_url from a comma-separated string to a list
    splunk_urls = splunk_url.split(",")

    # Filter the list of URLs to keep only the reachable ones
    reachable_urls = [url for url in splunk_urls if is_reachable(session, url)]

    # Randomly choose a reachable URL or return False if no valid URLs are found
    selected_url = random.choice(reachable_urls) if reachable_urls else False

    return selected_url


def get_bearer_token(storage_passwords, account):
    # realm
    credential_realm = "__REST_CREDENTIAL__#trackme#configs/conf-trackme_account"
    credential_name = f"{credential_realm}:{account}``"

    # extract as raw json
    bearer_token_rawvalue = ""

    for credential in storage_passwords:
        if credential.content.get("realm") == str(
            credential_realm
        ) and credential.name.startswith(credential_name):
            bearer_token_rawvalue = bearer_token_rawvalue + str(
                credential.content.clear_password
            )

    # extract a clean json object
    bearer_token_rawvalue_match = re.search(
        '\{"bearer_token":\s*"(.*)"\}', bearer_token_rawvalue
    )
    if bearer_token_rawvalue_match:
        bearer_token = bearer_token_rawvalue_match.group(1)
    else:
        bearer_token = None

    return bearer_token


def establish_sdk_remote_service(parsed_url, bearer_token, app_namespace, account):
    try:
        service = client.connect(
            host=parsed_url.hostname,
            splunkToken=str(bearer_token),
            owner="nobody",
            app=app_namespace,
            port=parsed_url.port,
            autologin=True,
            timeout=300,
        )

        remote_apps = [app.label for app in service.apps]
        if remote_apps:
            logging.info(
                f'Remote search connectivity check to host="{parsed_url.hostname}" on port="{parsed_url.port}" was successful'
            )
            return service

    except Exception as e:
        error_msg = f'Remote search for account="{account}" has failed at connectivity check, host="{parsed_url.hostname}" on port="{parsed_url.port}" with exception="{str(e)}"'
        raise Exception(error_msg)

    return None


# Test remote account connectivity, for a least privileges approach, this function uses a system_authtoken
def trackme_test_remote_account(reqinfo, account):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo.server_rest_port,
        token=reqinfo.system_authtoken,
        timeout=300,
    )

    # Splunk credentials store
    storage_passwords = service.storage_passwords

    # get all acounts
    accounts = []
    conf_file = "trackme_account"

    # if there are no account, raise an exception, otherwise what we would do here?
    try:
        confs = service.confs[str(conf_file)]
    except Exception as e:
        error_msg = (
            "splunkremotesearch was called but we have no remote account configured yet"
        )
        raise Exception(error_msg)

    for stanza in confs:
        # get all accounts
        for name in stanza.name:
            accounts.append(stanza.name)
            break

    # account configuration
    isfound = False
    splunk_url = None
    app_namespace = None

    # get account
    for stanza in confs:
        if stanza.name == str(account):
            isfound = True
            for key, value in stanza.content.items():
                if key == "splunk_url":
                    splunk_url = value
                if key == "app_namespace":
                    app_namespace = value

    # Create a session within the generate function
    session = requests.Session()

    # Call target selector and pass the session as an argument
    selected_url = select_url(session, splunk_url)

    # end of get configuration

    # Stop here if we cannot find the submitted account
    if not isfound:
        error_msg = 'The account="{}" has not been configured on this instance, cannot proceed!'.format(
            account
        )
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "account": account,
            }
        )

    # If none of the endpoints could be reached
    elif not selected_url:
        error_msg = "None of the endpoints provided in the account URLs could be reached successfully, verify your network connectivity!"
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "account": account,
                "targets": splunk_url,
            }
        )

    # check the license
    try:
        check_license = trackme_check_license(
            reqinfo.server_rest_uri, reqinfo.session_key
        )
        license_is_valid = check_license.get("license_is_valid")
        logging.debug(
            'function check_license called, response="{}"'.format(
                json.dumps(check_license, indent=2)
            )
        )

    except Exception as e:
        license_is_valid = 0
        logging.error('function check_license exception="{}"'.format(str(e)))

    # try and return
    if license_is_valid != 1 and len(accounts) >= 2 and accounts[0] != account:
        error_msg = "This TrackMe deployment is running in Free limited edition and you have reached the maximum number of 1 remote deployment, only the first remote account ({}) can be used".format(
            accounts[0]
        )
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "account": account,
            }
        )

    else:
        # Enforce https and remove trailing slash in the URL, if any
        selected_url = f"https://{selected_url.replace('https://', '').rstrip('/')}"

        # Splunk remote application namespace where searches are going to be executed, default to search if not defined
        if not app_namespace:
            app_namespace = "search"

        # else get the bearer token stored encrypted
        else:
            bearer_token = get_bearer_token(storage_passwords, account)

        if not bearer_token:
            error_msg = f'The bearer token for the account="{account}" could not be retrieved, cannot proceed!'
            raise Exception(
                {
                    "status": "failure",
                    "message": error_msg,
                    "account": account,
                    "host": parsed_url.hostname,
                    "port": parsed_url.port,
                }
            )

        else:
            # Use urlparse to extract relevant info from target
            parsed_url = urllib.parse.urlparse(selected_url)

            # Establish the remote service
            try:
                remoteservice = establish_sdk_remote_service(
                    parsed_url, bearer_token, app_namespace, account
                )

            except Exception as e:
                error_msg = 'remote search for account="{}" has failed at connectivitity check, host="{}" on port="{}" for Splunk remote account="{}" with exception="{}"'.format(
                    account, parsed_url.hostname, parsed_url.port, account, str(e)
                )
                logging.error(error_msg)
                raise Exception(
                    {
                        "message": "remote search check failed at connectivity verification",
                        "account": account,
                        "host": parsed_url.hostname,
                        "port": parsed_url.port,
                    }
                )

            if remoteservice:
                logging.info(
                    'remote search connectivity check to host="{}" on port="{}" for Splunk remote account="{}" was successful'.format(
                        parsed_url.hostname, parsed_url.port, account
                    )
                )
                return {
                    "status": "success",
                    "message": "remote search connectivity check was successful, service was established",
                    "account": account,
                    "host": parsed_url.hostname,
                    "port": parsed_url.port,
                }

            else:
                error_msg = "remote search connectivity check has failed to retrieve the list of applications on the remote system"
                logging.error(error_msg)
                raise Exception(
                    {
                        "message": "remote search check failed at connectivity verification",
                        "account": account,
                        "host": parsed_url.hostname,
                        "port": parsed_url.port,
                    }
                )


# Test remote connectivity before the account is created, expects a dict account object containing required information to test the connectivity
def trackme_test_remote_connectivity(connection_info):
    splunk_url = connection_info.get("target_endpoints")
    app_namespace = connection_info.get("app_namespace", "search")
    bearer_token = connection_info.get("bearer_token")

    # Create a session within the generate function
    session = requests.Session()

    # Call target selector and pass the session as an argument
    selected_url = select_url(session, splunk_url)

    # end of get configuration

    # Stop here if none of the submitted endpoints can be reached
    if not selected_url:
        error_msg = "None of the endpoints provided in the account URLs could be reached successfully, verify your network connectivity!"
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "targets": splunk_url,
            }
        )

    # Enforce https and remove trailing slash in the URL, if any
    selected_url = f"https://{selected_url.replace('https://', '').rstrip('/')}"

    if not bearer_token:
        error_msg = f"The bearer token was not provided, cannot proceed!"
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "host": parsed_url.hostname,
                "port": parsed_url.port,
            }
        )

    else:
        # Use urlparse to extract relevant info from target
        parsed_url = urllib.parse.urlparse(selected_url)

        # Establish the remote service
        try:
            remoteservice = establish_sdk_remote_service(
                parsed_url, bearer_token, app_namespace, "connection_test"
            )

        except Exception as e:
            error_msg = f'remote search has failed at connectivitity check, host="{parsed_url.hostname}" on port="{parsed_url.port}" with exception="{str(e)}"'
            logging.error(error_msg)
            raise Exception(
                {
                    "message": "remote search check failed at connectivity verification",
                    "host": parsed_url.hostname,
                    "port": parsed_url.port,
                    "exception": str(e),
                }
            )

        if remoteservice:
            logging.info(
                f'remote search connectivity check to host="{parsed_url.hostname}" on port="{parsed_url.port}" was successful'
            )
            return {
                "status": "success",
                "message": "remote search connectivity check was successful, service was established",
                "host": parsed_url.hostname,
                "port": parsed_url.port,
            }

        else:
            error_msg = "remote search connectivity check has failed to retrieve the list of applications on the remote system"
            logging.error(error_msg)
            raise Exception(
                {
                    "message": error_msg,
                    "host": parsed_url.hostname,
                    "port": parsed_url.port,
                }
            )


# Get remote account credentials, designed to be used for a least privileges approach in a programmatic approach
def trackme_get_remote_account(reqinfo, account):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo.server_rest_port,
        token=reqinfo.system_authtoken,
        timeout=300,
    )

    # Splunk credentials store
    storage_passwords = service.storage_passwords

    # get all acounts
    accounts = []
    conf_file = "trackme_account"

    # if there are no account, raise an exception, otherwise what we would do here?
    try:
        confs = service.confs[str(conf_file)]
    except Exception as e:
        error_msg = (
            "splunkremotesearch was called but we have no remote account configured yet"
        )
        raise Exception(error_msg)

    for stanza in confs:
        # get all accounts
        for name in stanza.name:
            accounts.append(stanza.name)
            break

    # account configuration
    isfound = False
    splunk_url = None
    app_namespace = None
    rbac_roles = None

    # get account
    for stanza in confs:
        if stanza.name == str(account):
            isfound = True
            for key, value in stanza.content.items():
                if key == "splunk_url":
                    splunk_url = value
                if key == "app_namespace":
                    app_namespace = value
                if key == "rbac_roles":
                    rbac_roles = value

    # end of get configuration

    # Stop here if we cannot find the submitted account
    if not isfound:
        error_msg = 'The account="{}" has not been configured on this instance, cannot proceed!'.format(
            account
        )
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "account": account,
            }
        )

    # check the license
    try:
        check_license = trackme_check_license(
            reqinfo.server_rest_uri, reqinfo.session_key
        )
        license_is_valid = check_license.get("license_is_valid")
        logging.debug(
            'function check_license called, response="{}"'.format(
                json.dumps(check_license, indent=2)
            )
        )

    except Exception as e:
        license_is_valid = 0
        logging.error('function check_license exception="{}"'.format(str(e)))

    # try and return
    if license_is_valid != 1 and len(accounts) >= 2 and accounts[0] != account:
        error_msg = "This TrackMe deployment is running in Free limited edition and you have reached the maximum number of 1 remote deployment, only the first remote account ({}) can be used".format(
            accounts[0]
        )
        raise Exception(
            {
                "status": "failure",
                "message": error_msg,
                "account": account,
            }
        )

    else:
        # Splunk remote application namespace where searches are going to be executed, default to search if not defined
        if not app_namespace:
            app_namespace = "search"

        # RBAC: the user must be a member of the grante roles for this account, for retro-compatibility purposes,
        # if this was not defined yet, use builtin TrackMe roles
        if not rbac_roles:
            rbac_roles = [
                "admin",
                "sc_admin",
                "trackme_user",
                "trackme_power",
                "trackme_admin",
            ]
        else:
            rbac_roles = rbac_roles.split(",")

        # get the bearer token stored encrypted
        bearer_token = get_bearer_token(storage_passwords, account)

        if not bearer_token:
            error_msg = f'The bearer token for the account="{account}" could not be retrieved, cannot proceed!'
            raise Exception(
                {
                    "status": "failure",
                    "message": error_msg,
                    "account": account,
                    "splunk_url": splunk_url,
                }
            )

        else:
            # render
            return {
                "status": "success",
                "message": "remote search connectivity check was successful, service was established",
                "account": account,
                "splunk_url": splunk_url,
                "app_namespace": app_namespace,
                "token": bearer_token,
                "rbac_roles": rbac_roles,
            }


#
#
#


def trackme_create_report(
    session_key,
    splunkd_uri,
    tenant_id,
    report_name,
    report_search,
    report_properties,
    report_acl,
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # create a new report
    logging.info(
        'tenant_id="{}", attempting to create report report_name="{}"'.format(
            tenant_id, report_name
        )
    )

    #
    # Splunkd API needs a couple of seconds to refresh while macros were created
    # In a programmatic context, this may lead the report creation to be failing
    # the function will check the KO status, and wait if needed for a certain amount of time

    # set max failed re-attempt
    max_failures_count = 24
    sleep_time = 5
    creation_success = False
    current_failures_count = 0

    while current_failures_count < max_failures_count and not creation_success:
        try:
            newtracker = service.saved_searches.create(
                str(report_name), str(report_search)
            )
            logging.info(
                'tenant_id="{}", action="success", report_name="{}"'.format(
                    tenant_id, report_name
                )
            )
            creation_success = True
            break

        except Exception as e:
            # We except this sentence in the exception if the API is not ready yet
            logging.warning(
                'tenant_id="{}", temporary failure, the report is not yet available, will sleep and re-attempt, report report_name="{}"'.format(
                    tenant_id, report_name
                )
            )
            time.sleep(sleep_time)
            current_failures_count += 1

            if current_failures_count >= max_failures_count:
                logging.error(
                    'tenant_id="{}", max attempt reached, failure to create report report_name="{}" with exception:"{}"'.format(
                        tenant_id, report_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", max attempt reached, failure to create report report_name="{}" with exception:"{}"'.format(
                        tenant_id, report_name, str(e)
                    )
                )

    # update the properties
    newtracker_update = service.saved_searches[str(report_name)]

    # Complete the report definition
    kwargs = report_properties

    # For optimization purposes, if the schedule is set to every 5 minutes, randomly choose an every 5 minutes schedule
    if kwargs.get("cron_schedule") == "*/5 * * * *":
        cron_random_list = [
            "*/5 * * * *",
            "1-56/5 * * * *",
            "2-57/5 * * * *",
            "3-58/5 * * * *",
            "4-59/5 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)
    elif kwargs.get("cron_schedule") == "*/10 * * * *":
        cron_random_list = [
            "*/10 * * * *",
            "1-59/10 * * * *",
            "2-59/10 * * * *",
            "3-59/10 * * * *",
            "4-59/10 * * * *",
            "5-59/10 * * * *",
            "6-59/10 * * * *",
            "7-59/10 * * * *",
            "8-59/10 * * * *",
            "9-59/10 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)
    elif kwargs.get("cron_schedule") == "*/15 * * * *":
        cron_random_list = [
            "*/15 * * * *",
            "1-59/15 * * * *",
            "2-59/15 * * * *",
            "3-59/15 * * * *",
            "4-59/15 * * * *",
            "5-59/15 * * * *",
            "6-59/15 * * * *",
            "7-59/15 * * * *",
            "8-59/15 * * * *",
            "9-59/15 * * * *",
            "10-59/15 * * * *",
            "11-59/15 * * * *",
            "12-59/15 * * * *",
            "13-59/15 * * * *",
            "14-59/15 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)

    elif kwargs.get("cron_schedule") == "*/20 * * * *":
        cron_random_list = [
            "*/20 * * * *",
            "1-59/20 * * * *",
            "2-59/20 * * * *",
            "3-59/20 * * * *",
            "4-59/20 * * * *",
            "5-59/20 * * * *",
            "6-59/20 * * * *",
            "7-59/20 * * * *",
            "8-59/20 * * * *",
            "9-59/20 * * * *",
            "10-59/20 * * * *",
            "11-59/20 * * * *",
            "12-59/20 * * * *",
            "13-59/20 * * * *",
            "14-59/20 * * * *",
            "15-59/20 * * * *",
            "16-59/20 * * * *",
            "17-59/20 * * * *",
            "18-59/20 * * * *",
            "19-59/20 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)

    elif (
        kwargs.get("cron_schedule") == "*/30 * * * *"
        or kwargs.get("cron_schedule") == "30 * * * *"
    ):
        cron_random_list = [
            "*/30 * * * *",
            "1,31 * * * *",
            "2,32 * * * *",
            "3,33 * * * *",
            "4,34 * * * *",
            "5,35 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)

    elif kwargs.get("cron_schedule") == "*/60 * * * *":
        cron_random_list = [
            "*/60 * * * *",
            "1 * * * *",
            "2 * * * *",
            "3 * * * *",
            "4 * * * *",
            "5 * * * *",
            "6 * * * *",
            "7 * * * *",
            "8 * * * *",
            "9 * * * *",
        ]
        kwargs["cron_schedule"] = random.choice(cron_random_list)

    # Update the server and refresh the local copy of the object
    logging.info(
        'tenant_id="{}", attempting to update report_name="{}" with kwargs="{}"'.format(
            tenant_id, report_name, json.dumps(kwargs, indent=1)
        )
    )
    try:
        #
        # Splunkd API needs a couple of seconds to refresh while macros were created
        # In a programmatic context, this may lead the report creation to be failing
        # the function will check the KO status, and wait if needed for a certain amount of time

        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5
        creation_success = False
        current_failures_count = 0

        while current_failures_count < max_failures_count and not creation_success:
            try:
                newtracker_update.update(**kwargs).refresh()
                logging.info(
                    'tenant_id="{}", action="success", report_name="{}" with kwargs="{}"'.format(
                        tenant_id, report_name, json.dumps(kwargs, indent=1)
                    )
                )
                newtracker_info = service.saved_searches[str(report_name)].get
                logging.debug(
                    'tenant_id="{}", report_name="{}", response="{}"'.format(
                        tenant_id, report_name, newtracker_info
                    )
                )
                creation_success = True
                break

            except Exception as e:
                logging.warning(
                    'tenant_id="{}", temporary failure, the report is not yet available, will sleep and re-attempt, report report_name="{}"'.format(
                        tenant_id, report_name
                    )
                )
                time.sleep(sleep_time)
                current_failures_count += 1

                if current_failures_count >= max_failures_count:
                    logging.error(
                        'tenant_id="{}", max attempt reached, failure to create report report_name="{}" with exception:"{}"'.format(
                            tenant_id, report_name, str(e)
                        )
                    )
                    raise Exception(str(e))

    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update report report_name="{}" with exception:"{}"'.format(
                tenant_id, report_name, str(e)
            )
        )
        raise Exception(str(e))

    # Handler RBAC
    logging.info(
        'tenant_id="{}", attempting to update report_name="{}" with kwargs="{}"'.format(
            tenant_id, report_name, json.dumps(report_acl, indent=1)
        )
    )
    try:
        service.post(
            "%s/%s" % (newtracker_update.links["alternate"], "acl"),
            body=urlencode(report_acl),
        )
        logging.info(
            'tenant_id="{}", action="success", report_name="{}" with kwargs="{}"'.format(
                tenant_id, report_name, json.dumps(report_acl, indent=1)
            )
        )

        return {
            "action": "success",
            "tenant_id": tenant_id,
            "report_name": report_name,
            "report_search": report_search,
            "report_owner": report_acl.get("owner"),
            "report_perms_read": report_acl.get("perms.read"),
            "report_perms_write": report_acl.get("perms.write"),
            "description": kwargs.get("description"),
            "is_scheduled": kwargs.get("is_scheduled"),
            "schedule_window": kwargs.get("schedule_window"),
            "dispatch.earliest_time": kwargs.get("dispatch.earliest_time"),
            "dispatch.latest_time": kwargs.get("dispatch.latest_time"),
            "cron_schedule": kwargs.get("cron_schedule"),
        }

    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update report report_name="{}" with exception:"{}"'.format(
                tenant_id, report_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_delete_report(session_key, splunkd_uri, tenant_id, report_name):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    try:
        # first retrieve the report definition for logging purposes
        savedsearch = service.saved_searches[report_name]
        savedsearch_search = savedsearch.content["search"]
        savedsearch_cron_schedule = savedsearch.content["cron_schedule"]
        savedsearch_description = savedsearch.content["description"]
        savedsearch_is_scheduled = savedsearch.content["is_scheduled"]
        savedsearch_earliest_time = savedsearch.content["dispatch.earliest_time"]
        savedsearch_latest_time = savedsearch.content["dispatch.latest_time"]

        # delete report
        logging.info(
            f'tenant_id="{tenant_id}", attempting to delete report report_name="{report_name}", report_search="{savedsearch_search}", report_cron_schedule="{savedsearch_cron_schedule}", report_description="{savedsearch_description}", report_is_scheduled="{savedsearch_is_scheduled}", report_earliest_time="{savedsearch_earliest_time}", report_latest_time="{savedsearch_latest_time}"'
        )

        service.saved_searches.delete(str(report_name))
        logging.info(
            'tenant_id="{}", action="success", report_name="{}"'.format(
                tenant_id, report_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to delete report report_name="{}" with exception:"{}"'.format(
                tenant_id, report_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_create_kvcollection(
    session_key, splunkd_uri, tenant_id, collection_name, collection_acl
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # create a new KVstore collection

    # if the collection is found, print it out
    # if not, then create the collection
    if collection_name not in service.kvstore:
        logging.info(
            'tenant_id="{}", attempting to create collection collection_name="{}"'.format(
                tenant_id, collection_name
            )
        )
        try:
            service.kvstore.create(
                collection_name, **{"app": "trackme", "owner": "nobody"}
            )
            logging.info(
                'tenant_id="{}", action="success", collection_name="{}"'.format(
                    tenant_id, collection_name
                )
            )
        except Exception as e:
            logging.error(
                'tenant_id="{}", failure to create collection collection_name="{}" with exception:"{}"'.format(
                    tenant_id, collection_name, str(e)
                )
            )
            raise Exception(
                'tenant_id="{}", failure to create collection collection_name="{}" with exception:"{}"'.format(
                    tenant_id, collection_name, str(e)
                )
            )

    record_url = f"{splunkd_uri}/servicesNS/nobody/trackme/storage/collections/config/{collection_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update collection collection_name="{}"'.format(
            tenant_id, collection_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=collection_acl,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", collection_name="{}"'.format(
                tenant_id, collection_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update collection collection_name="{}" with exception:"{}"'.format(
                tenant_id, collection_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_delete_kvcollection(session_key, splunkd_uri, tenant_id, collection_name):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    logging.info(
        'tenant_id="{}", attempting to delete collection collection_name="{}"'.format(
            tenant_id, collection_name
        )
    )
    try:
        service.kvstore.delete(collection_name)
        logging.info(
            'tenant_id="{}", action="success", collection_name="{}"'.format(
                tenant_id, collection_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to delete collection collection_name="{}" with exception:"{}"'.format(
                tenant_id, collection_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_create_kvtransform(
    session_key,
    splunkd_uri,
    tenant_id,
    transform_name,
    transform_fields,
    collection_name,
    transform_owner,
    transform_acl,
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # transforms
    transforms = service.confs["transforms"]

    logging.info(
        'tenant_id="{}", attempting to create transforms transforms_name="{}"'.format(
            tenant_id, transform_name
        )
    )
    try:
        transforms.create(
            name=str(transform_name),
            **{
                "app": "trackme",
                "sharing": "app",
                "external_type": "kvstore",
                "collection": str(collection_name),
                "fields_list": transform_fields,
                "owner": transform_owner,
            },
        )
        logging.info(
            'tenant_id="{}", action="success", transforms_name="{}"'.format(
                tenant_id, transform_name
            )
        )
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to create transforms transforms_name="{}" with exception:"{}"'.format(
                tenant_id, transform_name, str(e)
            )
        )
        raise Exception(
            'tenant_id="{}", failure to create transforms transforms_name="{}" with exception:"{}"'.format(
                tenant_id, transform_name, str(e)
            )
        )

    record_url = f"{splunkd_uri}/servicesNS/admin/trackme/data/transforms/lookups/{transform_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update transforms transforms_name="{}"'.format(
            tenant_id, transform_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=transform_acl,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", transforms_name="{}"'.format(
                tenant_id, transform_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update transforms transforms_name="{}" with exception:"{}"'.format(
                tenant_id, transform_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_delete_kvtransform(session_key, splunkd_uri, tenant_id, transform_name):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # transforms
    transforms = service.confs["transforms"]

    # proceed
    logging.info(
        'tenant_id="{}", attempting to delete transform transform_name="{}"'.format(
            tenant_id, transform_name
        )
    )
    try:
        transforms.delete(name=str(transform_name))
        logging.info(
            'tenant_id="{}", action="success", transform_name="{}"'.format(
                tenant_id, transform_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to delete transform transform_name="{}" with exception:"{}"'.format(
                tenant_id, transform_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_create_macro(
    session_key,
    splunkd_uri,
    tenant_id,
    macro_name,
    macro_definition,
    macro_owner,
    macro_acl,
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # macros
    macros = service.confs["macros"]

    logging.info(
        'tenant_id="{}", attempting to create macro macro_name="{}"'.format(
            tenant_id, macro_name
        )
    )
    try:
        macros.create(
            name=str(macro_name),
            **{
                "app": "trackme",
                "sharing": "app",
                "definition": str(macro_definition),
                "owner": str(macro_owner),
            },
        )
        logging.info(
            'tenant_id="{}", action="success", macro_name="{}"'.format(
                tenant_id, macro_name
            )
        )

    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to create macro macro_name="{}" with exception:"{}"'.format(
                tenant_id, macro_name, str(e)
            )
        )
        raise Exception(str(e))

    record_url = f"{splunkd_uri}/servicesNS/admin/trackme/data/macros/{macro_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update macro macro_name="{}"'.format(
            tenant_id, macro_name
        )
    )
    try:
        #
        # Splunkd API needs a couple of seconds to refresh while macros were created
        # In a programmatic context, this may lead the report creation to be failing
        # the function will check the KO status, and wait if needed for a certain amount of time

        # set max failed re-attempt
        max_failures_count = 24
        sleep_time = 5
        creation_success = False
        current_failures_count = 0

        while current_failures_count < max_failures_count and not creation_success:
            try:
                response = requests.post(
                    record_url,
                    headers=header,
                    data=macro_acl,
                    verify=False,
                    timeout=300,
                )
                logging.info(
                    'tenant_id="{}", action="success", macro_name="{}"'.format(
                        tenant_id, macro_name
                    )
                )
                new_macro = macros[str(macro_name)].get
                logging.debug(
                    'tenant_id="{}", macro_name="{}", response="{}"'.format(
                        tenant_id, macro_name, new_macro
                    )
                )
                creation_success = True
                break

            except Exception as e:
                logging.warning(
                    'tenant_id="{}", temporary failure, the macro is not yet available, will sleep and re-attempt, macro macro_name="{}"'.format(
                        tenant_id, macro_name
                    )
                )
                time.sleep(sleep_time)
                current_failures_count += 1

                if current_failures_count >= max_failures_count:
                    logging.error(
                        'tenant_id="{}", max attempt reached, failure to create macro macro_name="{}" with exception:"{}"'.format(
                            tenant_id, macro_name, str(e)
                        )
                    )
                    raise Exception(str(e))

        return "success"

    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update macro macro_name="{}" with exception:"{}"'.format(
                tenant_id, macro_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_delete_macro(session_key, splunkd_uri, tenant_id, macro_name):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # macros
    macros = service.confs["macros"]

    try:
        # get the definition first
        macro_definition = macros[macro_name].content["definition"]

        logging.info(
            f'tenant_id="{tenant_id}", attempting to delete macro macro_name="{macro_name}", macro_definition="{macro_definition}"'
        )

        # delete
        macros.delete(name=str(macro_name))

        logging.info(
            f'tenant_id="{tenant_id}", action="success", macro_name="{macro_name}'
        )

        return "success"

    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to delete macro macro_name="{}" with exception:"{}"'.format(
                tenant_id, macro_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_report_update_acl(
    session_key, splunkd_uri, tenant_id, report_name, report_acl_properties
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    record_url = f"{splunkd_uri}/servicesNS/nobody/trackme/saved/searches/{urllib.parse.quote(str(report_name))}/acl"

    logging.info(
        'tenant_id="{}", attempting to update report report_name="{}"'.format(
            tenant_id, report_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=report_acl_properties,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", report_name="{}"'.format(
                tenant_id, report_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update report report_name="{}" with exception:"{}"'.format(
                tenant_id, report_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_kvcollection_update_acl(
    session_key, splunkd_uri, tenant_id, collection_name, collection_acl_properties
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    record_url = f"{splunkd_uri}/servicesNS/nobody/trackme/storage/collections/config/{collection_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update collection collection_name="{}"'.format(
            tenant_id, collection_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=collection_acl_properties,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", collection_name="{}"'.format(
                tenant_id, collection_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update collection collection_name="{}" with exception:"{}"'.format(
                tenant_id, collection_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_macro_update_acl(
    session_key, splunkd_uri, tenant_id, macro_name, macro_acl_properties
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    record_url = f"{splunkd_uri}/servicesNS/admin/trackme/data/macros/{macro_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update macro macro_name="{}"'.format(
            tenant_id, macro_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=macro_acl_properties,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", macro_name="{}"'.format(
                tenant_id, macro_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update macro macro_name="{}" with exception:"{}"'.format(
                tenant_id, macro_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_transform_update_acl(
    session_key, splunkd_uri, tenant_id, transform_name, transform_acl_properties
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    record_url = f"{splunkd_uri}/servicesNS/admin/trackme/data/transforms/lookups/{transform_name}/acl"

    logging.info(
        'tenant_id="{}", attempting to update transforms transform_name="{}"'.format(
            tenant_id, transform_name
        )
    )
    try:
        response = requests.post(
            record_url,
            headers=header,
            data=transform_acl_properties,
            verify=False,
            timeout=300,
        )
        logging.info(
            'tenant_id="{}", action="success", transform_name="{}"'.format(
                tenant_id, transform_name
            )
        )
        return "success"
    except Exception as e:
        logging.error(
            'tenant_id="{}", failure to update transforms transform_name="{}" with exception:"{}"'.format(
                tenant_id, transform_name, str(e)
            )
        )
        raise Exception(str(e))


def trackme_report_update_enablement(
    session_key, splunkd_uri, tenant_id, report_name, action
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    if action not in ("enable", "disable"):
        raise Exception(
            'Invalid value for action="{}", valid options are: enable | disable'.format(
                action
            )
        )

    else:
        record_url = f"{splunkd_uri}/servicesNS/nobody/trackme/saved/searches/{urllib.parse.quote(str(report_name))}/{action}"
        logging.info(
            'tenant_id="{}", attempting to {} report report_name="{}"'.format(
                tenant_id, action, report_name
            )
        )
        try:
            response = requests.post(
                record_url, headers=header, verify=False, timeout=300
            )
            logging.info(
                'tenant_id="{}", action="success", report_name="{}"'.format(
                    tenant_id, report_name
                )
            )
            return "success"
        except Exception as e:
            logging.error(
                'tenant_id="{}", failure to update report report_name="{}" with exception:"{}"'.format(
                    tenant_id, report_name, str(e)
                )
            )
            raise Exception(str(e))


def trackme_macro_update_enablement(
    session_key, splunkd_uri, tenant_id, macro_name, action
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # get macros
    macros = service.confs["macros"]

    if action not in ("enable", "disable"):
        raise Exception(
            'Invalid value for action="{}", valid options are: enable | disable'.format(
                action
            )
        )

    else:
        if action == "enable":
            kwargs = {"disabled": "false"}
        elif action == "disable":
            kwargs = {"disabled": "true"}

        # update the properties
        macro_update = macros[str(macro_name)]
        logging.info(
            'tenant_id="{}", attempting to update macro macro_name="{}"'.format(
                tenant_id, macro_name
            )
        )
        try:
            macro_update.update(**kwargs).refresh()
            logging.info(
                'tenant_id="{}", action="success", macro_name="{}"'.format(
                    tenant_id, macro_name
                )
            )
            return "success"
        except Exception as e:
            logging.error(
                'tenant_id="{}", failure to update macro macro_name="{}" with exception:"{}"'.format(
                    tenant_id, macro_name, str(e)
                )
            )
            raise Exception(str(e))


def trackme_kvcollection_update_enablement(
    session_key, splunkd_uri, tenant_id, collection_name, action
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    if action not in ("enable", "disable"):
        raise Exception(
            'Invalid value for action="{}", valid options are: enable | disable'.format(
                action
            )
        )

    else:
        record_url = f"{splunkd_uri}/servicesNS/nobody/trackme/storage/collections/config/{collection_name}/{action}"

        logging.info(
            'tenant_id="{}", attempting to {} collection collection_name="{}"'.format(
                tenant_id, action, collection_name
            )
        )
        try:
            response = requests.post(
                record_url, headers=header, verify=False, timeout=300
            )
            logging.info(
                'tenant_id="{}", action="success", collection_name="{}"'.format(
                    tenant_id, collection_name
                )
            )
            return "success"
        except Exception as e:
            logging.error(
                'tenant_id="{}", failure to update collection collection_name="{}" with exception:"{}"'.format(
                    tenant_id, collection_name, str(e)
                )
            )
            raise Exception(str(e))


def trackme_transform_update_enablement(
    session_key, splunkd_uri, tenant_id, transform_name, action
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # get transforms
    transforms = service.confs["transforms"]

    if action not in ("enable", "disable"):
        raise Exception(
            'Invalid value for action="{}", valid options are: enable | disable'.format(
                action
            )
        )

    else:
        if action == "enable":
            kwargs = {"disabled": "false"}
        elif action == "disable":
            kwargs = {"disabled": "true"}

        # update the properties
        transform_update = transforms[str(transform_name)]
        logging.info(
            'tenant_id="{}", attempting to update transform="{}"'.format(
                tenant_id, transform_name
            )
        )
        try:
            transform_update.update(**kwargs).refresh()
            logging.info(
                'tenant_id="{}", action="success", transform_name="{}"'.format(
                    tenant_id, transform_name
                )
            )
            return "success"
        except Exception as e:
            logging.error(
                'tenant_id="{}", failure to update transforms transform_name="{}" with exception:"{}"'.format(
                    tenant_id, transform_name, str(e)
                )
            )
            raise Exception(str(e))


def trackme_audit_event(
    session_key,
    splunkd_uri,
    tenant_id,
    user,
    action,
    change_type,
    object_name,
    object_category,
    object_attrs,
    result,
    comment,
):
    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # Audit

    # set
    url = "%s/services/trackme/v2/audit/audit_event" % splunkd_uri
    data = {
        "action": str(action),
        "change_type": str(change_type),
        "object": str(object_name),
        "object_category": str(object_category),
        "object_attrs": str(object_attrs),
        "tenant_id": str(tenant_id),
        "user": str(user),
        "result": str(result),
        "comment": str(comment),
    }

    # Proceed
    try:
        response = requests.post(
            url, headers=header, data=json.dumps(data), verify=False, timeout=300
        )
        if response.ok:
            logging.debug(f'Success audit event, data="{response}"')
            response_json = response.json()
            return response_json
        else:
            error_message = f'Failed to generate an audit event, status_code={response.status_code}, response_text="{response.text}"'
            logging.error(error_message)
            raise Exception(error_message)

    except Exception as e:
        error_msg = 'trackme_audit_event has failed, exception="{}"'.format(str(e))
        raise Exception(error_msg)


def trackme_audit_flip(
    session_key,
    splunkd_uri,
    tenant_id,
    object,
    object_category,
    object_state,
    object_previous_state,
    latest_flip_time,
    latest_flip_state,
    result,
):
    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # set
    url = "%s/services/trackme/v2/audit/flip_event" % splunkd_uri
    data = {
        "tenant_id": str(tenant_id),
        "object": str(object),
        "object_category": str(object_category),
        "object_state": str(object_state),
        "object_previous_state": str(object_previous_state),
        "latest_flip_time": str(latest_flip_time),
        "latest_flip_state": str(latest_flip_state),
        "result": str(result),
    }

    # Proceed
    try:
        response = requests.post(
            url,
            headers=header,
            data=json.dumps(data, indent=1),
            verify=False,
            timeout=300,
        )
        if response.ok:
            logging.debug(f'Success flip event, data="{response}"')
            response_json = response.json()
            return response_json
        else:
            error_message = f'Failed to generate a flip event, status_code={response.status_code}, response_text="{response.text}", data="{json.dumps(data, indent=1)}"'
            logging.error(error_message)
            raise Exception(error_message)

    except Exception as e:
        error_msg = 'trackme_audit_flip has failed, exception="{}"'.format(str(e))
        raise Exception(error_msg)


def trackme_state_event(
    session_key, splunkd_uri, tenant_id, index, sourcetype, source, record
):
    # Define an header for requests authenticated communications with splunkd
    header = {
        "Authorization": "Splunk %s" % session_key,
        "Content-Type": "application/json",
    }

    # set
    url = "%s/services/trackme/v2/audit/state_event" % splunkd_uri
    data = {
        "tenant_id": str(tenant_id),
        "index": str(index),
        "sourcetype": str(sourcetype),
        "source": str(source),
        "record": record,
    }

    # Proceed
    try:
        response = requests.post(
            url, headers=header, data=json.dumps(data), verify=False, timeout=300
        )
        if response.ok:
            logging.debug(f'Success state event, data="{response}"')
            response_json = response.json()
            return response_json
        else:
            error_message = f'Failed to generate a state event, status_code={response.status_code}, response_text="{response.text}"'
            logging.error(error_message)
            raise Exception(error_message)

    except Exception as e:
        error_msg = 'trackme_state_event has failed, exception="{}"'.format(str(e))
        raise Exception(error_msg)


def trackme_components_register_gen_metrics(
    session_key, splunkd_uri, tenant_id, records
):
    # proceed
    try:
        # get the target index
        tenant_indexes = trackme_idx_for_tenant(
            session_key,
            splunkd_uri,
            tenant_id,
        )

        # Set a new log file handler
        filehandler = RotatingFileHandler(
            splunkhome + "/var/log/splunk/trackme_components_register_metrics.log",
            mode="a",
            maxBytes=100000000,
            backupCount=1,
        )
        formatter = JSONFormatter()
        logging.Formatter.converter = time.gmtime
        filehandler.setFormatter(formatter)
        log = logging.getLogger()  # root logger - Good to get it only once.
        orig_filehandler = None
        for hdlr in log.handlers[:]:  # remove the existing file handlers
            if isinstance(hdlr, logging.FileHandler):
                orig_filehandler = hdlr
                log.removeHandler(hdlr)
        log.addHandler(filehandler)  # set the new handler

        for record in records:
            logging.info(
                "Metrics - group=components_register_metrics",
                extra={
                    "target_index": tenant_indexes["trackme_metric_idx"],
                    "tenant_id": tenant_id,
                    "component": record.get("component"),
                    "tracker": record.get("tracker"),
                    "metrics_event": json.dumps(record.get("metrics_event")),
                },
            )

        # restore the original file handler
        if orig_filehandler:
            log.removeHandler(filehandler)
            log.addHandler(orig_filehandler)

        return True

    except Exception as e:
        raise Exception(str(e))


# register the tenant object summary status
def trackme_register_tenant_object_summary(
    session_key,
    splunkd_uri,
    tenant_id,
    component,
    report,
    last_status,
    last_exec,
    last_duration,
    last_result,
    earliest,
    latest,
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'Starting function trackme_register_tenant_object_summary, tenant_id="{}", component="{}", report="{}", last_exec="{}", last_status="{}", last_duration="{}", last_result="{}", earliest="{}", latest="{}"'.format(
            tenant_id,
            component,
            report,
            last_exec,
            last_status,
            last_duration,
            last_result,
            earliest,
            latest,
        )
    )

    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'The vtenant_key was successfully found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'The vtenant_key was not found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    if vtenant_key:
        # try to load the dict
        try:
            tenant_objects_exec_summary = json.loads(
                vtenant_record.get("tenant_objects_exec_summary")
            )
        except Exception as e:
            tenant_objects_exec_summary = None

        # logging debug
        logging.debug(
            'tenant_id="{}", component="{}", report="{}", Retrieve tenant_objects_exec_summary="{}"'.format(
                tenant_id, component, report, tenant_objects_exec_summary
            )
        )

        # add to existing disct
        if tenant_objects_exec_summary and tenant_objects_exec_summary != "None":
            try:
                # log
                logging.debug(
                    'tenant_id="{}", component="{}", report="{}", Updating the existing record in the dictionnary, summary_dict="{}"'.format(
                        tenant_id,
                        component,
                        report,
                        json.dumps(tenant_objects_exec_summary, indent=1),
                    )
                )

                report_dict = tenant_objects_exec_summary[report]

                # Update the existing record in the dict
                report_dict["component"] = str(component)
                report_dict["last_status"] = str(last_status)
                report_dict["last_exec"] = str(last_exec)
                report_dict["last_duration"] = round(float(last_duration), 3)
                report_dict["last_result"] = str(last_result)
                report_dict["earliest"] = str(earliest)
                report_dict["latest"] = str(latest)

                # sort report_dict alphabetically
                tenant_objects_exec_summary[report] = dict(sorted(report_dict.items()))

                # generate metrics
                components_register_metrics_gen_start = time.time()

                # convert the value of last_status from a string to a numerical:
                # - 1: success
                # - 2: failure
                # - 3: unknown
                if last_status == "success":
                    last_status_num = 1
                elif last_status == "failure":
                    last_status_num = 2
                else:
                    last_status_num = 3

                try:
                    components_register_metrics = (
                        trackme_components_register_gen_metrics(
                            session_key,
                            splunkd_uri,
                            tenant_id,
                            [
                                {
                                    "tenant_id": tenant_id,
                                    "component": component,
                                    "tracker": report,
                                    "metrics_event": {
                                        "status": last_status_num,
                                        "runtime": round(float(last_duration), 3),
                                    },
                                }
                            ],
                        )
                    )
                    logging.info(
                        f'context="components_register_gen_metrics", tenant_id="{tenant_id}", function trackme_components_register_gen_metrics success {components_register_metrics}, run_time={round(time.time()-components_register_metrics_gen_start, 3)}'
                    )
                except Exception as e:
                    logging.error(
                        f'context="components_register_gen_metrics", tenant_id="{tenant_id}", function trackme_components_register_gen_metrics failed with exception {str(e)}'
                    )

            except Exception as e:
                # set the dict
                summary_dict = {
                    report: {
                        "component": component,
                        "last_status": last_status,
                        "last_exec": last_exec,
                        "last_duration": last_duration,
                        "last_result": last_result,
                        "earliest": earliest,
                        "latest": latest,
                    }
                }

                # log
                logging.debug(
                    'tenant_id="{}", component="{}", report="{}", Adding a new record to the dictionnary, summary_dict="{}"'.format(
                        tenant_id, component, report, json.dumps(summary_dict, indent=1)
                    )
                )

                # Update with a new record
                tenant_objects_exec_summary.update(summary_dict)

        # Empty dict
        else:
            # Set the dict
            tenant_objects_exec_summary = {
                report: {
                    "component": component,
                    "last_status": last_status,
                    "last_exec": last_exec,
                    "last_duration": last_duration,
                    "last_result": last_result,
                    "earliest": earliest,
                    "latest": latest,
                }
            }

            # log
            logging.debug(
                'tenant_id="{}", component="{}", report="{}", Creating a new dictionnary, tenant_objects_exec_summary="{}"'.format(
                    tenant_id,
                    component,
                    report,
                    json.dumps(tenant_objects_exec_summary, indent=1),
                )
            )

        # logging debug
        logging.debug(
            'tenant_id="{}", component="{}", report="{}", Ended processing, tenant_objects_exec_summary="{}"'.format(
                tenant_id, component, report, tenant_objects_exec_summary
            )
        )

        try:
            vtenant_record["tenant_objects_exec_summary"] = json.dumps(
                tenant_objects_exec_summary, indent=2
            )
            collection_vtenants.data.update(
                str(vtenant_key), json.dumps(vtenant_record)
            )

        except Exception as e:
            logging.error(
                'failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                    str(e)
                )
            )


# return the tenant object summary status for the last execution registered
def trackme_return_tenant_object_summary(
    session_key, splunkd_uri, tenant_id, component, report
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'Starting function trackme_return_tenant_object_summary, tenant_id="{}", component="{}", report="{}"'.format(
            tenant_id, component, report
        )
    )

    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'The vtenant_key was successfully found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'The vtenant_key was not found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    if vtenant_key:
        # try to load the dict
        try:
            tenant_objects_exec_summary = json.loads(
                vtenant_record.get("tenant_objects_exec_summary")
            )
        except Exception as e:
            tenant_objects_exec_summary = None

        # load
        if tenant_objects_exec_summary and tenant_objects_exec_summary != "None":
            # logging debug
            logging.debug(
                'function trackme_return_tenant_object_summary, tenant_id="{}", component="{}", report="{}", Retrieve tenant_objects_exec_summary="{}"'.format(
                    tenant_id,
                    component,
                    report,
                    json.dumps(tenant_objects_exec_summary.get(report), indent=2),
                )
            )

            # return the dict
            return tenant_objects_exec_summary.get(report)

        # Empty dict
        else:
            return {
                "component": component,
                "last_status": "unknown",
                "last_exec": "unknown",
                "last_duration": "unknown",
                "last_result": "unknown",
                "earliest": "unknown",
                "latest": "unknown",
            }


# delete a tenant object summary record
def trackme_delete_tenant_object_summary(
    session_key, splunkd_uri, tenant_id, component, report
):
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'Starting function trackme_return_tenant_object_summary, tenant_id="{}", component="{}", report="{}"'.format(
            tenant_id, component, report
        )
    )

    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'The vtenant_key was successfully found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'The vtenant_key was not found in the collection, query_string="{}", originating_report="{}"'.format(
                query_string, report
            )
        )

    if vtenant_key:
        # try to load the dict
        try:
            tenant_objects_exec_summary = json.loads(
                vtenant_record.get("tenant_objects_exec_summary")
            )
        except Exception as e:
            tenant_objects_exec_summary = None

        # load
        if tenant_objects_exec_summary and tenant_objects_exec_summary != "None":
            # logging debug
            logging.debug(
                'function trackme_return_tenant_object_summary, tenant_id="{}", component="{}", report="{}", Retrieve tenant_objects_exec_summary="{}"'.format(
                    tenant_id,
                    component,
                    report,
                    json.dumps(tenant_objects_exec_summary.get(report), indent=2),
                )
            )

            # delete the record from the dict
            tenant_objects_exec_summary.pop(report, None)

            # update
            vtenant_record["tenant_objects_exec_summary"] = json.dumps(
                tenant_objects_exec_summary, indent=2
            )

            # update the KVstore record
            try:
                collection_vtenants.data.update(
                    str(vtenant_key), json.dumps(vtenant_record)
                )
                logging.info(
                    'function trackme_return_tenant_object_summary, tenant_id="{}", report="{}", register summary record="{}" was successfully purged'.format(
                        tenant_id, report, json.dumps(vtenant_record, indent=2)
                    )
                )
                return "success"

            except Exception as e:
                logging.error(
                    'function trackme_return_tenant_object_summary, tenant_id="{}", report="{}", Failure to remove the register summary record="{}", exception="{}"'.format(
                        tenant_id, report, json.dumps(vtenant_record, indent=2), str(e)
                    )
                )
                return "failure"

        # Empty dict
        else:
            logging.info(
                'function trackme_return_tenant_object_summary, found no record to be purged in the register object summary for tenant_id="{}"'.format(
                    tenant_id
                )
            )


# Return the Elastic Source search to be executed depending on the various options
def trackme_return_elastic_exec_search(
    tenant_id,
    object,
    search_constraint,
    search_mode,
    elastic_data_index,
    elastic_splk_dsmtype,
    earliest_time,
    latest_time,
    wrapper_name,
    register_component,
):
    if search_mode in ("tstats"):
        # search string aggreg
        search_string_aggreg = remove_leading_spaces(
            """\
            stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
            latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
            latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m, 
            max(data_last_ingest) as data_last_ingest, min(data_first_time_seen) as data_first_time_seen, max(data_last_time_seen) as data_last_time_seen, avg(data_last_ingestion_lag_seen) as data_last_ingestion_lag_seen, sum(data_eventcount) as data_eventcount
            | eval dcount_host=round(latest_dcount_host_5m, 2)
            | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
            """
        )

        elastic_report_root_search = remove_leading_spaces(
            f"""\
            | tstats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount where {search_constraint} by _time, index, sourcetype, host span=1s | eval data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen
            | eval spantime=data_last_ingest | eventstats max(data_last_time_seen) as data_last_time_seen | eval spantime=if(spantime>=(now()-300), spantime, null())
            | eventstats sum(data_eventcount) as eventcount_5m, avg(data_last_ingestion_lag_seen) as latency_5m, dc(host) as dcount_host_5m by spantime
            | {search_string_aggreg} | eval object="{object}", data_index="{elastic_data_index}", data_sourcetype="{elastic_splk_dsmtype}"
            | `trackme_elastic_dedicated_tracker("{tenant_id}")` | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
            | `register_tenant_component_summary({tenant_id}, dsm)`\
            """
        )

    elif search_mode in ("raw"):
        # search string aggreg
        search_string_aggreg = remove_leading_spaces(
            """\
            stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
            latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
            latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m, 
            max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, avg(ingest_lag) as data_last_ingestion_lag_seen, count as data_eventcount
            | eval dcount_host=round(latest_dcount_host_5m, 2)
            | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
            """
        )

        elastic_report_root_search = remove_leading_spaces(
            f"""\
            {search_constraint}
            | eval ingest_lag=_indextime-_time
            | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
            | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
            | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
            | {search_string_aggreg} | eval object="{object}", data_index="{elastic_data_index}", data_sourcetype="{elastic_splk_dsmtype}"
            | `trackme_elastic_dedicated_tracker("{tenant_id}")`
            | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
            | `register_tenant_component_summary({tenant_id}, dsm)`\
            """
        )

    elif search_mode in ("from"):
        if re.match("datamodel:", str(search_constraint)):
            # search string aggreg
            search_string_aggreg = remove_leading_spaces(
                f"""\
                stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
                latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
                latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m,
                max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, avg(ingest_lag) as data_last_ingestion_lag_seen, count as data_eventcount
                | eval dcount_host=round(latest_dcount_host_5m, 2)
                | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
                """
            )

            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | {search_mode} {search_constraint}
                | eval ingest_lag=_indextime-_time
                | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
                | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg} | eval object="{object}", data_index="{elastic_data_index}", data_sourcetype="{elastic_splk_dsmtype}"
                | `trackme_elastic_dedicated_tracker("{tenant_id}")`
                | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        if re.match("lookup:", str(search_constraint)):
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | {search_mode} {search_constraint}
                | eventstats max(_time) as indextime | eval _indextime=if(isnum(_indextime), _indextime, indextime) | fields - indextime
                | eval host=if(isnull(host), "none", host)
                | stats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount, dc(host) as dcount_host
                | eval latest_eventcount_5m=data_eventcount
                | eval object="{object}", data_index="{elastic_data_index}", data_sourcetype="{elastic_splk_dsmtype}" | `trackme_elastic_dedicated_tracker("{tenant_id}")`
                | eval tenant_id="{tenant_id}"
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

    elif search_mode in ("mstats"):
        elastic_report_root_search = remove_leading_spaces(
            f"""\
            | mstats latest(_value) as value where {search_constraint} by host, metric_name span=1m 
            | stats min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, dc(metric_name) as data_eventcount, dc(host) as dcount_host 
            | eval object="{object}", data_index="{elastic_data_index}", data_sourcetype="{elastic_splk_dsmtype}", data_last_ingest=data_last_time_seen, data_last_ingestion_lag_seen=now()-data_last_time_seen 
            | `trackme_elastic_dedicated_tracker("{tenant_id}")` 
            | eval tenant_id="{tenant_id}" 
            | stats count as report_entities_count by tenant_id 
            | `register_tenant_component_summary({tenant_id}, dsm)`\
            """
        )

    elif search_mode in ("remote_tstats"):
        # extract using rex
        logging.debug('search_constraint="{}"'.format(search_constraint))
        remote_matches = re.match(
            r"(account\=\s{0,}\"{0,1}[^\|]+\"{0,1})\s{0,}\|\s{0,}(.*)",
            search_constraint,
        )
        if remote_matches:
            remote_target = remote_matches.group(1).replace('\\"', '"')
            remote_constraint = remote_matches.group(2)
        else:
            logging.error(
                'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                    search_constraint
                )
            )
            raise Exception(
                'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                    search_constraint
                )
            )

        # escape back slashes
        remote_constraint = remote_constraint.replace("\\", "\\\\")

        # search string aggreg
        search_string_aggreg = remove_leading_spaces(
            f"""\
            stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
            latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
            latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m, 
            max(data_last_ingest) as data_last_ingest, min(data_first_time_seen) as data_first_time_seen, max(data_last_time_seen) as data_last_time_seen, avg(data_last_ingestion_lag_seen) as data_last_ingestion_lag_seen, sum(data_eventcount) as data_eventcount
            | eval dcount_host=round(latest_dcount_host_5m, 2)
            | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
            """
        )

        # final search
        if register_component == "True":
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="| tstats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount where {remote_constraint} by _time, index, sourcetype, host span=1s | eval data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen
                | eval spantime=data_last_ingest | eventstats max(data_last_time_seen) as data_last_time_seen | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats sum(data_eventcount) as eventcount_5m, avg(data_last_ingestion_lag_seen) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}" register_component="True" tenant_id="{tenant_id}" component="splk-dsm" report="{wrapper_name}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}"
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        else:
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="| tstats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount where {remote_constraint} by _time, index, sourcetype, host span=1s
                | eval data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen
                | eval spantime=data_last_ingest | eventstats max(data_last_time_seen) as data_last_time_seen | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats sum(data_eventcount) as eventcount_5m, avg(data_last_ingestion_lag_seen) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}"
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        # log
        logging.debug(
            'elastic_report_root_search="{}"'.format(elastic_report_root_search)
        )

    elif search_mode in ("remote_mstats"):
        # extract using rex
        logging.debug('search_constraint="{}"'.format(search_constraint))
        remote_matches = re.match(
            r"(account\=\s{0,}\"{0,1}[^\|]+\"{0,1})\s{0,}\|\s{0,}(.*)",
            search_constraint,
        )
        if remote_matches:
            remote_target = remote_matches.group(1).replace('\\"', '"')
            remote_constraint = remote_matches.group(2)
        else:
            logging.error(
                'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                    search_constraint
                )
            )
            raise Exception(
                str(
                    'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                        search_constraint
                    )
                )
            )

        # escape back slashes
        remote_constraint = remote_constraint.replace("\\", "\\\\")

        # final search
        if register_component == "True":
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="| mstats latest(_value) as value where {remote_constraint} by host, metric_name span=1m | stats min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, dc(metric_name) as data_eventcount, dc(host) as dcount_host | eval data_last_ingest=data_last_time_seen | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}" register_component="True" tenant_id="{tenant_id}" component="splk-dsm" report="{wrapper_name}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                 `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        else:
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="| mstats latest(_value) as value where {remote_constraint} by host, metric_name span=1m | stats min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, dc(metric_name) as data_eventcount, dc(host) as dcount_host | eval data_last_ingest=data_last_time_seen | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}"\n
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}"\n
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
            """
            )

        # log
        logging.debug(
            'elastic_report_root_search="{}"'.format(elastic_report_root_search)
        )

    elif search_mode in ("remote_raw"):
        # extract using rex
        logging.debug(
            'search_mode="{}", search_constraint="{}"'.format(
                search_mode, search_constraint
            )
        )

        remote_matches = re.match(
            r"(account\=\s{0,}\"{0,1}[^\|]+\"{0,1})\s{0,}\|\s{0,}(.*)",
            search_constraint,
        )
        if remote_matches:
            remote_target = remote_matches.group(1).replace('\\"', '"')
            remote_constraint = remote_matches.group(2)
        else:
            logging.error(
                'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                    search_constraint
                )
            )
            raise Exception(
                'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                    search_constraint
                )
            )

        # escape back slashes
        remote_constraint = remote_constraint.replace("\\", "\\\\")

        # search string aggreg
        search_string_aggreg = remove_leading_spaces(
            f"""\
            stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
            latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
            latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m, 
            max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, avg(ingest_lag) as data_last_ingestion_lag_seen, count as data_eventcount
            | eval dcount_host=round(latest_dcount_host_5m, 2)
            | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
            """
        )

        # final search
        if register_component == "True":
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="search {remote_constraint}
                | eval ingest_lag=_indextime-_time
                | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
                | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}" register_component="True" tenant_id="{tenant_id}" component="splk-dsm" report="{wrapper_name}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}"
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        else:
            elastic_report_root_search = remove_leading_spaces(
                f"""\
                | splunkremotesearch {remote_target} search="search {remote_constraint}
                | eval ingest_lag=_indextime-_time
                | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
                | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}"
                | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
            )

        # log
        logging.debug(
            'elastic_report_root_search="{}"'.format(elastic_report_root_search)
        )

    elif search_mode in ("remote_from"):
        # log
        logging.debug(
            'search_mode="{}", search_constraint="{}"'.format(
                search_mode, search_constraint
            )
        )

        # Extract the from mode
        remote_type_search = None
        remote_type_search_matches = re.match(
            r"account\=[^\|]*\s{0,}\|\s{0,}(datamodel|lookup)", search_constraint
        )
        if remote_type_search_matches:
            remote_type_search = remote_type_search_matches.group(1)

        # datamodel
        if remote_type_search == "datamodel":
            # log
            logging.debug("remote search is of datamodel type")

            # extract using rex
            remote_matches = re.match(
                r"((?:account)\=[^\|]*)\s{0,}\|\s{0,}(.*)", search_constraint
            )
            if remote_matches:
                remote_target = remote_matches.group(1).replace('\\"', '"')
                remote_constraint = remote_matches.group(2)
            else:
                logging.error(
                    'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                        search_constraint
                    )
                )
                raise Exception(
                    'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                        search_constraint
                    )
                )

            # escape back slashes
            remote_constraint = remote_constraint.replace("\\", "\\\\")

            # search string aggreg
            search_string_aggreg = remove_leading_spaces(
                f"""\
                stats latest(eventcount_5m) as latest_eventcount_5m, avg(eventcount_5m) as avg_eventcount_5m, stdev(eventcount_5m) as stdev_eventcount_5m, perc95(eventcount_5m) as perc95_eventcount_5m,
                latest(latency_5m) as latest_latency_5m, avg(latency_5m) as avg_latency_5m, stdev(latency_5m) as stdev_latency_5m, perc95(latency_5m) as perc95_latency_5m,
                latest(dcount_host_5m) as latest_dcount_host_5m, avg(dcount_host_5m) as avg_dcount_host_5m, stdev(dcount_host_5m) as stdev_dcount_host_5m, perc95(dcount_host_5m) as perc95_dcount_host_5m,
                max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, avg(ingest_lag) as data_last_ingestion_lag_seen, count as data_eventcount
                | eval dcount_host=round(latest_dcount_host_5m, 2)
                | eval data_last_ingestion_lag_seen=round(data_last_ingestion_lag_seen, 0)\
                """
            )

            # final search
            if register_component == "True":
                elastic_report_root_search = remove_leading_spaces(
                    f"""\
                | splunkremotesearch {remote_target} search="| from {remote_constraint}
                | eval ingest_lag=_indextime-_time
                | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
                | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}" register_component="True" tenant_id="{tenant_id}" component="splk-dsm" report="{wrapper_name}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
                )

            else:
                elastic_report_root_search = remove_leading_spaces(
                    f"""\
                | splunkremotesearch {remote_target} search="| from {remote_constraint}
                | eval ingest_lag=_indextime-_time
                | eventstats max(_indextime) as data_last_ingest, max(_time) as data_last_time_seen
                | eval spantime=data_last_ingest | eval spantime=if(spantime>=(now()-300), spantime, null())
                | eventstats count as eventcount_5m, avg(ingest_lag) as latency_5m, dc(host) as dcount_host_5m by spantime
                | {search_string_aggreg}
                | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}"
                | head 1
                | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
                )

            # log
            logging.debug(
                'elastic_report_root_search="{}"'.format(elastic_report_root_search)
            )

        elif remote_type_search == "lookup":
            # log
            logging.debug("remote search is of lookup type")

            # extract using rex
            remote_matches = re.match(
                r"((?:account)\=[^\|]*)\s{0,}\|\s{0,}(.*)", search_constraint
            )

            if remote_matches:
                remote_target = remote_matches.group(1).replace('\\"', '"')
                remote_constraint = remote_matches.group(2)
            else:
                logging.error(
                    'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                        search_constraint
                    )
                )
                return Exception(
                    'invalid search, account or search constraint could not be extracted, search_constraint="{}"'.format(
                        search_constraint
                    )
                )

            # escape back slashes
            remote_constraint = remote_constraint.replace("\\", "\\\\")

            # final search
            if register_component == "True":
                elastic_report_root_search = remove_leading_spaces(
                    f"""\
                    | splunkremotesearch {remote_target} search="| from {remote_constraint}
                    | eventstats max(_time) as indextime | eval _indextime=if(isnum(_indextime), _indextime, indextime)
                    | fields - indextime
                    | eval host=if(isnull(host), \\"none\\", host)
                    | stats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount, dc(host) as dcount_host
                    | eval latest_eventcount_5m=data_eventcount
                    | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}" register_component="True" tenant_id="{tenant_id}" component="splk-dsm" report="{wrapper_name}"
                    | head 1
                    | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                    | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                    | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
                )

            else:
                elastic_report_root_search = remove_leading_spaces(
                    f"""\
                    | splunkremotesearch {remote_target} search="| from {remote_constraint}
                    | eventstats max(_time) as indextime | eval _indextime=if(isnum(_indextime), _indextime, indextime) | fields - indextime
                    | eval host=if(isnull(host), \\"none\\", host)
                    | stats max(_indextime) as data_last_ingest, min(_time) as data_first_time_seen, max(_time) as data_last_time_seen, count as data_eventcount, dc(host) as dcount_host
                    | eval latest_eventcount_5m=data_eventcount
                    | eval object=\\"{object}\\", data_index=\\"{elastic_data_index}\\", data_sourcetype=\\"{elastic_splk_dsmtype}\\", data_last_ingestion_lag_seen=data_last_ingest-data_last_time_seen" earliest="{earliest_time}" latest="{latest_time}"
                    | head 1
                    | `trackme_elastic_dedicated_tracker("{tenant_id}", "{object}")`
                    | eval tenant_id="{tenant_id}" | stats count as report_entities_count by tenant_id
                    | `register_tenant_component_summary({tenant_id}, dsm)`\
                """
                )

    # return
    return remove_leading_spaces(elastic_report_root_search)


def trackme_register_tenant_component_summary(
    session_key, splunkd_port, tenant_id, component
):
    max_failures_count = 20
    sleep_time = 10

    # get service
    service = client.connect(
        owner="nobody", app="trackme", port=splunkd_port, token=session_key, timeout=300
    )

    # Set the search
    if component in ("dsm|dhm|mhm|flx|wlk"):
        search = (
            "| `register_tenant_component_summary("
            + str(tenant_id)
            + ", "
            + component
            + ")`"
        )
    elif component in ("cim"):
        search = (
            "| `register_tenant_component_summary_nofilter("
            + str(tenant_id)
            + ", "
            + component
            + ")`"
        )

    kwargs_oneshot = {"earliest_time": "-5m", "latest_time": "now", "count": 0}

    # automatic retry
    search_successful = False
    current_failures_count = 0
    while current_failures_count < max_failures_count and not search_successful:
        try:
            oneshotsearch_results = service.jobs.oneshot(search, **kwargs_oneshot)
            logging.info(
                'tenant_id="{}", component="{}", call to trackme_register_tenant_component_summary was successful'.format(
                    tenant_id, component
                )
            )
            search_successful = True
            break
        except Exception as e:
            current_failures_count += 1
            search_successful = False
            if current_failures_count >= max_failures_count:
                logging.error(
                    'tenant_id="{}", component="{}", Failed to call trackme_register_tenant_component_summary, permanent failure with exception="{}"'.format(
                        tenant_id, component, str(e)
                    )
                )
                break
            else:
                logging.warn(
                    'tenant_id="{}", component="{}", trackme_register_tenant_component_summary, temporary search failure with exception="{}"'.format(
                        tenant_id, component, str(e)
                    )
                )
                time.sleep(sleep_time)


# Update the dataset cache
def dataset_update_cache(session_key, splunkd_uri, tenant_id, component):
    # set max failed re-attempt
    max_failures_count = 24
    sleep_time = 5

    # parse splunkd_uri
    parsed_url = urllib.parse.urlparse(splunkd_uri)

    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=parsed_url.port,
        token=session_key,
        timeout=300,
    )

    logging.debug('starting cache operation component="{}"'.format(component))

    # simply run this search, the trackmedatasetgen does what we need automatically
    search = (
        "| makeresults\n"
        + '| search [ | trackmedatasetgen tenant_id="'
        + tenant_id
        + '" component="'
        + component
        + '" regex="false" target="allowlist" addprefix="true" store_cache="true" | return $search_str ]'
        + "\n"
        + '| where [ | trackmedatasetgen tenant_id="'
        + tenant_id
        + '" component="'
        + component
        + '" regex="true" target="allowlist" addprefix="true" store_cache="true" | return $search_str ]'
        + "\n"
        + '| search [ | trackmedatasetgen tenant_id="'
        + tenant_id
        + '" component="'
        + component
        + '" regex="false" target="blocklist" addprefix="true" store_cache="true" | return $search_str ]'
        + "\n"
        + '| where [ | trackmedatasetgen tenant_id="'
        + tenant_id
        + '" component="'
        + component
        + '" regex="true" target="blocklist" addprefix="true" store_cache="true" | return $search_str ]'
    )

    # fields variations per component
    if component == "dsm":
        search = (
            search
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="blocklist" addprefix="true" fields="index,sourcetype,object" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="blocklist" addprefix="true" fields="index,sourcetype,object" store_cache="true" | return $search_str ]'
        )

    elif component == "dhm":
        search = search = (
            "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="blocklist" addprefix="true" fields="index,sourcetype,object" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="blocklist" addprefix="true" fields="index,sourcetype,object" store_cache="true" | return $search_str ]'
        )

    elif component == "mhm":
        search = (
            search
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="allowlist" addprefix="true" fields="index" store_cache="true" | return $search_str ]'
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="blocklist" addprefix="true" fields="index,metric_category,object" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="blocklist" addprefix="true" fields="index,metric_category,object" store_cache="true" | return $search_str ]'
            + "\n"
            + '| search [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="false" target="blocklist" addprefix="true" fields="category" store_cache="true" | return $search_str ]'
            + "\n"
            + '| where [ | trackmedatasetgen tenant_id="'
            + tenant_id
            + '" component="'
            + component
            + '" regex="true" target="blocklist" addprefix="true" fields="category" store_cache="true" | return $search_str ]'
        )

    logging.debug('search="{}"'.format(search))

    kwargs_oneshot = {
        "earliest_time": "-5m",
        "latest_time": "now",
        "output_mode": "json",
        "count": 0,
    }

    # automatic retry
    search_successful = False
    current_failures_count = 0
    while current_failures_count < max_failures_count and not search_successful:
        try:
            oneshotsearch_results = service.jobs.oneshot(search, **kwargs_oneshot)
            reader = results.JSONResultsReader(oneshotsearch_results)

            for item in reader:
                if isinstance(item, dict):
                    logging.debug('search_results="{}"'.format(item))
            return "success"

        except Exception as e:
            if re.search(
                "The maximum number of concurrent historical searches on this instance has been reached",
                str(e),
            ):
                current_failures_count += 1
                search_successful = False
                if current_failures_count >= max_failures_count:
                    raise Exception(str(e))
                else:
                    logging.warn(
                        'temporary failure due with exception="{}"'.format(str(e))
                    )
                    time.sleep(sleep_time)
            else:
                raise Exception(str(e))


def trackme_send_to_tcm(session_key, splunkd_uri, resp_dict, http_mode, http_service):
    """
    Send the transaction to TrackMe Configuration Manager
    """

    # Ensure splunkd_uri starts with "https://"
    if not splunkd_uri.startswith("https://"):
        splunkd_uri = f"https://{splunkd_uri}"

    # Build header and target URL
    headers = CaseInsensitiveDict()
    headers["Authorization"] = f"Splunk {session_key}"
    headers["Content-Type"] = "application/json"
    target_url = f"{splunkd_uri}/services/trackme_conf_manager/v1/conf_manager_receiver"

    # Create a requests session for better performance
    session = requests.Session()
    session.headers.update(headers)

    data = {
        "transaction_request": resp_dict,
        "transaction_http_mode": http_mode,
        "transaction_http_service": http_service,
    }

    try:
        # Use a context manager to handle the request
        with session.post(target_url, data=json.dumps(data), verify=False) as response:
            if response.ok:
                logging.debug(
                    f'Success sending the transaction to TCM, data="{response}"'
                )
                response_json = response.json()
                return response_json
            else:
                error_message = f'Failed to send the transaction to TCM, status_code={response.status_code}, response_text="{response.text}"'
                logging.error(error_message)
                raise Exception(error_message)

    except Exception as e:
        error_message = f'Failed to send the transaction to TCM, exception="{str(e)}"'
        logging.error(error_message)
        raise Exception(error_message)
