#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2021-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import os
import sys
import json
import logging
import requests
import urllib3
import re
import time

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk libs
import splunklib.client as client
import splunklib.results as results

# import trackme libs
from trackme_libs import (
    trackme_create_report,
    trackme_delete_report,
    trackme_create_kvtransform,
    trackme_delete_kvtransform,
    trackme_delete_macro,
)

# import trackme libs utils
from trackme_libs_utils import remove_leading_spaces

# import the collections dict
from collections_data import collections_dict

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


# update the schema version to a certain release number
def trackme_schema_get_version(reqinfo, tenant_id, schema_version):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_get_version, tenant_id="{}"'.format(
            tenant_id
        )
    )

    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", trackme_schema_get_version, the vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", trackme_schema_get_version, the vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # try to get the schema version
        try:
            tenant_schema_version = vtenant_record.get("schema_version")
        except Exception as e:
            tenant_schema_version = None

        # update as needed
        if not tenant_schema_version:
            logging.info(
                'context="schema_upgrade", tenant_id="{}", trackme_schema_get_version, this tenant does not have a schema version defined yet, processing now.'.format(
                    tenant_id
                )
            )
            return None
        else:
            logging.info(
                'context="schema_upgrade", tenant_id="{}", trackme_schema_get_version, current schema_version="{}"'.format(
                    tenant_id, tenant_schema_version
                )
            )
            return tenant_schema_version


# update the schema version to a certain release number
def trackme_schema_update_version(reqinfo, tenant_id, schema_version):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_update_vesion, tenant_id="{}"'.format(
            tenant_id
        )
    )

    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # try to get the schema version
        try:
            tenant_schema_version = vtenant_record.get("schema_version")
        except Exception as e:
            tenant_schema_version = None

        # logging debug
        logging.debug(
            'context="schema_upgrade", tenant_id="{}", current schema_version="{}"'.format(
                tenant_id, tenant_schema_version
            )
        )

        # update as needed
        if not tenant_schema_version:
            logging.info(
                'context="schema_upgrade", tenant_id="{}", schema version check, this tenant does not have a schema version defined yet, processing its update now.'.format(
                    tenant_id
                )
            )

            try:
                vtenant_record["schema_version"] = schema_version
                collection_vtenants.data.update(
                    str(vtenant_key), json.dumps(vtenant_record)
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", schema version upgraded successfully, new schema_version="{}"'.format(
                        tenant_id, schema_version
                    )
                )
                return {
                    "action": "success",
                    "response": "schema version was successfully upgraded",
                    "schema_version": schema_version,
                }

            except Exception as e:
                logging.error(
                    'context="schema_upgrade", failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                        str(e)
                    )
                )
                raise Exception(
                    'failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                        str(e)
                    )
                )

        elif tenant_schema_version != schema_version:
            logging.info(
                'context="schema_upgrade", tenant_id="{}", schema version check, this tenant needs to be upgraded, processing its update now.'.format(
                    tenant_id
                )
            )

            try:
                vtenant_record["schema_version"] = schema_version
                collection_vtenants.data.update(
                    str(vtenant_key), json.dumps(vtenant_record)
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", schema version upgraded successfully, new schema_version="{}"'.format(
                        tenant_id, schema_version
                    )
                )
                return {
                    "action": "success",
                    "response": "schema version was successfully upgraded",
                    "schema_version": schema_version,
                }

            except Exception as e:
                logging.error(
                    'context="schema_upgrade", failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                        str(e)
                    )
                )
                raise Exception(
                    'failure while trying to update the vtenant KVstore record, exception="{}"'.format(
                        str(e)
                    )
                )

        else:
            logging.info(
                'context="schema_upgrade", tenant_id="{}", schema version check, this tenant schema is up to date, nothing to do.'.format(
                    tenant_id
                )
            )
            return {
                "action": "success",
                "schema_version": schema_version,
            }


# process update version 2.0.9
# changes: in version 2.0.9, we introduced a new Ack Metadata, ack_type, which requires an update of the KV transform


def trackme_schema_upgrade_209(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_209, tenant_id="{}"'.format(
            tenant_id
        )
    )

    # dict of collections
    collections_dict = dict(
        [
            (
                "trackme_common_alerts_ack",
                "_key, object, object_category, ack_mtime, ack_expiration, ack_state, ack_type, ack_comment",
            ),
        ]
    )

    # name of the object
    object_name = "trackme_common_alerts_ack"

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        transform_name = "trackme_common_alerts_ack_tenant_" + str(tenant_id)
        collection_name = "kv_trackme_common_alerts_ack_tenant_" + str(tenant_id)
        transform_fields = collections_dict[object_name]
        transform_acl = {
            "owner": vtenant_record.get("tenant_owner"),
            "sharing": "app",
            "perms.write": vtenant_record.get("tenant_roles_admin"),
            "perms.read": vtenant_record.get("tenant_roles_user"),
        }

        # delete the transform
        try:
            action_delete = trackme_delete_kvtransform(
                reqinfo["session_key"],
                reqinfo["server_rest_uri"],
                tenant_id,
                transform_name,
            )
            logging.info(
                'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                    tenant_id, transform_name
                )
            )
        except Exception as e:
            logging.error(
                'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                    tenant_id, transform_name, str(e)
                )
            )
            raise Exception(
                'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                    tenant_id, transform_name, str(e)
                )
            )

        # create the transform
        try:
            action_create = trackme_create_kvtransform(
                reqinfo["session_key"],
                reqinfo["server_rest_uri"],
                tenant_id,
                transform_name,
                transform_fields,
                collection_name,
                vtenant_record.get("tenant_owner"),
                transform_acl,
            )
            logging.info(
                'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                    tenant_id, transform_name
                )
            )
        except Exception as e:
            logging.error(
                'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                    tenant_id, transform_name, str(e)
                )
            )
            raise Exception(
                'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                    tenant_id, transform_name, str(e)
                )
            )


def trackme_schema_upgrade_2015(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2015, tenant_id="{}"'.format(
            tenant_id
        )
    )

    # dict of collections
    collections_dict = dict(
        [
            (
                "trackme_dsm",
                "_key, mtime, search_mode, tenant_id, object_category, alias, data_index, data_last_lag_seen, data_last_ingestion_lag_seen, data_eventcount, data_last_lag_seen_idx, data_first_time_seen, data_last_time_seen, data_last_ingest, data_last_time_seen_idx, data_max_lag_allowed, data_max_delay_allowed, data_lag_alert_kpis, monitored_state, object, data_sourcetype, data_monitoring_wdays, isUnderMonitoringDays, data_monitoring_hours_ranges, isUnderMonitoringHours, data_override_lagging_class, object_state, tracker_runtime, object_previous_state, data_previous_tracker_runtime, dcount_host, min_dcount_host, isAnomaly, data_sample_lastrun, tags, latest_flip_state, latest_flip_time, priority, status_message, anomaly_reason",
            ),
            (
                "trackme_cim",
                "_key, mtime, tenant_id, object, alias, monitored_state, cim_tracking_rules, cim_tracking_results, object_state, account, tracker_name, tracker_runtime, tracker_last_duration, latest_flip_state, latest_flip_time, priority, anomaly_reason",
            ),
            (
                "trackme_flx",
                "_key, mtime, tenant_id, group, object, object_category, object_state, object_description, alias, monitored_state, account, tracker_name, tracker_runtime, status, status_description, metrics, outliers_metrics, monitoring_wdays, isUnderMonitoringDays, monitoring_hours_ranges, isUnderMonitoringHours, latest_flip_state, latest_flip_time, priority, status_message, anomaly_reason",
            ),
            (
                "trackme_dhm",
                "_key, mtime, search_mode, tenant_id, object_category, object, alias, data_index, data_sourcetype, data_last_lag_seen, data_last_ingestion_lag_seen, data_eventcount, data_first_time_seen, data_last_time_seen, data_last_ingest, data_max_lag_allowed, data_max_delay_allowed, data_lag_alert_kpis, monitored_state, data_monitoring_wdays, isUnderMonitoringDays, data_monitoring_hours_ranges, isUnderMonitoringHours, data_override_lagging_class, object_state, tracker_runtime, object_previous_state, data_previous_tracker_runtime, splk_dhm_st_summary, splk_dhm_alerting_policy, latest_flip_state, latest_flip_time, priority, status_message, anomaly_reason",
            ),
            (
                "trackme_mhm",
                "_key, mtime, tenant_id, object_category, object, alias, metric_index, metric_category, metric_details, metric_last_lag_seen, metric_first_time_seen, metric_last_time_seen, metric_max_lag_allowed, monitored_state, metric_monitoring_wdays, metric_override_lagging_class, object_state, tracker_runtime, object_previous_state, metric_previous_tracker_runtime, latest_flip_state, latest_flip_time, priority, status_message, anomaly_reason",
            ),
        ]
    )

    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm")

        if vtenant_record.get("tenant_mhm_enabled") == 1:
            objects_to_process.append("trackme_mhm")

        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx")

        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append("trackme_cim")

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": vtenant_record.get("tenant_roles_admin"),
                "perms.read": vtenant_record.get("tenant_roles_user"),
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )


def trackme_schema_upgrade_2016(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2016, tenant_id="{}"'.format(
            tenant_id
        )
    )

    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append(
                "trackme_dsm_outliers_mltrain_tracker_tenant_%s" % tenant_id
            )

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append(
                "trackme_dhm_outliers_mltrain_tracker_tenant_%s" % tenant_id
            )

        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append(
                "trackme_flx_outliers_mltrain_tracker_tenant_%s" % tenant_id
            )

        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append(
                "trackme_cim_outliers_mltrain_tracker_tenant_%s" % tenant_id
            )

        for report_name in objects_to_process:
            # create the report
            component = None

            if report_name.startswith("trackme_dsm"):
                component = "dsm"
            elif report_name.startswith("trackme_dhm"):
                component = "dhm"
            elif report_name.startswith("trackme_flx"):
                component = "flx"
            elif report_name.startswith("trackme_cim"):
                component = "cim"

            report_search = (
                '| trackmesplkoutlierstrainhelper tenant_id="%s" component="%s"'
                % (tenant_id, component)
            )
            report_properties = {
                "description": "This scheduled report generate and trains Machine Learning models for the tenant",
                "is_scheduled": True,
                "schedule_window": "auto",
                "cron_schedule": "*/60 * * * *",
                "dispatch.earliest_time": "-5m",
                "dispatch.latest_time": "now",
                "schedule_window": "5",
            }

            # for read permissions, concatenate admin and guest
            tenant_roles_read_perms = "%s,%s" % (
                str(vtenant_record.get("tenant_roles_admin")),
                str(vtenant_record.get("tenant_roles_user")),
            )

            # delete the report
            try:
                action = trackme_delete_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                )
                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "delete",
                    "result": "success",
                }
                logging.info(
                    'context="schema_upgrade", schema migration 2016 report deleted successfully, result="{}"'.format(
                        result, indent=2
                    )
                )

            except Exception as e:
                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "delete",
                    "result": "failure",
                    "exception": str(e),
                }
                logging.info(
                    'context="schema_upgrade", schema migration 2016, report failed to be deleted, result="{}"'.format(
                        result, indent=2
                    )
                )

                # stop here
                break

            # create the report
            try:
                report_acl = {
                    "owner": str(vtenant_record.get("tenant_owner")),
                    "sharing": "app",
                    "perms.write": str(vtenant_record.get("tenant_roles_admin")),
                    "perms.read": str(tenant_roles_read_perms),
                }

                action = trackme_create_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                    report_search,
                    report_properties,
                    report_acl,
                )

                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "success",
                }

                logging.info(
                    'context="schema_upgrade", schema migration 2016 report created successfully, result="{}"'.format(
                        result, indent=2
                    )
                )

            except Exception as e:
                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "failure",
                    "exception": str(e),
                }

                logging.info(
                    'context="schema_upgrade", schema migration 2016 report failed to be created, result="{}"'.format(
                        result, indent=2
                    )
                )


def trackme_schema_upgrade_2020(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2020, tenant_id="{}"'.format(
            tenant_id
        )
    )

    components_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            components_to_process.append("dsm")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            components_to_process.append("dhm")

        # loop
        for component in components_to_process:
            # for read permissions, concatenate admin and guest
            tenant_roles_read_perms = "%s,%s" % (
                str(vtenant_record.get("tenant_roles_admin")),
                str(vtenant_record.get("tenant_roles_user")),
            )

            report_acl = {
                "owner": str(vtenant_record.get("tenant_owner")),
                "sharing": "app",
                "perms.write": str(vtenant_record.get("tenant_roles_admin")),
                "perms.read": str(tenant_roles_read_perms),
            }

            # create the wrapper
            report_name = (
                f"trackme_{component}_delayed_entities_tracker_tenant_{tenant_id}"
            )
            if component in ("dsm"):
                report_search = f'| trackmesplkfeedsdelayed tenant_id="{tenant_id}" component="{component}" earliest="-24h" latest="+4h" max_runtime_sec="3600" max_ingest_age_sec="86400" min_time_since_inspection_sec="900" dsm_tstats_root_breakby_include_splunk_server=True dsm_tstats_root_breakby_include_host=True'
            elif component in ("dhm"):
                report_search = f'| trackmesplkfeedsdelayed tenant_id="{tenant_id}" component="{component}" earliest="-24h" latest="+4h" max_runtime_sec="3600" max_ingest_age_sec="86400" min_time_since_inspection_sec="900" dhm_tstats_root_breakby_include_splunk_server=True'
            report_properties = {
                "description": "This scheduled report handled delayed entities for the tenant and component",
                "is_scheduled": True,
                "schedule_window": "auto",
                "cron_schedule": "*/60 * * * *",
                "dispatch.earliest_time": "-5m",
                "dispatch.latest_time": "now",
                "schedule_window": "5",
            }

            try:
                report_acl = {
                    "owner": str(vtenant_record.get("tenant_owner")),
                    "sharing": "app",
                    "perms.write": str(vtenant_record.get("tenant_roles_admin")),
                    "perms.read": str(tenant_roles_read_perms),
                }

                action = trackme_create_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                    report_search,
                    report_properties,
                    report_acl,
                )

                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "success",
                }

            except Exception as e:
                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "failure",
                    "exception": str(e),
                }

        logging.info(
            'context="schema_upgrade", schema migration 2020 terminated, result="{}"'.format(
                result, indent=2
            )
        )

        return result


def trackme_schema_upgrade_2026(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2026, tenant_id="{}"'.format(
            tenant_id
        )
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # init the wlk component to be disabled
        vtenant_record["tenant_wlk_enabled"] = 0

        # update
        try:
            collection_vtenants.data.update(
                str(vtenant_key), json.dumps(vtenant_record)
            )
            logging.info(
                'context="schema_upgrade", schema migration 2026, Virtual Tenant updated successfully, result="{}"'.format(
                    json.dumps(vtenant_record, indent=2)
                )
            )
            return "success"

        except Exception as e:
            error_msg = 'context="schema_upgrade", function trackme_schema_upgrade_2026, failed to update the Virtual Tenant record, exception="{}"'.format(
                str(e)
            )
            logging.error(error_msg)
            raise Exception(error_msg)


def trackme_schema_upgrade_2034(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2034, tenant_id="{}"'.format(
            tenant_id
        )
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # first, introduce the tenant_replica boolean and update the tenant record
        vtenant_record["tenant_replica"] = False
        try:
            collection_vtenants.data.update(
                str(vtenant_key), json.dumps(vtenant_record)
            )
            logging.info(
                'context="schema_upgrade", vtenant record was updated successfully, tenant_replica was set to False'
            )
        except Exception as e:
            logging.error(
                'context="schema_upgrade", vtenant record update has failed, tenant_replica has not been set, exception="{}"'.format(
                    str(e)
                )
            )

        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm")

        if vtenant_record.get("tenant_mhm_enabled") == 1:
            objects_to_process.append("trackme_mhm")

        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx")

        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk")

        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append("trackme_cim")

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": vtenant_record.get("tenant_roles_admin"),
                "perms.read": vtenant_record.get("tenant_roles_user"),
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            'context="schema_upgrade", function trackme_schema_upgrade_2034 terminated, tenant_id="{}"'.format(
                tenant_id
            )
        )


def trackme_schema_upgrade_2034_least_privileges(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2034_least_privileges, tenant_id="{}"'.format(
            tenant_id
        )
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # get the admin value
        try:
            current_tenant_roles_admin = vtenant_record["tenant_roles_admin"]
        except Exception as e:
            current_tenant_roles_admin = "trackme_admin"

        # set a value
        tenant_roles_power = None
        if current_tenant_roles_admin != "trackme_admin":
            tenant_roles_power = current_tenant_roles_admin
        else:
            tenant_roles_power = "trackme_power"

        # update the record
        vtenant_record["tenant_roles_power"] = tenant_roles_power

        # update
        try:
            collection_vtenants.data.update(
                str(vtenant_key), json.dumps(vtenant_record)
            )
            logging.info(
                'context="schema_upgrade", schema migration 2034 (least privileges), Virtual Tenant updated successfully, result="{}"'.format(
                    json.dumps(vtenant_record, indent=2)
                )
            )

        except Exception as e:
            error_msg = 'context="schema_upgrade", function trackme_schema_upgrade_2034_leat_privileges, failed to update the Virtual Tenant record, exception="{}"'.format(
                str(e)
            )
            logging.error(error_msg)
            raise Exception(error_msg)

        # RBAC update

        # Define an header for requests authenticated communications with splunkd
        header = {
            "Authorization": "Splunk %s" % reqinfo["session_key"],
            "Content-Type": "application/json",
        }

        url = "%s/services/trackme/v2/vtenants/admin/update_tenant_rbac" % (
            reqinfo["server_rest_uri"]
        )
        data = {
            "tenant_id": tenant_id,
            "tenant_roles_admin": vtenant_record["tenant_roles_admin"],
            "tenant_roles_power": vtenant_record["tenant_roles_power"],
            "tenant_roles_user": vtenant_record["tenant_roles_user"],
            "tenant_owner": vtenant_record["tenant_owner"],
        }

        # create the account
        try:
            response = requests.post(
                url,
                headers=header,
                data=json.dumps(data, indent=1),
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.error(
                    'context="schema_upgrade", schema migration 2034 (least privileges), RBAC update has failed, response.status_code="{}", response.text="{}"'.format(
                        response.status_code, response.text
                    )
                )
            else:
                logging.info(
                    'context="schema_upgrade", schema migration 2034 (least privileges), RBAC update) was operated successfully, response="{}"'.format(
                        json.dumps(response.json(), indent=2)
                    )
                )
        except Exception as e:
            logging.error(
                'context="schema_upgrade", schema migration 2034 (least privileges), RBAC update has failed, exception="{}"'.format(
                    str(e)
                )
            )


def trackme_schema_upgrade_2036(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2036, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Create the vtenant account configuration
        #

        url = f'{reqinfo["server_rest_uri"]}/servicesNS/nobody/trackme/trackme_vtenants'
        data = {
            "name": vtenant_record.get("tenant_id"),
            "description": vtenant_record.get("tenant_desc"),
            "ui_min_object_width": 300,
            "data_sampling_set_state": 1,
            "outliers_set_state": 1,
            "data_sampling_obfuscation": 0,
        }

        # create the account
        try:
            response = requests.post(
                url,
                headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                data=json.dumps(data),
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.error(
                    f'context="schema_upgrade", schema migration 2036, create vtenant account has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                )
            else:
                logging.info(
                    f'context="schema_upgrade", schema migration 2036, create vtenant account was operated successfully, response.status_code="{response.status_code}"'
                )
        except Exception as e:
            logging.error(
                f'context="schema_upgrade", schema migration 2036, create vtenant account has failed, exception="{str(e)}"'
            )

        #
        # Decomission of the data sampling obfuscation macro
        #

        if vtenant_record.get("tenant_dsm_enabled") == 1:
            # for read permissions, concatenate admin, power and user
            tenant_roles_read_perms = f"{vtenant_record.get('tenant_roles_admin')},{vtenant_record.get('tenant_roles_power')},{vtenant_record.get('tenant_roles_user')}"

            # for write permissions, concatenate admin, power
            tenant_roles_write_perms = f"{vtenant_record.get('tenant_roles_admin')},{vtenant_record.get('tenant_roles_power')}"

            # deletion of the sampling tracker
            report_name = f"trackme_dsm_data_sampling_tracker_tenant_{tenant_id}"

            # do not raise an exception shall we fail
            try:
                action_delete = trackme_delete_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully deleted report definition, report="{report_name}"'
                )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", failed to delete the report definition, macro="{report_name}", exception="{str(e)}"'
                )

            # attempt to create the new fresh tracker
            report_search = (
                f'| trackmesamplingexecutor tenant_id="{tenant_id}"'
                + "\n``` To avoid reaching KVstore document max size, limit the content of raw events to be stored ```"
                + f"\n| eval raw_sample=substr(raw_sample, 1, 50000)"
                + "\n``` Call the data sampling engine macro ```"
                + f"\n| `trackme_data_sampling_engine({tenant_id})`"
                + "\n``` Apply data obfuscation mode ```"
                + f"\n| `trackme_data_sampling_set_obfuscation_mode({tenant_id})`"
                + "\n``` Update the KVstore content ```"
                + f"\n| outputlookup append=t key_field=key trackme_dsm_data_sampling_tenant_{tenant_id}"
                + "\n| stats c by object"
            )
            report_properties = {
                "description": "TrackMe DSM Data Sampling tracker",
                "is_scheduled": True,
                "schedule_window": "auto",
                "cron_schedule": "*/20 * * * *",
                "dispatch.earliest_time": "-24h",
                "dispatch.latest_time": "-4h",
            }
            try:
                report_acl = {
                    "owner": vtenant_record.get("tenant_owner"),
                    "sharing": "app",
                    "perms.write": str(tenant_roles_write_perms),
                    "perms.read": str(tenant_roles_read_perms),
                }
                trackme_create_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                    report_search,
                    report_properties,
                    report_acl,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully created the report definition, report="{report_name}"'
                )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", failed to create the report definition, report="{report_name}", exception="{str(e)}"'
                )

            # decomission the macro
            macro_name = f"trackme_dsm_data_sampling_obfuscation_tenant_{tenant_id}"
            # do not raise an exception shall we fail
            try:
                action_delete = trackme_delete_macro(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    macro_name,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully deleted macro definition, macro="{macro_name}"'
                )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", failed to delete the macro definition, macro="{macro_name}", exception="{str(e)}"'
                )

        #
        # Replica tenant
        #

        # then, introduce the tenant_replica boolean and update the tenant record
        vtenant_record["tenant_replica"] = False
        try:
            collection_vtenants.data.update(
                str(vtenant_key), json.dumps(vtenant_record)
            )
            logging.info(
                'context="schema_upgrade", vtenant record was updated successfully, tenant_replica was set to False'
            )
        except Exception as e:
            logging.error(
                f'context="schema_upgrade", vtenant record update has failed, tenant_replica has not been set, exception="{str(e)}"'
            )

        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm")

        if vtenant_record.get("tenant_mhm_enabled") == 1:
            objects_to_process.append("trackme_mhm")

        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx")

        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk")

        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append("trackme_cim")

        for object_name in objects_to_process:
            transform_name = f"{object_name}_tenant_{tenant_id}"
            collection_name = f"kv_{object_name}_tenant_{tenant_id}"
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": vtenant_record.get("tenant_roles_admin"),
                "perms.read": vtenant_record.get("tenant_roles_user"),
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully deleted transform definition, transform="{transform_name}"'
                )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )
                raise Exception(
                    f'tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully created transform definition, transform="{transform_name}"'
                )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )
                raise Exception(
                    f'tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2036 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2038(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2038, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Vtenant account preferences
        #

        #
        # Update the vtenant account configuration
        #
        tenant_id = vtenant_record.get("tenant_id")

        url = f'{reqinfo["server_rest_uri"]}/servicesNS/nobody/trackme/trackme_vtenants/{tenant_id}'
        data = {
            "ui_expand_metrics": 0,
        }

        # update the account
        try:
            response = requests.post(
                url,
                headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                data=data,
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.error(
                    f'context="schema_upgrade", schema migration 2038, update vtenant account has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                )
            else:
                logging.info(
                    f'context="schema_upgrade", schema migration 2038, update vtenant account was operated successfully, response.status_code="{response.status_code}"'
                )
        except Exception as e:
            logging.error(
                f'context="schema_upgrade", schema migration 2038, update vtenant account has failed, exception="{str(e)}"'
            )

        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm")

        if vtenant_record.get("tenant_mhm_enabled") == 1:
            objects_to_process.append("trackme_mhm")

        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2038 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2043(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2043, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm_outliers_entity_rules")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm_outliers_entity_rules")

        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append("trackme_cim_outliers_entity_rules")

        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx_outliers_entity_rules")

        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk_outliers_entity_rules")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2043 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2044(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2044, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2044 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2045(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2045, tenant_id="{}"'.format(
            tenant_id
        )
    )

    components_to_process = []
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_flx_enabled") == 1:
            components_to_process.append("flx")
            objects_to_process.append("trackme_flx")

        #
        # objects
        #

        for object_name in objects_to_process:
            # get RBAC
            tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
            tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
            tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

            # for read permissions, concatenate admin, power and user
            tenant_roles_read_perms = (
                f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
            )

            # for write permissions, concatenate admin, power
            tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        #
        # components
        #

        for component in components_to_process:
            # get RBAC
            tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
            tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
            tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

            # for read permissions, concatenate admin, power and user
            tenant_roles_read_perms = (
                f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
            )

            # for write permissions, concatenate admin, power
            tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

            report_acl = {
                "owner": str(vtenant_record.get("tenant_owner")),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # create the wrapper
            report_name = (
                f"trackme_{component}_inactive_entities_tracker_tenant_{tenant_id}"
            )
            report_search = f'| trackmesplkflxinactiveinspector tenant_id="{tenant_id}" register_component="True" report="{report_name}" max_sec_since_inactivity_before_status_update=3600 max_days_since_inactivity_before_purge=7'
            report_properties = {
                "description": "This scheduled report handles inactive entities for the splk-flx component",
                "is_scheduled": True,
                "schedule_window": "auto",
                "cron_schedule": "*/15 * * * *",
                "dispatch.earliest_time": "-5m",
                "dispatch.latest_time": "now",
                "schedule_window": "5",
            }

            try:
                report_acl = {
                    "owner": str(vtenant_record.get("tenant_owner")),
                    "sharing": "app",
                    "perms.write": str(vtenant_record.get("tenant_roles_admin")),
                    "perms.read": str(tenant_roles_read_perms),
                }

                action = trackme_create_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                    report_search,
                    report_properties,
                    report_acl,
                )

                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "success",
                }

            except Exception as e:
                result = {
                    "object": report_name,
                    "object_type": "report",
                    "action": "create",
                    "result": "failure",
                    "exception": str(e),
                }

        logging.info(
            'context="schema_upgrade", schema migration 2045 terminated, result="{}"'.format(
                result, indent=2
            )
        )

        return result


def trackme_schema_upgrade_2054(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2054, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2054 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2064(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2064, tenant_id="{}"'.format(
            tenant_id
        )
    )

    components_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            components_to_process.append("dsm")

        if vtenant_record.get("tenant_dhm_enabled") == 1:
            components_to_process.append("dhm")

        # Retrieve the exec summary record
        tenant_objects_exec_summary = vtenant_record.get("tenant_objects_exec_summary")

        # load tenant_objects_exec_summary as an object
        try:
            tenant_objects_exec_summary = json.loads(tenant_objects_exec_summary)
        except Exception as e:
            tenant_objects_exec_summary = {}

        # results
        results = []

        # loop
        for component in components_to_process:
            # create the wrapper
            report_name = (
                f"trackme_{component}_delayed_entities_tracker_tenant_{tenant_id}"
            )

            # do not raise an exception shall we fail
            try:
                action_delete = trackme_delete_report(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    report_name,
                )
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", successfully deleted report definition, report="{report_name}"'
                )
                results.append(
                    f'tenant_id="{tenant_id}", successfully deleted report definition, report="{report_name}"'
                )

                # attempt to clear the tenant_objects_exec_summary record from this object
                try:
                    tenant_objects_exec_summary.pop(report_name)
                    vtenant_record["tenant_objects_exec_summary"] = json.dumps(
                        tenant_objects_exec_summary, indent=2
                    )
                    collection_vtenants.data.update(
                        str(vtenant_key), json.dumps(vtenant_record)
                    )
                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", successfully cleared the tenant_objects_exec_summary record, report="{report_name}"'
                    )
                    results.append(
                        f'tenant_id="{tenant_id}", successfully cleared the tenant_objects_exec_summary record, report="{report_name}"'
                    )
                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", failed to clear the tenant_objects_exec_summary record, report="{report_name}", exception="{str(e)}"'
                    )
                    results.append(
                        f'tenant_id="{tenant_id}", failed to clear the tenant_objects_exec_summary record, report="{report_name}", exception="{str(e)}"'
                    )

            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", context="schema_upgrade", failed to delete the report definition, macro="{report_name}", exception="{str(e)}"'
                )

        logging.info(
            f'context="schema_upgrade", schema migration 2064 terminated, results="{json.dumps(results, indent=2)}"'
        )

        return results


def trackme_schema_upgrade_2067(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        'context="schema_upgrade", Starting function trackme_schema_upgrade_2067, tenant_id="{}"'.format(
            tenant_id
        )
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{}"'.format(
                query_string
            )
        )

    if vtenant_key:
        # check components and add accordingly
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            process_schema_upgrade = True
        else:
            process_schema_upgrade = False

        if process_schema_upgrade:
            # Define the query
            search = f'| inputlookup trackme_wlk_tenant_{tenant_id} | eval key = _key | where match(object, "\\\\\\\\") | table key, object'

            kwargs_oneshot = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # set max failed re-attempt
            max_failures_count = 5
            sleep_time = 5

            # A list to store all executions
            results_list = []

            # A dict to store the results
            results_dict = {}

            # automatic retry
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                try:
                    oneshotsearch_results = service.jobs.export(
                        search, **kwargs_oneshot
                    )
                    reader = results.JSONResultsReader(oneshotsearch_results)

                    for item in reader:
                        if isinstance(item, dict):
                            results_dict[item["key"]] = item["object"]
                            logging.info(
                                f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", detected non ASCII entity to be purged, object="{item["object"]}", key="{item["key"]}"'
                            )

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            logging.error(
                                f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", could not verify the presence of non ASCII entities, search permanently failed, exception="{str(e)}"'
                            )
                            break

                        else:
                            logging.error(
                                f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", could not verify the presence of non ASCII entities, search temporarily failed, exception="{str(e)}"'
                            )
                            time.sleep(sleep_time)
                    else:
                        logging.error(
                            f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", could not verify the presence of non ASCII entities, search permanently failed, exception="{str(e)}"'
                        )
                        break

            # if the results_dict is not empty, we need to purge the entities
            if results_dict:
                # Data collection
                collection_name = "kv_trackme_wlk_tenant_" + str(tenant_id)
                collection = service.kvstore[collection_name]

                # loop through the results dict and remove any entity from the KVstore
                for key, object in results_dict.items():
                    try:
                        collection.data.delete(json.dumps({"_key": key}))
                        logging.info(
                            f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", successfully purged non ASCII entity, object="{object}", key="{key}"'
                        )
                        results.append(
                            {
                                "object": object,
                                "key": key,
                                "result": "successfully purged non ASCII entity",
                            }
                        )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", failed to purge non ASCII entity, object="{object}", key="{key}", exception="{str(e)}"'
                        )
                        results_list.append(
                            {
                                "object": object,
                                "key": key,
                                "result": "failure to purge non ASCII entity",
                                "exception": str(e),
                            }
                        )

            # log a message if there were no entities to be purged
            else:
                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", no non ASCII entities to be purged'
                )

        logging.info(
            f'context="schema_upgrade", schema migration 2067 terminated, results="{json.dumps(results_list, indent=2)}"'
        )

        return results_list


def trackme_schema_upgrade_2070(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2070, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2070 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2071(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2071, tenant_id="{tenant_id}"'
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", processing with the schema upgrade'
            )

            # retrieve the list of scheduler reports to be processed
            collection_trackers_name = (
                f"kv_trackme_wlk_hybrid_trackers_tenant_{tenant_id}"
            )
            collection_trackers = service.kvstore[collection_trackers_name]

            trackers_dict = {}

            trackers_to_process = []
            trackers_to_process_kos = []

            # get records from the KVstore
            trackers_records = collection_trackers.data.query()

            for tracker_record in trackers_records:
                tracker_name = tracker_record.get("tracker_name")
                tracker_kos = tracker_record.get("knowledge_objects")

                # if the tracker_name starts by scheduler_
                if re.search("^scheduler_", tracker_name):
                    trackers_to_process.append(tracker_name)
                    trackers_to_process_kos.append(tracker_kos)

                    trackers_dict[tracker_name] = {
                        "knowledge_objects": json.loads(tracker_kos),
                    }

            # loop through the trackers
            for tracker_shortname in trackers_to_process:
                tracker_name = (
                    f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"
                )

                tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                # get the current search definition
                tracker_current = service.saved_searches[tracker_name]
                tracker_current_search = tracker_current.content.get("search")
                tracker_account = tracker_kos["properties"][0]["account"]

                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                )

                # in tracker_current_search, replace the sentence:
                # | fields - collection_app, collection_user
                # with: (note that account need to replaced by the actual account name)
                # | fields - collection_app, collection_user
                # ``` Verify the report owner from rest if it is not yet known to TrackMe ```
                # | trackmesplkwlkgetreportowner account="{account}"

                try:
                    tracker_new_search = re.sub(
                        r"\| fields - collection_app, collection_user",
                        f'| fields - collection_app, collection_user\n\n``` Verify the report owner from rest if it is not yet known to TrackMe, this custom command will perform a get Metadata operation attempt if necessary ```\n| trackmesplkwlkgetreportowner account="{tracker_account}"\n',
                        tracker_current_search,
                    )

                    # update the search definition
                    tracker_current.update(search=tracker_new_search)

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, the tracker was updated successfully, new_search="{tracker_new_search}"'
                    )

                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, failed to update the tracker, exception="{str(e)}"'
                    )

        # check components and add accordingly
        objects_to_process = []
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2071 terminated, tenant_id="{tenant_id}"'
        )


"""
In this function:
- A bug had affected the definition of outliers rules for wlk, activating lower bound breaches for elapsed which is unwanted
- Enhance the scheduler execution errors detection by taking into account searches in delegated errors
- Update transforms for dsm/dhm for the purposes of allow_adaptive_delay
- Update ML Outliers rules transforms to add the confidence and confidence_reason fields
- Initialize ML Outliers values for confidence and confidence_reason
"""


def trackme_schema_upgrade_2072(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2072, tenant_id="{tenant_id}"'
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        #
        # prepare transforms updates
        #

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        # TrackMe sharing level
        trackme_default_sharing = reqinfo["trackme_conf"]["trackme_general"][
            "trackme_default_sharing"
        ]

        #
        # main transforms updates
        #

        # check components and add accordingly
        objects_to_process = []
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm")
        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm")

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": trackme_default_sharing,
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/delete_kvtransform'
            data = {
                "tenant_id": tenant_id,
                "transform_name": transform_name,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to delete the transform definition, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully deleted transform definition, transform="{transform_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

            # create the transform
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/create_kvtransform'
            data = {
                "tenant_id": tenant_id,
                "transform_name": transform_name,
                "transform_fields": transform_fields,
                "collection_name": collection_name,
                "transform_acl": transform_acl,
                "owner": vtenant_record.get("tenant_owner"),
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", create transform definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully created transform definition, transform="{transform_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

        #
        # ML rules outliers transforms update
        #

        # check components and add accordingly
        objects_to_process = []
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            objects_to_process.append("trackme_dsm_outliers_entity_rules")
        if vtenant_record.get("tenant_dhm_enabled") == 1:
            objects_to_process.append("trackme_dhm_outliers_entity_rules")
        if vtenant_record.get("tenant_flx_enabled") == 1:
            objects_to_process.append("trackme_flx_outliers_entity_rules")
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            objects_to_process.append("trackme_wlk_outliers_entity_rules")

        for object_name in objects_to_process:
            transform_name = f"{object_name}_tenant_{tenant_id}"
            collection_name = f"kv_{object_name}_tenant_{tenant_id}"
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": trackme_default_sharing,
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/delete_kvtransform'
            data = {
                "tenant_id": tenant_id,
                "transform_name": transform_name,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to delete the transform definition, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully deleted transform definition, transform="{transform_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

            # create the transform
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/create_kvtransform'
            data = {
                "tenant_id": tenant_id,
                "transform_name": transform_name,
                "transform_fields": transform_fields,
                "collection_name": collection_name,
                "transform_acl": transform_acl,
                "owner": vtenant_record.get("tenant_owner"),
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", create transform definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully created transform definition, transform="{transform_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
                )

        #
        # ML: initiate confidence and confidence_reason fields for existing tenants
        #

        components_to_process = []
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            components_to_process.append("dsm")
        if vtenant_record.get("tenant_dhm_enabled") == 1:
            components_to_process.append("dhm")
        if vtenant_record.get("tenant_flx_enabled") == 1:
            components_to_process.append("flx")
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            components_to_process.append("wlk")

        # Get the value for splk_outliers_min_days_history
        splk_outliers_min_days_history = reqinfo["trackme_conf"][
            "splk_outliers_detection"
        ]["splk_outliers_min_days_history"]

        for component in components_to_process:
            # set kwargs
            kwargs_confidence = {
                "earliest_time": "-90d",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # define the gen search
            metric_root = None
            if component in ("dsm", "dhm"):
                metric_root = f"trackme.splk.feeds"
            else:
                metric_root = f"trackme.splk.{component}"

            ml_confidence_search = remove_leading_spaces(
                f"""\
                | mstats latest({metric_root}.*) as * where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" object_category="splk-{component}" object=* by object span=1d
                | stats min(_time) as first_time by object
                | eval metrics_duration=now()-first_time
                | eval days_required = {splk_outliers_min_days_history}
                | eval confidence=if(metrics_duration<(days_required*86400), "low", "normal")
                | eval metrics_duration=tostring(metrics_duration, "duration")
                | eval confidence_reason=if(metrics_duration<(days_required*86400), "ML has insufficient historical metrics to proceed (metrics_duration=" . metrics_duration . ", required=" . days_required + "days)", "ML has sufficient historical metrics to proceed (metrics_duration=" . metrics_duration . ", required=" . days_required . "days)")
                | fields object, confidence, confidence_reason
                | lookup local=t trackme_{component}_outliers_entity_rules_tenant_{tenant_id} object OUTPUT _key as keyid, entities_outliers, is_disabled, last_exec, mtime, object_category | where (isnotnull(keyid) AND isnotnull(confidence))
                | outputlookup append=t key_field=keyid trackme_{component}_outliers_entity_rules_tenant_{tenant_id}
                """
            )

            ml_confidence_update_counter = 0
            ml_confidence_updated_entities = []

            # automatic retry
            max_failures_count = 24
            sleep_time = 5
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                # track the search runtime
                start = time.time()

                # proceed
                try:
                    ml_confidence_results = service.jobs.export(
                        ml_confidence_search, **kwargs_confidence
                    )
                    reader = results.JSONResultsReader(ml_confidence_results)

                    for item in reader:
                        if isinstance(item, dict):
                            ml_confidence_update_counter += 1
                            ml_confidence_updated_entities.append(item.get("object"))

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                tenant_id,
                                component,
                                current_failures_count,
                                max_failures_count,
                                str(e),
                            )
                            logging.error(msg)
                            raise Exception(msg)
                        else:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", temporary search failure with exception="{}", run_time="{}"'.format(
                                tenant_id,
                                component,
                                str(e),
                                str(time.time() - start),
                            )
                            logging.warn(msg)
                            time.sleep(sleep_time)
                    else:
                        msg = 'context="schema_upgrade", tenant_id="{}", component="{}", search failed with exception="{}", run_time="{}"'.format(
                            tenant_id,
                            component,
                            str(e),
                            str(time.time() - start),
                        )
                        logging.error(msg)
                        raise Exception(msg)

            # log
            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", component="{component}", ml_confidence_update_counter="{ml_confidence_update_counter}", ml_confidence_updated_entities="{json.dumps(ml_confidence_updated_entities, indent=2)}"'
            )

        #
        # Adaptive delay tracker
        #

        components_to_process = []
        if vtenant_record.get("tenant_dsm_enabled") == 1:
            components_to_process.append("dsm")
        if vtenant_record.get("tenant_dhm_enabled") == 1:
            components_to_process.append("dhm")

        for component in components_to_process:
            #
            # Create the adaptive delay tracker
            #

            report_acl = {
                "owner": str(vtenant_record.get("tenant_owner")),
                "sharing": trackme_default_sharing,
                "perms.write": str(vtenant_record.get("tenant_roles_admin")),
                "perms.read": str(tenant_roles_read_perms),
            }

            # create the wrapper
            report_name = (
                f"trackme_{component}_adaptive_delay_tracker_tenant_{tenant_id}"
            )
            report_search = f'| trackmesplkadaptivedelay tenant_id="{tenant_id}" component="{component}" min_delay_sec=3600 min_historical_metrics_days=7 earliest_time_mstats="-30d" max_runtime=900'
            report_properties = {
                "description": "This scheduled report manages adaptive delay tresholds for TrackMe",
                "is_scheduled": True,
                "schedule_window": "auto",
                "cron_schedule": "*/15 * * * *",
                "dispatch.earliest_time": "-5m",
                "dispatch.latest_time": "now",
                "schedule_window": "5",
            }

            # create the report
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/create_report'
            data = {
                "tenant_id": tenant_id,
                "report_name": report_name,
                "report_search": report_search,
                "report_properties": report_properties,
                "report_acl": report_acl,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", create report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully created report definition, report="{report_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to create the report definition, report="{report_name}", exception="{str(e)}"'
                )

        #
        # Workload
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_wlk_enabled") == 1:
            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", processing with the schema upgrade'
            )

            #
            # fix outliers rules
            #

            # access the KV
            collection_outliers_rules_name = (
                f"kv_trackme_wlk_outliers_entity_rules_tenant_{tenant_id}"
            )
            collection_outliers_rules = service.kvstore[collection_outliers_rules_name]

            # get all records
            collection_records = []
            collection_records_keys = set()

            end = False
            skip_tracker = 0
            while end == False:
                process_collection_records = collection_outliers_rules.data.query(
                    skip=skip_tracker
                )
                if len(process_collection_records) != 0:
                    for item in process_collection_records:
                        if item.get("_key") not in collection_records_keys:
                            collection_records.append(item)
                            collection_records_keys.add(item.get("_key"))
                    skip_tracker += 5000
                else:
                    end = True

            # loop through the records
            for outlier_record in collection_records:
                entities_outliers = json.loads(outlier_record.get("entities_outliers"))
                for model_id in entities_outliers:
                    kpi_metric = entities_outliers[model_id].get("kpi_metric")
                    if kpi_metric == "splk.wlk.elapsed":
                        entities_outliers[model_id]["alert_lower_breached"] = 0

                # update the record
                outlier_record["entities_outliers"] = json.dumps(
                    entities_outliers, indent=2
                )

                # update the KVstore
                try:
                    collection_outliers_rules.data.update(
                        outlier_record.get("_key"),
                        json.dumps(outlier_record, indent=2),
                    )
                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", failed to update the entity rules, exception="{str(e)}"'
                    )

            #
            # enhance scheduler execution errors detection
            #

            # retrieve the list of scheduler reports to be processed
            collection_trackers_name = (
                f"kv_trackme_wlk_hybrid_trackers_tenant_{tenant_id}"
            )
            collection_trackers = service.kvstore[collection_trackers_name]

            trackers_dict = {}

            trackers_to_process = []
            trackers_to_process_kos = []

            # get records from the KVstore
            trackers_records = collection_trackers.data.query()

            for tracker_record in trackers_records:
                tracker_name = tracker_record.get("tracker_name")
                tracker_kos = tracker_record.get("knowledge_objects")

                # if the tracker_name starts by scheduler_
                if re.search("^scheduler_", tracker_name):
                    trackers_to_process.append(tracker_name)
                    trackers_to_process_kos.append(tracker_kos)

                    trackers_dict[tracker_name] = {
                        "knowledge_objects": json.loads(tracker_kos),
                    }

            # loop through the trackers
            for tracker_shortname in trackers_to_process:
                tracker_name = (
                    f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"
                )

                tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                # get the current search definition
                tracker_current = service.saved_searches[tracker_name]
                tracker_current_search = tracker_current.content.get("search")
                tracker_account = tracker_kos["properties"][0]["account"]

                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                )

                # in tracker_current_search, replace the sentence:
                # len(errmsg)>0,\"error\"
                # with: (note that account need to replaced by the actual account name)
                # len(errmsg)>0,\"error\" OR status == \"delegated_remote_error\"

                if tracker_account == "local":
                    tracker_new_search = re.sub(
                        r'len\(errmsg\)>0,"error"',
                        r'len(errmsg)>0 OR status == "delegated_remote_error","error"',
                        tracker_current_search,
                    )
                else:
                    tracker_new_search = re.sub(
                        r'len\(errmsg\)>0,\\"error\\"',
                        r'len(errmsg)>0 OR status == \\"delegated_remote_error\\",\\"error\\"',
                        tracker_current_search,
                    )

                # update the search definition
                url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                data = {
                    "tenant_id": tenant_id,
                    "report_name": tracker_name,
                    "report_search": tracker_new_search,
                }

                try:
                    response = requests.post(
                        url,
                        headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                        data=json.dumps(data),
                        verify=False,
                        timeout=300,
                    )
                    if response.status_code not in (200, 201, 204):
                        logging.error(
                            f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                        )
                    else:
                        logging.info(
                            f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                        )
                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                    )

            #
            # Update the metadata search to include an explicit call to filters_get_last_updates="host=*"
            #

            # retrieve the list of scheduler reports to be processed
            collection_trackers_name = (
                f"kv_trackme_wlk_hybrid_trackers_tenant_{tenant_id}"
            )
            collection_trackers = service.kvstore[collection_trackers_name]

            trackers_dict = {}

            trackers_to_process = []
            trackers_to_process_kos = []

            # get records from the KVstore
            trackers_records = collection_trackers.data.query()

            for tracker_record in trackers_records:
                tracker_name = tracker_record.get("tracker_name")
                tracker_kos = tracker_record.get("knowledge_objects")

                # if the tracker_name starts by scheduler_
                if re.search("^metadata_", tracker_name):
                    trackers_to_process.append(tracker_name)
                    trackers_to_process_kos.append(tracker_kos)

                    trackers_dict[tracker_name] = {
                        "knowledge_objects": json.loads(tracker_kos),
                    }

            # loop through the trackers
            for tracker_shortname in trackers_to_process:
                tracker_name = (
                    f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"
                )

                tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                # get the current search definition
                tracker_current = service.saved_searches[tracker_name]
                tracker_current_search = tracker_current.content.get("search")
                tracker_account = tracker_kos["properties"][0]["account"]

                logging.info(
                    f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                )

                # in tracker_current_search, replace the sentence:
                # max_runtime_sec="900"
                # with:
                # max_runtime_sec="900" filters_get_last_updates="host=*"

                tracker_new_search = re.sub(
                    r'max_runtime_sec="900"',
                    r'max_runtime_sec="900" filters_get_last_updates="host=*"',
                    tracker_current_search,
                )

                # update the search definition
                url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                data = {
                    "tenant_id": tenant_id,
                    "report_name": tracker_name,
                    "report_search": tracker_new_search,
                }

                try:
                    response = requests.post(
                        url,
                        headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                        data=json.dumps(data),
                        verify=False,
                        timeout=300,
                    )
                    if response.status_code not in (200, 201, 204):
                        logging.error(
                            f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                        )
                    else:
                        logging.info(
                            f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                        )
                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                    )

            #
            # ML Outliers: update the default Outliers rules macro
            #

            macro_name = f"trackme_wlk_set_outliers_metrics_tenant_{tenant_id}"
            macro_current = service.confs["macros"][macro_name]
            macro_current_definition = macro_current.content.get("definition")

            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", macro_name="{macro_name}", macro_current_definition="{macro_current_definition}", processing macro update'
            )

            """
            In macro_current_defintion, replace:
            
            eval outliers_metrics="{'elapsed': {'alert_lower_breached': 0, 'alert_upper_breached': 1}}"
            
            by:
            
            eval outliers_metrics="{'elapsed': {'alert_lower_breached': 0, 'alert_upper_breached': 1, 'time_factor': 'none'}}"

            """

            macro_new_definition = "eval outliers_metrics=\"{'elapsed': {'alert_lower_breached': 0, 'alert_upper_breached': 1, 'time_factor': 'none'}}\""

            # update the search definition
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_macro'
            data = {
                "tenant_id": tenant_id,
                "macro_name": macro_name,
                "macro_definition": macro_new_definition,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", update macro definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", successfully updated macro definition, transform="{macro_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2072, tenant_id="{tenant_id}", failed to update the macro definition, transform="{macro_name}", exception="{str(e)}"'
                )

            #
            # ML Outliers: mass update rules for wlk
            #

            # set kwargs
            kwargs_wlk_outliers_rules_update = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # define the gen search
            wlk_outliers_rules_update_search = remove_leading_spaces(
                f"""\
                | inputlookup trackme_wlk_outliers_entity_rules_tenant_{tenant_id} | eval keyid=_key
                ``` mass update the time_factor ```
                | rex field=entities_outliers mode=sed "s/\\\"time_factor\\\": \\\"%H\\\"/\\\"time_factor\\\": \\\"none\\\"/g"
                ``` force update the confidence and confidence_reason ```
                | eval confidence="low", confidence_reason="upgrade to TrackMe 2.0.72 for splk-wlk requires models to be trained again due to the change of the time_factor field, which will happen automatically as soon as possible."
                | outputlookup append=t key_field=keyid trackme_wlk_outliers_entity_rules_tenant_{tenant_id}
                """
            )

            wlk_outliers_rules_update_counter = 0
            wlk_outliers_rules_updated_entities = []

            # automatic retry
            max_failures_count = 24
            sleep_time = 5
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                # track the search runtime
                start = time.time()

                # proceed
                try:
                    wlk_outliers_rules_results = service.jobs.export(
                        wlk_outliers_rules_update_search,
                        **kwargs_wlk_outliers_rules_update,
                    )
                    reader = results.JSONResultsReader(wlk_outliers_rules_results)

                    for item in reader:
                        if isinstance(item, dict):
                            wlk_outliers_rules_update_counter += 1
                            wlk_outliers_rules_updated_entities.append(
                                item.get("object")
                            )

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                tenant_id,
                                component,
                                current_failures_count,
                                max_failures_count,
                                str(e),
                            )
                            logging.error(msg)
                            raise Exception(msg)
                        else:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", temporary search failure with exception="{}", run_time="{}"'.format(
                                tenant_id,
                                component,
                                str(e),
                                str(time.time() - start),
                            )
                            logging.warn(msg)
                            time.sleep(sleep_time)
                    else:
                        msg = 'context="schema_upgrade", tenant_id="{}", component="{}", search failed with exception="{}", run_time="{}"'.format(
                            tenant_id,
                            component,
                            str(e),
                            str(time.time() - start),
                        )
                        logging.error(msg)
                        raise Exception(msg)

            # log
            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", component="{component}", wlk_outliers_rules_update_counter="{wlk_outliers_rules_update_counter}", wlk_outliers_rules_updated_entities="{json.dumps(wlk_outliers_rules_updated_entities, indent=2)}", search="{wlk_outliers_rules_update_search}"'
            )

        #
        # End version 2.0.72
        #

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2072 terminated, tenant_id="{tenant_id}"'
        )


"""
In this function:
- Update the vtenant accounts to define adaptive_delay
- Create a new KVstore collection for wlk to store last seen activity per type of data
- Support overlap in the searches with deduplication for a safer collection of Wlk metrics
"""


def trackme_schema_upgrade_2075(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2075, tenant_id="{tenant_id}"'
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Vtenant account preferences
        #

        #
        # Update the vtenant account configuration
        #
        tenant_id = vtenant_record.get("tenant_id")

        url = f'{reqinfo["server_rest_uri"]}/servicesNS/nobody/trackme/trackme_vtenants/{tenant_id}'
        data = {
            "adaptive_delay": 1,
        }

        # update the account
        try:
            response = requests.post(
                url,
                headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                data=data,
                verify=False,
                timeout=300,
            )
            if response.status_code not in (200, 201, 204):
                logging.error(
                    f'context="schema_upgrade", schema migration 2075, update vtenant account has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                )
            else:
                logging.info(
                    f'context="schema_upgrade", schema migration 2075, update vtenant account was operated successfully, response.status_code="{response.status_code}"'
                )
        except Exception as e:
            logging.error(
                f'context="schema_upgrade", schema migration 2075, update vtenant account has failed, exception="{str(e)}"'
            )

        #
        # Components updates
        #

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        #
        # Adaptive trackers
        #

        components_to_process = []
        if (
            vtenant_record.get("tenant_dsm_enabled") == 1
            and vtenant_record.get("tenant_replica") == 0
        ):
            components_to_process.append("dsm")
        if (
            vtenant_record.get("tenant_dhm_enabled") == 1
            and vtenant_record.get("tenant_replica") == 0
        ):
            components_to_process.append("dhm")

        for component in components_to_process:
            #
            # Update the adaptive delay tracker
            #

            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", component="{component}", processing with adaptive delay tracker update'
            )

            # report name
            tracker_name = (
                f"trackme_{component}_adaptive_delay_tracker_tenant_{tenant_id}"
            )

            # get the current search definition
            tracker_current = service.saved_searches[tracker_name]
            tracker_current_search = tracker_current.content.get("search")

            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", tracker_current_search="{tracker_current_search}"'
            )

            # add our new parameter
            tracker_new_search = f"{tracker_current_search} max_auto_delay_sec=604800"

            # update the search definition
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
            data = {
                "tenant_id": tenant_id,
                "report_name": tracker_name,
                "report_search": tracker_new_search,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                )

        #
        # WLK: create new collection & transforms definition
        #

        # check components and add accordingly
        if (
            vtenant_record.get("tenant_wlk_enabled") == 1
            and vtenant_record.get("tenant_replica") == 0
        ):
            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", component="wlk", processing with Workload tracker updates'
            )

            transform_name = f"trackme_wlk_last_seen_activity_tenant_{tenant_id}"
            collection_name = f"kv_trackme_wlk_last_seen_activity_tenant_{tenant_id}"
            transform_fields = collections_dict["trackme_wlk_last_seen_activity"]
            definition_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # create the KVstore collection
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/create_kvcollection'
            data = {
                "tenant_id": tenant_id,
                "collection_name": collection_name,
                "collection_acl": definition_acl,
                "owner": vtenant_record.get("tenant_owner"),
            }

            kvstore_created = False
            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", create KVstore collection has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully KVstore collection, collection="{collection_name}"'
                    )
                    kvstore_created = True
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to create the KVstore collection, collection="{collection_name}", exception="{str(e)}"'
                )

            #
            # Only continue if successful
            #

            # create the transform
            if kvstore_created:
                url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/create_kvtransform'
                data = {
                    "tenant_id": tenant_id,
                    "transform_name": transform_name,
                    "transform_fields": transform_fields,
                    "collection_name": collection_name,
                    "transform_acl": definition_acl,
                    "owner": vtenant_record.get("tenant_owner"),
                }

                try:
                    response = requests.post(
                        url,
                        headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                        data=json.dumps(data),
                        verify=False,
                        timeout=300,
                    )
                    if response.status_code not in (200, 201, 204):
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", create transform definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                        )
                    else:
                        logging.info(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully created transform definition, transform="{transform_name}"'
                        )
                except Exception as e:
                    logging.error(
                        f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
                    )

                #
                # Trackers updates
                #

                # retrieve the list of scheduler reports to be processed
                collection_trackers_name = (
                    f"kv_trackme_wlk_hybrid_trackers_tenant_{tenant_id}"
                )
                collection_trackers = service.kvstore[collection_trackers_name]

                #
                # enhance scheduler execution to support deduplication
                #

                trackers_dict = {}
                trackers_to_process = []
                trackers_to_process_kos = []

                # get records from the KVstore
                trackers_records = collection_trackers.data.query()

                for tracker_record in trackers_records:
                    tracker_name = tracker_record.get("tracker_name")
                    tracker_kos = tracker_record.get("knowledge_objects")

                    # if the tracker_name starts by scheduler_
                    if re.search("^scheduler_", tracker_name):
                        trackers_to_process.append(tracker_name)
                        trackers_to_process_kos.append(tracker_kos)

                        trackers_dict[tracker_name] = {
                            "knowledge_objects": json.loads(tracker_kos),
                        }

                # loop through the trackers (wrapper)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # in tracker_current_search, replace the sentence: (if account is not local, double escape the quotes)
                    #  _index_earliest="-5m" _index_latest="now"
                    # with:
                    #  _index_earliest="-20m" _index_latest="now"
                    # | bucket _time span=5m

                    if tracker_account == "local":
                        tracker_new_search = re.sub(
                            r'_index_earliest="-5m" _index_latest="now"',
                            r'_index_earliest="-20m" _index_latest="now" earliest="-20m" latest="now"\n| eval orig_time = _time | bucket _time span=5m',
                            tracker_current_search,
                        )

                    else:
                        tracker_new_search = re.sub(
                            r'_index_earliest=\\"-5m\\" _index_latest=\\"now\\"',
                            r'_index_earliest=\\"-20m\\" _index_latest=\\"now\\" earliest=\\"-20m\\" latest=\\"now\\"\n| eval orig_time = _time | bucket _time span=5m',
                            tracker_current_search,
                        )

                    # only if remote, replace the sentence:
                    #  earliest="-10m" latest="now"
                    # with:
                    #  earliest="-20m" latest="now"

                    if tracker_account != "local":
                        tracker_new_search = re.sub(
                            r'earliest="-10m" latest="now"',
                            r'earliest="-20m" latest="now"',
                            tracker_new_search,
                        )

                    # replace the sentence:
                    # | stats count(eval(status
                    # with:
                    # | stats max(_time) as _time, count(eval(status

                    tracker_new_search = re.sub(
                        r"\| stats count\(eval\(status",
                        r"| stats max(_time) as _time, count(eval(status",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # by app, user, savedsearch_name
                    # with:
                    # by _time, app, user, savedsearch_name

                    tracker_new_search = re.sub(
                        r"by app, user, savedsearch_name",
                        r"by orig_time, app, user, savedsearch_name",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # | fields app, user, savedsearch_name, count_completed
                    # with:
                    # | fields last_seen, app, user, savedsearch_name, count_completed

                    tracker_new_search = re.sub(
                        r"\| fields app, user, savedsearch_name, count_completed",
                        r"| fields _time, app, user, savedsearch_name, count_completed",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # context="live"
                    # with:
                    # context="live" check_last_seen=True check_last_seen_field=last_seen_scheduler

                    tracker_new_search = re.sub(
                        r"context=\"live\"",
                        r"context=live check_last_seen=True check_last_seen_field=last_seen_scheduler",
                        tracker_new_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                # loop through the trackers (tracker)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_tracker_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # replace the sentence:
                    # earliest="-10m" latest="now"
                    # with:
                    # earliest="-20m" latest="now"

                    tracker_new_search = re.sub(
                        r'earliest="-10m" latest="now"',
                        r'earliest="-20m" latest="now"',
                        tracker_current_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                #
                # enhance introspection execution to support deduplication
                #

                trackers_dict = {}
                trackers_to_process = []
                trackers_to_process_kos = []

                # get records from the KVstore
                trackers_records = collection_trackers.data.query()

                for tracker_record in trackers_records:
                    tracker_name = tracker_record.get("tracker_name")
                    tracker_kos = tracker_record.get("knowledge_objects")

                    # if the tracker_name starts by scheduler_
                    if re.search("^introspection_", tracker_name):
                        trackers_to_process.append(tracker_name)
                        trackers_to_process_kos.append(tracker_kos)

                        trackers_dict[tracker_name] = {
                            "knowledge_objects": json.loads(tracker_kos),
                        }

                # loop through the trackers (wrapper)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # in tracker_current_search, replace the sentence: (if account is not local, double escape the quotes)
                    #  _index_earliest="-5m" _index_latest="now"
                    # with:
                    #  _index_earliest="-20m" _index_latest="now"
                    # | bucket _time span=5m

                    if tracker_account == "local":
                        tracker_new_search = re.sub(
                            r'_index_earliest="-5m" _index_latest="now"',
                            r'_index_earliest="-20m" _index_latest="now" earliest="-20m" latest="now"',
                            tracker_current_search,
                        )

                    else:
                        tracker_new_search = re.sub(
                            r'_index_earliest=\\"-5m\\" _index_latest=\\"now\\"',
                            r'_index_earliest=\\"-20m\\" _index_latest=\\"now\\" earliest=\\"-20m\\" latest=\\"now\\"',
                            tracker_current_search,
                        )

                    # only if remote, replace the sentence:
                    #  earliest="-10m" latest="now"
                    # with:
                    #  earliest="-20m" latest="now"

                    if tracker_account != "local":
                        tracker_new_search = re.sub(
                            r'earliest="-10m" latest="now"',
                            r'earliest="-20m" latest="now"',
                            tracker_new_search,
                        )

                    # replace the sentence:
                    # | stats max(_time) as last_time, avg(elapsed) as elapsed, avg(pct_cpu) as pct_cpu
                    # with:
                    # | eval orig_time = _time | bucket orig_time span=5m | stats max(_time) as last_time, avg(elapsed) as elapsed, avg(pct_cpu) as pct_cpu

                    tracker_new_search = re.sub(
                        r"\| stats max\(_time\) as last_time, avg\(elapsed\) as elapsed, avg\(pct_cpu\) as pct_cpu",
                        r"| eval orig_time = _time | bucket orig_time span=5m | stats max(_time) as _time, avg(elapsed) as elapsed, avg(pct_cpu) as pct_cpu",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # by object, app, savedsearch_name
                    # with:
                    # by _time, object, app, savedsearch_name

                    tracker_new_search = re.sub(
                        r"by object, app, savedsearch_name",
                        r"by orig_time, object, app, savedsearch_name",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # | fields object, object_id, app, savedsearch_name, last_time, elapsed, pct_cpu, pct_memory, scan_count, type, user, group
                    # with:
                    # | fields object, object_id, app, savedsearch_name, last_seen, elapsed, pct_cpu, pct_memory, scan_count, type, user, group

                    tracker_new_search = re.sub(
                        r"\| fields object, object_id, app, savedsearch_name, last_time, elapsed, pct_cpu, pct_memory, scan_count, type, user, group",
                        r"| fields _time, object, object_id, app, savedsearch_name, elapsed, pct_cpu, pct_memory, scan_count, type, user, group",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # context="live"
                    # with:
                    # context="live" check_last_seen=True check_last_seen_field=last_seen_introspection

                    tracker_new_search = re.sub(
                        r"context=\"live\"",
                        r"context=live check_last_seen=True check_last_seen_field=last_seen_introspection",
                        tracker_new_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                # loop through the trackers (tracker)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_tracker_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # replace the sentence:
                    # earliest="-10m" latest="now"
                    # with:
                    # earliest="-20m" latest="now"

                    tracker_new_search = re.sub(
                        r'earliest="-10m" latest="now"',
                        r'earliest="-20m" latest="now"',
                        tracker_current_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                #
                # enhance notable execution to support deduplication
                #

                trackers_dict = {}
                trackers_to_process = []
                trackers_to_process_kos = []

                # get records from the KVstore
                trackers_records = collection_trackers.data.query()

                for tracker_record in trackers_records:
                    tracker_name = tracker_record.get("tracker_name")
                    tracker_kos = tracker_record.get("knowledge_objects")

                    # if the tracker_name starts by scheduler_
                    if re.search("^notable_", tracker_name):
                        trackers_to_process.append(tracker_name)
                        trackers_to_process_kos.append(tracker_kos)

                        trackers_dict[tracker_name] = {
                            "knowledge_objects": json.loads(tracker_kos),
                        }

                # loop through the trackers (wrapper)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # in tracker_current_search, replace the sentence: (if account is not local, double escape the quotes)
                    #  _index_earliest="-5m" _index_latest="now"
                    # with:
                    #  _index_earliest="-20m" _index_latest="now"
                    # | bucket _time span=5m

                    if tracker_account == "local":
                        tracker_new_search = re.sub(
                            r'_index_earliest="-5m" _index_latest="now"',
                            r'_index_earliest="-20m" _index_latest="now" earliest="-20m" latest="now"',
                            tracker_current_search,
                        )

                    else:
                        tracker_new_search = re.sub(
                            r'_index_earliest=\\"-5m\\" _index_latest=\\"now\\"',
                            r'_index_earliest=\\"-20m\\" _index_latest=\\"now\\" earliest=\\"-20m\\" latest=\\"now\\"',
                            tracker_current_search,
                        )

                    # only if remote, replace the sentence:
                    #  earliest="-10m" latest="now"
                    # with:
                    #  earliest="-20m" latest="now"

                    if tracker_account != "local":
                        tracker_new_search = re.sub(
                            r'earliest="-10m" latest="now"',
                            r'earliest="-20m" latest="now"',
                            tracker_new_search,
                        )

                    # replace the sentence:
                    # by source
                    # with:
                    # by _time, source span=5m

                    tracker_new_search = re.sub(
                        r"by source",
                        r"by _time, source span=5m",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # | fields app, user, savedsearch_name, count_ess_notable, group, object, object_id
                    # with:
                    # | fields last_seen, app, user, savedsearch_name, count_ess_notable, group, object, object_id

                    tracker_new_search = re.sub(
                        r"\| fields app, user, savedsearch_name, count_ess_notable, group, object, object_id",
                        r"| fields _time, app, user, savedsearch_name, count_ess_notable, group, object, object_id",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # context="live"
                    # with:
                    # context="live" check_last_seen=True check_last_seen_field=last_seen_notable

                    tracker_new_search = re.sub(
                        r"context=\"live\"",
                        r"context=live check_last_seen=True check_last_seen_field=last_seen_notable",
                        tracker_new_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                # loop through the trackers (tracker)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_tracker_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # replace the sentence:
                    # earliest="-10m" latest="now"
                    # with:
                    # earliest="-20m" latest="now"

                    tracker_new_search = re.sub(
                        r'earliest="-10m" latest="now"',
                        r'earliest="-20m" latest="now"',
                        tracker_current_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                #
                # enhance splunkcloud_svc execution to support deduplication
                #

                trackers_dict = {}
                trackers_to_process = []
                trackers_to_process_kos = []

                # get records from the KVstore
                trackers_records = collection_trackers.data.query()

                for tracker_record in trackers_records:
                    tracker_name = tracker_record.get("tracker_name")
                    tracker_kos = tracker_record.get("knowledge_objects")

                    # if the tracker_name starts by scheduler_
                    if re.search("^splunkcloud_svc_", tracker_name):
                        trackers_to_process.append(tracker_name)
                        trackers_to_process_kos.append(tracker_kos)

                        trackers_dict[tracker_name] = {
                            "knowledge_objects": json.loads(tracker_kos),
                        }

                # loop through the trackers (wrapper)
                for tracker_shortname in trackers_to_process:
                    tracker_name = f"trackme_wlk_hybrid_{tracker_shortname}_wrapper_tenant_{tenant_id}"

                    tracker_kos = trackers_dict[tracker_shortname]["knowledge_objects"]

                    # get the current search definition
                    tracker_current = service.saved_searches[tracker_name]
                    tracker_current_search = tracker_current.content.get("search")
                    tracker_account = tracker_kos["properties"][0]["account"]

                    logging.info(
                        f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", account={tracker_account}, tracker_current_search="{tracker_current_search}", tracker_kos="{json.dumps(tracker_kos, indent=2)}"'
                    )

                    # in tracker_current_search, replace the sentence: (if account is not local, double escape the quotes)
                    #  _index_earliest="-15m" _index_latest="now"
                    # with:
                    #  _index_earliest="-3h" _index_latest="now"

                    if tracker_account == "local":
                        tracker_new_search = re.sub(
                            r'_index_earliest="-15m" _index_latest="now"',
                            r'_index_earliest="-3h" _index_latest="now" earliest="--3h" latest="now"',
                            tracker_current_search,
                        )

                    else:
                        tracker_new_search = re.sub(
                            r'_index_earliest=\\"-15m\\" _index_latest=\\"now\\"',
                            r'_index_earliest=\\"-3h\\" _index_latest=\\"now\\" earliest=\\"-3h\\" latest=\\"now\\"',
                            tracker_current_search,
                        )

                    # replace the sentence:
                    # | fields app, group, object, object_id, savedsearch_name, user, svc_usage
                    # with:
                    # | fields _time, app, group, object, object_id, savedsearch_name, user, svc_usage

                    tracker_new_search = re.sub(
                        r"\| fields app, group, object, object_id, savedsearch_name, user, svc_usage",
                        r"| fields _time, app, group, object, object_id, savedsearch_name, user, svc_usage",
                        tracker_new_search,
                    )

                    # replace the sentence:
                    # context="live"
                    # with:
                    # context="live" check_last_seen=True check_last_seen_field=last_seen_splunkcloud_svc

                    tracker_new_search = re.sub(
                        r"context=\"live\"",
                        r"context=live check_last_seen=True check_last_seen_field=last_seen_splunkcloud_svc",
                        tracker_new_search,
                    )

                    # update the search definition
                    url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
                    data = {
                        "tenant_id": tenant_id,
                        "report_name": tracker_name,
                        "report_search": tracker_new_search,
                        "earliest_time": "-20m",
                        "latest_time": "now",
                    }

                    try:
                        response = requests.post(
                            url,
                            headers={
                                "Authorization": f'Splunk {reqinfo["session_key"]}'
                            },
                            data=json.dumps(data),
                            verify=False,
                            timeout=300,
                        )
                        if response.status_code not in (200, 201, 204):
                            logging.error(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                            )
                        else:
                            logging.info(
                                f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                            )
                    except Exception as e:
                        logging.error(
                            f'context="schema_upgrade", schema migration 2075, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                        )

                # Notes: updating trackers for splunkcloud_svc is not required

            #
            # ML Outliers: mass update rules for wlk
            #

            # set kwargs
            kwargs_wlk_outliers_rules_update = {
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }

            # define the gen search
            wlk_outliers_rules_update_search = remove_leading_spaces(
                f"""\
                | inputlookup trackme_wlk_outliers_entity_rules_tenant_{tenant_id} | eval keyid=_key
                ``` mass update the time_factor ```
                | where match(entities_outliers, "%H")
                | rex field=entities_outliers mode=sed "s/\\\"time_factor\\\": \\\"%H\\\"/\\\"time_factor\\\": \\\"none\\\"/g"
                ``` force update the confidence and confidence_reason ```
                | eval confidence="low", confidence_reason="upgrade to TrackMe 2.0.72 for splk-wlk requires models to be trained again due to the change of the time_factor field, which will happen automatically as soon as possible."
                | outputlookup append=t key_field=keyid trackme_wlk_outliers_entity_rules_tenant_{tenant_id}
                """
            )

            wlk_outliers_rules_update_counter = 0
            wlk_outliers_rules_updated_entities = []

            # automatic retry
            max_failures_count = 24
            sleep_time = 5
            search_successful = False
            current_failures_count = 0
            while current_failures_count < max_failures_count and not search_successful:
                # track the search runtime
                start = time.time()

                # proceed
                try:
                    wlk_outliers_rules_results = service.jobs.export(
                        wlk_outliers_rules_update_search,
                        **kwargs_wlk_outliers_rules_update,
                    )
                    reader = results.JSONResultsReader(wlk_outliers_rules_results)

                    for item in reader:
                        if isinstance(item, dict):
                            wlk_outliers_rules_update_counter += 1
                            wlk_outliers_rules_updated_entities.append(
                                item.get("object")
                            )

                    # break while
                    search_successful = True
                    break

                except Exception as e:
                    if re.search(
                        "The maximum number of concurrent historical searches on this instance has been reached",
                        str(e),
                    ):
                        current_failures_count += 1
                        search_successful = False
                        if current_failures_count >= max_failures_count:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", permanent search failure, current_failures_count="{}", max_failures_count="{}", exception="{}"'.format(
                                tenant_id,
                                component,
                                current_failures_count,
                                max_failures_count,
                                str(e),
                            )
                            logging.error(msg)
                            raise Exception(msg)
                        else:
                            msg = 'context="schema_upgrade", tenant_id="{}", component="{}", temporary search failure with exception="{}", run_time="{}"'.format(
                                tenant_id,
                                component,
                                str(e),
                                str(time.time() - start),
                            )
                            logging.warn(msg)
                            time.sleep(sleep_time)
                    else:
                        msg = 'context="schema_upgrade", tenant_id="{}", component="{}", search failed with exception="{}", run_time="{}"'.format(
                            tenant_id,
                            component,
                            str(e),
                            str(time.time() - start),
                        )
                        logging.error(msg)
                        raise Exception(msg)

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2075 terminated, tenant_id="{tenant_id}"'
        )


def trackme_schema_upgrade_2077(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2077, tenant_id="{tenant_id}"'
    )
    objects_to_process = []

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        # check components and add accordingly
        if vtenant_record.get("tenant_cim_enabled") == 1:
            objects_to_process.append("trackme_cim")

        # get RBAC
        tenant_roles_admin = str(vtenant_record.get("tenant_roles_admin"))
        tenant_roles_user = str(vtenant_record.get("tenant_roles_user"))
        tenant_roles_power = str(vtenant_record.get("tenant_roles_power"))

        # for read permissions, concatenate admin, power and user
        tenant_roles_read_perms = (
            f"{tenant_roles_admin},{tenant_roles_power},{tenant_roles_user}"
        )

        # for write permissions, concatenate admin, power
        tenant_roles_write_perms = f"{tenant_roles_admin},{tenant_roles_power}"

        for object_name in objects_to_process:
            transform_name = "%s_tenant_%s" % (object_name, tenant_id)
            collection_name = "kv_%s_tenant_%s" % (object_name, tenant_id)
            transform_fields = collections_dict[object_name]
            transform_acl = {
                "owner": vtenant_record.get("tenant_owner"),
                "sharing": "app",
                "perms.write": tenant_roles_read_perms,
                "perms.read": tenant_roles_write_perms,
            }

            # delete the transform
            try:
                action_delete = trackme_delete_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully deleted transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to delete the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

            # create the transform
            try:
                action_create = trackme_create_kvtransform(
                    reqinfo["session_key"],
                    reqinfo["server_rest_uri"],
                    tenant_id,
                    transform_name,
                    transform_fields,
                    collection_name,
                    vtenant_record.get("tenant_owner"),
                    transform_acl,
                )
                logging.info(
                    'context="schema_upgrade", tenant_id="{}", successfully created transform definition, transform="{}"'.format(
                        tenant_id, transform_name
                    )
                )
            except Exception as e:
                logging.error(
                    'context="schema_upgrade", tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )
                raise Exception(
                    'tenant_id="{}", failed to create the transform definition, transform="{}", exception="{}"'.format(
                        tenant_id, transform_name, str(e)
                    )
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2077 terminated, tenant_id="{tenant_id}"'
        )


"""
In this function:
- Update the adaptive delay report to add the new option max_changes_past_7days
"""


def trackme_schema_upgrade_2078(reqinfo, tenant_id):
    # get service
    service = client.connect(
        owner="nobody",
        app="trackme",
        port=reqinfo["server_rest_port"],
        token=reqinfo["session_key"],
        timeout=300,
    )

    # Register the object summary in the vtenant collection
    collection_vtenants_name = "kv_trackme_virtual_tenants"
    collection_vtenants = service.kvstore[collection_vtenants_name]

    # Define the KV query search string
    query_string = {
        "tenant_id": tenant_id,
    }

    # log
    logging.debug(
        f'context="schema_upgrade", Starting function trackme_schema_upgrade_2078, tenant_id="{tenant_id}"'
    )

    # get the tenant record
    try:
        vtenant_record = collection_vtenants.data.query(query=json.dumps(query_string))[
            0
        ]
        vtenant_key = vtenant_record.get("_key")
        logging.debug(
            f'context="schema_upgrade", The vtenant_key was successfully found in the collection, query_string="{query_string}"'
        )

    except Exception as e:
        vtenant_key = None
        logging.error(
            f'context="schema_upgrade", The vtenant_key was not found in the collection, query_string="{query_string}"'
        )

    if vtenant_key:
        #
        # Components updates
        #

        #
        # Adaptive trackers
        #

        components_to_process = []
        if (
            vtenant_record.get("tenant_dsm_enabled") == 1
            and vtenant_record.get("tenant_replica") == 0
        ):
            components_to_process.append("dsm")
        if (
            vtenant_record.get("tenant_dhm_enabled") == 1
            and vtenant_record.get("tenant_replica") == 0
        ):
            components_to_process.append("dhm")

        for component in components_to_process:
            #
            # Update the adaptive delay tracker
            #

            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", component="{component}", processing with adaptive delay tracker update'
            )

            # report name
            tracker_name = (
                f"trackme_{component}_adaptive_delay_tracker_tenant_{tenant_id}"
            )

            # get the current search definition
            tracker_current = service.saved_searches[tracker_name]
            tracker_current_search = tracker_current.content.get("search")

            logging.info(
                f'context="schema_upgrade", tenant_id="{tenant_id}", tracker_name="{tracker_name}", tracker_current_search="{tracker_current_search}"'
            )

            # add our new parameter
            tracker_new_search = f"{tracker_current_search} max_changes_past_7days=10"

            # update the search definition
            url = f'{reqinfo["server_rest_uri"]}/services/trackme/v2/configuration/admin/update_report'
            data = {
                "tenant_id": tenant_id,
                "report_name": tracker_name,
                "report_search": tracker_new_search,
            }

            try:
                response = requests.post(
                    url,
                    headers={"Authorization": f'Splunk {reqinfo["session_key"]}'},
                    data=json.dumps(data),
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):
                    logging.error(
                        f'context="schema_upgrade", schema migration 2078, tenant_id="{tenant_id}", update report definition has failed, response.status_code="{response.status_code}", response.text="{response.text}"'
                    )
                else:
                    logging.info(
                        f'context="schema_upgrade", schema migration 2078, tenant_id="{tenant_id}", successfully updated report definition, report="{tracker_name}"'
                    )
            except Exception as e:
                logging.error(
                    f'context="schema_upgrade", schema migration 2078, tenant_id="{tenant_id}", failed to update the report definition, report="{tracker_name}", exception="{str(e)}"'
                )

        # log
        logging.info(
            f'context="schema_upgrade", function trackme_schema_upgrade_2078 terminated, tenant_id="{tenant_id}"'
        )
