#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import os
import sys
import requests
import json
import hashlib
from collections import OrderedDict
import time
import logging
from logging.handlers import RotatingFileHandler
from urllib.parse import urlencode
import urllib.parse
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import TrackMe libs
from trackme_libs import JSONFormatter

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


def trackme_flx_gen_metrics(
    tenant_id, object_value, object_id, metric_index, metrics_event
):
    # proceed
    try:
        # load the event as an object
        if not isinstance(metrics_event, dict):
            metrics_event = json.loads(metrics_event)

        # Set a new log file handler
        filehandler = RotatingFileHandler(
            splunkhome + "/var/log/splunk/trackme_flx_metrics.log",
            mode="a",
            maxBytes=100000000,
            backupCount=1,
        )
        formatter = JSONFormatter()
        logging.Formatter.converter = time.gmtime
        filehandler.setFormatter(formatter)
        log = logging.getLogger()  # root logger - Good to get it only once.
        orig_filehandler = None
        for hdlr in log.handlers[:]:  # remove the existing file handlers
            if isinstance(hdlr, logging.FileHandler):
                orig_filehandler = hdlr
                log.removeHandler(hdlr)
        log.addHandler(filehandler)  # set the new handler

        logging.info(
            "Metrics - group=wlk_metrics",
            extra={
                "target_index": metric_index,
                "tenant_id": tenant_id,
                "object": object_value,
                "object_id": object_id,
                "object_category": "splk-flx",
                "metrics_event": json.dumps(metrics_event),
            },
        )

        # restore the original file handler
        if orig_filehandler:
            log.removeHandler(filehandler)
            log.addHandler(orig_filehandler)

    except Exception as e:
        raise Exception(str(e))


# return main searches logics for that entity
def splk_flx_return_searches(tenant_id, entity_info):
    # log debug
    logging.debug(
        'Starting function=splk_flx_return_searches with entity_info="{}"'.format(
            json.dumps(entity_info, indent=2)
        )
    )

    # define required searches dynamically based on the upstream entity information
    splk_flx_mctalog_search = None
    splk_flx_metrics_report = None
    splk_flx_mpreview = None
    splk_flx_metrics_populate_search = None

    # get the object_id
    object_id = hashlib.md5(entity_info["object"].encode("utf-8")).hexdigest()

    try:
        ########
        # mstats
        ########

        # mcatalog
        splk_flx_mctalog_search = (
            "| mcatalog values(metric_name) as metrics, values(_dims) as dims where `trackme_metrics_idx("
            + tenant_id
            + ')` tenant_id="'
            + tenant_id
            + '" object_category="splk-flx" object="'
            + entity_info["object"]
            + '" metric_name=* by index'
        )

        # metrics report
        splk_flx_metrics_report = (
            "| mstats latest(_value) as latest_value, avg(_value) as avg_value, max(_value) as max_value, perc95(_value) as perc95_value, stdev(_value) as stdev_value where `trackme_metrics_idx("
            + tenant_id
            + ')` metric_name=* tenant_id="'
            + tenant_id
            + '" object_category="splk-flx" object="'
            + entity_info["object"]
            + '" by index, object, metric_name'
            + "\n| foreach *_value [ eval <<FIELD>> = if(match(metric_name, \"\.status\"), round('<<FIELD>>', 0), round('<<FIELD>>', 3)) ]"
        )

        # mpreview
        splk_flx_mpreview = (
            "| mpreview `trackme_metrics_idx("
            + tenant_id
            + ')` filter="tenant_id='
            + tenant_id
            + ' object_category="splk-flx" object='
            + entity_info["object"]
            + '"'
        )

        # metrics popuating search
        splk_flx_metrics_populate_search = (
            "| mcatalog values(metric_name) as metrics where `trackme_metrics_idx("
            + tenant_id
            + ')` tenant_id="'
            + tenant_id
            + '" object_category="splk-flx" object="'
            + entity_info["object"]
            + '" metric_name=* | mvexpand metrics | rename metrics as metric_name'
            + ' | rex field=metric_name "^trackme\.splk.flx\.(?<label>.*)" | eval order=if(metric_name=="trackme.splk.flx.status", 0, 1) | sort 0 order | fields - order'
        )

        # return
        return {
            "splk_flx_mctalog_search": "search?q="
            + urllib.parse.quote(splk_flx_mctalog_search),
            "splk_flx_mctalog_search_litsearch": splk_flx_mctalog_search,
            "splk_flx_metrics_report": "search?q="
            + urllib.parse.quote(splk_flx_metrics_report),
            "splk_flx_metrics_report_litsearch": splk_flx_metrics_report,
            "splk_flx_mpreview": "search?q=" + urllib.parse.quote(splk_flx_mpreview),
            "splk_flx_mpreview_litsearch": splk_flx_mpreview,
            "splk_flx_metrics_populate_search": splk_flx_metrics_populate_search,
        }

    except Exception as e:
        logging.error(
            'function splk_flx_return_searches, an exception was encountered, exception="{}"'.format(
                str(e)
            )
        )
        raise Exception(e)
