#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2021-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

import os
import sys
import requests
import json
import hashlib
from collections import OrderedDict
import time
import logging
from logging.handlers import RotatingFileHandler
from urllib.parse import urlencode
import urllib.parse
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import TrackMe libs
from trackme_libs import JSONFormatter

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


def trackme_ingest_version(index, source, sourcetype, event):
    try:
        filehandler = RotatingFileHandler(
            "%s/var/log/splunk/trackme_wlk_version.log" % splunkhome,
            mode="a",
            maxBytes=100000000,
            backupCount=1,
        )
        formatter = JSONFormatter()
        logging.Formatter.converter = time.gmtime
        filehandler.setFormatter(formatter)
        log = logging.getLogger()  # root logger
        orig_filehandler = None
        for hdlr in log.handlers[:]:  # remove the existing file handlers
            if isinstance(hdlr, logging.FileHandler):
                orig_filehandler = hdlr
                log.removeHandler(hdlr)
        log.addHandler(filehandler)  # set the new handler

        logging.info(
            "TrackMe State Events",
            extra={
                "target_index": index,
                "target_sourcetype": sourcetype,
                "target_source": source,
                "event": event,
            },
        )

        if orig_filehandler:
            log.removeHandler(filehandler)
            log.addHandler(orig_filehandler)

    except Exception as e:
        raise Exception(str(e))


def trackme_wlk_gen_metrics(
    tenant_id,
    overgroup,
    group,
    app,
    user,
    account,
    savedsearch_name,
    object_value,
    object_id,
    version_id,
    metric_index,
    metrics_event,
):
    try:
        filehandler = RotatingFileHandler(
            "%s/var/log/splunk/trackme_wlk_metrics.log" % splunkhome,
            mode="a",
            maxBytes=100000000,
            backupCount=1,
        )
        formatter = JSONFormatter()
        logging.Formatter.converter = time.gmtime
        filehandler.setFormatter(formatter)
        log = logging.getLogger()  # root logger
        orig_filehandler = None
        for hdlr in log.handlers[:]:  # remove the existing file handlers
            if isinstance(hdlr, logging.FileHandler):
                orig_filehandler = hdlr
                log.removeHandler(hdlr)
        log.addHandler(filehandler)  # set the new handler

        logging.info(
            "Metrics - group=wlk_metrics",
            extra={
                "target_index": metric_index,
                "tenant_id": tenant_id,
                "overgroup": overgroup,
                "group": group,
                "app": app,
                "user": user,
                "account": account,
                "savedsearch_name": savedsearch_name,
                "object": object_value,
                "object_id": object_id,
                "object_category": "splk-wlk",
                "version_id": version_id,
                "metrics_event": json.dumps(metrics_event),
            },
        )

        if orig_filehandler:
            log.removeHandler(filehandler)
            log.addHandler(orig_filehandler)

    except Exception as e:
        raise Exception(str(e))


# return main searches logics for that entity
def splk_wlk_return_searches(tenant_id, entity_info):
    # log debug
    logging.debug(
        'Starting function=splk_wlk_return_searches with entity_info="{}"'.format(
            json.dumps(entity_info, indent=2)
        )
    )

    # define required searches dynamically based on the upstream entity information
    splk_wlk_mctalog_search = None
    splk_wlk_metrics_report = None
    splk_wlk_mpreview = None
    splk_wlk_metrics_populate_search = None
    splk_wlk_scheduler_skipping_search = None
    splk_wlk_scheduler_errors_search = None
    splk_wlk_scheduler_skipping_search_sample = None
    splk_wlk_scheduler_errors_search_sample = None
    splk_wlk_check_orphan = None
    splk_wlk_get_metadata = None

    # get the object_id
    object_id = hashlib.md5(entity_info["object"].encode("utf-8")).hexdigest()
    object_name = entity_info["object"]

    try:
        ########
        # mstats
        ########

        # mcatalog
        splk_wlk_mctalog_search = (
            "| mcatalog values(metric_name) as metrics, values(_dims) as dims where `trackme_metrics_idx("
            + tenant_id
            + ')` tenant_id="'
            + tenant_id
            + '" object_category="splk-wlk" object_id="'
            + entity_info["object_id"]
            + '" metric_name=* by index'
        )

        # metrics report
        splk_wlk_metrics_report = (
            "| mstats latest(_value) as latest_value, avg(_value) as avg_value, max(_value) as max_value, perc95(_value) as perc95_value, stdev(_value) as stdev_value where `trackme_metrics_idx("
            + tenant_id
            + ')` metric_name=* tenant_id="'
            + tenant_id
            + '" object_category="splk-wlk" object_id="'
            + entity_info["object_id"]
            + '" by index, object, object_id, metric_name'
            + "\n| foreach *_value [ eval <<FIELD>> = if(match(metric_name, \"\.status\"), round('<<FIELD>>', 0), round('<<FIELD>>', 3)) ]"
        )

        # mpreview
        splk_wlk_mpreview = (
            "| mpreview `trackme_metrics_idx("
            + tenant_id
            + ')` filter="tenant_id='
            + tenant_id
            + ' object_category="splk-wlk" object_id=\\"'
            + entity_info["object_id"]
            + '\\""'
        )

        # metrics popuating search
        splk_wlk_metrics_populate_search = (
            "| mcatalog values(metric_name) as metrics where `trackme_metrics_idx("
            + tenant_id
            + ')` tenant_id="'
            + tenant_id
            + '" object_category="splk-wlk" object_id="'
            + entity_info["object_id"]
            + '" metric_name=* | mvexpand metrics | rename metrics as metric_name'
            + ' | rex field=metric_name "^trackme\.splk.wlk\.(?<label>.*)" | eval order=if(metric_name=="trackme.splk.wlk.status", 0, 1) | sort 0 order | fields - order'
        )

        # skipping & errors searches
        scheduler_base_search = (
            "(index=_internal sourcetype=scheduler)"
            + "\n``` in some error cases, we need to manage extractions and status ```"
            + '\n| rex field=savedsearch_id "^(?<user>[^\;]*)\;(?<app>[^\;]*)\;(?<savedsearch_name>.*)"'
            + "\n| eval user=coalesce(user, user_alt), app=coalesce(app, app_alt), savedsearch_name=coalesce(savedsearch_name, savedsearch_name_alt)"
            + '\n| search savedsearch_name="'
            + entity_info["savedsearch_name"]
            + '"'
            + '\n| eval errmsg=case(len(errmsg)>0, errmsg, match(log_level, "(?i)error") AND len(message)>0, message)'
            + '\n| eval alert_actions=if((isnull(alert_actions) OR (alert_actions == "")), "none", alert_actions)'
            + '\n| eval status=case(((status == "success") OR (status == "completed")),"completed",(status == "skipped"),"skipped",(status == "continued"),"deferred",len(errmsg)>0 OR status == "delegated_remote_error","error")'
        )

        # skipping
        splk_wlk_scheduler_skipping_search = (
            scheduler_base_search + '\n| where status="skipped"'
        )
        splk_wlk_scheduler_skipping_search_sample = (
            splk_wlk_scheduler_skipping_search + "\n| head 10"
        )

        if entity_info["account"] != "local":
            splk_wlk_scheduler_skipping_search = (
                '| splunkremotesearch account="'
                + entity_info["account"]
                + '" search="'
                + splk_wlk_scheduler_skipping_search.replace('"', '\\"')
                + '"'
            )
            splk_wlk_scheduler_skipping_search_sample = (
                '| splunkremotesearch account="'
                + entity_info["account"]
                + '" search="'
                + splk_wlk_scheduler_skipping_search_sample.replace('"', '\\"')
                + '"'
            )
        else:
            splk_wlk_scheduler_skipping_search_sample = f"search {splk_wlk_scheduler_skipping_search_sample}"  # required for the search to be executed through SmartStatus

        # errors count
        splk_wlk_scheduler_errors_search = (
            scheduler_base_search + '\n| where status="error"'
        )
        splk_wlk_scheduler_errors_search_sample = (
            splk_wlk_scheduler_errors_search + "\n| head 10"
        )

        if entity_info["account"] != "local":
            splk_wlk_scheduler_errors_search = (
                '| splunkremotesearch account="'
                + entity_info["account"]
                + '" search="'
                + splk_wlk_scheduler_errors_search.replace('"', '\\"')
                + '"'
            )
            splk_wlk_scheduler_errors_search_sample = (
                '| splunkremotesearch account="'
                + entity_info["account"]
                + '" search="'
                + splk_wlk_scheduler_errors_search_sample.replace('"', '\\"')
                + '"'
            )
        else:
            splk_wlk_scheduler_errors_search_sample = f"search {splk_wlk_scheduler_errors_search_sample}"  # required for the search to be executed through SmartStatus

        # orphan check
        splk_wlk_check_orphan = (
            "| rest timeout=1800 splunk_server=local /servicesNS/-/-/saved/searches add_orphan_field=yes count=0"
            + "\n| rename title as object, eai:acl.owner AS user, eai:acl.app AS app"
            + "\n| fields object, user, app, orphan"
            + '\n| eval user=if(user=="nobody" OR user=="splunk-system-user", "system", user)'
            + '\n| eval object = app . ":" . user . ":" . object, key=md5(object)'
            + f'\n| search object="{object_name}"'
            + "\n| table key, object, app, user, orphan"
        )

        if entity_info["account"] != "local":
            splk_wlk_check_orphan = (
                '| splunkremotesearch account="'
                + entity_info["account"]
                + '" search="'
                + splk_wlk_check_orphan.replace('"', '\\"')
                + '"'
            )

        # metadata
        splk_wlk_get_metadata = (
            f'| inputlookup trackme_wlk_tenant_{tenant_id} where object="{object_name}" | eval object_id=_key'
            + "| fields tenant_id, account, app, user, savedsearch_name, search, object, object_id, metrics"
            + "``` Call the inspector backend ```"
            + f'| trackmesplkwlkgetreportsdefstream tenant_id="{tenant_id}" context="live" max_runtime_sec="900"'
            + '```| trackmeprettyjson fields="json_data"```'
            + "| where isnotnull(json_data)"
            + "| table account, app, owner, sharing, savedsearch_name, object cron_schedule, cron_exec_sequence_sec, disabled, is_scheduled, schedule_window, earliest_time, latest_time, description, search"
            + f'| join type=outer [ | mstats latest(_value) as value where `trackme_metrics_idx({tenant_id})` tenant_id="{tenant_id}" earliest="-7d" latest="now" object_category="splk-wlk" object="{object_name}" metric_name=trackme.splk.wlk.count_execution by object span=5m | eval last_execution=strftime(_time, "%c") | head 1 | fields object, last_execution ]'
            + '| eval last_execution=if(isnull(last_execution), "No execution in the past 7 days", last_execution)'
        )

        # return
        return {
            "splk_wlk_mctalog_search": "search?q="
            + urllib.parse.quote(splk_wlk_mctalog_search),
            "splk_wlk_mctalog_search_litsearch": splk_wlk_mctalog_search,
            "splk_wlk_metrics_report": "search?q="
            + urllib.parse.quote(splk_wlk_metrics_report),
            "splk_wlk_metrics_report_litsearch": splk_wlk_metrics_report,
            "splk_wlk_mpreview": "search?q=" + urllib.parse.quote(splk_wlk_mpreview),
            "splk_wlk_mpreview_litsearch": splk_wlk_mpreview,
            "splk_wlk_metrics_populate_search": splk_wlk_metrics_populate_search,
            "splk_wlk_scheduler_skipping_search": "search?q="
            + urllib.parse.quote(splk_wlk_scheduler_skipping_search),
            "splk_wlk_scheduler_skipping_search_sample": splk_wlk_scheduler_skipping_search_sample,
            "splk_wlk_scheduler_errors_search": "search?q="
            + urllib.parse.quote(splk_wlk_scheduler_errors_search),
            "splk_wlk_scheduler_errors_search_sample": splk_wlk_scheduler_errors_search_sample,
            "splk_wlk_check_orphan": splk_wlk_check_orphan,
            "splk_wlk_get_metadata": splk_wlk_get_metadata,
        }

    except Exception as e:
        logging.error(
            'function splk_wlk_return_searches, an exception was encountered, exception="{}"'.format(
                str(e)
            )
        )
        raise Exception(e)
